using SimpleJSON;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.Text.RegularExpressions;
using System;
using UnityEngine;
using static EntryPointUtils;

sealed class AutoUpdate : MVRScript
{
    string _packagePrefix;

    public override void Init() => StartCoroutine(InitCo());

    IEnumerator InitCo()
    {
        yield return new WaitForEndOfFrame();
        while(SuperController.singleton.isLoading)
        {
            yield return null;
        }

        string packageId = GetPackageId(manager, GetPluginId(name));
        _packagePrefix = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";

        var naturalisBM = this.FirstOwnAtomStorableByRegex(NaturalisBMRegex);
        var naturalisTM = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
        if(naturalisBM || naturalisTM)
        {
            yield return MigrateFromNaturalis(naturalisBM, naturalisTM);
            yield return new WaitForSeconds(1.5f);
            manager.RemovePluginWithUID(GetPluginId(name));
        }
        else
        {
            AddNaturalis(manager, _packagePrefix);
            manager.RemovePluginWithUID(GetPluginId(name));
        }
    }

    IEnumerator MigrateFromNaturalis(JSONStorable naturalisBM, JSONStorable naturalisTM)
    {
        var sourceStorable = naturalisBM ? naturalisBM : naturalisTM;
        string pluginId = GetPluginId(sourceStorable.name);
        var bootyMagicJSON = naturalisBM ? naturalisBM.GetJSON(true, true, true) : null;
        var tittyMagicJSON = naturalisTM ? naturalisTM.GetJSON(true, true, true) : null;
        manager.RemovePluginWithUID(pluginId);
        pluginLabelJSON.val = $"Adding {Constant.NATURALIS}...";
        AddNaturalis(manager, _packagePrefix);
        yield return new WaitForSeconds(1);
        yield return RestorePlugins(bootyMagicJSON, tittyMagicJSON);
    }

    IEnumerator RestorePlugins(JSONClass bootyMagicJSON, JSONClass tittyMagicJSON)
    {
        var bootyMagicStorable = this.FirstOwnAtomStorableByRegex(NaturalisBMRegex);
        var tittyMagicStorable = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
        while(!bootyMagicStorable)
        {
            yield return new WaitForSeconds(0.1f);
            bootyMagicStorable = this.FirstOwnAtomStorableByRegex(NaturalisBMRegex);
            tittyMagicStorable = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
        }

        while(
            bootyMagicStorable.GetBoolParamValue(Constant.INTERNAL_IS_CALIBRATING) ||
            tittyMagicStorable.GetBoolParamValue(Constant.INTERNAL_IS_CALIBRATING)
        )
        {
            yield return null;
        }

        yield return new WaitForSeconds(0.05f);
        if(bootyMagicJSON != null)
        {
            bootyMagicStorable.RestoreFromJSON(bootyMagicJSON);
        }

        if(tittyMagicJSON != null)
        {
            tittyMagicStorable.RestoreFromJSON(tittyMagicJSON);
        }
    }
}

// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global

static class EntryPointUtils
{
    public static Regex NaturalisBMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.BOOTY_MAGIC}", RegexOptions.Compiled);
    public static Regex NaturalisTMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.TITTY_MAGIC}", RegexOptions.Compiled);

    public static float GetFloatValue(JSONClass jc, string key, float defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        float value = jc[key].AsFloat;
        jc.Remove(key);
        return value;
    }

    public static bool GetBoolValue(JSONClass jc, string key, bool defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        bool value = jc[key].AsBool;
        jc.Remove(key);
        return value;
    }

    public static string GetPluginId(string pluginStorableName) =>
        pluginStorableName.Substring(0, pluginStorableName.IndexOf('_'));

    public static string GetPackageId(MVRPluginManager manager, string pluginUid)
    {
        string filename = manager.GetJSON()["plugins"][pluginUid].Value;
        int idx = filename.IndexOf(":/", StringComparison.Ordinal);
        return idx >= 0 ? filename.Substring(0, idx) : null;
    }

    public static void AddNaturalis(MVRPluginManager manager, string packagePrefix)
    {
        var plugin = manager.CreatePlugin();
        plugin.pluginURLJSON.val = $"{packagePrefix}Custom/Scripts/{Constant.EVERLASTER}/{Constant.NATURALIS}/{Constant.NATURALIS}.cslist";
    }
}

static class JSONStorableExtensions
{
    public static JSONStorable FirstOwnAtomStorableByRegex(this JSONStorable self, Regex regex)
    {
        var storableIds = self.containingAtom.GetStorableIDs();
        foreach(string id in storableIds)
        {
            if(regex.IsMatch(id))
            {
                var storable = self.containingAtom.GetStorableByID(id);
                if(storable && storable.containingAtom && storable != self)
                {
                    return storable;
                }
            }
        }

        return null;
    }
}

[SuppressMessage("ReSharper", "InconsistentNaming")]
static class Constant
{
    public const string EVERLASTER = "everlaster";
    public const string NATURALIS = "Naturalis";
    public const string BOOTY_MAGIC = "BootyMagic";
    public const string TITTY_MAGIC = "TittyMagic";
    [SuppressMessage("ReSharper", "InconsistentNaming")]
    const string INTERNAL = "internal_";
    public readonly static string INTERNAL_IS_CALIBRATING = $"{INTERNAL}isCalibrating";
    public readonly static string INTERNAL_IS_MIGRATION_SCRIPT = $"{INTERNAL}isMigrationScript";
}
