using SimpleJSON;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.Text.RegularExpressions;
using System;
using UnityEngine;
using static EntryPointUtils;

sealed class AutoUpgradeFromTittyMagic5 : MVRScript
{
    readonly Regex _tittyMagicRegex = new Regex($@"^plugin#\d+_{Constant.TITTY_MAGIC}.Script", RegexOptions.Compiled);

    string _packagePrefix;

    public override void Init() => StartCoroutine(InitCo());

    IEnumerator InitCo()
    {
        yield return new WaitForEndOfFrame();
        while(SuperController.singleton.isLoading)
        {
            yield return null;
        }

        string packageId = GetPackageId(manager, GetPluginId(name));
        _packagePrefix = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";

        var tittyMagic = this.FirstOwnAtomStorableByRegex(_tittyMagicRegex);
        if(tittyMagic)
        {
            yield return MigrateFromTittyMagic(tittyMagic);
            SuperController.LogMessage(
                "Successfully upgraded from TittyMagic 5 to Naturalis." +
                " Any triggers in the scene targeting TittyMagic 5 parameters will need to be cleaned up or fixed manually."
            );
            yield return new WaitForSeconds(1.5f);
            manager.RemovePluginWithUID(GetPluginId(name));
        }
        else
        {
            pluginLabelJSON.val = $"{Constant.TITTY_MAGIC} not found on this atom.";
        }
    }

    IEnumerator MigrateFromTittyMagic(JSONStorable tittyMagic)
    {
        string versionStr = tittyMagic.IsStringJSONParam("version") ? tittyMagic.GetStringParamValue("version") : "";
        if(string.IsNullOrEmpty(versionStr))
        {
            pluginLabelJSON.val = $"{Constant.TITTY_MAGIC} version not identified / too old";
            yield break;
        }

        if(versionStr.StartsWith("v"))
        {
            versionStr = versionStr.Substring(1);
        }

        if(!versionStr.StartsWith("5."))
        {
            pluginLabelJSON.val = $"Only upgrade from {Constant.TITTY_MAGIC} 5 is supported";
            yield break;
        }

        if(versionStr.Split('-').Length > 1)
        {
            pluginLabelJSON.val = $"Upgrading pre-release v{versionStr} is unsupported";
            yield break;
        }

        var tittyMagicJSON = tittyMagic.GetJSON(true, true, true);
        manager.RemovePluginWithUID(GetPluginId(tittyMagic.name));
        pluginLabelJSON.val = "Adding Naturalis...";
        AddNaturalis(manager, _packagePrefix);

        yield return new WaitForSeconds(1);
        var naturalisTMNew = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
        while(!naturalisTMNew || naturalisTMNew.GetBoolParamValue(Constant.INTERNAL_IS_CALIBRATING))
        {
            yield return null;
        }

        yield return new WaitForSeconds(0.1f);
        pluginLabelJSON.val = "Restoring plugin params...";
        var jc = ConvertToNaturalisTittyMagicJSON(tittyMagicJSON);
        naturalisTMNew.RestoreFromJSON(jc);
    }

    /* Converts TittyMagic 5 JSON to Naturalis TittyMagic JSON.
     * - LDP is disabled (even if you have it enabled by default in Naturalis User Preferences).
     * - Auto-Adjust Gravity Counterforce is enabled (default in Naturalis).
     * - Mass Offset is not applied, set to 0 instead since the mass calculation is different in Naturalis.
     * - All physics parameter offsets are also not applied, except for Angle Targets.
     *   The default physics values differ too much with Naturalis that the result could be quite different with the TittyMagic 5 offset values.
     *   With soft physics parameters, the gradient also complicates things.
     * - Quickness is converted from the 0 to 100 range to the -1 to 1 range, with 70 in TittyMagic corresponding to 0 in Naturalis.
     * - The overall gravity physics multipliers are converted to Naturalis' parameter specific multipliers.
     * - Everything else should be migrated as is - softness, soft physics on/off, self-collision, adaptive friction, nipple erection,
     *   gravity offset morphing, and hard collider properties.
     */
    static JSONClass ConvertToNaturalisTittyMagicJSON(JSONClass jc)
    {
        float baseGravity = GetFloatValue(jc, "gravityPhysicsBase", 1f);
        float forwardGravity = GetFloatValue(jc, "gravityPhysicsForward", 1f) * baseGravity;
        float backGravity = GetFloatValue(jc, "gravityPhysicsBack", 1f) * baseGravity;
        float leftRightGravity = GetFloatValue(jc, "gravityPhysicsLeftRight", 1f) * baseGravity;

        float legacyQuickness = GetFloatValue(jc, "breastQuickness", 70f);
        float quickness = legacyQuickness <= 70f
            ? Mathf.Lerp(-1f, 0f, legacyQuickness / 70f)
            : Mathf.Lerp(0f, 1f, (legacyQuickness - 70f) / 30f);

        var result = new JSONClass
        {
            ["id"] = "plugin#0_Naturalis.TittyMagic",
            ["pluginLabel"] = jc["pluginLabel"].Value,
            ["enabled"] = { AsBool = true },
            ["Breast Soft Physics Enabled"] = { AsBool = GetBoolValue(jc, "on", false) },
            ["Local Distortion Physics Enabled"] = { AsBool = false },
            ["Breast Soft Physics Self Collide"] = { AsBool = GetBoolValue(jc, "allowSelfCollision", false) },
            ["Use Adaptive Friction"] = { AsBool = GetBoolValue(jc, "enableAdaptiveFriction", false) },
            ["Up/Down Angle Target Offset"] = { AsFloat = GetFloatValue(jc, "targetRotationXOffset", 0f) },
            ["Right/Left Angle Target Offset"] = { AsFloat = GetFloatValue(jc, "targetRotationYOffset", 0f) },
            ["Twist Angle Target"] = { AsFloat = GetFloatValue(jc, "targetRotationZ", 1) },
            ["Damper Back Gravity"] = { AsFloat = backGravity },
            ["Damper Forward Gravity"] = { AsFloat = forwardGravity },
            ["Damper Sideways In Gravity"] = { AsFloat = leftRightGravity },
            ["In/Out Spring Back Gravity"] = { AsFloat = backGravity },
            ["In/Out Spring Forward Gravity"] = { AsFloat = forwardGravity },
            ["Force Physics Base Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsBase", 1) },
            ["Force Physics Depth Out Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsForward", 1) },
            ["Force Physics Depth In Multiplier"] = { AsFloat = GetFloatValue(jc, "forcePhysicsBack", 1) },
            ["Friction Offset"] = { AsFloat = GetFloatValue(jc, "frictionOffset", 0f) },
            ["Dry Skin Friction"] = { AsFloat = 1f },
            ["Nipple Erection"] = { AsFloat = GetFloatValue(jc, "nippleErection", 0f) },
            ["Breast Softness"] = { AsFloat = GetFloatValue(jc, "breastSoftness", 70f) },
            ["Breast Quickness Offset"] = { AsFloat = quickness },
        };

        return result;
    }
}

// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global

static class EntryPointUtils
{
    public static Regex NaturalisBMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.BOOTY_MAGIC}", RegexOptions.Compiled);
    public static Regex NaturalisTMRegex { get; } = new Regex($@"^plugin#\d+_{Constant.NATURALIS}.{Constant.TITTY_MAGIC}", RegexOptions.Compiled);

    public static float GetFloatValue(JSONClass jc, string key, float defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        float value = jc[key].AsFloat;
        jc.Remove(key);
        return value;
    }

    public static bool GetBoolValue(JSONClass jc, string key, bool defaultValue)
    {
        if(!jc.HasKey(key))
        {
            return defaultValue;
        }

        bool value = jc[key].AsBool;
        jc.Remove(key);
        return value;
    }

    public static string GetPluginId(string pluginStorableName) =>
        pluginStorableName.Substring(0, pluginStorableName.IndexOf('_'));

    public static string GetPackageId(MVRPluginManager manager, string pluginUid)
    {
        string filename = manager.GetJSON()["plugins"][pluginUid].Value;
        int idx = filename.IndexOf(":/", StringComparison.Ordinal);
        return idx >= 0 ? filename.Substring(0, idx) : null;
    }

    public static void AddNaturalis(MVRPluginManager manager, string packagePrefix)
    {
        var plugin = manager.CreatePlugin();
        plugin.pluginURLJSON.val = $"{packagePrefix}Custom/Scripts/{Constant.EVERLASTER}/{Constant.NATURALIS}/{Constant.NATURALIS}.cslist";
    }
}

static class JSONStorableExtensions
{
    public static JSONStorable FirstOwnAtomStorableByRegex(this JSONStorable self, Regex regex)
    {
        var storableIds = self.containingAtom.GetStorableIDs();
        foreach(string id in storableIds)
        {
            if(regex.IsMatch(id))
            {
                var storable = self.containingAtom.GetStorableByID(id);
                if(storable && storable.containingAtom && storable != self)
                {
                    return storable;
                }
            }
        }

        return null;
    }
}

[SuppressMessage("ReSharper", "InconsistentNaming")]
static class Constant
{
    public const string EVERLASTER = "everlaster";
    public const string NATURALIS = "Naturalis";
    public const string BOOTY_MAGIC = "BootyMagic";
    public const string TITTY_MAGIC = "TittyMagic";
    [SuppressMessage("ReSharper", "InconsistentNaming")]
    const string INTERNAL = "internal_";
    public readonly static string INTERNAL_IS_CALIBRATING = $"{INTERNAL}isCalibrating";
    public readonly static string INTERNAL_IS_MIGRATION_SCRIPT = $"{INTERNAL}isMigrationScript";
}
