//
using NAT.Common;
using NAT.Common.Components;
using NAT.Common.Storables;
using NAT.Common.UI;
using NAT.Store;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using UnityEngine;

namespace NAT
{
    sealed class Bootstrap : MonoBehaviour, IModParent
    {
        internal const string NATURALIS_VERSION = "1.2.3";
        internal SemVer NaturalisVersion { get; private set; }

        internal bool IsDevelopmentEnv { get; private set; }

        internal ScriptBase BootyMagic { get; private set; }
        internal ScriptBase TittyMagic { get; private set; }
        internal string PackageId { get; private set; }
        internal string PackageFullPath { get; private set; }
        internal Person Person { get; private set; }
        internal PersonMonitor PersonMonitor { get; private set; }
        internal ClothingStore ClothingStore { get; private set; }
        internal HairStore HairStore { get; private set; }
        internal Camera Camera { get; private set; }

        const string PREFS_FILE = "preferences.json";

        internal bool? Done { get; private set; }
        internal bool IsGrabbingSubscene { get; private set; }
        public bool IsEnabledAndActive() => (TittyMagic != null && TittyMagic.IsEnabledAndActive()) || (BootyMagic != null && BootyMagic.IsEnabledAndActive());

        static bool IsMigrationScript(List<JSONStorable> storables)
        {
            foreach(var storable in storables)
            {
                if(storable.GetBoolParamValueNullSafe(Constant.INTERNAL_IS_MIGRATION_SCRIPT))
                {
                    return true;
                }
            }

            return false;
        }

        public Transform GetParentUI() => Person.AtomUIContent;

        internal Bootstrap Init(ScriptBase bootyMagic)
        {
            BootyMagic = bootyMagic;
            NaturalisVersion = new SemVer(NATURALIS_VERSION, true);
            #if ENV_DEVELOPMENT
            {
                IsDevelopmentEnv = true;
            }
            #endif

            StartCoroutine(SetupRoutine());
            return this;
        }

        internal JSONStorableString SavedPrefsInfoJss { get; private set; }
        internal JSONStorableBool PauseFemaleMorphUpdatesJsb { get; private set; }
        internal JSONStorableBool PauseMaleMorphUpdatesJsb { get; private set; }
        internal StorableBool FreezeMotionSoundJsb { get; private set; }
        internal StorableBool PauseSceneAnimationJsb { get; private set; }
        internal JSONStorableBool GlutesLdpDefaultEnabledJsb { get; private set; }
        internal JSONStorableBool BreastsLdpDefaultEnabledJsb { get; private set; }
        internal JSONStorableStringChooser LdpUpdateRateJssc { get; private set; }
        internal JSONStorableBool CalibrateOnUIClosedJsb { get; private set; }
        internal JSONStorableBool CalibrateOnPresetLoadedJsb { get; private set; }
        internal JSONStorableBool CheckForSizeChangesJsb { get; private set; }
        internal JSONStorableBool ForceEnableAtomSoftPhysicsJsb { get; private set; }
        internal JSONStorableStringChooser PhysicsQualityJssc { get; private set; }

        internal int SolverIterationsMultiplier { get; private set; }
        internal int ColliderSolverIterationsMultiplier { get; private set; }

        internal SharedUIModManager UIModManager { get; private set; }

        internal bool ForceUpdatesPaused() => UIModManager.IsMorphsTabVisibleFor1Sec() && PauseFemaleMorphUpdatesJsb.val;

        IEnumerator SetupRoutine()
        {
            yield return new WaitForEndOfFrame();
            while(SuperController.singleton.isLoading)
            {
                yield return null;
            }

            TittyMagic = (ScriptBase) BootyMagic.GetOwnScript("TittyMagic");

            try
            {
                var atom = BootyMagic.containingAtom;
                if(atom.type != "Person")
                {
                    Loggr.Error($"Add to a Person atom, not {atom.type}", false);
                    Done = false;
                    yield break;
                }

                var naturalisStorables = atom
                    .FindStorablesByRegexMatch(Utils.NewRegex($@"^plugin#\d+_{Constant.NATURALIS}"))
                    .Where(storable => storable && storable != BootyMagic && storable != TittyMagic)
                    .ToList();
                if(naturalisStorables.Count > 0 && !IsMigrationScript(naturalisStorables))
                {
                    Loggr.Error($"{Constant.NATURALIS} is already added", false);
                    Done = false;
                    yield break;
                }

                if(atom.StorableExistsByRegexMatch(Utils.NewRegex($@"^plugin#\d+_{Constant.TITTY_MAGIC}")))
                {
                    Loggr.Error($"Person already has an instance of {Constant.TITTY_MAGIC} - remove it and then reload {Constant.NATURALIS}", false);
                    Done = false;
                    yield break;
                }

                if(atom.StorableExistsByRegexMatch(Utils.NewRegex(@"^plugin#\d+_Stopper.FemaleChestPhysicsForMale")))
                {
                    Loggr.Message($"Remove Stopper.FemaleChestPhysicsForMale before adding {Constant.NATURALIS}");
                    Done = false;
                    yield break;
                }
            }
            catch(Exception e)
            {
                Done = false;
                Loggr.Error($"Init error: {e}");
            }

            yield return SetupPersonCo(BootyMagic.containingAtom);
            if(Done == false)
            {
                yield break;
            }

            try
            {
                ClothingStore = new ClothingStore(Person);
                HairStore = new HairStore(Person.Geometry);
                SetupStorables();
                ReadUserPreferencesFromFile();
                SetSolverIterationsMultiplier();
                PersonMonitor = gameObject.AddComponent<PersonMonitor>().Init(this);
                UIModManager = new SharedUIModManager(this);
                SetupSubsceneFreezeGrabCallbacks();
                Camera = Utils.GetCamera();
                Done = true;
            }
            catch(Exception e)
            {
                Loggr.Error($"Init error: {e}");
                Done = false;
            }
        }

        IEnumerator SetupPersonCo(Atom atom)
        {
            PackageId = BootyMagic.GetPackageId();
            Person = new Person(this, atom, PackageId);
            const float limit = 15f;
            yield return Person.WaitForGeometryAndSkinReady(limit);
            if(!Person.CharacterReady)
            {
                Loggr.Error($"Selected character {Person.Geometry.selectedCharacter.name} was not ready after {limit.ToString()} seconds of waiting");
                Done = false;
                yield break;
            }

            if(!Person.SkinReady)
            {
                Loggr.Error($"Person skin materials not found after {limit.ToString()} seconds of waiting");
                Done = false;
                yield break;
            }

            Person.Setup();
            Person.onScaleChangedHandlers += OnAtomScaleChanged;
            yield return Person.CycleSoftPhysicsEnabled();
            if(Person.SetupDone)
            {
                PackageFullPath = GetPackageFullPathFromBreastMorphLoadPath("Clothing/Reduce Size L");
            }
        }

        void OnAtomScaleChanged(float _) => StartCoroutine(Person.CycleBodyScaleMorphCo());

        /* Hack. Determine full package path from morph load path which includes the full path info, unlike the MVRScript itself. */
        string GetPackageFullPathFromBreastMorphLoadPath(string morphPath)
        {
            var morph = Person.GetBreastMorph(morphPath);
            int packageSeparatorIdx = morph.deltasLoadPath.IndexOf(":\\", StringComparison.Ordinal);
            return packageSeparatorIdx == -1
                ? null
                : morph.deltasLoadPath.Substring(0, packageSeparatorIdx);
        }

        void SetupStorables()
        {
            SavedPrefsInfoJss = new JSONStorableString("savedPrefsInfo", "");
            PauseFemaleMorphUpdatesJsb = new JSONStorableBool("Pause Morph Updates", true)
            {
                setCallbackFunction = value => PauseMaleMorphUpdatesJsb.valNoCallback = value,
            };
            PauseMaleMorphUpdatesJsb = new JSONStorableBool("Pause Morph Updates", true)
            {
                setCallbackFunction = value => PauseFemaleMorphUpdatesJsb.valNoCallback = value,
            };
            SetupPluginInstanceStorables();
            SetupUserPreferencesStorables();
        }

        void SetupPluginInstanceStorables()
        {
            FreezeMotionSoundJsb = new StorableBool("Calibration: Freeze Motion/Sound", true)
            {
                setCallbackFunction = value =>
                {
                    if(value)
                    {
                        PauseSceneAnimationJsb.val = false;
                    }
                },
            };
            PauseSceneAnimationJsb = new StorableBool("Calibration: Pause Scene Animation", false)
            {
                setCallbackFunction = value =>
                {
                    if(value && FreezeMotionSoundJsb.val)
                    {
                        PauseSceneAnimationJsb.valNoCallback = false;
                    }
                },
            };

            BootyMagic.PreRegisterGeneralParam(FreezeMotionSoundJsb);
            BootyMagic.PreRegisterGeneralParam(PauseSceneAnimationJsb);
        }

        void SetupUserPreferencesStorables()
        {
            CalibrateOnUIClosedJsb = new JSONStorableBool("Calibrate On Plugin UI Closed", false, (bool _) => SyncPreferencesFile());
            CalibrateOnPresetLoadedJsb = new JSONStorableBool("Calibrate On Preset Loaded", true, (bool _) => SyncPreferencesFile());
            GlutesLdpDefaultEnabledJsb = new JSONStorableBool("Glutes LDP Default Enabled", false, (bool _) => SyncPreferencesFile());
            BreastsLdpDefaultEnabledJsb = new JSONStorableBool("Breasts LDP Default Enabled", false, (bool _) => SyncPreferencesFile());

            var updateRates = new List<string> { "20 Hz", "30 Hz", "45 Hz", "60 Hz", "Use Physics Rate" };
            LdpUpdateRateJssc = new JSONStorableStringChooser(
                "Local Distortion Physics Update Rate",
                updateRates,
                "30 Hz",
                "LDP Update Rate"
            );
            LdpUpdateRateJssc.setCallbackFunction += _ => SyncPreferencesFile();

            CheckForSizeChangesJsb = new JSONStorableBool("Monitor: Check For Size Changes", true, value =>
            {
                if(value)
                {
                    BootyMagic.DeferProcessMorphsForListener();
                    TittyMagic.DeferProcessMorphsForListener();
                }
                else
                {
                    BootyMagic.calibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED, false);
                    TittyMagic.calibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED, false);
                }

                SyncPreferencesFile();
            });

            ForceEnableAtomSoftPhysicsJsb = new JSONStorableBool("Force-Enable Atom Soft Physics", false)
            {
                setCallbackFunction = value =>
                {
                    if(value && IsEnabledAndActive())
                    {
                        Person.SoftBodyPhysicsEnabler.SetBoolParamValue("enabled", true);
                    }

                    SyncPreferencesFile();
                },
            };

            var qualityOptions = new List<string> { "1", "2", "3", "4", "5" };
            PhysicsQualityJssc = new JSONStorableStringChooser(
                "Physics Quality",
                qualityOptions,
                qualityOptions[0],
                "Physics Quality"
            );
            PhysicsQualityJssc.setCallbackFunction += _ =>
            {
                SetSolverIterationsMultiplier();
                BootyMagic.OnHighQualityPhysicsChanged();
                TittyMagic.OnHighQualityPhysicsChanged();
                SyncPreferencesFile();
            };
        }

        void ReadUserPreferencesFromFile()
        {
            var prefsJSON = FileUtils.ReadDataJSON(PREFS_FILE);
            if(prefsJSON != null)
            {
                SetValueFromPrefsJSON(prefsJSON, CalibrateOnUIClosedJsb);
                SetValueFromPrefsJSON(prefsJSON, CalibrateOnPresetLoadedJsb);
                SetValueFromPrefsJSON(prefsJSON, LdpUpdateRateJssc);
                SetValueFromPrefsJSON(prefsJSON, CheckForSizeChangesJsb);
                SetValueFromPrefsJSON(prefsJSON, ForceEnableAtomSoftPhysicsJsb);

                if(prefsJSON.HasKey("Local Distortion Physics Default On")) // legacy setting
                {
                    bool ldpDefaultOn = prefsJSON["Local Distortion Physics Default On"].AsBool;
                    GlutesLdpDefaultEnabledJsb.val = GlutesLdpDefaultEnabledJsb.defaultVal = ldpDefaultOn;
                    BreastsLdpDefaultEnabledJsb.val = BreastsLdpDefaultEnabledJsb.defaultVal = ldpDefaultOn;
                }
                else
                {
                    SetValueFromPrefsJSON(prefsJSON, GlutesLdpDefaultEnabledJsb);
                    SetValueFromPrefsJSON(prefsJSON, BreastsLdpDefaultEnabledJsb);
                }

                SetValueFromPrefsJSON(prefsJSON, PhysicsQualityJssc);
            }
            else
            {
                SyncPreferencesFile();
            }
        }

        static void SetValueFromPrefsJSON(JSONClass prefsJSON, JSONStorableParam storable)
        {
            if(!prefsJSON.HasKey(storable.name))
            {
                return;
            }

            var jsb = storable as JSONStorableBool;
            if(jsb != null)
            {
                jsb.defaultVal = prefsJSON[jsb.name].AsBool;
                jsb.valNoCallback = jsb.defaultVal;
                return;
            }

            var jssc = storable as JSONStorableStringChooser;
            if(jssc != null)
            {
                jssc.defaultVal = prefsJSON[jssc.name];
                jssc.valNoCallback = jssc.defaultVal;
            }
        }

        void SetSolverIterationsMultiplier()
        {
            int multiplier;
            int.TryParse(PhysicsQualityJssc.val, out multiplier);
            SolverIterationsMultiplier = Mathf.Max(multiplier, 2);
            ColliderSolverIterationsMultiplier = multiplier;
        }

        /* If subscene is being grabbed and Freeze Physics While Grabbing is unchecked,
         * custom forces applied by the plugin will accumulate and spike when grab is ended.
         * Adding forces only if IsGrabbingSubscene is false prevents this. Might be hacky
         */
        void SetupSubsceneFreezeGrabCallbacks()
        {
            SuperController.singleton.GetAtoms().ForEach(OnAddAtom);
            SuperController.singleton.onAtomAddedHandlers += OnAddAtom;
            SuperController.singleton.onAtomRemovedHandlers += OnRemoveAtom;
        }

        void OnAddAtom(Atom atom)
        {
            if(!atom.isSubSceneType)
            {
                return;
            }

            atom.mainController.onGrabStartHandlers += OnStartGrabSubscene;
            atom.mainController.onGrabEndHandlers += OnEndGrabSubscene;
        }

        void OnRemoveAtom(Atom atom)
        {
            if(!atom || !atom.isSubSceneType)
            {
                return;
            }

            atom.mainController.onGrabStartHandlers += OnStartGrabSubscene;
            atom.mainController.onGrabEndHandlers += OnEndGrabSubscene;
        }

        void OnStartGrabSubscene(FreeControllerV3 controller)
        {
            if(controller.GetBoolParamValue("freezeAtomPhysicsWhenGrabbed"))
            {
                IsGrabbingSubscene = true;
            }
        }

        void OnEndGrabSubscene(FreeControllerV3 controller) => IsGrabbingSubscene = false;

        internal void CreatePauseToggle(UIMod uiMod, JSONStorableBool jsb)
        {
            var rightSide = uiMod.Target.Find("LeftSide");
            var customTransform = BootyMagic.InstantiateToggle(rightSide);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = new Vector2(0, 249);
            rectTransform.sizeDelta = new Vector2(-10, 50);
            uiMod.AddCustomObject(customTransform.gameObject);
            var uiDynamic = customTransform.GetComponent<UIDynamicToggle>();
            jsb.toggle = uiDynamic.toggle;
            uiDynamic.label = $"{Constant.NATURALIS}: Pause morph updates";
        }

        void SyncPreferencesFile()
        {
            var jc = FileUtils.ReadDataJSON(PREFS_FILE) ?? new JSONClass();
            jc.Remove("Local Distortion Physics Default On"); // legacy setting
            jc[CalibrateOnUIClosedJsb.name].AsBool = CalibrateOnUIClosedJsb.val;
            jc[CalibrateOnPresetLoadedJsb.name].AsBool = CalibrateOnPresetLoadedJsb.val;
            jc[GlutesLdpDefaultEnabledJsb.name].AsBool = GlutesLdpDefaultEnabledJsb.val;
            jc[BreastsLdpDefaultEnabledJsb.name].AsBool = BreastsLdpDefaultEnabledJsb.val;
            jc[LdpUpdateRateJssc.name] = LdpUpdateRateJssc.val;
            jc[CheckForSizeChangesJsb.name].AsBool = CheckForSizeChangesJsb.val;
            jc[ForceEnableAtomSoftPhysicsJsb.name].AsBool = ForceEnableAtomSoftPhysicsJsb.val;
            jc[PhysicsQualityJssc.name] = PhysicsQualityJssc.val;
            FileUtils.WriteDataJSON(
                PREFS_FILE,
                jc,
                () => StartCoroutine(ShowSavedInfoTextCo())
            );
        }

        bool _inShowSavedInfoTextCo;
        bool _doClearSavedInfoText = true;

        IEnumerator ShowSavedInfoTextCo()
        {
            if(Done != true)
            {
                yield break;
            }

            if(_inShowSavedInfoTextCo)
            {
                _doClearSavedInfoText = false;
            }

            _inShowSavedInfoTextCo = true;
            SavedPrefsInfoJss.val = "Changes saved";
            yield return new WaitForSeconds(1.0f);
            if(_doClearSavedInfoText)
            {
                SavedPrefsInfoJss.val = "";
            }

            _doClearSavedInfoText = true;
            _inShowSavedInfoTextCo = false;
        }

        void Update()
        {
            if(Done != true)
            {
                return;
            }

            try
            {
                ClothingStore.Execute();
                HairStore.Execute();
            }
            catch(Exception e)
            {
                Loggr.Error($"Update: {e}");
            }
        }

        void OnDestroy()
        {
            try
            {
                SuperController.singleton.GetAtoms().ForEach(OnRemoveAtom);
                SuperController.singleton.onAtomAddedHandlers -= OnAddAtom;
                SuperController.singleton.onAtomRemovedHandlers -= OnRemoveAtom;
                Destroy(PersonMonitor);
                PersonMonitor = null;

                Person?.Destroy();
                ClothingStore?.Destroy();
                HairStore?.Destroy();
                UIModManager?.Destroy();
            }
            catch(Exception e)
            {
                if(Done == true)
                {
                    SuperController.LogError($"OnDestroy: {e}");
                }
                else
                {
                    Debug.LogError($"OnDestroy: {e}");
                }
            }
        }
    }
}
