using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using NAT.Common;
using NAT.Common.Components;
using NAT.Common.Models;
using System;
using UnityEngine;

namespace NAT
{
    sealed class Person
    {
        readonly Bootstrap _bootstrap;
        internal readonly Atom atom;
        readonly string _packageId;
        internal readonly CalibrationState calibrationState;

        internal DAZCharacterSelector Geometry { get; private set; }
        internal DAZSkinV2 Skin { get; private set; }

        DAZCharacterRun _characterRun;
        GenerateDAZMorphsControlUI _morphsControlUI;
        internal JSONStorable SoftBodyPhysicsEnabler { get; private set; }
        internal JSONStorableBool CollisionEnabledJsb { get; private set; }
        internal JSONStorableFloat ScaleJsf { get; private set; }
        internal float oneOverScale;
        internal Action<float> onScaleChangedHandlers;

        internal AdjustJoints GluteControl { get; private set; }
        internal AdjustJoints BreastControl { get; private set; }
        internal DAZPhysicsMesh GlutePhysicsMesh { get; private set; }
        internal DAZPhysicsMesh BreastPhysicsMesh { get; private set; }

        internal readonly List<AutoColliderGroup> autoColliderGroups = new List<AutoColliderGroup>();

        internal Rigidbody HipRb { get; private set; }
        internal Rigidbody ChestRb { get; private set; }

        internal BodyPart LeftBreast { get; private set; }
        internal BodyPart RightBreast { get; private set; }
        internal BodyPart LeftGlute { get; private set; }
        internal BodyPart RightGlute { get; private set; }

        string _morphsPath;
        string _packagePrefix;
        PresetManagerControl _posePmc;
        PresetManager _posePm;
        PresetManager _morphPm;

        internal Transform AtomUIContent { get; private set; }

        internal bool CharacterReady { get; private set; }
        internal bool SkinReady { get; private set; }
        internal bool IsFemale { get; private set; }
        internal bool IsFreezeGrabbing { get; private set; }
        internal bool SetupDone { get; private set; }
        bool _futaSetupDone;

        internal Person(Bootstrap bootstrap, Atom atom, string packageId)
        {
            _bootstrap = bootstrap;
            this.atom = atom;
            _packageId = packageId;
            calibrationState = new CalibrationState(this);
        }

        internal JSONStorable GetStorable(string id) => atom.GetStorableByID(id);

        internal bool PectoralRbCollisionsEnabled(bool enabled) => enabled
            ? LeftBreast.rigidbody.detectCollisions || RightBreast.rigidbody.detectCollisions
            : !LeftBreast.rigidbody.detectCollisions || !RightBreast.rigidbody.detectCollisions;

        internal bool HasDummySkin() => Geometry.selectedCharacter.name == "femaledummy";

        internal IEnumerator WaitForGeometryAndSkinReady(float limit)
        {
            float timeout = Time.unscaledTime + limit;
            while(!(CharacterReady && SkinReady) && Time.unscaledTime < timeout)
            {
                Geometry = (DAZCharacterSelector) atom.GetStorableByID("geometry");
                CharacterReady = Geometry.selectedCharacter.ready;
                SkinReady = HasDummySkin() || atom.GetStorableByID("skin");
                yield return new WaitForSecondsRealtime(0.5f);
            }
        }

        internal IEnumerator RefreshSkinCo()
        {
            CharacterReady = false;
            SkinReady = false;
            const float limit = 15f;
            yield return WaitForGeometryAndSkinReady(limit);

            if(!CharacterReady)
            {
                Loggr.Error($"Selected character {Geometry.selectedCharacter.name} was not ready after {limit.ToString()} seconds of waiting. Try reloading");
                yield break;
            }

            if(!SkinReady)
            {
                Loggr.Error($"Person skin materials not found after {limit.ToString()} seconds of waiting. Try reloading");
                yield break;
            }

            Skin = atom.GetComponentInChildren<DAZCharacter>().skin;
        }

        internal bool IsSelectedCharacterReady() => Geometry.selectedCharacter.ready;

        internal void Setup()
        {
            try
            {
                IsFemale = !Geometry.selectedCharacter.isMale;
                Skin = atom.GetComponentInChildren<DAZCharacter>().skin;
                _characterRun = atom.GetComponentInChildren<DAZCharacterRun>();
                _morphsControlUI = IsFemale ? Geometry.morphsControlUI : Geometry.morphsControlUIOtherGender;
                SoftBodyPhysicsEnabler = atom.GetStorableByID("SoftBodyPhysicsEnabler");
                CollisionEnabledJsb = atom.GetStorableByID("AtomControl").GetBoolJSONParam("collisionEnabled");
                ScaleJsf = atom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");
                ScaleJsf.setCallbackFunction += OnScaleChanged;
                OnScaleChanged(ScaleJsf.val);

                if(IsFemale)
                {
                    GluteControl = (AdjustJoints) atom.GetStorableByID("GluteControl");
                    GlutePhysicsMesh = (DAZPhysicsMesh) atom.GetStorableByID("LowerPhysicsMesh");
                    HipRb = Utils.FindRigidbody(atom, "hip");
                    LeftGlute = new BodyPart(GluteControl.joint1);
                    RightGlute = new BodyPart(GluteControl.joint2);
                }

                BreastControl = (AdjustJoints) atom.GetStorableByID(IsFemale ? "BreastControl" : "PectoralControl");
                BreastPhysicsMesh = (DAZPhysicsMesh) atom.GetStorableByID("BreastPhysicsMesh");

                ChestRb = Utils.FindRigidbody(atom, "chest");
                LeftBreast = new BodyPart(BreastControl.joint2, atom.rigidbodies.FirstOrDefault(rb => rb.name == "lNipple"));
                RightBreast = new BodyPart(BreastControl.joint1, atom.rigidbodies.FirstOrDefault(rb => rb.name == "rNipple"));

                autoColliderGroups.AddRange(atom.autoColliderBatchUpdaters.First().GetComponentsInChildren<AutoColliderGroup>(true));
                SaveOriginalUseColliders();

                string packagePath = string.IsNullOrEmpty(_packageId) ? "" : $"{_packageId}:/";
                _morphsPath = $"{packagePath}Custom/Atom/Person/Morphs/female/{Constant.EVERLASTER}";
                _packagePrefix = string.IsNullOrEmpty(_packageId) ? "" : $"{_packageId}:";

                _posePmc = atom.presetManagerControls.Find(pmc => pmc.name == "PosePresets");
                _posePm = atom.GetStorableByID("PosePresets").GetComponentInChildren<PresetManager>();
                _posePm.postLoadEvent.AddListener(OnPosePresetLoaded);

                _morphPm = atom.GetStorableByID("MorphPresets").GetComponentInChildren<PresetManager>();
                _morphPm.postLoadEvent.AddListener(ZeroAllDuplicateMorphs);

                AtomUIContent = atom.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content");

                atom.mainController.onGrabStartHandlers -= OnStartGrab;
                atom.mainController.onGrabEndHandlers -= OnEndGrab;
                atom.mainController.onGrabStartHandlers += OnStartGrab;
                atom.mainController.onGrabEndHandlers += OnEndGrab;

                SetupDone = true;
            }
            catch(Exception e)
            {
                Loggr.Error($"Error in Person setup: {e}");
            }
        }

        void OnScaleChanged(float value)
        {
            oneOverScale = 1 / value;
            onScaleChangedHandlers?.Invoke(value);
        }

        /* Ensures all skin vertex positions can be tracked */
        internal IEnumerator CycleSoftPhysicsEnabled()
        {
            bool globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;
            UserPreferences.singleton.softPhysics = true;

            bool atomSoftPhysicsOn = SoftBodyPhysicsEnabler.GetBoolParamValue("enabled");
            SoftBodyPhysicsEnabler.SetBoolParamValue("enabled", true);

            bool breastSoftPhysicsOn = BreastPhysicsMesh.on;
            BreastPhysicsMesh.on = true;

            bool gluteSoftPhysicsOn = false;
            if(IsFemale)
            {
                gluteSoftPhysicsOn = GlutePhysicsMesh.on;
                GlutePhysicsMesh.on = true;
            }

            /* Necessary */
            yield return null;
            yield return null;
            yield return null;
            yield return null;
            yield return null;

            if(IsFemale)
            {
                GlutePhysicsMesh.on = gluteSoftPhysicsOn;
            }

            BreastPhysicsMesh.on = breastSoftPhysicsOn;
            SoftBodyPhysicsEnabler.SetBoolParamValue("enabled", atomSoftPhysicsOn);
            UserPreferences.singleton.softPhysics = globalSoftPhysicsOn;
        }

        void SetupFuta()
        {
            TransplantBreastDazMesh();
            ParentBreastPhysicsMesh("MaleMorphers");
            EnableAutoColliderGroup("FemaleAutoColliderslPectoral");
            EnableAutoColliderGroup("FemaleAutoCollidersrPectoral");
            IgnoreMaleChestCollidersOnFemalePectorals();
            _futaSetupDone = true;
        }

        void TransplantBreastDazMesh()
        {
            var skin = atom.GetComponentInChildren<DAZMergedSkinV2>();
            var breastDazMesh = BreastPhysicsMesh.GetComponent<DAZPhysicsMesh>();
            IgnoreMaleChestCollidersOnPhysicsMesh(breastDazMesh);
            breastDazMesh.InitColliders();
            breastDazMesh.skin = skin;
            RegisterMesh(breastDazMesh);
            breastDazMesh.Init();
            breastDazMesh.enabled = true;
        }

        void IgnoreMaleChestCollidersOnPhysicsMesh(DAZPhysicsMesh breastDazMesh)
        {
            try
            {
                var maleChest = autoColliderGroups.First(x => x.name == "MaleAutoColliderschest");
                foreach(var vertGroup in breastDazMesh.softVerticesGroups)
                {
                    var ignoreAutoColliderGroups = vertGroup.ignoreAutoColliderGroups.ToList();
                    if(!ignoreAutoColliderGroups.Contains(maleChest))
                    {
                        ignoreAutoColliderGroups.Add(maleChest);
                    }

                    vertGroup.ignoreAutoColliderGroups = ignoreAutoColliderGroups.ToArray();
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"IgnoreMaleChestCollidersOnPhysicsMesh: {e}");
            }
        }

        void RegisterMesh(DAZPhysicsMesh mesh)
        {
            if(!_characterRun)
            {
                Loggr.Error("RegisterMesh: DAZCharacterRun not found");
                return;
            }

            var newMeshes = _characterRun.physicsMeshes.ToList();
            if(!newMeshes.Contains(mesh))
            {
                newMeshes.Add(mesh);
                _characterRun.physicsMeshes = newMeshes.ToArray();
            }
        }

        void ParentBreastPhysicsMesh(string parentName)
        {
            var parent = Geometry.transform.Find(parentName);
            if(!parent)
            {
                Loggr.Error($"Transform {parentName} not found in geometry");
                return;
            }

            BreastPhysicsMesh.transform.parent = parent;
        }

        void EnableAutoColliderGroup(string name)
        {
            try
            {
                var group = autoColliderGroups.First(x => x.name == name);
                group.enabled = true;
                group.gameObject.SetActive(true);
                foreach(var ac in group.GetAutoColliders())
                {
                    ac.enabled = true;
                    ac.gameObject.SetActive(true);
                    ac.jointRB.detectCollisions = CollisionEnabledJsb.val;
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"EnableAutoColliderGroup {name}: {e}");
            }
        }

        void IgnoreMaleChestCollidersOnFemalePectorals()
        {
            try
            {
                var maleChest = autoColliderGroups.First(x => x.name == "MaleAutoColliderschest");
                IgnoreColliderGroup("FemaleAutoColliderslPectoral", maleChest);
                IgnoreColliderGroup("FemaleAutoCollidersrPectoral", maleChest);
            }
            catch(Exception e)
            {
                Loggr.Error($"IgnoreMaleChestCollidersOnFemalePectorals: {e}");
            }
        }

        void IgnoreColliderGroup(string name, AutoColliderGroup ignoreGroup)
        {
            try
            {
                var group = autoColliderGroups.First(x => x.name == name);
                var ignoreColliderGroups = group.ignoreColliderGroups.ToList();
                if(!ignoreColliderGroups.Contains(ignoreGroup))
                {
                    ignoreColliderGroups.Add(ignoreGroup);
                }

                group.ignoreColliderGroups = ignoreColliderGroups.ToArray();
                group.InitColliders();
            }
            catch(Exception e)
            {
                Loggr.Error($"IgnoreColliderGroup {name}: {e}");
            }
        }

        bool _originalUseAdvancedColliders;
        bool _originalUseAuxBreastColliders;

        void SaveOriginalUseColliders()
        {
            _originalUseAdvancedColliders = Geometry.useAdvancedColliders;
            _originalUseAuxBreastColliders = Geometry.useAuxBreastColliders;
        }

        internal void SetupUseColliders()
        {
            /* Advanced colliders must be enabled for collider visualizer, force morphing and hard collider handler */
            Geometry.useAdvancedColliders = true;
            Geometry.useAuxBreastColliders = true;
        }

        void OnStartGrab(FreeControllerV3 controller)
        {
            if(controller.GetBoolParamValue("freezeAtomPhysicsWhenGrabbed") && !controller.GetBoolParamValue("detachControl"))
            {
                IsFreezeGrabbing = true;
            }
        }

        void OnPosePresetLoaded()
        {
            if(_posePmc.GetStringParamValue("presetName").StartsWith($"{Constant.EVERLASTER}/{Constant.NATURALIS}/"))
            {
                return;
            }

            ZeroAllDuplicateMorphs();
        }

        void OnEndGrab(FreeControllerV3 controller) => IsFreezeGrabbing = false;

        internal void SetGlutesUseGravity(bool value)
        {
            LeftGlute.rigidbody.useGravity = value;
            RightGlute.rigidbody.useGravity = value;
        }

        internal void SetBreastsUseGravity(bool value)
        {
            LeftBreast.rigidbody.useGravity = value;
            RightBreast.rigidbody.useGravity = value;
        }

        internal void SetPectoralRbCollisions(bool value)
        {
            LeftBreast.rigidbody.detectCollisions = value;
            RightBreast.rigidbody.detectCollisions = value;
        }

        internal void SetGlutesHighQualityPhysics(bool value) => SetHighQualityPhysics(HipRb, LeftGlute, RightGlute, value);
        internal void SetBreastsHighQualityPhysics(bool value) => SetHighQualityPhysics(ChestRb, LeftBreast, RightBreast, value);

        void SetHighQualityPhysics(Rigidbody chestOrHipRb, BodyPart left, BodyPart right, bool value)
        {
            int solverIterations = SuperController.singleton.solverIterations;
            int colliderSolverIterations = SuperController.singleton.solverIterations;
            if(value && !UserPreferences.singleton.physicsHighQuality)
            {
                solverIterations *= _bootstrap.SolverIterationsMultiplier;
                colliderSolverIterations *= _bootstrap.ColliderSolverIterationsMultiplier;
            }

            chestOrHipRb.solverIterations = solverIterations;
            foreach(var rb in left.colliderRigidbodies)
            {
                rb.solverIterations = colliderSolverIterations;
            }

            foreach(var rb in right.colliderRigidbodies)
            {
                rb.solverIterations = colliderSolverIterations;
            }
        }

        internal DAZMorph GetBreastMorph(string fileName) => GetOwnMorph($"{Constant.TITTY_MAGIC}/{fileName}.vmi");
        internal DAZMorph GetGluteMorph(string fileName) => GetOwnMorph($"{Constant.BOOTY_MAGIC}/{fileName}.vmi");
        DAZMorph GetOwnMorph(string subPath) => GetMorph($"{_morphsPath}/{subPath}");

        DAZMorph GetMorph(string uid)
        {
            var dazMorph = _morphsControlUI.GetMorphByUid(uid);

            if(dazMorph == null)
            {
                Loggr.Error($"Morph with uid '{uid}' not found");
            }

            return dazMorph;
        }

        /* Triggers JointPhysicsHandler's CheckBoneDirtyChanged */
        internal IEnumerator CycleBodyScaleMorphCo()
        {
            var morph = _morphsControlUI.GetMorphByUid("Body Scale");
            if(morph == null)
            {
                yield break;
            }

            morph.morphValue += 0.001f;
            yield return null;
            morph.morphValue -= 0.001f;
        }

        internal PresetParams GetPosePresetParams() => PresetParams.GetParamsFromControl(_posePmc);
        internal void SetPosePresetParams(PresetParams presetParams) => presetParams.SetParamsToControl(_posePmc);

        internal void SavePosePreset(
            string name,
            bool storePoseMorphsInPose = true,
            bool storeBonesInPose = true,
            bool storeRootNodeInPose = false
        )
        {
            new PresetParams(
                $"{Constant.EVERLASTER}/{Constant.NATURALIS}/{name}",
                storeRootNodeInPose,
                storeBonesInPose,
                storePoseMorphsInPose
            ).SetParamsToControl(_posePmc);
            _posePmc.CallAction("StorePreset");
        }

        internal void LoadPosePreset(string name, bool fromPackage)
        {
            string prefix = fromPackage ? _packagePrefix : "";
            _posePmc.SetStringParamValue("presetName", $"{prefix}{Constant.EVERLASTER}/{Constant.NATURALIS}/{name}");
            _posePmc.CallAction("LoadPreset");
        }

        internal void DeletePosePreset(string name)
        {
            _posePmc.SetStringParamValue("presetName", $"{Constant.EVERLASTER}/{Constant.NATURALIS}/{name}");
            _posePm.DeletePreset();
        }

        const string BM_MORPH_NAME_PREFIX = "BM_";
        const string TM_MORPH_NAME_PREFIX = "TM_";
        const string TM3_MORPH_NAME_PREFIX = "TM3_";
        const string TM5_MORPH_NAME_PREFIX = "TM5_";
        readonly string _naturalisPackageUidPrefix = $"{Constant.EVERLASTER}.{Constant.NATURALIS}.";
        readonly string _tittymagicPackageUidPrefix = $"{Constant.EVERLASTER}.{Constant.TITTY_MAGIC}.";

        public void ZeroAllDuplicateMorphs()
        {
            try
            {
                var morphs = Geometry.morphBank1.morphs;
                for(int i = 0; i < morphs.Count; i++)
                {
                    var morph = morphs[i];
                    if(morph.isInPackage)
                    {
                        ZeroDuplicatePackageMorph(morph);
                    }
                    else if(morph.isTransient || morph.isRuntime)
                    {
                        ZeroDuplicateLocalCustomMorph(morph);
                    }
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"{nameof(ZeroAllDuplicateMorphs)}: {e}");
            }
        }

        void ZeroDuplicatePackageMorph(DAZMorph morph)
        {
            string packageUid = morph.packageUid;
            if(packageUid == _packageId)
            {
                return;
            }

            string morphName = morph.morphName;
            if(packageUid.HasSubstring(0, _naturalisPackageUidPrefix))
            {
                if(
                    morphName.HasSubstring(0, BM_MORPH_NAME_PREFIX) ||
                    morphName.HasSubstring(0, TM_MORPH_NAME_PREFIX) ||
                    morphName.HasSubstring(0, TM5_MORPH_NAME_PREFIX)
                )
                {
                    morph.morphValue = 0;
                }
            }
            else if(packageUid.HasSubstring(0, _tittymagicPackageUidPrefix))
            {
                if(
                    morphName.HasSubstring(0, TM3_MORPH_NAME_PREFIX) ||
                    morphName.HasSubstring(0, TM5_MORPH_NAME_PREFIX)
                )
                {
                    morph.morphValue = 0;
                }
            }
        }

        static void ZeroDuplicateLocalCustomMorph(DAZMorph morph)
        {
            string morphName = morph.morphName;
            if(
                morphName.HasSubstring(0, BM_MORPH_NAME_PREFIX) ||
                morphName.HasSubstring(0, TM_MORPH_NAME_PREFIX) ||
                morphName.HasSubstring(0, TM3_MORPH_NAME_PREFIX) ||
                morphName.HasSubstring(0, TM5_MORPH_NAME_PREFIX)
            )
            {
                morph.morphValue = 0;
            }
        }

        internal Vector3 SkinVertsAveragePosition(int[] indices)
        {
            var sum = Vector3.zero;
            foreach(int t in indices)
            {
                sum += Skin.rawSkinnedVerts[t];
            }

            return sum / indices.Length;
        }

        internal void EnableCustomHipPhysics()
        {
            SetupUseColliders();
            SetGlutesUseGravity(false);
            SetGlutesHighQualityPhysics(true);
        }

        internal void EnableCustomChestPhysics()
        {
            SetupUseColliders();
            if(!IsFemale)
            {
                if(!_futaSetupDone)
                {
                    SetupFuta();
                }
                else
                {
                    ParentBreastPhysicsMesh("MaleMorphers");
                }
            }

            SetPectoralRbCollisions(false);
            SetBreastsUseGravity(false);
            SetBreastsHighQualityPhysics(true);
        }

        internal void DisableCustomHipPhysics()
        {
            SetGlutesUseGravity(true);
            SetGlutesHighQualityPhysics(false);
        }

        internal void DisableCustomChestPhysics()
        {
            if(!IsFemale)
            {
                ParentBreastPhysicsMesh("FemaleMorphers");
            }

            SetPectoralRbCollisions(atom.collisionEnabled);
            SetBreastsUseGravity(true);
            SetBreastsHighQualityPhysics(false);
        }

        internal void Destroy()
        {
            if(ScaleJsf != null)
            {
                ScaleJsf.setCallbackFunction -= OnScaleChanged;
            }

            if(_posePm)
            {
                _posePm.postLoadEvent.RemoveListener(OnPosePresetLoaded);
            }

            if(_morphPm)
            {
                _morphPm.postLoadEvent.RemoveListener(ZeroAllDuplicateMorphs);
            }

            if(IsFemale && Geometry)
            {
                Geometry.useAdvancedColliders = _originalUseAdvancedColliders;
                Geometry.useAuxBreastColliders = _originalUseAuxBreastColliders;
            }

            if(atom && atom.mainController)
            {
                atom.mainController.onGrabStartHandlers -= OnStartGrab;
                atom.mainController.onGrabEndHandlers -= OnEndGrab;
            }
        }
    }
}
