using NAT.Common;
using System.Collections;
using UnityEngine;

namespace NAT
{
    sealed class PersonMonitor : MonoBehaviour
    {
        Bootstrap _bootstrap;
        Person _person;
        ScriptBase _bootyMagic;
        ScriptBase _tittyMagic;
        JSONStorableBool _softBodyPhysicsJsb;
        JSONStorableBool _breastInOutJsb;
        DAZCharacter _selectedCharacter;

        float _fixedDeltaTime;
        bool _physicsHighQuality;

        internal bool IsBreastSoftPhysicsOn() => _globalSoftPhysicsOn && _atomSoftPhysicsOn && _breastSoftPhysicsOn;
        internal bool IsGluteSoftPhysicsOn() => _globalSoftPhysicsOn && _atomSoftPhysicsOn && _gluteSoftPhysicsOn;

        bool _globalSoftPhysicsOn;
        bool _atomSoftPhysicsOn;
        bool _breastSoftPhysicsOn;
        bool _gluteSoftPhysicsOn;

        internal bool Initialized { get; private set; }

        internal PersonMonitor Init(Bootstrap bootstrap)
        {
            enabled = false; // will be enabled after first-time calibration
            _bootstrap = bootstrap;
            _bootyMagic = bootstrap.BootyMagic;
            _tittyMagic = bootstrap.TittyMagic;
            _person = bootstrap.Person;
            _selectedCharacter = _person.Geometry.selectedCharacter;
            _fixedDeltaTime = Time.fixedDeltaTime;
            _physicsHighQuality = UserPreferences.singleton.physicsHighQuality;

            _softBodyPhysicsJsb = _person.SoftBodyPhysicsEnabler.GetBoolJSONParam("enabled");
            if(_bootstrap.ForceEnableAtomSoftPhysicsJsb.val)
            {
                _softBodyPhysicsJsb.val = true;
            }

            _softBodyPhysicsJsb.setJSONCallbackFunction += OnAtomSoftPhysicsChanged;
            _atomSoftPhysicsOn = _softBodyPhysicsJsb.val;
            _globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;

            if(_person.IsFemale)
            {
                _breastInOutJsb = _person.GetStorable("BreastInOut").GetBoolJSONParam("enabled");
                if(_tittyMagic.IsEnabledNullSafe() && _person.IsFemale)
                {
                    /* prevent breasts being flattened due to breastInOut morphs on scene load with plugin already present */
                    _breastInOutJsb.val = true;
                    _breastInOutJsb.val = false;
                }

                _breastInOutJsb.setJSONCallbackFunction += OnBreastInOutChanged;
            }

            StartCoroutine(FinishInit());
            return this;
        }

        IEnumerator FinishInit()
        {
            if(_person.IsFemale)
            {
                if(_bootyMagic.IsEnabledNullSafe())
                {
                    while(_bootyMagic.softPhysicsHandler?.softPhysicsOnJsb == null)
                    {
                        yield return null;
                    }

                    _gluteSoftPhysicsOn = _bootyMagic.softPhysicsHandler.softPhysicsOnJsb.val;
                }
            }

            if(_tittyMagic.IsEnabledNullSafe())
            {
                while(_tittyMagic.softPhysicsHandler?.softPhysicsOnJsb == null)
                {
                    yield return null;
                }

                _breastSoftPhysicsOn = _tittyMagic.softPhysicsHandler.softPhysicsOnJsb.val;
            }

            Initialized = true;
        }

        void OnAtomSoftPhysicsChanged(JSONStorableBool jsb)
        {
            if(!_bootstrap.IsEnabledAndActive() || !_bootstrap.ForceEnableAtomSoftPhysicsJsb.val)
            {
                return;
            }

            if(!jsb.val)
            {
                jsb.val = true;
                Loggr.Message("Prevented disabling Soft Body Physics in Control & Physics 1. (Change this behavior in Naturalis User Preferences.)");
            }
        }

        void OnBreastInOutChanged(JSONStorableBool jsb)
        {
            if(!_tittyMagic.IsEnabledNullSafe() || !_person.IsFemale)
            {
                return;
            }

            if(jsb.val)
            {
                jsb.val = false;
                Loggr.Message("Prevented enabling Auto Breast In/Out Morphs.");
            }
        }

        internal void Check(bool runCallbacks = false)
        {
            if(!Initialized || !enabled)
            {
                return;
            }

            if(_tittyMagic.IsEnabledNullSafe())
            {
                CheckBreasts();
            }

            if(_bootyMagic.IsEnabledNullSafe())
            {
                CheckGlutes();
            }

            if(_tittyMagic.IsEnabledNullSafe() || _bootyMagic.IsEnabledNullSafe())
            {
                CheckCommon();

                if(_selectedCharacter != _person.Geometry.selectedCharacter)
                {
                    StartCoroutine(RefreshCharacterCo());
                }
            }

            CheckSoftPhysicsToggled();
            CheckPhysicsRateChanged();
            CheckHighQualityPhysicsToggled();

            if(runCallbacks)
            {
                OnBreastInOutChanged(_breastInOutJsb);
                OnAtomSoftPhysicsChanged(_softBodyPhysicsJsb);
            }
        }

        void CheckBreasts()
        {
            _person.SetBreastsUseGravity(false);
            if(_person.PectoralRbCollisionsEnabled(true))
            {
                _person.SetPectoralRbCollisions(false);
            }

            if(!_person.Geometry.useAuxBreastColliders)
            {
                _person.Geometry.useAuxBreastColliders = true;
                Loggr.Message("Breast Hard Colliders enabled.");
            }
        }

        void CheckGlutes() => _person.SetGlutesUseGravity(false);

        void CheckCommon()
        {
            if(_person.IsFemale)
            {
                if(!_person.Geometry.useAdvancedColliders)
                {
                    _person.Geometry.useAdvancedColliders = true;
                    Loggr.Message("Advanced Colliders enabled - they are necessary for directional force morphing and hard colliders to work.");
                }
            }
        }

        void CheckSoftPhysicsToggled()
        {
            bool atomSoftPhysicsOn = _softBodyPhysicsJsb.val;
            bool globalSoftPhysicsOn = UserPreferences.singleton.softPhysics;

            if(_bootyMagic.IsEnabledNullSafe())
            {
                bool gluteSoftPhysicsOn = _person.GlutePhysicsMesh.on;
                string location = LocationWhereStillDisabled(gluteSoftPhysicsOn, _gluteSoftPhysicsOn, atomSoftPhysicsOn, globalSoftPhysicsOn);
                if(!string.IsNullOrEmpty(location))
                {
                    Loggr.Message($"Soft physics is still disabled in {location}");
                }

                bool softPhysicsOn = globalSoftPhysicsOn && atomSoftPhysicsOn && gluteSoftPhysicsOn;
                if(softPhysicsOn != IsGluteSoftPhysicsOn())
                {
                    _bootyMagic.OnSoftPhysicsToggled(softPhysicsOn);
                }

                _gluteSoftPhysicsOn = gluteSoftPhysicsOn;
            }

            if(_tittyMagic.IsEnabledNullSafe())
            {
                bool breastSoftPhysicsOn = _person.BreastPhysicsMesh.on;
                string location = LocationWhereStillDisabled(breastSoftPhysicsOn, _breastSoftPhysicsOn, atomSoftPhysicsOn, globalSoftPhysicsOn);
                if(!string.IsNullOrEmpty(location))
                {
                    Loggr.Message($"Soft physics is still disabled in {location}");
                }

                bool softPhysicsOn = globalSoftPhysicsOn && atomSoftPhysicsOn && breastSoftPhysicsOn;
                if(softPhysicsOn != IsBreastSoftPhysicsOn())
                {
                    _tittyMagic.OnSoftPhysicsToggled(softPhysicsOn);
                }

                _breastSoftPhysicsOn = breastSoftPhysicsOn;
            }

            _atomSoftPhysicsOn = atomSoftPhysicsOn;
            _globalSoftPhysicsOn = globalSoftPhysicsOn;
        }

        void CheckPhysicsRateChanged()
        {
            float value = Time.fixedDeltaTime;
            if(!value.IsEqual(_fixedDeltaTime))
            {
                _tittyMagic.OnPhysicsRateChanged();
                _bootyMagic.OnPhysicsRateChanged();
            }

            _fixedDeltaTime = value;
        }

        void CheckHighQualityPhysicsToggled()
        {
            bool value = UserPreferences.singleton.physicsHighQuality;
            if(value != _physicsHighQuality)
            {
                _tittyMagic.OnHighQualityPhysicsChanged();
                _bootyMagic.OnHighQualityPhysicsChanged();
            }

            _physicsHighQuality = value;
        }

        bool _inRefreshCharacterCo;

        IEnumerator RefreshCharacterCo()
        {
            if(_inRefreshCharacterCo)
            {
                yield break;
            }

            _inRefreshCharacterCo = true;
            while(!_person.IsSelectedCharacterReady())
            {
                yield return null;
            }

            if(_selectedCharacter.isMale == _person.Geometry.selectedCharacter.isMale)
            {
                yield return _person.RefreshSkinCo();
                yield return _person.CycleSoftPhysicsEnabled();
                _tittyMagic.OnCharacterChanged();
                _bootyMagic.OnCharacterChanged();
                _selectedCharacter = _person.Geometry.selectedCharacter;
            }
            else
            {
                Loggr.Message("Changing gender while the plugin is active is not supported. Disable the plugin, change the gender and then reload the plugin.");
                _person.Geometry.selectedCharacter = _selectedCharacter;
            }

            _inRefreshCharacterCo = false;
        }

        string LocationWhereStillDisabled(bool currEnabled, bool prevEnabled, bool atomSoftPhysicsOn, bool globalSoftPhysicsOn)
        {
            if(currEnabled && !prevEnabled)
            {
                if(!atomSoftPhysicsOn && !globalSoftPhysicsOn)
                {
                    return "Control & Physics 1 and in User Preferences";
                }

                if(!atomSoftPhysicsOn)
                {
                    return "Control & Physics 1";
                }

                if(!globalSoftPhysicsOn)
                {
                    return "User Preferences";
                }
            }
            else if(atomSoftPhysicsOn && !_atomSoftPhysicsOn)
            {
                if(!currEnabled && !globalSoftPhysicsOn)
                {
                    return "the plugin UI and in User Preferences";
                }

                if(!currEnabled)
                {
                    return "the plugin UI";
                }

                if(!globalSoftPhysicsOn)
                {
                    return "User Preferences";
                }
            }
            else if(globalSoftPhysicsOn && !_globalSoftPhysicsOn)
            {
                if(!currEnabled && !atomSoftPhysicsOn)
                {
                    return "the plugin UI and in Control & Physics 1";
                }

                if(!currEnabled)
                {
                    return "the plugin UI";
                }

                if(!atomSoftPhysicsOn)
                {
                    return "Control & Physics 1";
                }
            }

            return null;
        }

        void Update()
        {
            if(Time.frameCount % 60 == 0)
            {
                Check();
            }
        }

        void OnDestroy()
        {
            if(_breastInOutJsb != null)
            {
                _breastInOutJsb.setJSONCallbackFunction -= OnBreastInOutChanged;
            }

            if(_softBodyPhysicsJsb != null)
            {
                _softBodyPhysicsJsb.setJSONCallbackFunction -= OnAtomSoftPhysicsChanged;
            }
        }
    }
}
