using NAT.Common;
using NAT.Common.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace NAT.Store
{
    sealed class ClothingStore
    {
        const string UID_PREFIX = "clothing:";
        readonly DAZCharacterSelector _geometry;
        readonly bool _isFemale;
        internal readonly Dictionary<string, ClothingItem> items = new Dictionary<string, ClothingItem>();
        readonly Dictionary<object, ClothingItem> _itemsByRef = new Dictionary<object, ClothingItem>();
        internal readonly List<ClothingItem> enabledItems = new List<ClothingItem>();
        readonly List<ClothingItem> _retryItems = new List<ClothingItem>();

        internal Action onItemsReloadedHandlers;
        internal Action<ClothingItem> onItemToggledHandlers;
        internal Action<ClothingItem, bool> onItemAllowDetachToggledHandlers;

        internal ClothingStore(Person person)
        {
            _geometry = person.Geometry;
            _isFemale = person.IsFemale;
            SyncItems();
            enabledItems.AddRange(items.Values.Where(item => item.enabledJsb?.val ?? false));
        }

        internal void ReloadItems()
        {
            SyncItems();
            foreach(var pair in items)
            {
                var item = pair.Value;
                if(item.enabledJsb.val)
                {
                    if(!enabledItems.Contains(item))
                    {
                        enabledItems.Add(item);
                    }
                }
                else
                {
                    enabledItems.Remove(item);
                }
            }

            onItemsReloadedHandlers?.Invoke();
        }

        void SyncItems()
        {
            var duplicates = new Dictionary<string, int>();
            var stringBuilder = new StringBuilder();
            foreach(string name in _geometry.GetBoolParamNames())
            {
                if(!name.StartsWith(UID_PREFIX))
                {
                    continue;
                }

                string uid = name.Replace(UID_PREFIX, "");
                if(items.ContainsKey(uid) || uid == "Clothing Creator")
                {
                    continue;
                }

                var dazItem = _geometry.GetClothingItem(uid);
                if(
                    (_isFemale && dazItem.gender != DAZDynamicItem.Gender.Female) ||
                    (!_isFemale && dazItem.gender != DAZDynamicItem.Gender.Male)
                )
                {
                    continue;
                }

                var enabledJsb = _geometry.GetBoolJSONParam(name);
                string label = BuildLabel(duplicates, dazItem);
                var item = new ClothingItem(dazItem, label, enabledJsb);
                item.SetPresetSubDir(stringBuilder);

                enabledJsb.setJSONCallbackFunction += OnToggleItem;
                item.RefreshControls();
                AddOnItemAllowDetachHandler(item);
                items[uid] = item;
            }

            _retryItems.Clear();
            var needRetry = items
                .Where(pair => pair.Value.needsRetry)
                .Select(pair => pair.Value);
            _retryItems.AddRange(needRetry);
        }

        static string BuildLabel(Dictionary<string, int> duplicates, DAZClothingItem dazItem)
        {
            string label = $"{dazItem.displayName}";
            if(dazItem.IsInPackage)
            {
                label += $" [{dazItem.packageUid}]";
                return label;
            }

            label += $" [{dazItem.creatorName}]";
            int count;
            if(duplicates.ContainsKey(label))
            {
                duplicates[label]++;
                count = duplicates[label];
            }
            else
            {
                duplicates[label] = 1;
                count = 1;
            }

            return count > 1 ? $"{label} ({count})" : label;
        }

        void OnToggleItem(JSONStorableBool jsb)
        {
            try
            {
                string uid = jsb.name.Replace(UID_PREFIX, "");
                var item = items[uid];
                if(item.enabledJsb.val)
                {
                    enabledItems.Add(item);
                    item.RefreshControls();
                    if(item.needsRetry)
                    {
                        _retryItems.Add(item);
                    }
                    else
                    {
                        AddOnItemAllowDetachHandler(item);
                    }
                }
                else
                {
                    enabledItems.Remove(item);
                }

                onItemToggledHandlers?.Invoke(item);
            }
            catch(Exception e)
            {
                Loggr.Error($"OnToggleItem: {e}");
            }
        }

        void AddOnItemAllowDetachHandler(ClothingItem item)
        {
            if(!item.SimControl)
            {
                return;
            }

            var jsb = item.SimControl.GetBoolJSONParam("allowDetach");
            _itemsByRef[jsb] = item;
            jsb.setJSONCallbackFunction -= OnToggleItemAllowDetach;
            jsb.setJSONCallbackFunction += OnToggleItemAllowDetach;
        }

        void OnToggleItemAllowDetach(JSONStorableBool jsb)
        {
            try
            {
                onItemAllowDetachToggledHandlers?.Invoke(_itemsByRef[jsb], jsb.val);
            }
            catch(Exception e)
            {
                Loggr.Error($"OnToggleItemAllowDetach: {e}");
            }
        }

        internal void Execute()
        {
            if(Time.frameCount % 60 != 0)
            {
                return;
            }

            for(int i = _retryItems.Count - 1; i >= 0; i--)
            {
                var item = _retryItems[i];
                item.RefreshControls();
                if(!item.needsRetry)
                {
                    _retryItems.RemoveAt(i);
                    AddOnItemAllowDetachHandler(item);
                }
            }
        }

        internal void Destroy()
        {
            foreach(var pair in items)
            {
                var item = pair.Value;
                if(item.enabledJsb != null)
                {
                    item.enabledJsb.setJSONCallbackFunction -= OnToggleItem;
                }

                if(item.SimControl != null)
                {
                    item.SimControl.GetBoolJSONParam("allowDetach").setJSONCallbackFunction -= OnToggleItemAllowDetach;
                }
            }
        }
    }
}
