using NAT.Common.Models;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Store
{
    sealed class HairStore
    {
        const string UID_PREFIX = "hair:";
        readonly DAZCharacterSelector _geometry;
        readonly Dictionary<string, HairItem> _items = new Dictionary<string, HairItem>();
        internal readonly List<HairItem> enabledItems = new List<HairItem>();
        readonly List<HairItem> _retryItems = new List<HairItem>();

        internal HairStore(DAZCharacterSelector geometry)
        {
            _geometry = geometry;
            SyncItems();
            SyncEnabledItems();
        }

        void SyncItems()
        {
            foreach(string name in _geometry.GetBoolParamNames())
            {
                if(!name.StartsWith(UID_PREFIX))
                {
                    continue;
                }

                string uid = name.Replace(UID_PREFIX, "");
                if(_items.ContainsKey(uid))
                {
                    continue;
                }

                var enabledJsb = _geometry.GetBoolJSONParam(name);
                var item = new HairItem(_geometry.GetHairItem(uid), enabledJsb);
                enabledJsb.setJSONCallbackFunction += OnItemToggled;
                item.RefreshControls();
                _items[uid] = item;
            }

            _retryItems.Clear();
            var needRetry = _items
                .Where(pair => pair.Value.needsRetry)
                .Select(pair => pair.Value);
            _retryItems.AddRange(needRetry);
        }

        void OnItemToggled(JSONStorableBool jsb)
        {
            SyncEnabledItems();
            if(jsb.val)
            {
                string uid = jsb.name.Replace(UID_PREFIX, "");
                var item = _items[uid];
                item.RefreshControls();
                if(item.needsRetry)
                {
                    _retryItems.Add(item);
                }
            }
        }

        void SyncEnabledItems()
        {
            var enabledItemsDict = _items
                .Where(pair => pair.Value.enabledJsb?.val ?? false)
                .ToDictionary(pair => pair.Key, pair => pair.Value);
            enabledItems.Clear();
            enabledItems.AddRange(enabledItemsDict.Values);
        }

        internal void Execute()
        {
            if(Time.frameCount % 30 != 0)
            {
                return;
            }

            for(int i = _retryItems.Count - 1; i >= 0; i--)
            {
                var item = _retryItems[i];
                item.RefreshControls();
                if(!item.needsRetry)
                {
                    _retryItems.RemoveAt(i);
                }
            }
        }

        internal void Destroy()
        {
            foreach(var pair in _items)
            {
                if(pair.Value.enabledJsb != null)
                {
                    pair.Value.enabledJsb.setJSONCallbackFunction -= OnItemToggled;
                }
            }
        }
    }
}
