using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using NAT;
using NAT.BM;
using NAT.BM.Handlers;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.Common.UI;
using NAT.Common.UI.Windows;
using SimpleJSON;
using UnityEngine;

namespace Naturalis
{
    sealed class BootyMagic : ScriptBase
    {
        public override bool ShouldIgnore() => false;

        protected override void OnUIEnabled()
        {
            if(!person.IsFemale)
            {
                return;
            }

            uiHandler.gravityControlsSection.Show(true);

            if(uiHandler.IsMainTabActive())
            {
                if(!uiHandler.IsNestedWindowActive())
                {
                    uiHandler.mainWindow.Refresh();
                }
                else if(uiHandler.GetActiveNestedWindow() is DevWindow)
                {
                    softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(enabled);
                }
            }
            else if(uiHandler.IsHardCollidersTabActive())
            {
                if(!enabled)
                {
                    uiHandler.GoToMainWindow();
                }

                hardColliderHandler.colliderVisualizer.SetEnabledNullSafe(enabled);
                hardColliderHandler.colliderDrawService.SetEnabledNullSafe(enabled);
            }
            else if(uiHandler.IsPhysicsTabActive())
            {
                if(!uiHandler.IsNestedWindowActive())
                {
                    if(softPhysicsHandler.softPhysicsSynced)
                    {
                        uiHandler.physicsWindow.Refresh();
                    }
                }
                else if(uiHandler.GetActiveNestedWindow() is JointAngleWindow)
                {
                    // jointPhysicsHandler.jointDrawService.SetEnabledNullSafe(enabled);
                }
                else if(uiHandler.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(enabled);
                    var softParameterWindow = (SoftParameterWindow) uiHandler.GetActiveNestedWindow();
                    softParameterWindow.Refresh();
                    softParameterWindow.SetShowColliderPreviews();
                }
            }
            else if(uiHandler.IsClothingTabActive())
            {
                // TODO if soft physics section is active
                softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(enabled);
            }
        }

        protected override void OnUIDisabled()
        {
            if(!person.IsFemale)
            {
                return;
            }

            gravityPhysicsHandler.StopGravitySimulationAction.actionCallback();
            uiHandler.gravityControlsSection.Hide();

            if(bootstrap.CalibrateOnUIClosedJsb.val && calibrationStateHandler.Pending)
            {
                StartCalibration();
            }

            softPhysicsHandler.colliderVisualizer.SetEnabledNullSafe(false);
            softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(false);
            hardColliderHandler.colliderVisualizer.SetEnabledNullSafe(false);
            hardColliderHandler.colliderDrawService.SetEnabledNullSafe(false);
            calibrationStateHandler.EnableCalibrationStateJsb.val = false;

            uiHandler.ClosePopups();

            if(uiHandler.IsMainTabActive())
            {
                if(uiHandler.GetActiveNestedWindow() is SettingsWindow)
                {
                    uiHandler.mainWindow.OnReturn();
                }
            }
        }

        internal override bool IsLdpUpdateOn => softPhysicsHandler.ldpOn && bootstrap.PersonMonitor.IsGluteSoftPhysicsOn();
        internal override bool IsSoftPhysicsOn => softPhysicsHandler.softPhysicsOn && bootstrap.PersonMonitor.IsGluteSoftPhysicsOn();

        public override void Init()
        {
            calibrationDataVersion = "1.2.0";
            SetupDomain(Constant.BOOTY_MAGIC, Colors.peachPink);
            StartOrPostponeCoroutine(SetupRoutine());
        }

        IEnumerator SetupRoutine()
        {
            yield return InitBootstrapCo();
            if(bootstrap.Done == false)
            {
                initialized = false;
                yield break;
            }

            if(!bootstrap.Person.IsFemale)
            {
                var errorJss = new JSONStorableString("Info", $"<b>{Constant.BOOTY_MAGIC} is only supported on a female character.</b>");
                var textField = CreateTextField(errorJss);
                textField.backgroundColor = Color.clear;
                initialized = false;
                enabledJSON.valNoCallback = false;
                yield break;
            }

            person = bootstrap.Person;
            leftCalibrationTarget = new CalibrationTarget(person.HipRb, person.LeftGlute.rigidbody, person);
            rightCalibrationTarget = new CalibrationTarget(person.HipRb, person.RightGlute.rigidbody, person);

            if(enabled)
            {
                person.EnableCustomHipPhysics();
            }

            /* Used to communicate version to other plugin instances */
            var versionJss = new StorableString(Constant.NATURALIS_VERSION, "")
            {
                val = Bootstrap.NATURALIS_VERSION,
            };
            PreRegisterInternalParam(versionJss, false);

            angleCalculator = new AngleCalculator(person.LeftGlute.rigidbody, person.RightGlute.rigidbody, person.HipRb);
            SetupHandlers();
            SetupGluteMorphListener();
            SetupStorables();
            SetupCallbacks();
            calibrationStateHandler = AddCalibrationStateHandler();

            bindings = bindings ?? new Bindings(this); // Might already be setup in OnBindingsListRequested.
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            uiModManager = new UIModManager(this);
            uiHandler = new UIHandler(this);
            RegisterParams();

            if(!restoringFromJSON)
            {
                presetHandler.LoadDefaultPreset(person.IsFemale, true);
                if(enabled)
                {
                    foreach(var handler in handlers)
                    {
                        handler.Enable();
                    }

                    StartCalibration(true);
                }
            }

            initialized = true;
        }

        IEnumerator InitBootstrapCo()
        {
            bootstrap = gameObject.AddComponent<Bootstrap>().Init(this);
            while(!bootstrap || bootstrap.Done == null)
            {
                yield return null;
            }
        }

        void SetupHandlers()
        {
            presetHandler = new PluginPresetHandler(this, $"{DomainName}Presets");

            handlers.Add(jointPhysicsHandler = new GluteJointPhysicsHandler(this));
            handlers.Add(softPhysicsHandler = new GluteSoftPhysicsHandler(this));
            handlers.Add(gravityPhysicsHandler = new GluteGravityPhysicsHandler(this));
            handlers.Add(forcePhysicsHandler = new GluteForcePhysicsHandler(this));
            handlers.Add(frictionHandler = new GluteFrictionHandler(this));
            handlers.Add(hardColliderHandler = new HardColliderHandler(this));
            handlers.Add(forceMorphHandler = new GluteForceMorphHandler(this));
            handlers.Add(clothingHandler = new HipClothingHandler(this));

            bootstrap.ClothingStore.onItemsReloadedHandlers += clothingHandler.OnItemsReloaded;
            bootstrap.ClothingStore.onItemToggledHandlers += clothingHandler.OnItemToggled;
            bootstrap.ClothingStore.onItemAllowDetachToggledHandlers += clothingHandler.OnItemAllowDetachToggled;
        }

        void SetupGluteMorphListener()
        {
            morphListener = new MorphListener(
                new HashSet<int>(
                    jointPhysicsHandler.leftContext.volumeVertexIndices.Concat(jointPhysicsHandler.rightContext.volumeVertexIndices)
                ),
                new HashSet<string>
                {
                    "FBMFitnessDetails",
                    "FBMHeight",
                    "NyssaH",
                    "Norma",
                    "Alessa",
                    "Eisa",
                    "Izarra",
                    "bellybulge",
                    "Athletic",
                    "Body Tone",
                    "Bodybuilder Details",
                    "Ribs Width",
                    "Distort 1",
                    "Distort 2",
                    "FBMBodyTone",
                    "FBMBodybuilderDetails",
                    "Center Gap Smooth",
                    "PBMSacralDimples",
                    "Glute Line Sharpness",
                    "G3F_ThighsInnerSlim",
                    "Inner Thigh Gap2",
                    "Inner Thigh Gap",
                    "!Bend Helper",
                    "Pubic Area Width",
                    "Centre Crease",
                    "WaistBellyIn",
                    "BackLessCurve",
                    "TeethFixer",
                    "Thigh inner contour",
                    "G3F_Hip Leg Space",
                    "Doggy Style Adjust",
                    "Glute crease",
                    "deepbulge_bot",
                }
            );

            if(bootstrap.CheckForSizeChangesJsb.val)
            {
                ProcessMorphsForListener();
            }
        }

        void SetupStorables()
        {
            softnessJsf = new StorableFloat("Glute Softness", 70f, 0f, 100f)
            {
                setCallbackFunction = _ => SetCalibrationPending(softnessJsf),
            };
            quicknessJsf = new StorableFloat("Glute Quickness Offset", 0.00f, -1.00f, 1.00f)
            {
                setCallbackFunction = _ => SetCalibrationPending(softnessJsf),
            };

            leftRightAngleOffsetJsf = new StorableFloat("Left/Right Angle Offset", 0.0f, -30.0f, 30.0f)
            {
                setCallbackFunction = _ => SyncMorphsAngleOffsets(),
            };
            upDownAngleOffsetJsf = new StorableFloat("Up/Down Angle Offset", 0.0f, -15.0f, 15.0f)
            {
                setCallbackFunction = _ => SyncMorphsAngleOffsets(),
            };
            depthOffsetJsf = new StorableFloat("Depth Offset", 0.0f, -15.0f, 15.0f)
            {
                setCallbackFunction = _ => SyncMorphsAngleOffsets(),
            };

            calibrationAction = new StorableAction(Constant.CALIBRATE, () => StartCalibration(true));
            var calibrateOnNextPendingAction = new StorableAction(
                Constant.CALIBRATE_ON_NEXT_PENDING,
                () => calibrationStateHandler.calibrateOnNextPending = true
            );

            PreRegisterGeneralParam(softnessJsf);
            PreRegisterGeneralParam(quicknessJsf);
            PreRegisterGeneralParam(calibrationAction);
            PreRegisterGeneralParam(calibrateOnNextPendingAction);
        }

        internal override void SyncMorphsAngleOffsets(float? timer = null)
        {
            if(!enabled)
            {
                return;
            }

            float x = leftRightAngleOffsetJsf.val;
            float y = -upDownAngleOffsetJsf.val + 5.0f;
            float z = -0.001f * depthOffsetJsf.val;

            var leftMorphsOffsets = clothingHandler.LeftMorphsOffsets + clothingHandler.LeftMorphsOffsetsCompensation;
            var rightMorphsOffsets = clothingHandler.RightMorphsOffsets + clothingHandler.RightMorphsOffsetsCompensation;

            var offsetLeft = new Vector3(-x + leftMorphsOffsets.x, y + leftMorphsOffsets.y, z + leftMorphsOffsets.z);
            var offsetRight = new Vector3(x + rightMorphsOffsets.x, y + rightMorphsOffsets.y, z + rightMorphsOffsets.z);

            if(timer != null && timer.Value >= Time.deltaTime)
            {
                var transition = CoroutineUtils.ExponentialVector3ArrayTransition(
                    timer.Value,
                    new Vector3TransitionParams(leftMeasurementOffset, offsetLeft, value => leftMeasurementOffset = value),
                    new Vector3TransitionParams(rightMeasurementOffset, offsetRight, value => rightMeasurementOffset = value)
                );
                StartCoroutine(transition);
            }
            else
            {
                leftMeasurementOffset = offsetLeft;
                rightMeasurementOffset = offsetRight;
            }
        }

        void SetupCallbacks()
        {
            person.onScaleChangedHandlers += OnAtomScaleChanged;
            SuperController.singleton.onBeforeSceneSaveHandlers += OnBeforeSceneSave;
            SuperController.singleton.onSceneSavedHandlers += OnSceneSaved;
        }

        internal override void SetCalibrationTargetFunctions(bool softPhysicsOn)
        {
            if(softPhysicsOn)
            {
                leftCalibrationTarget.SetGluteFunctionsWhenSoftPhysicsOn(
                    person,
                    softPhysicsHandler.GetLeftPositionalRigidbodies(SoftVerticesGroupName.OUTER),
                    softPhysicsHandler.GetLeftPositionalRigidbodies(SoftVerticesGroupName.MAIN),
                    softPhysicsHandler.leftContext.centerSoftJointsRigidbodies
                );
                rightCalibrationTarget.SetGluteFunctionsWhenSoftPhysicsOn(
                    person,
                    softPhysicsHandler.GetRightPositionalRigidbodies(SoftVerticesGroupName.OUTER),
                    softPhysicsHandler.GetRightPositionalRigidbodies(SoftVerticesGroupName.MAIN),
                    softPhysicsHandler.rightContext.centerSoftJointsRigidbodies // TODO other groups as well
                );
            }
            else
            {
                leftCalibrationTarget.SetGluteFunctionsWhenSoftPhysicsOff();
                rightCalibrationTarget.SetGluteFunctionsWhenSoftPhysicsOff();
            }
        }

        void Update()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                if(!isUpdatePaused)
                {
                    UpdateBase();
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"Update: {e}");
                isUpdatePaused = true;
            }
        }

        void FixedUpdate()
        {
            if(initialized != true ||
                restoringFromJSON ||
                !person.Skin ||
                person.IsFreezeGrabbing ||
                bootstrap.IsGrabbingSubscene ||
                !SuperController.singleton.autoSimulation
            )
            {
                return;
            }

            if(isUpdatePaused)
            {
                if(inCalibrationCo && gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }

                return;
            }

            try
            {
                leftCalibrationTarget.Measure(leftMeasurementOffset);
                rightCalibrationTarget.Measure(rightMeasurementOffset);

                if(gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }
                else
                {
                    angleCalculator.CalculateAngles();
                    gravityPhysicsHandler.FixedUpdate();

                    if(!bootstrap.ForceUpdatesPaused() && !calibrationStateHandler.EnableCalibrationStateJsb.val)
                    {
                        forcePhysicsHandler.OnFixedUpdate();
                        forceMorphHandler.OnFixedUpdate();
                    }
                }

                hardColliderHandler.OnFixedUpdate();
                clothingHandler.OnFixedUpdate();
                jointPhysicsHandler.OnFixedUpdate();
                softPhysicsHandler.OnFixedUpdate();
            }
            catch(Exception e)
            {
                Loggr.LogError($"FixedUpdate: {e}");
                isUpdatePaused = true;
            }
        }

        bool _updatingMass;

        void SimulateGravity()
        {
            string direction = gravityPhysicsHandler.GravitySimulationDirection;
            angleCalculator.OverrideAngles(direction);
            gravityPhysicsHandler.AddGravitySimulationForce(!inCalibrationCo);

            if(!_updatingMass)
            {
                if(inCalibrationCo && direction != Direction.DOWN)
                {
                    leftCalibrationTarget.Measure(Vector3.zero);
                    rightCalibrationTarget.Measure(Vector3.zero);
                }

                gravityPhysicsHandler.FixedUpdate();
                forcePhysicsHandler.OnFixedUpdate();
                forceMorphHandler.OnFixedUpdate();
            }
        }

        protected override IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            calibrationStateHandler.EnableCalibrationStateJsb.val = false;
            inCalibrationCo = true;
            isUpdatePaused = true;
            gravityPhysicsHandler.StopGravitySimulationAction.actionCallback();
            /* Ensure everything is ready for calibration to proceed in case of e.g. appearance preset load */
            {
                /* Ensures glute soft physics enabled value is correct on first calibration when soft physics enabled is synced to SoftPhysicsHandler */
                while(!bootstrap.PersonMonitor.Initialized)
                {
                    yield return null;
                }

                /* Can occur when loading a look and VaM pauses to load assets */
                while(!person.IsSelectedCharacterReady())
                {
                    yield return null;
                }

                /* Must be unfrozen before checking for hard collider rigidbodies */
                yield return calibrationStateHandler.WaitForAtomPhysicsUnfrozen();

                bootstrap.PersonMonitor.Check(true);
                bootstrap.PersonMonitor.enabled = false;

                hardColliderHandler.InitAutoColliders();

                /* Sanity check. Might occur (?) even after hard colliders are enabled by settings monitor */
                float timeout = Time.unscaledTime + 5;
                while(!hardColliderHandler.RigidbodiesFound() && Time.unscaledTime < timeout)
                {
                    yield return null;
                }

                if(!hardColliderHandler.RigidbodiesFound())
                {
                    Loggr.LogError("Calibration error: collider rigidbodies not found after 5 seconds of waiting");
                    yield break;
                }
            }

            bool doFullCalibration =
                calibrationData == null ||
                !calibrationData.HasKey("mass") ||
                !calibrationData.HasKey("left_glute") ||
                !calibrationData.HasKey("right_glute") ||
                !calibrationData.HasKey("soft_joints");

            if(!doFullCalibration)
            {
                try
                {
                    CalibrateFromJSON(calibrationData);
                }
                catch(Exception e)
                {
                    doFullCalibration = true;
                    Loggr.LogError($"Error loading calibration data from JSON, continuing with full calibration. Error: {e}");
                }
            }

            if(doFullCalibration)
            {
                yield return FullCalibrationCo();
            }

            hardColliderHandler.SyncAll();
            SaveCalibratedValues();
            if(!restoredFromJSON && !firstCalibrationDone)
            {
                clothingHandler.AutoLoadDefaultProfiles();
            }

            clothingHandler.SetAllAdjustmentValues();
            clothingHandler.SyncAdjustments();
            SyncMorphsAngleOffsets();

            calibrationStateHandler.updateMassWhenCalibrating = false;
            isUpdatePaused = false;
            inCalibrationCo = false;
            firstCalibrationDone = true;

            uiHandler.GetActiveWindow()?.Refresh();
        }

        void CalibrateFromJSON(JSONClass jc)
        {
            softPhysicsHandler.SyncSoftPhysicsOn();
            jointPhysicsHandler.UpdateMassValueAndAmounts(jc["mass"].AsFloat);

            softness = CalculateSoftness();
            quickness = quicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            leftCalibrationTarget.RestoreFromJSON(jc["left_glute"].AsObject);
            rightCalibrationTarget.RestoreFromJSON(jc["right_glute"].AsObject);
            softPhysicsHandler.CalibrateFromJSON(jc["soft_joints"].AsObject);

            calibrationStateHandler.EnableSettingsMonitor();
        }

        IEnumerator FullCalibrationCo()
        {
            softPhysicsHandler.softJointDrawService.ClearLines();
            hardColliderHandler.colliderDrawService.ClearLines();
            hardColliderHandler.colliderVisualizer.showPreviewsJSON.val = false;
            /* Must be called before calibration pose is applied because the current pose must be stored without duplicate morphs having >0 values. */
            ZeroDynamicAdjustments();
            person.ZeroAllDuplicateMorphs();

            yield return calibrationStateHandler.Begin();
            if(calibrationStateHandler.cancelling)
            {
                calibrationStateHandler.cancelling = false;
                yield break;
            }

            softPhysicsHandler.SyncSoftPhysicsOn();

            {
                gravityPhysicsHandler.EnableGravitySimulation(Direction.DOWN);
                yield return UpdateMassAndPhysicsCo();
                jointPhysicsHandler.SetGravitySimulationDamper();
                yield return new WaitForSeconds(0.25f);

                float minTime = Time.time + 0.40f;
                float maxTime = Time.time + 1.00f;
                const float tolerance = 0.003f;
                while(Time.time < minTime ||
                    ((leftCalibrationTarget.Moved(tolerance) || rightCalibrationTarget.Moved(tolerance)) && Time.time < maxTime)
                )
                {
                    yield return null;
                }

                leftCalibrationTarget.Calibrate();
                rightCalibrationTarget.Calibrate();
                if(IsLdpUpdateOn)
                {
                    softPhysicsHandler.CalibrateSoftJoints(Direction.DOWN);
                }
            }

            if(IsLdpUpdateOn)
            {
                foreach(string direction in new[] { Direction.FORWARD, Direction.UP, Direction.LEFT, Direction.RIGHT, Direction.BACK })
                {
                    gravityPhysicsHandler.EnableGravitySimulation(direction);
                    yield return new WaitForSeconds(0.25f);

                    float timeout = Time.time + 1.50f;
                    while(softPhysicsHandler.AnySoftJointMoved() && Time.time < timeout)
                    {
                        yield return null;
                    }

                    softPhysicsHandler.CalibrateSoftJoints(direction);
                }
            }

            gravityPhysicsHandler.StopGravitySimulation();
            jointPhysicsHandler.UnsetGravitySimulationDamper();
            hardColliderHandler.colliderVisualizer.showPreviewsJSON.val = true;
            yield return calibrationStateHandler.Finish();
        }

        void ZeroDynamicAdjustments()
        {
            leftCalibrationTarget.Reset();
            rightCalibrationTarget.Reset();
            foreach(var handler in handlers)
            {
                handler.ResetAll();
            }
        }

        float CalculateSoftness() => Calc.SoftnessBaseCurve(softnessJsf.val / 100f);

        IEnumerator UpdateMassAndPhysicsCo()
        {
            _updatingMass = true;
            if(calibrationStateHandler.updateMassWhenCalibrating)
            {
                softness = 0;
                quickness = 0;
                jointPhysicsHandler.UpdateBaseValues(false);
                softPhysicsHandler.UpdatePhysics(false);

                yield return new WaitForSeconds(0.25f);
                /* Mass is calculated multiple times because each new mass value changes the exact glute
                 * shape and therefore the estimated volume.
                 */
                jointPhysicsHandler.CalculateMass();
                yield return Utils.WaitAndRepeat(
                    () => jointPhysicsHandler.CalculateMass(),
                    (int) Mathf.Lerp(4, 12, jointPhysicsHandler.normalizedRealMass),
                    0.13f
                );
            }

            softness = CalculateSoftness();
            quickness = quicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            _updatingMass = false;
        }

        protected override void SetMultipliersAndUpdatePhysics()
        {
            foreach(var handler in handlers)
            {
                handler.SetMultipliers();
            }

            jointPhysicsHandler.UpdateBaseValues();
            softPhysicsHandler.UpdatePhysics();
            frictionHandler.CalculateFriction();
        }

        void SaveCalibratedValues()
        {
            softnessJsf.SaveCalibrated();
            quicknessJsf.SaveCalibrated();
            morphListener.SaveCalibrated();
            calibratedScale = person.ScaleJsf.val;
            foreach(var handler in handlers)
            {
                handler.SaveCalibrated();
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if(person.IsFemale)
            {
                if(_isSavingScene && firstCalibrationDone && !calibrationStateHandler.Pending && includePhysical && includeAppearance)
                {
                    jc["stored_in_scene"].AsBool = true; // Unused, might be useful later
                    var calibrationData = new JSONClass
                    {
                        ["version"] = calibrationDataVersion,
                        ["mass"] = { AsFloat = jointPhysicsHandler.calibratedMass },
                        ["left_glute"] = leftCalibrationTarget.Serialize(),
                        ["right_glute"] = rightCalibrationTarget.Serialize(),
                        ["soft_joints"] = softPhysicsHandler.SerializeCalibrationData(),
                    };
                    jc["calibration_data"] = calibrationData;
                }

                var activeClothingProfiles = new JSONArray();
                foreach(var profile in clothingHandler.activeProfiles)
                {
                    if(!string.IsNullOrEmpty(profile.loadedFullPath))
                    {
                        activeClothingProfiles.Add(new JSONClass
                        {
                            ["presetName"] = profile.loadedFullPath,
                        });
                    }
                }

                jc["Active Clothing Profiles"] = activeClothingProfiles;
            }

            needsStore = true;
            return jc;
        }

        protected override void FinishRestoreFromJSON(JSONClass jc, string scenePackageId, JSONClass calibrationData)
        {
            if(!firstCalibrationDone)
            {
                // ReSharper disable once ExpressionIsAlwaysNull
                clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), scenePackageId);
            }

            if(enabled)
            {
                foreach(var handler in handlers)
                {
                    handler.Enable();
                }

                StartCalibration(true, calibrationData);
            }
        }

        void OnAtomScaleChanged(float value)
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            if(!calibrationStateHandler.Pending && !calibrationStateHandler.calibrateOnNextPending)
            {
                Loggr.LogMessage("Person atom scale changed - calibration set to pending.");
            }

            calibrationStateHandler.SetCalibrationPending(Constant.SCALE_CHANGED, !value.IsEqual(calibratedScale));
        }

        bool _isSavingScene;

        void OnBeforeSceneSave()
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            try
            {
                _isSavingScene = true;
                isUpdatePaused = true;
                forceMorphHandler.ResetAll();
                jointPhysicsHandler.ResetAll();
                clothingHandler.ResetMorphs();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnBeforeSceneSave: {e}");
            }
        }

        void OnSceneSaved()
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            try
            {
                _isSavingScene = false;
                if(!calibrationStateHandler.CalibratingJsb.val)
                {
                    isUpdatePaused = false;
                    clothingHandler.SyncMorphsOffsetsCompensation();
                    SyncMorphsAngleOffsets();
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnSceneSaved: {e}");
            }
        }

        internal override void OnSoftPhysicsToggled(bool value)
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            try
            {
                softPhysicsHandler.ReverseSyncSoftPhysicsOn();

                if(!inCalibrationCo)
                {
                    isUpdatePaused = true;
                    calibrationStateHandler.SetCalibrationPending(Constant.SOFT_PHYSICS_CHANGED);
                    if(!calibrationStateHandler.calibrateOnNextPending)
                    {
                        Loggr.LogMessage($"Soft physics {(value ? "enabled" : "disabled")} - updates paused, calibration set to pending.");
                    }
                }

                softPhysicsHandler.softPhysicsSynced = false;

                if(uiHandler.IsPhysicsTabActive())
                {
                    if(uiHandler.GetActiveNestedWindow() is SoftParameterWindow)
                    {
                        if(!value)
                        {
                            uiHandler.physicsWindow.OnReturn();
                        }
                    }
                    else if(!value)
                    {
                        uiHandler.physicsWindow.Refresh();
                    }
                }

                if(!value)
                {
                    softPhysicsHandler.softJointDrawService.visualizeLdpJsb.val = false;
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnSoftPhysicsToggled: {e}");
            }
        }

        void OnEnable()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                BaseOnEnable();
                person.EnableCustomHipPhysics();

                if(!restoringFromJSON)
                {
                    foreach(var handler in handlers)
                    {
                        handler.Enable();
                    }

                    StartCalibration(true);
                }

                uiModManager.Enable();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnEnable: {e}");
            }
        }

        void OnDisable()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                if(calibrationCo != null)
                {
                    StopCoroutine(calibrationCo);
                }

                foreach(var handler in handlers)
                {
                    handler.Disable();
                }

                uiModManager.Disable();
                calibrationStateHandler.Finish(bootstrap.TittyMagic.IsEnabledNullSafe());
                person.DisableCustomHipPhysics();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnDisable: {e}");
            }
        }

        void OnDestroy()
        {
            try
            {
                BaseOnDestroy();
                foreach(var handler in handlers)
                {
                    handler.Destroy();
                }

                Destroy(bootstrap);
                bootstrap = null;

                SuperController.singleton.onSceneSavedHandlers -= OnSceneSaved;
                SuperController.singleton.onBeforeSceneSaveHandlers -= OnBeforeSceneSave;
                SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            }
            catch(Exception e)
            {
                if(initialized == true)
                {
                    SuperController.LogError($"OnDestroy: {e}");
                }
                else
                {
                    Debug.LogError($"OnDestroy: {e}");
                }
            }
        }
    }
}
