using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.BM.Handlers
{
    sealed class GluteForcePhysicsHandler : ForceHandler
    {
        List<JointPhysicsParam> _leftParams, _rightParams;

        internal GluteForcePhysicsHandler(ScriptBase script) : base(script)
        {
            baseJsf = NewMultiplierJsf("Force Physics Base Multiplier");
            upJsf = NewMultiplierJsf("Force Physics Up Multiplier");
            downJsf = NewMultiplierJsf("Force Physics Down Multiplier");
            depthOutJsf = NewMultiplierJsf("Force Physics Depth Out Multiplier");
            depthInJsf = NewMultiplierJsf("Force Physics Depth In Multiplier");
            sidewaysInJsf = NewMultiplierJsf("Force Physics Sideways In Multiplier");
            sidewaysOutJsf = NewMultiplierJsf("Force Physics Sideways Out Multiplier");
            LoadSettings();
        }

        protected override void OnMultiplierValueChanged(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            if(Script.IsLdpUpdateOn)
            {
                Script.SetCalibrationPending(storable);
            }
        }

        protected override void LoadSettings()
        {
            var centerOfGravity = Script.jointPhysicsHandler.centerOfGravity;
            var inOutDamper = Script.jointPhysicsHandler.inOutDamper;
            var upDownAngleTarget = Script.jointPhysicsHandler.upDownAngleTarget;
            var rightLeftAngleTarget = Script.jointPhysicsHandler.rightLeftAngleTarget;

            centerOfGravity.SetForceCalculators(CenterOfGravityCalculators());
            inOutDamper.SetForceCalculators(InOutDamperCalculators());
            upDownAngleTarget.SetForceCalculators(UpDownAngleTargetCalculators());
            rightLeftAngleTarget.SetForceCalculators(RightLeftAngleTargetCalculators());

            _leftParams = new List<JointPhysicsParam>
            {
                centerOfGravity.Left,
                inOutDamper.Left,
                // upDownAngleTarget.Left,
                // rightLeftAngleTarget.Left,
            };

            _rightParams = new List<JointPhysicsParam>
            {
                centerOfGravity.Right,
                inOutDamper.Right,
                // upDownAngleTarget.Right,
                // rightLeftAngleTarget.Right,
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> CenterOfGravityCalculators()
        {
            var backCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -0.0975f,
                    multiplierB = -0.065f,
                    curveA = x => 1 - x,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var forwardCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 0.0975f,
                    multiplierB = 0.065f,
                    curveA = x => 1 - x,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backCalculator },
                { FORWARD, forwardCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, backCalculator },
                { FORWARD, forwardCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> InOutDamperCalculators()
        {
            var depthInCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 1.5f,
                    multiplierB = -4f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var depthOutCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 1.5f,
                    multiplierB = -4f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInCalculator },
                { FORWARD, depthOutCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { BACK, depthInCalculator },
                { FORWARD, depthOutCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> UpDownAngleTargetCalculators()
        {
            var upCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = 5.33f,
                    multiplierB = 4.0f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var downCalculator = new PhysicsCalculator
            {
                primaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                    multiplierA = -1.0f,
                    multiplierB = -0.75f,
                    curveB = Calc.ForcePhysicsSoftnessCurve,
                },
            };
            var left = new Dictionary<string, PhysicsCalculator>
            {
                { UP, upCalculator },
                { DOWN, downCalculator },
            };
            var right = new Dictionary<string, PhysicsCalculator>
            {
                { UP, upCalculator },
                { DOWN, downCalculator },
            };
            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        static Dictionary<string, Dictionary<string, PhysicsCalculator>> RightLeftAngleTargetCalculators() =>
            new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                {
                    Side.LEFT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                primaryConfig = new MultiplierConfig
                                {
                                    multiplierA = 5.33f,
                                    multiplierB = 4.0f,
                                    curveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                primaryConfig = new MultiplierConfig
                                {
                                    multiplierA = -5.33f,
                                    multiplierB = -4.0f,
                                    curveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                    }
                },
                {
                    Side.RIGHT, new Dictionary<string, PhysicsCalculator>
                    {
                        {
                            SIDEWAYS_IN, new PhysicsCalculator
                            {
                                primaryConfig = new MultiplierConfig
                                {
                                    multiplierA = -5.33f,
                                    multiplierB = -4.0f,
                                    curveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                        {
                            SIDEWAYS_OUT, new PhysicsCalculator
                            {
                                primaryConfig = new MultiplierConfig
                                {
                                    multiplierA = 5.33f,
                                    multiplierB = 4.0f,
                                    curveB = Calc.ForcePhysicsSoftnessCurve,
                                },
                            }
                        },
                    }
                },
            };

        internal override void PreRegisterParams()
        {
            Script.PreRegisterForcePhysicsParam(baseJsf);
            // Script.PreRegisterForcePhysicsParam(UpJsf);
            // Script.PreRegisterForcePhysicsParam(DownJsf);
            Script.PreRegisterForcePhysicsParam(depthOutJsf);
            Script.PreRegisterForcePhysicsParam(depthInJsf);
            // Script.PreRegisterForcePhysicsParam(SidewaysInJsf);
            // Script.PreRegisterForcePhysicsParam(SidewaysOutJsf);
        }

        /* TODO duplicates ForceMorphHandler */

        float _mass;
        float _softness;

        float _upDownLimit;
        float _depthInLimit;
        float _depthOutLimit;
        float _sidewaysLimit;

        float _upMultiplier;
        float _downMultiplier;
        float _depthInMultiplier;
        float _depthOutMultiplier;
        float _sidewaysMultiplier;

        static float SoftnessMultiplier(float x) => Calc.InverseSmoothStep(x, 0.80f, 0.03f);

        internal override void SetMultipliers()
        {
            _mass = Script.jointPhysicsHandler.normalizedRealMass;
            _softness = SoftnessMultiplier(Script.softness);
            SetUpDownMultipliers();
            SetDepthInMultiplier();
            SetDepthOutMultiplier();
            SetSidewaysMultiplier();
        }

        void SetUpDownMultipliers()
        {
            _upDownLimit = 50 * Mathf.Lerp(1, 1.20f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.25f, _softness);
            _upMultiplier = softnessMultiplier * 1.50f * Mathf.Lerp(1, 1.80f, _mass);
            _downMultiplier = softnessMultiplier * 1.80f * Mathf.Lerp(1, 1.80f, _mass);
        }

        void SetDepthInMultiplier()
        {
            _depthInLimit = 0.045f * Mathf.Lerp(1, 3.00f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.00f, _softness);
            _depthInMultiplier = softnessMultiplier * 4.40f * Mathf.Lerp(1, 0.50f, _mass);
        }

        void SetDepthOutMultiplier()
        {
            _depthOutLimit = 0.045f * Mathf.Lerp(1, 3.00f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.65f, 1.125f, _softness);
            _depthOutMultiplier = softnessMultiplier * 2.20f * Mathf.Lerp(1, 1.50f, _mass);
        }

        void SetSidewaysMultiplier()
        {
            _sidewaysLimit = 50 * Mathf.Lerp(1, 1.20f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.25f, _softness);
            _sidewaysMultiplier = softnessMultiplier * 1.10f * Mathf.Lerp(1, 1.80f, _mass);
        }

        /* TODO END */

        internal override void OnFixedUpdate()
        {
            AdjustUpPhysics();
            AdjustDownPhysics();
            AdjustDepthOutPhysics();
            AdjustDepthInPhysics();
            AdjustSidewaysPhysics();
        }

        static float UpDownForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float UpEffect(float angle) =>
            _upMultiplier *
            UserUpMultiplier() *
            UpDownForceEffectCurve(Mathf.Abs(angle) / _upDownLimit);

        void AdjustUpPhysics()
        {
            if(leftTarget.angleY <= 0)
            {
                UpdatePhysicsLeft(UP, UpEffect(leftTarget.angleY));
            }
            else
            {
                ResetPhysicsLeft(UP);
            }

            if(rightTarget.angleY <= 0)
            {
                UpdatePhysicsRight(UP, UpEffect(rightTarget.angleY));
            }
            else
            {
                ResetPhysicsRight(UP);
            }
        }

        float DownEffect(float angle) => _downMultiplier *
            UserDownMultiplier() *
            UpDownForceEffectCurve(Mathf.Abs(angle) / _upDownLimit);

        void AdjustDownPhysics()
        {
            if(leftTarget.angleY <= 0)
            {
                ResetPhysicsLeft(DOWN);
            }
            else
            {
                UpdatePhysicsLeft(DOWN, DownEffect(leftTarget.angleY));
            }

            if(rightTarget.angleY <= 0)
            {
                ResetPhysicsRight(DOWN);
            }
            else
            {
                UpdatePhysicsRight(DOWN, DownEffect(rightTarget.angleY));
            }
        }

        static float DepthForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float DepthInEffect(float depthDiff) =>
            _depthInMultiplier *
            UserDepthInMultiplier() *
            DepthForceEffectCurve(Mathf.Abs(depthDiff) / _depthInLimit);

        void AdjustDepthInPhysics()
        {
            if(leftTarget.depthDiff >= 0)
            {
                ResetPhysicsLeft(BACK);
            }
            else
            {
                UpdatePhysicsLeft(BACK, DepthInEffect(leftTarget.depthDiff));
            }

            if(rightTarget.depthDiff >= 0)
            {
                ResetPhysicsRight(BACK);
            }
            else
            {
                UpdatePhysicsRight(BACK, DepthInEffect(rightTarget.depthDiff));
            }
        }

        float DepthOutEffect(float depthDiff) => _depthOutMultiplier *
            UserDepthOutMultiplier() *
            DepthForceEffectCurve(Mathf.Abs(depthDiff) / _depthOutLimit);

        void AdjustDepthOutPhysics()
        {
            if(leftTarget.depthDiff >= 0)
            {
                UpdatePhysicsLeft(FORWARD, DepthOutEffect(leftTarget.depthDiff));
            }
            else
            {
                ResetPhysicsLeft(FORWARD);
            }

            if(rightTarget.depthDiff >= 0)
            {
                UpdatePhysicsRight(FORWARD, DepthOutEffect(rightTarget.depthDiff));
            }
            else
            {
                ResetPhysicsRight(FORWARD);
            }
        }

        static float SidewaysForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float SidewaysInEffect(float angle) =>
            _sidewaysMultiplier *
            UserSidewaysInMultiplier() *
            SidewaysForceEffectCurve(Mathf.Abs(angle) / _sidewaysLimit);

        float SidewaysOutEffect(float angle) =>
            _sidewaysMultiplier *
            UserSidewaysOutMultiplier() *
            SidewaysForceEffectCurve(Mathf.Abs(angle) / _sidewaysLimit);

        void AdjustSidewaysPhysics()
        {
            if(leftTarget.angleX <= 0)
            {
                // right force on left glute
                ResetPhysicsLeft(SIDEWAYS_OUT);
                UpdatePhysicsLeft(SIDEWAYS_IN, SidewaysInEffect(leftTarget.angleX));
            }
            else
            {
                // left force on left glute
                ResetPhysicsLeft(SIDEWAYS_IN);
                UpdatePhysicsLeft(SIDEWAYS_OUT, SidewaysOutEffect(leftTarget.angleX));
            }

            if(rightTarget.angleX <= 0)
            {
                // right force on right glute
                ResetPhysicsRight(SIDEWAYS_IN);
                UpdatePhysicsRight(SIDEWAYS_OUT, SidewaysOutEffect(rightTarget.angleX));
            }
            else
            {
                // left force on right glute
                ResetPhysicsRight(SIDEWAYS_OUT);
                UpdatePhysicsRight(SIDEWAYS_IN, SidewaysInEffect(rightTarget.angleX));
            }
        }

        void UpdatePhysicsLeft(string direction, float effect)
        {
            foreach(var param in _leftParams)
            {
                param.UpdateForceValue(direction, effect, _mass, _softness);
            }
        }

        void UpdatePhysicsRight(string direction, float effect)
        {
            foreach(var param in _rightParams)
            {
                param.UpdateForceValue(direction, effect, _mass, _softness);
            }
        }

        internal override void ResetAll()
        {
            foreach(string direction in new[]
            {
                LEFT,
                RIGHT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetPhysicsLeft(direction);
                ResetPhysicsRight(direction);
            }
        }

        void ResetPhysicsLeft(string direction)
        {
            foreach(var param in _leftParams)
            {
                param.ResetForceValue(direction);
            }
        }

        void ResetPhysicsRight(string direction)
        {
            foreach(var param in _rightParams)
            {
                param.ResetForceValue(direction);
            }
        }
    }
}
