using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.BM.Handlers
{
    sealed class GluteGravityPhysicsHandler : GravityPhysicsHandler
    {
        internal GluteGravityPhysicsHandler(ScriptBase script) : base(script)
        {
            LoadSettings();
        }

        protected override void LoadSettings()
        {
            leftContext.calculators = Calculators(Side.LEFT);
            rightContext.calculators = Calculators(Side.RIGHT);

            leftContext.calculators.Add(
                jointPhysicsHandler.rightLeftAngleTarget,
                new Dictionary<string, PhysicsCalculator>
                {
                    {
                        SIDEWAYS_IN, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_IN),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = 1.00f,
                                multiplierB = 2.00f,
                            },
                        }
                    },
                    {
                        SIDEWAYS_OUT, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_OUT),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = -1.50f,
                                multiplierB = -3.00f,
                            },
                        }
                    },
                }
            );

            rightContext.calculators.Add(
                jointPhysicsHandler.rightLeftAngleTarget,
                new Dictionary<string, PhysicsCalculator>
                {
                    {
                        SIDEWAYS_IN, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_IN),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = -1.00f,
                                multiplierB = -2.00f,
                            },
                        }
                    },
                    {
                        SIDEWAYS_OUT, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_OUT),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = 1.50f,
                                multiplierB = 3.00f,
                            },
                        }
                    },
                }
            );

            foreach(var pair in leftContext.calculators)
            {
                var paramPair = pair.Key;
                var calculators = pair.Value;
                paramPair.Left.gravityCalculators = calculators;
                paramPair.SetGravityMultiplierCallbacks();
                leftContext.updateHandlers += (direction, effect) => paramPair.Left.UpdateGravityValue(direction, effect, mass, softness);
                leftContext.resetHandlers += direction => paramPair.Left.ResetGravityValue(direction);
            }

            foreach(var pair in rightContext.calculators)
            {
                var paramPair = pair.Key;
                var calculators = pair.Value;
                paramPair.Right.gravityCalculators = calculators;
                paramPair.SetGravityMultiplierCallbacks();
                rightContext.updateHandlers += (direction, effect) => paramPair.Right.UpdateGravityValue(direction, effect, mass, softness);
                rightContext.resetHandlers += direction => paramPair.Right.ResetGravityValue(direction);
            }
        }

        Dictionary<JointPhysicsParamPair, Dictionary<string, PhysicsCalculator>> Calculators(string side) =>
            new Dictionary<JointPhysicsParamPair, Dictionary<string, PhysicsCalculator>>
            {
                {
                    jointPhysicsHandler.rotationSpring, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, BackSpringCalculator(side, BACK) },
                        { FORWARD, ForwardSpringCalculator(side) },
                    }
                },
                {
                    jointPhysicsHandler.inOutSpring, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, InOutSpringCalculator(side, BACK) },
                        { FORWARD, InOutSpringCalculator(side, FORWARD) },
                    }
                },
                {
                    jointPhysicsHandler.upDownAngleTarget, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpAngleTargetCalculator(side) },
                        { DOWN, DownAngleTargetCalculator(side) },
                    }
                },
            };

        static PhysicsCalculator BackSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_SPRING.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = -0.75f,
                multiplierB = -3.75f,
                curveA = x => 1 - x,
            },
        };

        static PhysicsCalculator ForwardSpringCalculator(string side) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_SPRING.NAME, side), FORWARD),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = 7.5f,
                multiplierB = 2.5f,
                curveA = x => 1 - x,
            },
        };

        static PhysicsCalculator InOutSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(IN_OUT_SPRING.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = -25f,
                multiplierA = -100f,
                multiplierB = -125f,
                curveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                curveB = Calc.InOutSpringSoftnessCurve,
            },
        };

        static PhysicsCalculator UpAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(GLUTE_UP_DOWN_ANGLE_TARGET.NAME, side), UP),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 2.50f,
            },
        };

        static PhysicsCalculator DownAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(GLUTE_UP_DOWN_ANGLE_TARGET.NAME, side), DOWN),
            primaryConfig = new MultiplierConfig
            {
                baseValue = -2.50f,
            },
        };

        internal override void SetMultipliers() // TODO
        {
            mass = jointPhysicsHandler.normalizedRealMass;
            upGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            downGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            forwardGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            backGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            sidewaysInGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            sidewaysOutGravityMultiplier = Mathf.Lerp(0.5f, 0.5f, mass);
            SetGravityValues();
        }
    }
}
