using NAT.Common;
using NAT.Common.Contexts;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System;
using System.Text;
using UnityEngine;

namespace NAT.BM.Handlers
{
    sealed class GluteJointPhysicsHandler : JointPhysicsHandler
    {
        MorphWrapper _downAngleMorph;
        MorphWrapper _upAngleMorph;
        MorphWrapper _sidewaysInMorph;
        MorphWrapper _sidewaysOutMorph;

        internal GluteJointPhysicsHandler(ScriptBase script) : base(script)
        {
            control = Person.GluteControl;
            leftContext.bodyPart = Person.LeftGlute;
            rightContext.bodyPart = Person.RightGlute;

            leftContext.volumeVertexIndices = new[]
            {
                7877, 283, 7821, 7833, 2391, 288, 2392, 7878, 7875, 7876, 7864, 7825, 7826,
                7845, 274, 2390, 8249, 296, 281, 7849, 7824, 8212, 265, 2388, 4036, 287, 4039,
            };
            rightContext.volumeVertexIndices = new[]
            {
                18557, 11211, 18501, 18513, 13221, 11216, 13222, 18558, 18555, 18556, 18544, 18505, 18506, 18525,
                11202, 13220, 18920, 11224, 11209, 18529, 18504, 18883, 11193, 13218, 14812, 11215, 14815,
            };

            script.leftCalibrationTarget.centerVertexIndices = new[] { 7860, 7861, 7875, 7878, 7825, 7826 };
            script.rightCalibrationTarget.centerVertexIndices = new[] { 18505, 18506, 18540, 18541, 18555, 18558 };

            LoadSettings();
        }

        internal override void CalculateMass()
        {
            float leftVolume = CalculateVolume(leftContext.volumeVertexIndices);
            float rightVolume = CalculateVolume(rightContext.volumeVertexIndices);
            float volume = (leftVolume + rightVolume) / 2;
            /* Interpolates volume range [0.77, 2.70] to mass range [1.00, 4.00] */
            UpdateMassValueAndAmounts(1.55f * volume - 0.19f);
        }

        internal override void UpdateMassValueAndAmounts(float mass)
        {
            calibratedMass = mass;
            const float min = GLUTE_JOINT_MASS.MIN;
            const float max = GLUTE_JOINT_MASS.MAX;
            float jointToCollidersRatio = Mathf.Lerp(0.92f, 0.78f, Mathf.InverseLerp(min, max, mass));
            jointMass.UpdateMassValue(jointToCollidersRatio * mass);
            jointMass.UpdateOffsetMinMax();
            Script.hardColliderHandler.collidersMassAmount = (1 - jointToCollidersRatio) * mass;
            normalizedMass = Mathf.InverseLerp(min, max, jointMass.Left.Value);
            normalizedRealMass = Mathf.InverseLerp(min, max, jointMass.Left.baseValueJsf.val);
        }

        float CalculateVolume(int[] vertexIndices)
        {
            var skinVerts = Person.Skin.rawSkinnedVerts;
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(Person.HipRb, skinVerts[vertexIndices[i]]);
            }

            return 1000 * 0.5f * VolumeCalculator.SphereVolume(positions, Person.ScaleJsf.val);
        }

        protected override void LoadSettings()
        {
            _upAngleMorph = new MorphWrapper(Person.GetGluteMorph("Glute Up Angle"));
            _downAngleMorph = new MorphWrapper(Person.GetGluteMorph("Glute Down Angle"));
            _sidewaysInMorph = new MorphWrapper(Person.GetGluteMorph("Glute Sideways In Angle"));
            _sidewaysOutMorph = new MorphWrapper(Person.GetGluteMorph("Glute Sideways Out Angle"));

            jointMass = new JointPhysicsParamPair(Script, GLUTE_JOINT_MASS.NAME, Constant.GLUTE_WEIGHT)
            {
                ValueFormat = GLUTE_JOINT_MASS.FORMAT,
                Left = GluteMassParam(leftContext),
                Right = GluteMassParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(GLUTE_JOINT_MASS.NAME)
                    .Append(" is the mass of the joint in kg.\n\n")
                    .Append("Since mass represents glute size, other physics parameters are adjusted based on its value.")
                    .ToString(),
            };

            paramPairs.Add(centerOfGravity = new JointPhysicsParamPair(Script, CENTER_OF_GRAVITY.NAME)
            {
                ValueFormat = CENTER_OF_GRAVITY.FORMAT,
                Left = CenterOfGravityParam(leftContext),
                Right = CenterOfGravityParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(CENTER_OF_GRAVITY.NAME)
                    .Append(" is the position of the joint's center of mass.\n\n")
                    .Append("At 0, the center of mass is inside the body at the joint. At 1, it is at the skin surface.")
                    .ToString(),
            });

            paramPairs.Add(rotationSpring = new JointPhysicsParamPair(Script, ROTATION_SPRING.NAME)
            {
                ValueFormat = ROTATION_SPRING.FORMAT,
                Left = RotationSpringParam(leftContext),
                Right = RotationSpringParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(ROTATION_SPRING.NAME)
                    .Append(" pushes the joint towards its angle target.\n\n")
                    .Append("The angle target is defined by the Up/Down and Left/Right Angle Target parameters.")
                    .ToString(),
            });

            paramPairs.Add(rotationDamper = new JointPhysicsParamPair(Script, ROTATION_DAMPER.NAME)
            {
                ValueFormat = ROTATION_DAMPER.FORMAT,
                Left = RotationDamperParam(leftContext),
                Right = RotationDamperParam(rightContext),
                dependsOnPhysicsRate = true,
                infoText = new StringBuilder()
                    .AppendBold(ROTATION_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint's angle target.\n\n")
                    .Append("The higher the damper, the quicker glutes will stop swinging.")
                    .ToString(),
            });

            paramPairs.Add(inOutSpring = new JointPhysicsParamPair(Script, IN_OUT_SPRING.NAME)
            {
                ValueFormat = IN_OUT_SPRING.FORMAT,
                Left = InOutSpringParam(leftContext),
                Right = InOutSpringParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(IN_OUT_SPRING.NAME)
                    .Append(" pushes the joint towards its position target along the Z axis.\n\n")
                    .Append("Directional force morphing along the forward-back axis depends on In/Out Spring being suitably low for the given glute mass.")
                    .ToString(),
            });

            paramPairs.Add(inOutDamper = new JointPhysicsParamPair(Script, IN_OUT_DAMPER.NAME)
            {
                ValueFormat = IN_OUT_DAMPER.FORMAT,
                Left = InOutDamperParam(leftContext),
                Right = InOutDamperParam(rightContext),
                dependsOnPhysicsRate = true,
                infoText = new StringBuilder()
                    .AppendBold(IN_OUT_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint position target along the Z axis.")
                    .ToString(),
            });

            paramPairs.Add(upDownBaseAngle = new JointPhysicsParamPair(Script, GLUTE_UP_DOWN_BASE_ANGLE.NAME, "Up/Down Angle")
            {
                ValueFormat = GLUTE_UP_DOWN_BASE_ANGLE.FORMAT,
                Left = UpDownBaseAngleParam(leftContext),
                Right = UpDownBaseAngleParam(rightContext),
                usesRealMass = true,
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(GLUTE_UP_DOWN_BASE_ANGLE.NAME)
                    .Append(" adjusts the glute joint's vertical rotation using morphs.\n\n")
                    .Append("Use the calibration state to pause real-time morphing.")
                    .ToString(),
            });

            paramPairs.Add(upAngleLimit = new JointPhysicsParamPair(Script, UP_ANGLE_LIMIT.NAME)
            {
                ValueFormat = UP_ANGLE_LIMIT.FORMAT,
                Left = UpAngleLimitParam(leftContext),
                Right = UpAngleLimitParam(rightContext),
                needsCalibration = NeedsCalibration.IF_LDP_ENABLED,
                infoText = new StringBuilder()
                    .AppendBold(UP_ANGLE_LIMIT.NAME)
                    .Append(" is the maximum angle the joint can rotate up.")
                    .ToString(),
            });

            paramPairs.Add(downAngleLimit = new JointPhysicsParamPair(Script, DOWN_ANGLE_LIMIT.NAME)
            {
                ValueFormat = DOWN_ANGLE_LIMIT.FORMAT,
                Left = DownAngleLimitParam(leftContext),
                Right = DownAngleLimitParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(DOWN_ANGLE_LIMIT.NAME)
                    .Append(" is the maximum angle the joint can rotate down.")
                    .ToString(),
            });

            paramPairs.Add(upDownAngleTarget = new JointPhysicsParamPair(Script, GLUTE_UP_DOWN_ANGLE_TARGET.NAME)
            {
                ValueFormat = GLUTE_UP_DOWN_ANGLE_TARGET.FORMAT,
                Left = UpDownAngleTargetParam(leftContext),
                Right = UpDownAngleTargetParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(GLUTE_UP_DOWN_ANGLE_TARGET.NAME)
                    .Append(" is the joint's vertical target angle. Negative values pull glutes down, positive values push them up.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftBaseAngle = new JointPhysicsParamPair(Script, RIGHT_LEFT_BASE_ANGLE.NAME, "Right/Left Angle")
            {
                ValueFormat = RIGHT_LEFT_BASE_ANGLE.FORMAT,
                Left = RightLeftBaseAngleParam(leftContext),
                Right = RightLeftBaseAngleParam(rightContext),
                usesRealMass = true,
                needsCalibration = NeedsCalibration.ALWAYS,
                rightInverted = true,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_BASE_ANGLE.NAME)
                    .Append(" adjusts the glute joint's sideways rotation using morphs.\n\n")
                    .Append("Use the calibration state to pause real-time morphing.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftAngleLimit = new JointPhysicsParamPair(Script, RIGHT_LEFT_ANGLE_LIMIT.NAME)
            {
                ValueFormat = RIGHT_LEFT_ANGLE_LIMIT.FORMAT,
                Left = RightLeftAngleLimitParam(leftContext),
                Right = RightLeftAngleLimitParam(rightContext),
                needsCalibration = NeedsCalibration.IF_LDP_ENABLED,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_ANGLE_LIMIT.NAME)
                    .Append(" is the maximum angle the joint can rotate sideways.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftAngleTarget = new JointPhysicsParamPair(Script, RIGHT_LEFT_ANGLE_TARGET.NAME)
            {
                ValueFormat = RIGHT_LEFT_ANGLE_TARGET.FORMAT,
                Left = RightLeftAngleTargetParam(leftContext, upDownAngleTarget.Left),
                Right = RightLeftAngleTargetParam(rightContext, upDownAngleTarget.Right),
                needsCalibration = NeedsCalibration.ALWAYS,
                rightInverted = true,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_ANGLE_TARGET.NAME)
                    .Append(" is the joint's sideways target angle. A negative offset pulls glutes apart, while a positive offset pushes them together.")
                    .ToString(),
            });

            jointMass.SetCallbackFunctions();
            foreach(var pair in paramPairs)
            {
                pair.SetCallbackFunctions();
            }
        }

        JointPhysicsParam GluteMassParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            GLUTE_JOINT_MASS.NAME,
            GLUTE_JOINT_MASS.MIN,
            GLUTE_JOINT_MASS.MAX,
            value => SyncMass(context.bodyPart.rigidbody, value)
        );

        JointPhysicsParam CenterOfGravityParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            CENTER_OF_GRAVITY.NAME,
            CENTER_OF_GRAVITY.MIN,
            CENTER_OF_GRAVITY.MAX,
            value => SyncCenterOfGravity(context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.33f,
                multiplierA = 0.75f,
                multiplierB = 1.33f,
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = -0.06f,
                multiplierA = 0.60f,
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.06f,
                multiplierA = 0.60f,
            },
        };

        JointPhysicsParam RotationSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            ROTATION_SPRING.NAME,
            ROTATION_SPRING.MIN,
            ROTATION_SPRING.MAX,
            value => SyncRotationSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 66f,
                multiplierA = 0.80f,
                multiplierB = -0.46f,
                curveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
                curveB = x => Calc.InverseSmoothStep(x, 0.10f, 0.13f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = 6f,
                multiplierA = 0.67f,
                curveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = -9f,
                multiplierA = 0.67f,
                curveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
        };

        JointPhysicsParam RotationDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            ROTATION_DAMPER.NAME,
            ROTATION_DAMPER.MIN,
            ROTATION_DAMPER.MAX,
            value => SyncRotationDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.58f,
                multiplierA = 1.00f,
                multiplierB = -0.55f,
                curveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
                curveB = x => Calc.InverseSmoothStep(x, 0.24f, 0.61f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = -0.03f,
                multiplierA = 8.375f,
                curveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.015f,
                multiplierA = 8.375f,
                curveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
        };

        // JointPhysicsParam RightLeftSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
        //     Script,
        //     context.id,
        //     RIGHT_LEFT_SPRING.NAME,
        //     RIGHT_LEFT_SPRING.MIN,
        //     RIGHT_LEFT_SPRING.MAX,
        //     value => SyncRightLeftSpring(context.bodyPart.joint, context.bodyPart.Rigidbody, value)
        // )
        // {
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 380f,
        //     },
        // };

        // JointPhysicsParam RightLeftDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
        //     Script,
        //     context.id,
        //     RIGHT_LEFT_DAMPER.NAME,
        //     RIGHT_LEFT_DAMPER.MIN,
        //     RIGHT_LEFT_DAMPER.MAX,
        //     value => SyncRightLeftDamper(context.bodyPart.joint, context.bodyPart.Rigidbody, value)
        // )
        // {
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 10f,
        //     },
        // };

        JointPhysicsParam InOutSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            IN_OUT_SPRING.NAME,
            IN_OUT_SPRING.MIN,
            IN_OUT_SPRING.MAX,
            value => SyncInOutSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 900f,
                multiplierA = -0.15f,
                multiplierB = -0.30f,
                curveA = x => 1 - x, // TODO smoothstep
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = 35f,
                multiplierA = -0.28f,
                curveA = x => 1 - x,
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = -52f,
                multiplierA = -0.28f,
                curveA = x => 1 - x,
            },
        };

        JointPhysicsParam InOutDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            IN_OUT_DAMPER.NAME,
            IN_OUT_DAMPER.MIN,
            IN_OUT_DAMPER.MAX,
            value => SyncInOutDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 9.50f,
                multiplierA = 0.67f,
                multiplierB = -0.40f,
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = -2.00f,
                multiplierA = 1.33f,
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = 3.00f,
                multiplierA = 1.33f,
            },
        };

        JointPhysicsParam UpDownBaseAngleParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            GLUTE_UP_DOWN_BASE_ANGLE.NAME,
            GLUTE_UP_DOWN_BASE_ANGLE.MIN,
            GLUTE_UP_DOWN_BASE_ANGLE.MAX,
            context.id == Side.LEFT ? (Action<float>) SyncUpDownBaseAngle : null
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        JointPhysicsParam UpAngleLimitParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            UP_ANGLE_LIMIT.NAME,
            UP_ANGLE_LIMIT.MIN,
            UP_ANGLE_LIMIT.MAX,
            value => SyncUpAngleLimit(context.bodyPart.joint, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 30.0f,
            },
        };

        JointPhysicsParam DownAngleLimitParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            DOWN_ANGLE_LIMIT.NAME,
            DOWN_ANGLE_LIMIT.MIN,
            DOWN_ANGLE_LIMIT.MAX,
            value => SyncDownAngleLimit(context.bodyPart.joint, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = -30.0f,
            },
        };

        static JointPhysicsParam UpDownAngleTargetParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            GLUTE_UP_DOWN_ANGLE_TARGET.NAME,
            GLUTE_UP_DOWN_ANGLE_TARGET.MIN,
            GLUTE_UP_DOWN_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        JointPhysicsParam RightLeftBaseAngleParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_BASE_ANGLE.NAME,
            RIGHT_LEFT_BASE_ANGLE.MIN,
            RIGHT_LEFT_BASE_ANGLE.MAX,
            context.id == Side.LEFT ? (Action<float>) SyncRightLeftBaseAngle : null
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        JointPhysicsParam RightLeftAngleLimitParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_ANGLE_LIMIT.NAME,
            RIGHT_LEFT_ANGLE_LIMIT.MIN,
            RIGHT_LEFT_ANGLE_LIMIT.MAX,
            value => SyncRightLeftAngleLimit(context.bodyPart.joint, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 20.0f,
            },
        };

        JointPhysicsParam RightLeftAngleTargetParam(JointPhysicsContext context, JointPhysicsParam upDownAngleTarget) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_ANGLE_TARGET.NAME,
            RIGHT_LEFT_ANGLE_TARGET.MIN,
            RIGHT_LEFT_ANGLE_TARGET.MAX,
            value => SyncAngleTarget(context.bodyPart.bone, upDownAngleTarget.Value, value)
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        void SyncUpDownBaseAngle(float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            if(value > 0)
            {
                _upAngleMorph.SetValue(Calc.RoundToDecimals(Mathf.InverseLerp(0, GLUTE_UP_DOWN_BASE_ANGLE.MAX, value)));
                _downAngleMorph.Reset();
            }
            else
            {
                _downAngleMorph.SetValue(Calc.RoundToDecimals(Mathf.InverseLerp(0, -GLUTE_UP_DOWN_BASE_ANGLE.MIN, -value)));
                _upAngleMorph.Reset();
            }
        }

        void SyncRightLeftBaseAngle(float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            if(value > 0)
            {
                _sidewaysInMorph.SetValue(Calc.RoundToDecimals(Mathf.InverseLerp(0, RIGHT_LEFT_BASE_ANGLE.MAX, value)));
                _sidewaysOutMorph.Reset();
            }
            else
            {
                _sidewaysOutMorph.SetValue(Calc.RoundToDecimals(Mathf.InverseLerp(0, -RIGHT_LEFT_BASE_ANGLE.MIN, -value)));
                _sidewaysInMorph.Reset();
            }
        }

        void SyncAngleTarget(DAZBone bone, float x, float y)
        {
            if(!Script.enabled)
            {
                return;
            }

            SetBoneRotation(bone, x, y);
        }

        static void SetBoneRotation(DAZBone bone, float x, float y)
        {
            var rotation = bone.baseJointRotation;
            rotation.x = x;
            rotation.y = y;
            if(rotation.IsEqual(bone.baseJointRotation))
            {
                return;
            }

            bone.baseJointRotation = rotation;
        }

        internal override void ResetAll()
        {
            _downAngleMorph.Reset();
            _upAngleMorph.Reset();
            _sidewaysInMorph.Reset();
            _sidewaysOutMorph.Reset();
        }

        internal override void Disable()
        {
            if(Script.initialized != true)
            {
                return;
            }

            ResetAll();
            RestoreOriginalPhysics();
            RestoreLimits(leftContext.bodyPart.joint);
            RestoreLimits(rightContext.bodyPart.joint);
        }

        static void RestoreLimits(ConfigurableJoint joint)
        {
            var highXLimit = joint.highAngularXLimit;
            var lowXLimit = joint.lowAngularXLimit;
            var yLimit = joint.angularYLimit;
            highXLimit.limit = 30;
            lowXLimit.limit = -30;
            yLimit.limit = 20;
            joint.highAngularXLimit = highXLimit;
            joint.lowAngularXLimit = lowXLimit;
            joint.angularYLimit = yLimit;
        }
    }
}
