using ColliderEditor;
using NAT.BM.Contexts;
using NAT.BM.Models;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.Common.UI.Windows;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using static NAT.SoftVerticesGroupName;

namespace NAT.BM.Handlers
{
    sealed class GluteSoftPhysicsHandler : SoftPhysicsHandler
    {
        readonly GluteSoftPhysicsContext _leftContext = new GluteSoftPhysicsContext(Side.LEFT);
        readonly GluteSoftPhysicsContext _rightContext = new GluteSoftPhysicsContext(Side.RIGHT);
        bool _pluginSoftPhysicsToggled;

        internal GluteSoftPhysicsHandler(ScriptBase script) : base(script)
        {
            physicsMesh = Person.GlutePhysicsMesh;
            leftContext = _leftContext;
            rightContext = _rightContext;
            SetupSoftVertices();
            softJointsFixedUpdate = new CustomFixedUpdate(Script.bootstrap.LdpUpdateRateJssc);
            softJointsFixedUpdate.actions += () => UpdateSoftJoints();
            SetupStorables();
            LoadSettings();
            SetupSoftJoints(_leftContext);
            SetupSoftJoints(_rightContext);
            colliderVisualizer = AddColliderVisualizer();
            softPhysicsOn = softPhysicsOnJsb.val;
            softJointDrawService = AddDrawService();
        }

        ColliderVisualizer AddColliderVisualizer() => Script.gameObject.AddComponent<ColliderVisualizer>(c =>
        {
            c.Init(
                leftContext.allSoftJoints.SelectMany(softJoint => softJoint.GetColliders()),
                rightContext.allSoftJoints.SelectMany(softJoint => softJoint.GetColliders()),
                new List<Group>
                {
                    new Group("Off", "$off"), //match nothing
                    new Group("Both glutes", @"PhysicsMeshJoint(left|right) glute?\d+$"),
                    new Group("Left glute", @"PhysicsMeshJointleft glute?\d+$"),
                    new Group("Right glute", @"PhysicsMeshJointright glute?\d+$"),
                }
            );
            c.previewOpacityJSON.val = 0.40f;
            c.previewOpacityJSON.defaultVal = 0.40f;
            c.selectedPreviewOpacityJSON.val = 1;
            c.selectedPreviewOpacityJSON.defaultVal = 1;
            c.groupsJSON.label = "Show Colliders";
            c.groupsJSON.val = "Left glute";
            c.groupsJSON.defaultVal = "Left glute";
            c.highlightMirrorJSON.val = true;
            c.showPreviewsJSON.val = true;

            foreach(string option in new[] { Constant.SELECT, "Other", "All" })
            {
                c.groupsJSON.choices.Remove(option);
            }
        });

        void SetupSoftVertices()
        {
            SetupSoftVerticesGroups();
            SetupSoftVerticesSets(_leftContext);
            SetupSoftVerticesSets(_rightContext);

            leftContext.positionalIndices[OUTER] = new[]
            {
                265, 287, 294, 338, 2388, 4036, 7844, 7872, 7881, 8188, 8191, 8193, 8212, 8246, 22817, 22856,
            };
            leftContext.positionalIndices[MAIN] = new[]
            {
                272, 281, 296, 2390, 2392, 4039, 4042, 7172, 7175, 7177, 7824, 7825, 7826, 7833, 7849, 7866, 7877, 8250, 8729,
            };
            rightContext.positionalIndices[OUTER] = new[]
            {
                11193, 11215, 11222, 11266, 13218, 14812, 18524, 18552, 18561, 18859, 18862, 18864, 18883, 18917, 22819, 22855,
            };
            rightContext.positionalIndices[MAIN] = new[]
            {
                11200, 11209, 11224, 13220, 13222, 14815, 14818, 17879, 17882, 17884,
                18504, 18505, 18506, 18513, 18529, 18546, 18557, 18921, 19400,
            };
        }

        void SetupSoftVerticesGroups()
        {
            var groups = physicsMesh.softVerticesGroups;
            _leftContext.softVerticesGroup = groups.Find(group => group.name == "left glute");
            _rightContext.softVerticesGroup = groups.Find(group => group.name == "right glute");
        }

        static void SetupSoftVerticesSets(GluteSoftPhysicsContext context)
        {
            context.softVerticesSets = new Dictionary<string, DAZPhysicsMeshSoftVerticesSet[]>();
            var sets = context.softVerticesGroup.softVerticesSets;
            foreach(string groupName in new[] { OUTER, MAIN, CENTER, LOWER })
            {
                var indices = VertexIndices.AllSoftVerts(context.id, groupName);
                context.softVerticesSets[groupName] = sets.Where(set => indices.Contains(set.targetVertex)).ToArray();
            }
        }

        void SetupSoftJoints(GluteSoftPhysicsContext context)
        {
            var allSJs = new List<SoftJoint>();
            var outerSJs = new List<SoftJoint>();
            var mainSJs = new List<SoftJoint>();
            var centerSJs = new List<SoftJoint>();
            var lowerSJs = new List<SoftJoint>();

            foreach(var pair in context.softVerticesSets)
            {
                string category = pair.Key;
                var softVerticesSets = pair.Value;
                foreach(var set in softVerticesSets)
                {
                    string id = $"{context.id}.{category}.{set.targetVertex.ToString()}";
                    var softJoint = new SoftJoint(id, set);
                    softJoint.SetPhysicsParams(
                        fatMass,
                        fatSpring,
                        fatDamper,
                        fatColliderRadius,
                        fatColliderDepth,
                        fatColliderFriction,
                        fatDistanceLimit
                    );
                    switch(category)
                    {
                        case OUTER:
                            outerSJs.Add(softJoint);
                            break;
                        case MAIN:
                            mainSJs.Add(softJoint);
                            break;
                        case CENTER:
                            centerSJs.Add(softJoint);
                            break;
                        case LOWER:
                            lowerSJs.Add(softJoint);
                            break;
                    }

                    allSJs.Add(softJoint);
                }
            }

            context.allSoftJoints = allSJs.ToArray();
            context.outerSoftJoints = outerSJs.ToArray();
            context.mainSoftJoints = mainSJs.ToArray();
            context.centerSoftJoints = centerSJs.ToArray();
            context.centerSoftJointsRigidbodies = context.centerSoftJoints.Select(softJoint => softJoint.jointRb).ToArray();
            context.lowerNonGradientJoints = lowerSJs.ToArray();

            var nearbyVerts = VertexIndices.NearbyVerts(context.id);
            var adjacentVerts = VertexIndices.AdjacentVerts(context.id);
            var baseSpringMultiplierVerts = VertexIndices.BaseSpringMultiplierVerts(context.id);
            foreach(var softJoint in allSJs)
            {
                int target = softJoint.softVerticesSet.targetVertex;
                int[] vert;
                softJoint.nearbyJoints = nearbyVerts.TryGetValue(target, out vert)
                    ? allSJs.Where(x => vert.Contains(x.softVerticesSet.targetVertex)).ToArray()
                    : new SoftJoint[] { }; // lower verts do not have nearby verts

                int[] adjacentVert;
                softJoint.adjacentJoints = adjacentVerts.TryGetValue(target, out adjacentVert)
                    ? allSJs.Where(x => adjacentVert.Contains(x.softVerticesSet.targetVertex)).ToArray()
                    : new SoftJoint[] { }; // outer, center and lower verts do not have adjacent verts

                float multiplierVert;
                if(baseSpringMultiplierVerts.TryGetValue(target, out multiplierVert))
                {
                    softJoint.baseSpringMultiplier = multiplierVert;
                }
            }

            var gradientGroupsList = new List<GradientGroup>();
            foreach(int[] indexGroup in VertexIndices.CenterToOuterVerts(context.id))
            {
                var gradientGroup = new GluteGradientGroup
                {
                    center = centerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[0])),
                    main1 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[1])),
                };

                switch(indexGroup.Length)
                {
                    case 3:
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        break;
                    case 4:
                        gradientGroup.main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3]));
                        break;
                    case 5:
                        gradientGroup.main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        gradientGroup.main3 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3]));
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                        break;
                    case 6:
                        gradientGroup.main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2]));
                        gradientGroup.main3 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3]));
                        gradientGroup.main4 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                        gradientGroup.outer = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[5]));
                        break;
                }

                gradientGroupsList.Add(gradientGroup);
            }

            context.gradientGroups = gradientGroupsList;

            var mainNonGradientIndices = VertexIndices.MainNonGradientVerts(context.id);
            context.mainNonGradientJoints = context.allSoftJoints
                .Where(x => mainNonGradientIndices.Contains(x.softVerticesSet.targetVertex))
                .ToArray();

            var lowerReferenceGradientIndices = VertexIndices.LowerReferenceGradientVerts(context.id);
            context.lowerReferenceGradientJoints = context.allSoftJoints
                .Where(x => lowerReferenceGradientIndices.Contains(x.softVerticesSet.targetVertex))
                .ToArray();
        }

        static bool MatchByVertex(SoftJoint softJoint, int vertex) => softJoint.softVerticesSet.targetVertex == vertex;

        void SetupStorables()
        {
            softPhysicsOnJsb = new StorableBool(Constant.GLUTE_SOFT_PHYSICS_ENABLED, physicsMesh.on);
            bool ldpEnabled = Script.bootstrap.GlutesLdpDefaultEnabledJsb.val && physicsMesh.on;
            ldpEnabledJsb = new StorableBool("Local Distortion Physics Enabled", ldpEnabled);
            showColliderPreviewsJsb = new JSONStorableBool("Show Collider Previews", true);

            softPhysicsOnJsb.setCallbackFunction = OnToggleSoftPhysics;
            ldpEnabledJsb.setJSONCallbackFunction = OnToggleLdp;
            showColliderPreviewsJsb.setCallbackFunction = _ =>
            {
                if(Script.uiHandler.IsPhysicsTabActive() && Script.uiHandler.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) Script.uiHandler.GetActiveNestedWindow()).SetShowColliderPreviews();
                }
            };
        }

        void OnToggleSoftPhysics(bool value)
        {
            if(Script.calibrationStateHandler.CalibratingJsb.val)
            {
                Script.Loggr.LogMessage("Cannot enable/disable soft physics while calibrating.");
                softPhysicsOnJsb.valNoCallback = !value;
                return;
            }

            if(value)
            {
                ldpEnabledJsb.val = ldpEnabledJsb.val || Script.bootstrap.GlutesLdpDefaultEnabledJsb.val;
            }
            else
            {
                ldpEnabledJsb.val = false;
            }

            if(!_reverseSyncingSoftPhysics)
            {
                _pluginSoftPhysicsToggled = true;
                if(Script.IsEnabledAndActive())
                {
                    physicsMesh.on = value;
                }
                else
                {
                    Script.Loggr.LogMessage($"Enable the plugin to set {softPhysicsOnJsb.name}.");
                }

                if(Script.uiHandler.IsMainTabActive())
                {
                    Script.uiHandler.mainWindow.Refresh();
                }
            }

            _reverseSyncingSoftPhysics = false;
        }

        void OnToggleLdp(JSONStorableBool jsb)
        {
            if(!Script.IsEnabledAndActive())
            {
                Script.Loggr.LogMessage($"Enable the plugin to set {jsb.name}.");
                return;
            }

            var storable = (StorableBool) jsb;
            if(storable.val)
            {
                if(!softPhysicsOnJsb.val)
                {
                    Script.Loggr.LogMessage("Enable glute soft physics to use local distortion physics.");
                }
                else
                {
                    Script.SetCalibrationPending(storable);
                }
            }
            else
            {
                if(softPhysicsOn)
                {
                    Script.SetCalibrationPending(storable);
                }
            }

            softJointDrawService.visualizeLdpJsb.val = jsb.val;

            if(Script.uiHandler.IsMainTabActive())
            {
                Script.uiHandler.mainWindow.Refresh();
            }
        }

        protected override void LoadSettings()
        {
            softParamPairs.Add(fatSpring = new SoftPhysicsParamPair(Script, this, FAT_SPRING.NAME)
            {
                Left = FatSpringParam(_leftContext),
                Right = FatSpringParam(_rightContext),
                ValueFormat = FAT_SPRING.FORMAT,
                visualizeRelativeChange = true,
                visualizationLimit = 125f,
                clothingAdjustmentVisualizationLimit = 1000f,
                distortionLimit = 500f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_SPRING.NAME)
                    .AppendLine(" holds each soft joint in its target position.")
                    .AppendLine()
                    .Append("Low fat spring makes glute fat soft and slow. High fat spring makes it rigid and quick to return to its normal shape.")
                    .ToString(),
            });

            softParamPairs.Add(fatDamper = new SoftPhysicsParamPair(Script, this, FAT_DAMPER.NAME)
            {
                Left = FatDamperParam(_leftContext),
                Right = FatDamperParam(_rightContext),
                ValueFormat = FAT_DAMPER.FORMAT,
                visualizeRelativeChange = true,
                visualizationLimit = 2.50f,
                clothingAdjustmentVisualizationLimit = 5.00f,
                distortionLimit = 5.00f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_DAMPER.NAME)
                    .AppendLine(" reduces oscillation of each soft joint around its target position.")
                    .AppendLine()
                    .Append("Low fat damper makes glute fat jiggle more easily.")
                    .ToString(),
            });

            softParamPairs.Add(fatMass = new SoftPhysicsParamPair(Script, this, FAT_MASS.NAME)
            {
                Left = FatMassParam(_leftContext),
                Right = FatMassParam(_rightContext),
                ValueFormat = FAT_MASS.FORMAT,
                visualizeRelativeChange = true,
                visualizationLimit = 0.150f,
                clothingAdjustmentVisualizationLimit = 0.150f,
                distortionLimit = 0.300f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_MASS.NAME)
                    .Append(" determines the mass of each soft joint and the overall density of glute fat.")
                    .Append(" Increasing glute size without increasing fat mass reduces density.")
                    .ToString(),
            });

            softParamPairs.Add(fatColliderRadius = new SoftPhysicsParamPair(Script, this, FAT_COLLIDER_RADIUS.NAME)
            {
                Left = FatColliderRadiusParam(_leftContext),
                Right = FatColliderRadiusParam(_rightContext),
                ValueFormat = FAT_COLLIDER_RADIUS.FORMAT,
                usesRealMass = true,
                hasSoftColliderVisualization = true,
                visualizeRelativeChange = false,
                visualizationLimit = 0.0100f,
                clothingAdjustmentVisualizationLimit = 0.0200f,
                distortionLimit = 0.0200f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_RADIUS.NAME)
                    .AppendLine(" is a scaling factor for the size of the colliders connected to each soft joint.")
                    .AppendLine()
                    .Append("The actual size varies per collider and automatically scales with glute size.")
                    .ToString(),
            });

            softParamPairs.Add(fatColliderDepth = new SoftPhysicsParamPair(Script, this, FAT_COLLIDER_DEPTH.NAME)
            {
                ValueFormat = FAT_COLLIDER_DEPTH.FORMAT,
                Left = FatColliderDepthParam(_leftContext),
                Right = FatColliderDepthParam(_rightContext),
                hasSoftColliderVisualization = true,
                visualizeRelativeChange = false,
                visualizationLimit = 0.0040f,
                clothingAdjustmentVisualizationLimit = 0.0040f,
                distortionLimit = 0.0040f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_DEPTH.NAME)
                    .AppendLine(" offsets collider positions relative to skin surface")
                    .AppendLine()
                    .Append("Negative values pull colliders out from the glute, positive values push them into the glute.")
                    .ToString(),
            });

            softParamPairs.Add(fatColliderFriction = new SoftPhysicsParamPair(Script, this, FAT_COLLIDER_FRICTION.NAME)
            {
                ValueFormat = FAT_COLLIDER_FRICTION.FORMAT,
                Left = FatColliderFrictionParam(_leftContext),
                Right = FatColliderFrictionParam(_rightContext),
                hasSoftColliderVisualization = true,
                visualizeRelativeChange = true,
                visualizationLimit = 0.500f,
                clothingAdjustmentVisualizationLimit = 0.500f,
                distortionLimit = 1.000f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_FRICTION.NAME)
                    .Append(" sets the stickiness of individual soft colliders. Click below to configure adaptive friction and to offset the amount of friction.")
                    .ToString(),
            });

            softParamPairs.Add(fatDistanceLimit = new SoftPhysicsParamPair(Script, this, FAT_DISTANCE_LIMIT.NAME)
            {
                ValueFormat = FAT_DISTANCE_LIMIT.FORMAT,
                Left = FatDistanceLimitParam(_leftContext),
                Right = FatDistanceLimitParam(_rightContext),
                usesRealMass = true,
                visualizeRelativeChange = true,
                visualizationLimit = 0.020f,
                clothingAdjustmentVisualizationLimit = 0.020f,
                distortionLimit = 0.020f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_DISTANCE_LIMIT.NAME)
                    .Append(" is the maximum distance each soft joint can move away from its target position.")
                    .ToString(),
            });

            paramPairs.Add(fatBackForce = new JointPhysicsParamPair(Script, FAT_BACK_FORCE.NAME)
            {
                ValueFormat = FAT_BACK_FORCE.FORMAT,
                Left = FatBackForceParam(_leftContext),
                Right = FatBackForceParam(_rightContext),
                infoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE.NAME)
                    .Append(" applies a force to the glute joint based on the movement of soft joints. A high value can cause an out of control feedback loop.")
                    .ToString(),
            });

            paramPairs.Add(fatBackForceMaxForce = new JointPhysicsParamPair(Script, FAT_BACK_FORCE_MAX_FORCE.NAME)
            {
                ValueFormat = FAT_BACK_FORCE_MAX_FORCE.FORMAT,
                Left = FatBackForceMaxForceParam(_leftContext),
                Right = FatBackForceMaxForceParam(_rightContext),
                infoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE_MAX_FORCE.NAME)
                    .Append(" sets an upper limit on the magnitude of back force.")
                    .ToString(),
            });

            paramPairs.Add(fatBackForceThreshold = new JointPhysicsParamPair(Script, FAT_BACK_FORCE_THRESHOLD.NAME)
            {
                ValueFormat = FAT_BACK_FORCE_THRESHOLD.FORMAT,
                Left = FatBackForceThresholdParam(_leftContext),
                Right = FatBackForceThresholdParam(_rightContext),
                infoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE_THRESHOLD.NAME)
                    .AppendLine(" sets a minimum distance each soft joint needs to move for back force to be applied.")
                    .AppendLine()
                    .AppendFormat("Along with {0}, this can be used to prevent an out of control feedback loop.", FAT_BACK_FORCE_MAX_FORCE.NAME)
                    .ToString(),
            });

            foreach(var pair in softParamPairs)
            {
                pair.SetCallbackFunctions();
            }

            foreach(var pair in paramPairs)
            {
                pair.SetCallbackFunctions();
            }
        }

        internal override Rigidbody[] GetLeftPositionalRigidbodies(string group) => GetPositionalRigidbodies(_leftContext, group);
        internal override Rigidbody[] GetRightPositionalRigidbodies(string group) => GetPositionalRigidbodies(_rightContext, group);

        static Rigidbody[] GetPositionalRigidbodies(GluteSoftPhysicsContext context, string group)
        {
            var rigidbodies = new List<Rigidbody>();
            int[] indices = context.positionalIndices[group];
            foreach(var set in context.softVerticesSets[group])
            {
                if(indices.Contains(set.targetVertex))
                {
                    rigidbodies.Add(set.jointRB);
                }
            }

            return rigidbodies.ToArray();
        }

        static float FatSpringSoftnessCurve(float x) => Calc.InverseSmoothStep(x, 0.00f, 0.11f);

        SoftPhysicsParam NewSoftPhysicsParameter(GluteSoftPhysicsContext context, string name, float min, float max) => new SoftPhysicsParam(
            Script.DomainName,
            context.id,
            name,
            min,
            max,
            (param, additionalOffset) => SyncGradientParam(context, param, additionalOffset)
        );

        SoftPhysicsParam FatSpringParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_SPRING.NAME, FAT_SPRING.MIN, FAT_SPRING.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.895f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.878f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                baseValue = 50f,
                multiplierA = 0.00f,
                multiplierB = -0.75f,
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                baseValue = -125f,
                multiplierA = 0.00f,
                multiplierB = -0.960f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.40f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.60f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.30f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.85f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.30f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.85f);

            return parameter;
        }

        SoftPhysicsParam FatDamperParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_DAMPER.NAME, FAT_DAMPER.MIN, FAT_DAMPER.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 2.16f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 2.16f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                baseValue = -1.92f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.96f,
                multiplierA = 0.18f,
                multiplierB = -0.59f,
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.20f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.60f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(-0.40f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(-0.40f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        SoftPhysicsParam FatMassParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_MASS.NAME, FAT_MASS.MIN, FAT_MASS.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.096f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 0.096f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                baseValue = -0.081f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.081f,
                multiplierA = 0.33f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.50f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(-0.25f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        SoftPhysicsParam FatColliderRadiusParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_COLLIDER_RADIUS.NAME, FAT_COLLIDER_RADIUS.MIN, FAT_COLLIDER_RADIUS.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.0236f,
                multiplierA = 1.03f,
                multiplierB = 0.20f,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 0.0272f,
                multiplierA = 1.03f,
                multiplierB = 0.20f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.10f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.10f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.70f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.40f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.70f);

            return parameter;
        }

        SoftPhysicsParam FatColliderDepthParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_COLLIDER_DEPTH.NAME, FAT_COLLIDER_DEPTH.MIN, FAT_COLLIDER_DEPTH.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = -0.0025f,
                multiplierB = 0.60f,
                curveB = x => -x,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = -0.0025f,
                multiplierB = 0.60f,
                curveB = x => -x,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(1.00f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.90f);

            return parameter;
        }

        SoftPhysicsParam FatColliderFrictionParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_COLLIDER_FRICTION.NAME, FAT_COLLIDER_FRICTION.MIN, FAT_COLLIDER_FRICTION.MAX);
            parameter.PrimaryConfig = new MultiplierConfig();
            parameter.SecondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.25f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.50f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.30f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.50f);

            return parameter;
        }

        SoftPhysicsParam FatDistanceLimitParam(GluteSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_DISTANCE_LIMIT.NAME, FAT_DISTANCE_LIMIT.MIN, FAT_DISTANCE_LIMIT.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.0204f,
                multiplierA = 1.17f,
                multiplierB = 0.26f,
                curveA = x => Calc.InverseSmoothStep(x, 0.22f, 0.10f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 0.0255f,
                multiplierA = 0.82f,
                multiplierB = 0.26f,
                curveA = x => Calc.InverseSmoothStep(x, 0.22f, 0.10f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.70f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        JointPhysicsParam FatBackForceParam(GluteSoftPhysicsContext context)
        {
            var parameter = new JointPhysicsParam(
                context.id,
                FAT_BACK_FORCE.NAME,
                FAT_BACK_FORCE.MIN,
                FAT_BACK_FORCE.MAX,
                value => SyncBackForce(context, value)
            )
            {
                PrimaryConfig = new MultiplierConfig
                {
                    baseValue = 2.88f,
                    multiplierA = 2.19f,
                    multiplierB = 8.75f,
                    curveA = x => Calc.InverseSmoothStep(x, 0.50f, 0.88f),
                    curveB = x => 1 - Calc.InverseSmoothStep(x, 0.58f, 0.07f),
                },
                SecondaryConfig = new MultiplierConfig
                {
                    baseValue = 1.44f,
                    multiplierA = 1.10f,
                    multiplierB = 17.50f,
                    curveA = x => Calc.InverseSmoothStep(x, 0.50f, 0.88f),
                    curveB = x => 1 - Calc.InverseSmoothStep(x, 0.58f, 0.07f),
                },
                QuicknessConfig = new MultiplierConfig
                {
                    baseValue = -0.36f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                    curveA = x => Calc.InverseSmoothStep(x, 0.50f, 0.88f),
                },
                SlownessConfig = new MultiplierConfig
                {
                    baseValue = 0.72f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        JointPhysicsParam FatBackForceMaxForceParam(GluteSoftPhysicsContext context)
        {
            var parameter = new JointPhysicsParam(
                context.id,
                FAT_BACK_FORCE_MAX_FORCE.NAME,
                FAT_BACK_FORCE_MAX_FORCE.MIN,
                FAT_BACK_FORCE_MAX_FORCE.MAX,
                value => SyncBackForceMaxForce(context, value)
            )
            {
                PrimaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                SecondaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        JointPhysicsParam FatBackForceThresholdParam(GluteSoftPhysicsContext context)
        {
            var parameter = new JointPhysicsParam(
                context.id,
                FAT_BACK_FORCE_THRESHOLD.NAME,
                FAT_BACK_FORCE_THRESHOLD.MIN,
                FAT_BACK_FORCE_THRESHOLD.MAX,
                value => SyncBackForceThresholdDistance(context, value)
            )
            {
                PrimaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                SecondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };
            return parameter;
        }

        void SyncGradientParam(GluteSoftPhysicsContext context, SoftPhysicsParam param, float additionalOffset)
        {
            if(Script.initialized != true || !Script.enabled || Script.restoringFromJSON)
            {
                // Syncing parameter before first time calibration (e.g. due to offset storable callback) results in NaN errors
                return;
            }

            string name = param.paramName;
            foreach(var group in context.gradientGroups)
            {
                group.SetGradientBaseValues(name, param.gradientConfig, param.baseValueJsf.val + param.offsetJsf.val);
                group.SetAdditionalOffsets(name, additionalOffset);
            }

            /* Calculate remaining physics values for joints not included in gradient.
             * Values are based on selected neighboring joints.
             */
            foreach(var joint in context.mainNonGradientJoints)
            {
                float average = 0f;
                foreach(var adjacentJoint in joint.adjacentJoints)
                {
                    average += adjacentJoint.GetBaseValue(name);
                }

                average /= joint.adjacentJoints.Length;

                joint.SetBaseValue(name, average);
                joint.SetAdditionalOffset(name, additionalOffset);
            }

            /* Calculate remaining physics value for lower non-gradient joints close to the vagina. */
            float averageLowerReferenceJointsValue = 0f;
            var joints = context.lowerReferenceGradientJoints;
            foreach(var referenceJoint in joints)
            {
                averageLowerReferenceJointsValue += referenceJoint.GetBaseValue(name);
            }

            averageLowerReferenceJointsValue /= joints.Length;

            foreach(var joint in context.lowerNonGradientJoints)
            {
                joint.SetBaseValue(name, averageLowerReferenceJointsValue);
                joint.SetAdditionalOffset(name, additionalOffset);
            }
        }

        void SyncBackForce(GluteSoftPhysicsContext context, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            context.softVerticesGroup.jointBackForce = value;
        }

        void SyncBackForceMaxForce(GluteSoftPhysicsContext context, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            context.softVerticesGroup.jointBackForceMaxForce = value;
        }

        void SyncBackForceThresholdDistance(GluteSoftPhysicsContext context, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            context.softVerticesGroup.jointBackForceThresholdDistance = value;
        }

        bool _reverseSyncingSoftPhysics;

        internal override void ReverseSyncSoftPhysicsOn()
        {
            if(!_pluginSoftPhysicsToggled)
            {
                _reverseSyncingSoftPhysics = true;
                softPhysicsOnJsb.val = physicsMesh.on;
            }

            _pluginSoftPhysicsToggled = false;
        }

        internal override void SyncSoftPhysicsOn()
        {
            softPhysicsOn = Script.bootstrap.PersonMonitor.IsGluteSoftPhysicsOn();
            ldpOn = softPhysicsOn && ldpEnabledJsb.val;
            if(!softPhysicsSynced)
            {
                if(softPhysicsOn)
                {
                    Script.hardColliderHandler.OnSoftPhysicsEnabled(leftContext.allSoftJoints, rightContext.allSoftJoints);
                }
                else
                {
                    Script.hardColliderHandler.OnSoftPhysicsDisabled();
                }

                Script.SetCalibrationTargetFunctions(softPhysicsOn);
            }

            softPhysicsSynced = true;
        }

        internal override void Enable()
        {
            EnableMultiplyFriction();
        }

        internal override void Disable()
        {
            if(Script.initialized != true)
            {
                return;
            }

            colliderVisualizer.enabled = false;
            softJointDrawService.enabled = false;

            RestoreOriginalFriction();
            RestoreOriginalPhysics();
        }
    }
}
