using NAT.Common;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using UnityEngine;

namespace NAT.BM.Models
{
    sealed class GluteGradientGroup : GradientGroup
    {
        // TODO when center is shared between multiple gradient groups
        internal SoftJoint center;
        internal SoftJoint main1;
        internal SoftJoint main2;
        internal SoftJoint main3;
        internal SoftJoint main4;
        internal SoftJoint outer;

        // ratios of distances to total distance
        float _main1Ratio;
        float _main2Ratio;
        float _main3Ratio;
        float _main4Ratio;

        internal override void CalculateDistances()
        {
            float total;
            float mainToOuter;
            float centerToMain1 = Vector3.Distance(center.jointRb.position, main1.jointRb.position);
            if(main2 != null)
            {
                float main1ToMain2 = Vector3.Distance(main1.jointRb.position, main2.jointRb.position);
                if(main3 != null)
                {
                    float main2ToMain3 = Vector3.Distance(main2.jointRb.position, main3.jointRb.position);
                    if(main4 != null)
                    {
                        float main3ToMain4 = Vector3.Distance(main3.jointRb.position, main4.jointRb.position);
                        mainToOuter = Vector3.Distance(main4.jointRb.position, outer.jointRb.position);
                        total = centerToMain1 + main1ToMain2 + main2ToMain3 + main3ToMain4 + mainToOuter;
                        _main4Ratio = (centerToMain1 + main1ToMain2 + main2ToMain3 + main3ToMain4) / total;
                    }
                    else
                    {
                        mainToOuter = Vector3.Distance(main3.jointRb.position, outer.jointRb.position);
                        total = centerToMain1 + main1ToMain2 + main2ToMain3 + mainToOuter;
                    }

                    _main3Ratio = (centerToMain1 + main1ToMain2 + main2ToMain3) / total;
                }
                else
                {
                    mainToOuter = Vector3.Distance(main2.jointRb.position, outer.jointRb.position);
                    total = centerToMain1 + main1ToMain2 + mainToOuter;
                }

                _main2Ratio = (centerToMain1 + main1ToMain2) / total;
            }
            else
            {
                mainToOuter = Vector3.Distance(main1.jointRb.position, outer.jointRb.position);
                total = centerToMain1 + mainToOuter;
            }

            _main1Ratio = centerToMain1 / total;
        }

        /* Value is parameter's overall base value based on softness, quickness and glute mass.
         * Individual joints get adjusted values based on the gradient.
         */
        internal override void SetGradientBaseValues(string paramName, GradientConfig gradientConfig, float value)
        {
            center.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 0));
            main1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main1Ratio));
            main2?.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main2Ratio));
            main3?.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main3Ratio));
            main4?.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main4Ratio));
            outer.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
        }

        // TODO called unnecessarily multiple times with same inputs, could cache values
        // Gradient starts from outer (ratio 1) and ends in center (ratio 0)
        static float CalculateJointBaseValue(GradientConfig config, float value, float ratio)
        {
            float middle = 1 -
                Mathf.Lerp(
                    0.01f,
                    0.99f,
                    Mathf.InverseLerp(config.middleShiftJsf.min, config.middleShiftJsf.max, config.middleShiftJsf.val)
                );

            float result;

            if(ratio > middle)
            {
                float smoothRatio = Calc.InverseSmoothStepWithMax(
                    ratio - middle,
                    1 - middle,
                    config.outerToMiddleCurvature,
                    config.outerToMiddleMidPoint
                );
                result = Mathf.Lerp(value, value * config.outerMultiplierJsf.val, smoothRatio);
            }
            else
            {
                result = config.centerMultiplierJsf.val * value;
            }

            return result;
        }

        internal override void SetAdditionalOffsets(string paramName, float value)
        {
            center.SetAdditionalOffset(paramName, value);
            main1.SetAdditionalOffset(paramName, value);
            main2?.SetAdditionalOffset(paramName, value);
            main3?.SetAdditionalOffset(paramName, value);
            main4?.SetAdditionalOffset(paramName, value);
            outer.SetAdditionalOffset(paramName, value);
        }
    }
}
