using ColliderEditor.Extensions;
using ColliderEditor.Models;
using UnityEngine;
using System;
using System.Collections.Generic;
using System.Linq;

namespace ColliderEditor
{
    /// <summary>
    /// Collider Visualizer
    /// Based on:
    ///
    /// Collider Editor
    /// By Acidbubbles, ProjectCanyon and via5
    /// Configures and customizes collisions (rigidbodies and colliders)
    /// Source: https://github.com/acidbubbles/vam-collider-editor
    /// </summary>
    sealed class ColliderVisualizer : MonoBehaviour
    {
        const string NO_SELECTION_LABEL = "Select...";
        const string ALL_LABEL = "All";

        // ReSharper disable NotAccessedField.Global MemberCanBePrivate.Global
        internal JSONStorableBool showPreviewsJSON;
        internal JSONStorableBool xRayPreviewsJSON;
        internal JSONStorableBool xRayPreviewsOffJSON;
        internal JSONStorableFloat previewOpacityJSON;
        internal JSONStorableFloat selectedPreviewOpacityJSON;
        internal JSONStorableFloat relativeXRayOpacityJSON;
        internal JSONStorableBool highlightMirrorJSON;
        internal JSONStorableStringChooser groupsJSON;
        internal JSONStorableStringChooser typesJSON;
        internal JSONStorableStringChooser selectablesJSON;
        internal SelectablesList SelectablesList { get; private set; }
        internal readonly ColliderPreviewConfig config = new ColliderPreviewConfig();
        // ReSharper restore NotAccessedField.Global MemberCanBePrivate.Global

        readonly List<IModel> _filteredSelectables = new List<IModel>();
        IModel _selected, _selectedMirror;
        bool _initialized;

        internal void Init(IEnumerable<Collider> leftColliders, IEnumerable<Collider> rightColliders, List<Group> groups)
        {
            try
            {
                SelectablesList = new SelectablesList(leftColliders.ToList(), rightColliders.ToList(), config, groups);
                SetupStorables();
                enabled = false;
                _initialized = true;
            }
            catch(Exception e)
            {
                LogError(nameof(Init), e.ToString());
            }
        }

        void SetupStorables()
        {
            showPreviewsJSON = new JSONStorableBool("showPreviews", ColliderPreviewConfig.DEFAULT_PREVIEWS_ENABLED, value =>
            {
                config.previewsEnabled = value;
                UpdateAllPreviewsFromConfig();
            });

            xRayPreviewsJSON = new JSONStorableBool("xRayPreviews", ColliderPreviewConfig.DEFAULT_X_RAY_PREVIEWS, value =>
            {
                config.xRayPreviews = value;
                UpdateAllPreviewsFromConfig();
            });

            xRayPreviewsOffJSON = new JSONStorableBool("xRayPreviewsOff", !ColliderPreviewConfig.DEFAULT_X_RAY_PREVIEWS,
                value => { xRayPreviewsJSON.val = !value; });

            previewOpacityJSON = new JSONStorableFloat("previewOpacity", ColliderPreviewConfig.DEFAULT_PREVIEWS_OPACITY, value =>
            {
                float alpha = value.ExponentialScale(ColliderPreviewConfig.EXPONENTIAL_SCALE_MIDDLE, 1f);
                config.previewsOpacity = alpha;
                UpdateAllPreviewsFromConfig();
            }, 0f, 1f);

            selectedPreviewOpacityJSON = new JSONStorableFloat("selectedPreviewOpacity", ColliderPreviewConfig.DEFAULT_SELECTED_PREVIEW_OPACITY,
                value =>
                {
                    float alpha = value.ExponentialScale(ColliderPreviewConfig.EXPONENTIAL_SCALE_MIDDLE, 1f);
                    config.selectedPreviewsOpacity = alpha;
                    UpdatePreviewsFromConfig(_selected);
                    UpdatePreviewsFromConfig(_selectedMirror);
                }, 0f, 1f);

            relativeXRayOpacityJSON = new JSONStorableFloat("relativeXRayOpacity", ColliderPreviewConfig.DEFAULT_RELATIVE_X_RAY_OPACITY, value =>
            {
                config.relativeXRayOpacity = value;
                UpdatePreviewsFromConfig(_selected);
                UpdatePreviewsFromConfig(_selectedMirror);
            }, 0f, 1f);

            highlightMirrorJSON = new JSONStorableBool(
                "highlightMirror",
                ColliderPreviewConfig.DEFAULT_HIGHLIGHT_MIRROR,
                value => config.highlightMirror = value
            );

            var groups = new List<string> { NO_SELECTION_LABEL };
            groups.AddRange(SelectablesList.groups.Select(e => e.name).Distinct());
            groups.Add(ALL_LABEL);
            groupsJSON = new JSONStorableStringChooser("Group", groups, groups[0], "Group")
            {
                setCallbackFunction = _ => UpdateFilter(),
            };

            var types = new List<string> { NO_SELECTION_LABEL };
            types.AddRange(SelectablesList.all.Select(e => e.Type).Distinct());
            types.Add(ALL_LABEL);
            typesJSON = new JSONStorableStringChooser("Type", types, types[0], "Type")
            {
                setCallbackFunction = _ => UpdateFilter(),
            };

            selectablesJSON = new JSONStorableStringChooser(
                "Select",
                new List<string>(),
                new List<string>(),
                "",
                "Select"
            )
            {
                setCallbackFunction = id =>
                {
                    if(id == null)
                    {
                        return;
                    }

                    IModel val;
                    SelectModel(SelectablesList.byUuid.TryGetValue(id, out val) ? val : null);
                },
            };
        }

        void SelectModel(IModel val)
        {
            Deselect(ref _selected);
            Deselect(ref _selectedMirror);

            if(val == null)
            {
                selectablesJSON.valNoCallback = "";
                return;
            }

            selectablesJSON.valNoCallback = val.Id;

            if(val.Shown)
            {
                Select(ref _selected, val);
            }

            if(config.highlightMirror && val.MirrorModel != null && val.MirrorModel.Shown)
            {
                Select(ref _selectedMirror, val.MirrorModel, false, true, false);
            }
        }

        void Deselect(ref IModel model)
        {
            if(model == null)
            {
                return;
            }

            model.Selected = false;
            model.Highlighted = false;
            model.Shown = _filteredSelectables.Contains(model);
            UpdatePreviewsFromConfig(model);
            model = null;
        }

        // ReSharper disable once RedundantAssignment
        void Select(ref IModel selected, IModel val, bool shown = true, bool highlighted = true, bool showUI = true)
        {
            selected = val;
            selected.Shown = shown || _filteredSelectables.Contains(selected);
            selected.Highlighted = highlighted;
            selected.Selected = showUI;
            UpdatePreviewsFromConfig(selected);
        }

        void UpdateFilter()
        {
            if(!_initialized || !enabled)
            {
                return;
            }

            try
            {
                _filteredSelectables.Clear();
                HideAllShown();

                if(groupsJSON.val != ALL_LABEL && groupsJSON.val != NO_SELECTION_LABEL)
                {
                    _filteredSelectables.AddRange(SelectablesList.all.Where(e => e.Groups.Any(group => group?.name == groupsJSON.val)));
                }

                if(typesJSON.val != ALL_LABEL && typesJSON.val != NO_SELECTION_LABEL)
                {
                    _filteredSelectables.AddRange(SelectablesList.all.Where(e => e.Type == typesJSON.val));
                }

                selectablesJSON.choices = _filteredSelectables.Select(x => x.Id).ToList();
                selectablesJSON.displayChoices = _filteredSelectables.Select(x => x.Label).ToList();

                foreach(var model in _filteredSelectables)
                {
                    model.Shown = true;
                    model.UpdatePreviewsFromConfig();
                }
            }
            catch(Exception e)
            {
                LogError(nameof(UpdateFilter), e.ToString());
            }
        }

        void HideAllShown()
        {
            foreach(var model in SelectablesList.all)
            {
                if(model.Shown)
                {
                    model.Shown = false;
                    model.UpdatePreviewsFromConfig();
                }
            }
        }

        void UpdateAllPreviewsFromConfig()
        {
            if(!_initialized || !enabled)
            {
                return;
            }

            foreach(var model in SelectablesList.all)
            {
                model.UpdatePreviewsFromConfig();
            }
        }

        void UpdatePreviewsFromConfig(IModel model)
        {
            if(!_initialized || !enabled)
            {
                return;
            }

            model?.UpdatePreviewsFromConfig();
        }

        void Update()
        {
            if(!_initialized || !enabled)
            {
                return;
            }

            if(showPreviewsJSON.val)
            {
                foreach(var model in SelectablesList.all)
                {
                    model.SyncPreviews();
                }
            }
        }

        void OnEnable()
        {
            if(!_initialized)
            {
                return;
            }

            try
            {
                UpdateFilter();
            }
            catch(Exception e)
            {
                LogError(nameof(OnEnable), e.ToString());
            }
        }

        void OnDisable()
        {
            if(!_initialized)
            {
                return;
            }

            try
            {
                foreach(var model in SelectablesList.all)
                {
                    model.DestroyPreviews();
                }
            }
            catch(Exception e)
            {
                LogError(nameof(OnDisable), e.ToString());
            }
        }

        void OnDestroy()
        {
            OnDisable(); // Should get called automatically but for some reason does not
        }

        static void LogError(string method, string message)
        {
            SuperController.LogError($"{nameof(ColliderVisualizer)}.{method}: {message}");
        }
    }
}
