using NAT.Common.UI;
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Components
{
    sealed class UIMod
    {
        readonly List<RectTransformReposition> _repositionedRects = new List<RectTransformReposition>();
        readonly List<GameObject> _inactivatedUIGameObjects = new List<GameObject>();
        readonly List<GameObject> _customUIGameObjects = new List<GameObject>();

        readonly IModParent _modParent;
        readonly string _targetName;
        UnityEventsListener _containerUIEventsListener;
        UnityEventsListener _targetUIEventsListener;

        internal bool IsVisible() => _targetUIEventsListener && _targetUIEventsListener.IsEnabled;

        readonly Func<UIMod, IEnumerator> _applyChanges;
        readonly bool _useDisabledStyle;

        internal UIMod(
            IModParent modParent,
            string targetTransformName,
            Func<UIMod, IEnumerator> changesFunc,
            bool useDisabledStyle = false
        )
        {
            _modParent = modParent;
            _targetName = targetTransformName;
            _applyChanges = changesFunc;
            _useDisabledStyle = useDisabledStyle;
        }

        internal void AddTargetUIEventsListener()
        {
            _targetUIEventsListener = Target.gameObject.AddComponent<UnityEventsListener>();
        }

        bool _isApplied;

        internal void DeferApply()
        {
            if(_isApplied)
            {
                return;
            }

            _isApplied = true;

            try
            {
                var containerGameObj = _modParent.GetParentUI().gameObject;
                if(containerGameObj.activeInHierarchy)
                {
                    _modParent.StartCoroutine(ApplyCo());
                }
                /* Add listener to wait for the transform to be active in hierarchy */
                else
                {
                    _containerUIEventsListener = containerGameObj.AddComponent<UnityEventsListener>(listener =>
                    {
                        listener.enabledHandlers += () => _modParent.StartCoroutine(ApplyCo());
                    });
                }
            }
            catch(Exception e)
            {
                Debug.Log($"UIMod failed to defer apply: {e}");
            }
        }

        bool _modified;
        internal Transform Target { get; private set; }

        IEnumerator ApplyCo()
        {
            if(_modified)
            {
                yield break;
            }

            UnityEngine.Object.DestroyImmediate(_containerUIEventsListener);
            _containerUIEventsListener = null;

            /* Set modified immediately to prevent double apply in case eventListener
             * gets enabled multiple times in a short period of time
             */
            _modified = true;

            if(_targetName != null)
            {
                float timeout = Time.unscaledTime + 1;
                while(Time.unscaledTime < timeout)
                {
                    yield return new WaitForSeconds(0.1f);
                    Target = _modParent.GetParentUI().Find(_targetName);
                }

                if(!Target)
                {
                    Loggr.Error($"Failed to modify UI on target {_targetName} - could not find transform");
                    yield break;
                }
            }

            yield return _applyChanges(this);
            if(!_modParent.IsEnabledAndActive())
            {
                Disable();
            }
        }

        internal void InactivateChildren()
        {
            foreach(Transform child in Target)
            {
                var go = child.gameObject;
                if(go.activeSelf)
                {
                    _inactivatedUIGameObjects.Add(go);
                    go.SetActive(false);
                }
            }
        }

        internal void AddCustomObject(GameObject go)
        {
            _customUIGameObjects.Add(go);
        }

        internal void RepositionTransform(Transform t, Vector2 originalPosition, Vector2 newPosition)
        {
            _repositionedRects.Add(new RectTransformReposition(t.GetComponent<RectTransform>(), originalPosition, newPosition));
        }

        internal void Enable()
        {
            if(!_modified)
            {
                DeferApply();
            }
            else
            {
                foreach(var go in _customUIGameObjects)
                {
                    if(_useDisabledStyle)
                    {
                        var elements = go.GetComponentsInChildren<UIDynamic>();
                        foreach(var element in elements)
                        {
                            element.SetActiveStyle(true, false, true);
                        }
                    }
                    else
                    {
                        go.SetActive(true);
                    }
                }

                foreach(var go in _inactivatedUIGameObjects)
                {
                    go.SetActive(false);
                }

                foreach(var repositionedRect in _repositionedRects)
                {
                    repositionedRect.Apply();
                }
            }
        }

        internal void Disable()
        {
            foreach(var go in _customUIGameObjects)
            {
                if(_useDisabledStyle)
                {
                    var elements = go.GetComponentsInChildren<UIDynamic>();
                    foreach(var element in elements)
                    {
                        element.SetActiveStyle(false, true);
                    }
                }
                else
                {
                    go.SetActive(false);
                }
            }

            foreach(var go in _inactivatedUIGameObjects)
            {
                go.SetActive(true);
            }

            foreach(var repositionedRect in _repositionedRects)
            {
                repositionedRect.RestoreOriginal();
            }
        }

        internal void Destroy()
        {
            UnityEngine.Object.DestroyImmediate(_containerUIEventsListener);
            _containerUIEventsListener = null;
            UnityEngine.Object.DestroyImmediate(_targetUIEventsListener);
            _targetUIEventsListener = null;
            foreach(var go in _customUIGameObjects)
            {
                UnityEngine.Object.Destroy(go);
            }
        }
    }
}
