using NAT.Common.Models;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Diagnostics
{
    sealed class ColliderDrawService : DrawService
    {
        ScriptBase _script;
        List<AutoColliderPair> _autoColliderPairs;
        Person _person;
        readonly Color _transparentWhite = new Color(1f, 1f, 1f, 0.600f);
        readonly Color _lightTransparentWhite = new Color(1f, 1f, 1f, 0.150f);
        readonly Color _orange = new Color(1f, 0.4f, 0.4f, 0.75f);
        readonly Color _transparentBlack = new Color(0f, 0f, 0f, 0.5f);

        internal ColliderDrawService Init(ScriptBase script, List<AutoColliderPair> autoColliderPairs)
        {
            _script = script;
            _autoColliderPairs = autoColliderPairs;
            _person = script.person;
            base.Init(script.bootstrap.Camera, true);
            return this;
        }

        internal void DrawAllColliderPairsDiagnostics(List<HardColliderPair> hardColliderPairs, string groupFilter, bool useXRayPreviews)
        {
            if(!enabled || !_script.person.Skin)
            {
                return;
            }

            foreach(var pair in hardColliderPairs)
            {
                DrawColliderPairDiagnostics(pair, groupFilter, useXRayPreviews);
            }
        }

        internal void DrawColliderPairDiagnostics(HardColliderPair pair, string groupFilter, bool useXRayPreviews)
        {
            if(!enabled || !_script.person.Skin)
            {
                return;
            }

            var radiusLineColor = useXRayPreviews ? _transparentWhite : _lightTransparentWhite;
            if(groupFilter.StartsWith("Left") || groupFilter.StartsWith("Both"))
            {
                DrawDiagnostics(pair.left, radiusLineColor);
            }

            if(groupFilter.StartsWith("Right") || groupFilter.StartsWith("Both"))
            {
                DrawDiagnostics(pair.right, radiusLineColor);
            }
        }

        void DrawDiagnostics(HardCollider hardCollider, Color radiusLineColor)
        {
            var autoCollider = hardCollider.autoCollider;
            var colliderPosition = hardCollider.CenterWorldPosition;
            var skinVertexPosition = _person.Skin.rawSkinnedVerts[autoCollider.targetVertex];
            var direction = (colliderPosition - skinVertexPosition).normalized;
            var surfacePoint = colliderPosition - direction * (autoCollider.colliderRadius * autoCollider.scale);
            lines.Add(new LineV3(colliderPosition, surfacePoint, radiusLineColor));
            DrawPositionMarker(skinVertexPosition, Color.yellow, 0.002f);
        }

        internal void DrawDevDiagnostics(int hoveredOverIndex, AutoCollider selectedCollider, Dictionary<string, List<int>> indices)
        {
            if(!enabled || !_person.Skin)
            {
                return;
            }

            var skinVerts = _person.Skin.rawSkinnedVerts;

            foreach(var pair in indices)
            {
                foreach(int index in pair.Value)
                {
                    var position = skinVerts[index];
                    DrawPositionMarker(position, _transparentBlack, 0.002f);
                }
            }

            if(hoveredOverIndex != -1)
            {
                var position = skinVerts[hoveredOverIndex];
                DrawRectangle(position, _orange, 0.003f);
            }

            foreach(var pair in _autoColliderPairs)
            {
                DrawAutoColliderLines(pair.left, selectedCollider, skinVerts);
                DrawAutoColliderLines(pair.right, selectedCollider, skinVerts);
            }
        }

        void DrawAutoColliderLines(AutoCollider autoCollider, AutoCollider selectedCollider, Vector3[] rawSkinnedVerts)
        {
            var targetPosition = rawSkinnedVerts[autoCollider.targetVertex];
            var anchor1Position = rawSkinnedVerts[autoCollider.anchorVertex1];
            var anchor2Position = rawSkinnedVerts[autoCollider.anchorVertex2];
            var colliderPosition = autoCollider.jointCollider.transform.position;

            if(autoCollider == selectedCollider)
            {
                DrawSelectedColliderLines(targetPosition, anchor1Position, anchor2Position, colliderPosition);
            }
            else
            {
                lines.Add(new LineV3(colliderPosition, targetPosition, _lightTransparentWhite));
            }
        }

        void DrawSelectedColliderLines(Vector3 targetPosition, Vector3 anchor1Position, Vector3 anchor2Position, Vector3 colliderPosition)
        {
            DrawPositionMarker(targetPosition, Color.yellow, 0.003f);
            lines.Add(new LineV3(colliderPosition, targetPosition, _transparentWhite));
            DrawPositionMarker(anchor1Position, Color.green, 0.002f);
            DrawPositionMarker(anchor2Position, Color.cyan, 0.002f);
            lines.Add(new LineV3(anchor1Position, anchor2Position, _transparentWhite));
        }

        void OnEnable()
        {
            ClearLines();
        }
    }
}
