using System.Collections.Generic;
using Battlehub.RTCommon;
using UnityEngine;
using UnityEngine.Rendering;

namespace NAT.Common.Diagnostics
{
    abstract class DrawService : MonoBehaviour, IGL
    {
        Material _drawMat;
        Camera _camera;
        CameraBehaviour _cameraBehaviour;
        protected readonly List<LineV3> lines = new List<LineV3>();

        protected void Init(Camera camera, bool xray = false)
        {
            var shader = Shader.Find(xray ? "Battlehub/RTHandles/VertexColor" : "Hidden/Internal-Colored");
            var lineMat = new Material(shader)
            {
                hideFlags = HideFlags.HideAndDontSave,
            };
            lineMat.SetInt("_SrcBlend", (int) BlendMode.SrcAlpha);
            lineMat.SetInt("_DstBlend", (int) BlendMode.OneMinusSrcAlpha);
            lineMat.SetInt("_Cull", (int) CullMode.Back);
            lineMat.SetInt("_ZWrite", 0);
            _drawMat = lineMat;
            _camera = camera;

            if(Utils.IsVR())
            {
                _cameraBehaviour = _camera.transform.gameObject.AddComponent<CameraBehaviour>();
                _cameraBehaviour.postRenderHandlers += DrawLines;
            }

            if(!GLRenderer.Instance)
            {
                new GameObject().AddComponent<GLRenderer>();
            }

            GLRenderer.Instance.Add(this);
            enabled = false;
        }

        public void Draw(int cullingMask)
        {
            DrawLines();
        }

        void DrawLines()
        {
            if(!enabled)
            {
                return;
            }

            GL.PushMatrix();
            _drawMat.SetPass(0);

            GL.Begin(GL.LINES);
            foreach(var line in lines)
            {
                GL.Color(line.color);
                GL.Vertex(line.from);
                GL.Vertex(line.to);
            }

            GL.End();
            GL.PopMatrix();
        }

        struct ToViewport
        {
            internal readonly Vector3 pos;
            internal readonly float sizeX;
            internal readonly float sizeY;

            internal ToViewport(Vector3 worldPos, Camera camera, float length)
            {
                var viewportPos = camera.WorldToViewportPoint(worldPos);
                length /= viewportPos.z;
                sizeX = camera.aspect < 1 ? length : length / camera.aspect;
                sizeY = camera.aspect < 1 ? length * camera.aspect : length;
                pos = viewportPos;
            }
        }

        protected void DrawPositionMarker(Vector3 pos, Color color, float length)
        {
            var vp = new ToViewport(pos, _camera, length);
            lines.Add(new LineV3
            (
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x - vp.sizeX, vp.pos.y, vp.pos.z)),
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x + vp.sizeX, vp.pos.y, vp.pos.z)),
                color
            ));
            lines.Add(new LineV3
            (
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x, vp.pos.y - vp.sizeY, vp.pos.z)),
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x, vp.pos.y + vp.sizeY, vp.pos.z)),
                color
            ));
        }

        protected void DrawRectangle(Vector3 pos, Color color, float length)
        {
            var vp = new ToViewport(pos, _camera, length);
            lines.Add(new LineV3
            (
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x - vp.sizeX, vp.pos.y - vp.sizeY, vp.pos.z)),
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x + vp.sizeX, vp.pos.y - vp.sizeY, vp.pos.z)),
                color
            ));
            lines.Add(new LineV3
            (
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x - vp.sizeX, vp.pos.y + vp.sizeY, vp.pos.z)),
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x + vp.sizeX, vp.pos.y + vp.sizeY, vp.pos.z)),
                color
            ));
            lines.Add(new LineV3
            (
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x - vp.sizeX, vp.pos.y - vp.sizeY, vp.pos.z)),
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x - vp.sizeX, vp.pos.y + vp.sizeY, vp.pos.z)),
                color
            ));
            lines.Add(new LineV3
            (
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x + vp.sizeX, vp.pos.y - vp.sizeY, vp.pos.z)),
                _camera.ViewportToWorldPoint(new Vector3(vp.pos.x + vp.sizeX, vp.pos.y + vp.sizeY, vp.pos.z)),
                color
            ));
        }

        internal void ClearLines() => lines.Clear();

        void OnDestroy()
        {
            if(GLRenderer.Instance)
            {
                GLRenderer.Instance.Remove(this);
            }

            Destroy(_drawMat);
            Destroy(_cameraBehaviour);
        }
    }
}
