using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Diagnostics
{
    sealed class SoftJointDrawService : DrawService
    {
        readonly Color _gridColor = new Color(1f, 1f, 1f, 0.050f);
        readonly Color _gradientColor = new Color(1f, 1f, 1f, 0.125f);
        readonly float _markerLength = Utils.IsVR() ? 0.001f : 0.003f;
        readonly float _rectangleLength = Utils.IsVR() ? 0.002f : 0.004f;

        SoftJoint[] _leftJoints, _rightJoints;

        internal JSONStorableBool visualizeGradientJsb;
        internal JSONStorableBool visualizeLdpJsb;

        internal bool isDrawingClothingCoverage;

        internal void Init(SoftJoint[] leftJoints, SoftJoint[] rightJoints, Camera camera)
        {
            base.Init(camera, true);
            _leftJoints = leftJoints;
            _rightJoints = rightJoints;
            visualizeGradientJsb = new JSONStorableBool("Visualize Gradient", true);
            visualizeLdpJsb = new JSONStorableBool("Visualize Local Distortion", false);
            enabled = false;
        }

        internal void DrawDiagnostics(SoftPhysicsParamPair paramPair)
        {
            if(!enabled || !(visualizeGradientJsb.val || visualizeLdpJsb.val))
            {
                return;
            }

            string paramName = paramPair.name;
            var param = paramPair.Left;
            float visualizationMin = -Mathf.Max(paramPair.visualizationLimit, param.offsetJsf.min - param.offsetJsf.val);
            float visualizationMax = Mathf.Min(paramPair.visualizationLimit, param.offsetJsf.max - param.offsetJsf.val);

            if(visualizeGradientJsb.val)
            {
                float baseWithOffset = param.baseValueJsf.val + param.offsetJsf.val;
                foreach(var softJoint in _leftJoints)
                {
                    var color = JointRelativeColor(
                        _gradientColor,
                        softJoint.GetBaseValue(paramName) - baseWithOffset,
                        Mathf.Abs(baseWithOffset),
                        visualizationMin,
                        visualizationMax
                    );
                    DrawPositionMarker(softJoint, color);
                }
            }

            if(visualizeLdpJsb.val)
            {
                if(paramPair.visualizeRelativeChange)
                {
                    foreach(var softJoint in _leftJoints)
                    {
                        float jointBaseValue = softJoint.GetBaseValue(paramName);
                        var color = JointRelativeColor(
                            _gridColor,
                            softJoint.GetDistortionValue(paramName),
                            jointBaseValue,
                            visualizationMin,
                            visualizationMax
                        );
                        DrawJointConnections(softJoint, color);
                    }
                }
                else
                {
                    foreach(var softJoint in _leftJoints)
                    {
                        var color = JointColor(
                            _gridColor,
                            softJoint.GetDistortionValue(paramName),
                            visualizationMin,
                            visualizationMax
                        );
                        DrawJointConnections(softJoint, color);
                    }
                }
            }
        }

        internal void DrawClothingCoverage(ClothingAdjustmentProfile profile)
        {
            if(!enabled)
            {
                return;
            }

            isDrawingClothingCoverage = true;
            foreach(var pair in profile.coveredSoftJoints)
            {
                foreach(var softJoint in pair.Value)
                {
                    DrawPositionMarker(softJoint, Colors.transparentGreen);
                }
            }

            foreach(var pair in profile.exposedSoftJoints)
            {
                foreach(var softJoint in pair.Value)
                {
                    DrawPositionMarker(softJoint, Colors.transparentLightGray);
                }
            }

            if(profile.hoveredOverJoint != null)
            {
                DrawRectangle(profile.hoveredOverJoint, Colors.desaturatedRed);
            }
        }

        internal void DrawClothingAdjustments(SoftPhysicsParamPair paramPair)
        {
            if(!enabled || paramPair == null)
            {
                return;
            }

            DrawClothingAdjustmentsSide(paramPair, paramPair.Left, _leftJoints);
            DrawClothingAdjustmentsSide(paramPair, paramPair.Right, _rightJoints);
        }

        void DrawClothingAdjustmentsSide(SoftPhysicsParamPair paramPair, SoftPhysicsParam param, IEnumerable<SoftJoint> softJoints)
        {
            string paramName = paramPair.name;
            float visualizationMin = -Mathf.Max(paramPair.clothingAdjustmentVisualizationLimit, param.offsetJsf.min - param.offsetJsf.val);
            float visualizationMax = Mathf.Min(paramPair.clothingAdjustmentVisualizationLimit, param.offsetJsf.max - param.offsetJsf.val);

            if(paramPair.visualizeRelativeChange)
            {
                foreach(var softJoint in softJoints)
                {
                    float jointBaseValue = softJoint.GetBaseValue(paramName);
                    var color = JointRelativeColor(
                        _gradientColor,
                        softJoint.GetClothingAdjustmentValue(paramName),
                        jointBaseValue,
                        visualizationMin,
                        visualizationMax
                    );
                    DrawPositionMarker(softJoint, color);
                }
            }
            else
            {
                foreach(var softJoint in softJoints)
                {
                    var color = JointColor(
                        _gradientColor,
                        softJoint.GetClothingAdjustmentValue(paramName),
                        visualizationMin,
                        visualizationMax
                    );
                    DrawPositionMarker(softJoint, color);
                }
            }
        }

        void DrawPositionMarker(SoftJoint softJoint, Color color) =>
            DrawPositionMarker(softJoint.jointRb.position, color, _markerLength);

        void DrawRectangle(SoftJoint softJoint, Color color) =>
            DrawRectangle(softJoint.jointRb.position, color, _rectangleLength);

        void DrawJointConnections(SoftJoint softJoint, Color color)
        {
            var origin = softJoint.jointRb.position;
            foreach(var target in softJoint.nearbyJoints)
            {
                lines.Add(new LineV3(origin, (origin + target.jointRb.position) / 2, color));
            }
        }

        static Color JointColor(Color baseColor, float diff, float min, float max) => diff < 0
            ? Color.Lerp(Color.cyan, baseColor, Mathf.InverseLerp(min, 0, diff))
            : Color.Lerp(baseColor, Color.red, Mathf.InverseLerp(0, max, diff));

        static Color JointRelativeColor(Color baseColor, float diff, float offsetLimit, float min, float max)
        {
            if(diff == 0)
            {
                return baseColor;
            }

            if(diff < 0)
            {
                min -= offsetLimit;
                return min >= 0
                    ? Color.cyan
                    : Color.Lerp(Color.cyan, baseColor, Mathf.InverseLerp(min, 0, diff));

            }

            max += offsetLimit;
            return max <= 0
                ? Color.red
                : Color.Lerp(baseColor, Color.red, Mathf.InverseLerp(0, max, diff));
        }

        void OnEnable()
        {
            isDrawingClothingCoverage = false;
            ClearLines();
        }

        void OnDisable()
        {
            isDrawingClothingCoverage = false;
        }
    }
}
