using System;
using UnityEngine;

namespace NAT.Common.Handlers
{
    class AngleCalculator
    {
        readonly Rigidbody _leftRb, _rightRb, _centerRb;

        // ReSharper disable MemberCanBePrivate.Global
        internal float LeftRoll { get; private set; }
        internal float LeftPitch { get; private set; }
        internal float RightRoll { get; private set; }
        internal float RightPitch { get; private set; }
        internal float CenterRoll { get; private set; }
        internal float CenterPitch { get; private set; }
        internal float CenterPitch2X { get; private set; }
        internal float LeftDiffFromHorizontal { get; private set; }
        internal float RightDiffFromHorizontal { get; private set; }

        internal float LeftForwardLeanAmount { get; private set; }
        internal float RightForwardLeanAmount { get; private set; }

        internal float SmoothLeftRoll { get; private set; }
        internal float SmoothRightRoll { get; private set; }
        internal float Smooth2XLeftPitch { get; private set; }
        internal float Smooth2XRightPitch { get; private set; }
        // ReSharper restore MemberCanBePrivate.Global

        internal AngleCalculator(Rigidbody leftRb, Rigidbody rightRb, Rigidbody centerRb)
        {
            _leftRb = leftRb;
            _rightRb = rightRb;
            _centerRb = centerRb;
        }

        internal void CalculateAngles()
        {
            var leftRotation = _leftRb.rotation;
            var rightRotation = _rightRb.rotation;
            var centerRotation = _centerRb.rotation;

            LeftRoll = Roll(leftRotation);
            LeftPitch = Pitch(leftRotation);
            RightRoll = Roll(rightRotation);
            RightPitch = Pitch(rightRotation);
            CenterRoll = Roll(centerRotation);
            CenterPitch = Pitch(centerRotation);
            CenterPitch2X = 2 * CenterPitch;

            CalculateDiffFromHorizontal();
            CalculateLeanAmounts();
            CalculateSmoothAngles();
        }

        static float PitchToForwardLean(float input)
        {
            if(input < 0.0f || input > 1.0f)
            {
                return 0.0f;
            }

            return input <= 0.5f ? 2.0f * input : 2.0f * (1.0f - input);
        }

        void CalculateDiffFromHorizontal()
        {
            LeftDiffFromHorizontal = DiffFromHorizontal(LeftPitch, LeftRoll);
            RightDiffFromHorizontal = DiffFromHorizontal(RightPitch, RightRoll);
        }

        void CalculateLeanAmounts()
        {
            LeftForwardLeanAmount = PitchToForwardLean(LeftPitch);
            RightForwardLeanAmount = PitchToForwardLean(RightPitch);
        }

        void CalculateSmoothAngles()
        {
            SmoothLeftRoll = Calc.SmoothStep(0.5f * (LeftRoll + CenterRoll));
            SmoothRightRoll = Calc.SmoothStep(0.5f * (RightRoll + CenterRoll));
            Smooth2XLeftPitch = 2 * Calc.SmoothStep(LeftPitch);
            Smooth2XRightPitch = 2 * Calc.SmoothStep(RightPitch);
        }

        static float Roll(Quaternion q) => 2 * InverseLerpToPi(Mathf.Asin(2 * q.x * q.y + 2 * q.z * q.w));

        static float Pitch(Quaternion q) => InverseLerpToPi(Mathf.Atan2(2 * q.x * q.w - 2 * q.y * q.z, 1 - 2 * q.x * q.x - 2 * q.z * q.z));

        static float InverseLerpToPi(float val) => val > 0
            ? Mathf.InverseLerp(0, Mathf.PI, val)
            : -Mathf.InverseLerp(0, Mathf.PI, -val);

        static float DiffFromHorizontal(float pitch, float roll) => Mathf.Abs(Mathf.Abs(2 * pitch) - 1) * RollMultiplier(roll);

        internal static float DepthAdjustByAngle(float pitch)
        {
            // leaning forward
            if(pitch >= 0)
            {
                // upright
                if(pitch < 1)
                {
                    return pitch;
                }

                // upside down
                return 2 - pitch;
            }

            // leaning back
            // upright
            if(pitch >= -1)
            {
                return -pitch;
            }

            // upside down
            return 2 + pitch;
        }

        internal static float UpDownAdjustByAngle(float pitch)
        {
            // leaning forward
            if(pitch >= 0)
            {
                // upright
                if(pitch < 1)
                {
                    return 1 - pitch;
                }

                // upside down
                return pitch - 1;
            }

            // leaning back
            // upright
            if(pitch >= -1)
            {
                return 1 + pitch;
            }

            // upside down
            return -pitch - 1;
        }

        internal static float RollMultiplier(float roll) => 1 - Mathf.Abs(roll);

        internal void OverrideAngles(string direction)
        {
            switch(direction)
            {
                case Direction.UP:
                {
                    CenterPitch = 1.00f;
                    LeftPitch = 0.88f;
                    RightPitch = 0.88f;
                    CenterRoll = 0;
                    LeftRoll = 0;
                    RightRoll = 0;
                    break;
                }
                case Direction.DOWN:
                {
                    CenterPitch = 0;
                    LeftPitch = 0;
                    RightPitch = 0;
                    CenterRoll = 0;
                    LeftRoll = 0;
                    RightRoll = 0;
                    break;
                }
                case Direction.FORWARD:
                {
                    CenterPitch = 0.50f;
                    LeftPitch = 0.50f;
                    RightPitch = 0.50f;
                    CenterRoll = 0.00f;
                    LeftRoll = -0.14f;
                    RightRoll = 0.14f;
                    break;
                }
                case Direction.BACK:
                {
                    CenterPitch = -0.48f;
                    LeftPitch = -0.54f;
                    RightPitch = -0.54f;
                    CenterRoll = 0;
                    LeftRoll = 0.30f;
                    RightRoll = -0.30f;
                    break;
                }
                case Direction.LEFT:
                {
                    CenterPitch = 1.00f;
                    LeftPitch = 1.00f;
                    RightPitch = 1.00f;
                    CenterRoll = 0.96f;
                    LeftRoll = 0.55f;
                    RightRoll = 0.99f;
                    break;
                }
                case Direction.RIGHT:
                {
                    CenterPitch = 1.00f;
                    LeftPitch = 1.00f;
                    RightPitch = 1.00f;
                    CenterRoll = -0.96f;
                    LeftRoll = -0.99f;
                    RightRoll = -0.55f;
                    break;
                }
                default: throw new ArgumentException($"Invalid direction {direction}");
            }

            CalculateDiffFromHorizontal();
            CalculateLeanAmounts();
            CalculateSmoothAngles();
        }
    }
}
