using NAT.Common.Contexts;
using NAT.Common.Models;
using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using NAT.Store;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class ClothingHandler : Handler
    {
        readonly ClothingStore _store;
        protected readonly JointPhysicsHandler jointPhysicsHandler;
        readonly SoftPhysicsHandler _softPhysicsHandler;
        readonly GravityPhysicsHandler _gravityPhysicsHandler;
        internal ClothingPresetHandler presetHandler;
        readonly Dictionary<string, Coroutine> _waitForWrapCoroutines = new Dictionary<string, Coroutine>();

        internal string ClothingProfilesDir { get; set; }
        protected string[] validRegionTags;
        protected string[] invalidRegionTags;

        readonly ClothingContext _leftContext = new ClothingContext(Side.LEFT);
        readonly ClothingContext _rightContext = new ClothingContext(Side.RIGHT);

        internal JSONStorableStringChooser ClothingItemJssc { get; }

        internal JSONStorableBool ItemEnabledJsb { get; }
        internal JSONStorableBool LeftActiveJsb { get; }
        internal JSONStorableBool RightActiveJsb { get; }
        internal JSONStorableBool SymmetricJsb { get; }
        internal JSONStorableFloat ProfileTransitionJsf { get; }
        internal JSONStorableFloat YForceJsf { get; }
        internal JSONStorableFloat XForceJsf { get; }
        internal JSONStorableFloat ZForceJsf { get; }
        internal JSONStorableFloat YXHoldPctJsf { get; }
        internal JSONStorableFloat ZHoldPctJsf { get; }
        internal JSONStorableFloat YXDampeningPctJsf { get; }
        internal JSONStorableFloat ZDampeningPctJsf { get; }
        internal JSONStorableFloat YMorphsOffsetJsf { get; }
        internal JSONStorableFloat XMorphsOffsetJsf { get; }
        internal JSONStorableFloat ZMorphsOffsetJsf { get; }
        internal JSONStorableBool ProfileOffWhenUndressingJsb { get; }
        internal JSONStorableFloat SoftJointSearchRadiusJsf { get; }
        internal JSONStorableStringChooser VisualizeClothingCoverageJssc { get; }
        internal JSONStorableBool AutoSelectVisualizedParamJsb { get; }
        internal SoftPhysicsParamPair VisualizedParamPair { get; private set; }

        internal Dictionary<string, JSONStorableFloat> SoftPhysicsJsfs { get; }
        internal readonly Dictionary<string, ClothingMorphConfig> morphConfigs = new Dictionary<string, ClothingMorphConfig>();

        internal StorableFloat MorphingTransitionJsf { get; }

        internal Vector3 LeftForce { get; private set; }
        internal Vector3 RightForce { get; private set; }

        internal Vector3 LeftMorphsOffsets { get; private set; }
        internal Vector3 RightMorphsOffsets { get; private set; }
        internal Vector3 LeftMorphsOffsetsCompensation { get; private set; }
        internal Vector3 RightMorphsOffsetsCompensation { get; private set; }

        internal float LeftYXHoldPct { get; private set; }
        internal float RightYXHoldPct { get; private set; }
        internal float LeftZHoldPct { get; private set; }
        internal float RightZHoldPct { get; private set; }
        internal float LeftYXDampeningPct { get; private set; }
        internal float RightYXDampeningPct { get; private set; }
        internal float LeftZDampeningPct { get; private set; }
        internal float RightZDampeningPct { get; private set; }

        protected float upDownMassMultiplier;
        protected float sidewaysMassMultiplier;
        protected float inOutMassMultiplier;

        readonly List<ClothingItem> _itemsToAutoLoadProfile = new List<ClothingItem>();
        internal ClothingItem SelectedItem { get; private set; }
        internal ClothingAdjustmentProfile SelectedProfile { get; private set; }
        readonly List<ClothingAdjustmentProfile> _profilesToPreRegister = new List<ClothingAdjustmentProfile>();
        internal readonly HashSet<ClothingAdjustmentProfile> activeProfiles = new HashSet<ClothingAdjustmentProfile>();
        ClothingAdjustmentParams _selectedParams;

        protected ClothingHandler(ScriptBase script) : base(script)
        {
            _store = Script.bootstrap.ClothingStore;
            jointPhysicsHandler = Script.jointPhysicsHandler;
            _softPhysicsHandler = Script.softPhysicsHandler;
            _gravityPhysicsHandler = Script.gravityPhysicsHandler;
            SetupContext();

            ClothingItemJssc = new JSONStorableStringChooser(
                "Clothing Item",
                new List<string> { Constant.SELECT },
                new List<string> { Constant.SELECT },
                Constant.SELECT,
                "Clothing Item",
                OnItemSelected
            );
            ItemEnabledJsb = new JSONStorableBool("Enabled", false);
            LeftActiveJsb = new JSONStorableBool("Profile Active", false);
            RightActiveJsb = new JSONStorableBool("Profile Active (Right)", false);
            SymmetricJsb = new JSONStorableBool("Symmetric", true);
            ProfileTransitionJsf = new StorableFloat("Profile Transition %", 0f, 0f, 100f);
            YForceJsf = new JSONStorableFloat("Y Force", 0.00f, -5.00f, 5.00f);
            XForceJsf = new JSONStorableFloat("X Force", 0.00f, -5.00f, 5.00f);
            ZForceJsf = new JSONStorableFloat("Z Force", 0.00f, -5.00f, 5.00f);
            YXHoldPctJsf = new JSONStorableFloat("YX Hold %", 0f, 0f, 100f);
            ZHoldPctJsf = new JSONStorableFloat("Z Hold %", 0f, 0f, 100f);
            YXDampeningPctJsf = new JSONStorableFloat("YX Dampening %", 0f, 0f, 100f);
            ZDampeningPctJsf = new JSONStorableFloat("Z Dampening %", 0f, 0f, 100f);
            YMorphsOffsetJsf = new JSONStorableFloat("Y Morphs Angle Offset", 0.0f, -15.0f, 15.0f);
            XMorphsOffsetJsf = new JSONStorableFloat("X Morphs Angle Offset", 0.0f, Script.IsTittyMagic ? -15.0f : -30.0f, Script.IsTittyMagic ? 15.0f : 30.0f);
            ZMorphsOffsetJsf = new JSONStorableFloat("Z Morphs Offset", 0.0f, -15.0f, 15.0f);
            ProfileOffWhenUndressingJsb = new JSONStorableBool("Set Profile Off When Undressing", false);

            ItemEnabledJsb.setCallbackFunction = value => SelectedItem?.enabledJsb.SetVal(value);
            LeftActiveJsb.setCallbackFunction = value => SelectedProfile?.LeftActive.SetVal(value);
            RightActiveJsb.setCallbackFunction = value => SelectedProfile?.RightActive.SetVal(value);
            SymmetricJsb.setCallbackFunction = value => SelectedProfile?.symmetric.SetVal(value);
            ProfileTransitionJsf.setCallbackFunction = value =>
            {
                if(SelectedProfile == null)
                {
                    return;
                }

                if(SelectedProfile.HasTarget)
                {
                    SelectedProfile?.Transition.SetVal(value);
                }
            };
            YForceJsf.setCallbackFunction = value => _selectedParams?.yForce.SetVal(value);
            XForceJsf.setCallbackFunction = value => _selectedParams?.xForce.SetVal(value);
            ZForceJsf.setCallbackFunction = value => _selectedParams?.zForce.SetVal(value);
            YXHoldPctJsf.setCallbackFunction = value => _selectedParams?.yxHoldPct.SetVal(value);
            ZHoldPctJsf.setCallbackFunction = value => _selectedParams?.zHoldPct.SetVal(value);
            YXDampeningPctJsf.setCallbackFunction = value => _selectedParams?.yxDampeningPct.SetVal(value);
            ZDampeningPctJsf.setCallbackFunction = value => _selectedParams?.zDampeningPct.SetVal(value);
            YMorphsOffsetJsf.setCallbackFunction = value => _selectedParams?.yMorphsOffset.SetVal(value);
            XMorphsOffsetJsf.setCallbackFunction = value => _selectedParams?.xMorphsOffset.SetVal(value);
            ZMorphsOffsetJsf.setCallbackFunction = value => _selectedParams?.zMorphsOffset.SetVal(value);
            ProfileOffWhenUndressingJsb.setCallbackFunction = value => SelectedProfile?.offWhenUndressing.SetVal(value);

            SoftJointSearchRadiusJsf = new JSONStorableFloat("Soft Joint Search Radius", 0.0100f, 0.0000f, Script.IsBootyMagic ? 0.0500f : 0.0250f);
            SoftPhysicsJsfs = new Dictionary<string, JSONStorableFloat>
            {
                [FAT_SPRING.NAME] = new JSONStorableFloat($"{FAT_SPRING.NAME} %", 0f, 0f, 50f),
                [FAT_DAMPER.NAME] = new JSONStorableFloat($"{FAT_DAMPER.NAME} %", 0f, 0f, 20f),
                [FAT_MASS.NAME] = new JSONStorableFloat($"{FAT_MASS.NAME} %", 0f, -100f, 0f),
                [FAT_COLLIDER_RADIUS.NAME] = new JSONStorableFloat($"{FAT_COLLIDER_RADIUS.NAME} %", 0f, -90f, 0f),
                [FAT_COLLIDER_DEPTH.NAME] = new JSONStorableFloat($"{FAT_COLLIDER_DEPTH.NAME} %", 0f, -100f, 100f),
                [FAT_DISTANCE_LIMIT.NAME] = new JSONStorableFloat($"{FAT_DISTANCE_LIMIT.NAME} %", 0f, -90f, 0f),
            };

            var options = new List<string> { "None", Constant.COVERAGE };
            options.AddRange(SoftPhysicsJsfs.Keys);
            VisualizeClothingCoverageJssc = new JSONStorableStringChooser(
                "Visualize Clothing Coverage",
                options,
                options[0],
                "Visualize"
            )
            {
                setCallbackFunction = value =>
                    VisualizedParamPair = _softPhysicsHandler.softParamPairs.FirstOrDefault(group => group.name == value),
            };
            AutoSelectVisualizedParamJsb = new JSONStorableBool("Auto-Select Visualized Parameter", true);

            foreach(var pair in SoftPhysicsJsfs)
            {
                pair.Value.setCallbackFunction = value =>
                {
                    if(SelectedProfile == null || _selectedParams == null)
                    {
                        return;
                    }

                    _selectedParams.softPhysicsParams[pair.Key].SetVal(value);
                    if(AutoSelectVisualizedParamJsb.val && SelectedProfile.isCoverageMapped)
                    {
                        VisualizeClothingCoverageJssc.val = pair.Key;
                    }
                };
            }

            MorphingTransitionJsf = new StorableFloat("Morphing Transition (s)", 0.10f, 0.00f, 1.00f);
        }

        void SetupContext()
        {
            if(Script.IsBootyMagic)
            {
                _leftContext.rigidbody = Person.LeftGlute.rigidbody;
                _rightContext.rigidbody = Person.RightGlute.rigidbody;

            }
            else if(Script.IsTittyMagic)
            {
                _leftContext.rigidbody = Person.LeftBreast.rigidbody;
                _rightContext.rigidbody = Person.RightBreast.rigidbody;
            }

            _leftContext.rigidbodyT = _leftContext.rigidbody.transform;
            _rightContext.rigidbodyT = _rightContext.rigidbody.transform;
        }

        protected override void LoadSettings()
        {
            InitializeProfiles();
            UpdateItemsToAutoLoadProfile();
            RebuildClothingOptions();
            InitActiveProfiles();
        }

        void InitializeProfiles()
        {
            foreach(var item in _store.enabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                var profile = GetProfile(item);
                if(profile == null)
                {
                    _waitForWrapCoroutines[item.label] = Script.StartCoroutine(GetSkinWrapCo(item));
                    InitializeProfile(item);
                }
            }
        }

        void UpdateItemsToAutoLoadProfile()
        {
            foreach(var item in _store.enabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                // Debug.Log($"UpdateItemsToAutoLoadProfile: {item.DazItem.uid}");
                _itemsToAutoLoadProfile.Add(item);
            }
        }

        protected void AddMorphConfigCallbacks()
        {
            foreach(var pair in morphConfigs)
            {
                pair.Value.storableFloat.setCallbackFunction = value => _selectedParams?.morphParams[pair.Key].SetVal(value);
            }
        }

        internal void SetAllAdjustmentValues()
        {
            foreach(var profile in activeProfiles)
            {
                profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            }
        }

        protected virtual void SyncJointPhysicsAdjustments()
        {
            if(!Script.enabled)
            {
                return;
            }

            LeftYXHoldPct = 0f;
            RightYXHoldPct = 0f;
            LeftZHoldPct = 0f;
            RightZHoldPct = 0f;
            LeftYXDampeningPct = 0f;
            RightYXDampeningPct = 0f;
            LeftZDampeningPct = 0f;
            RightZDampeningPct = 0f;

            foreach(var profile in activeProfiles)
            {
                if(profile.LeftActive.Value)
                {
                    LeftYXHoldPct = Mathf.Max(LeftYXHoldPct, profile.GetLeftYXHoldPct());
                    LeftZHoldPct = Mathf.Max(LeftZHoldPct, profile.GetLeftZHoldPct());
                    LeftYXDampeningPct = Mathf.Max(LeftYXDampeningPct, profile.GetLeftYXDampeningPct());
                    LeftZDampeningPct = Mathf.Max(LeftZDampeningPct, profile.GetLeftZDampeningPct());
                }

                if(profile.RightActive.Value)
                {
                    RightYXHoldPct = Mathf.Max(RightYXHoldPct, profile.GetRightYXHoldPct());
                    RightZHoldPct = Mathf.Max(RightZHoldPct, profile.GetRightZHoldPct());
                    RightYXDampeningPct = Mathf.Max(RightYXDampeningPct, profile.GetRightYXDampeningPct());
                    RightZDampeningPct = Mathf.Max(RightZDampeningPct, profile.GetRightZDampeningPct());
                }
            }

            float leftYXHoldAmount = CalculateHoldAmount(LeftYXHoldPct);
            float rightYXHoldAmount = CalculateHoldAmount(RightYXHoldPct);
            float leftZHoldAmount = CalculateHoldAmount(LeftZHoldPct);
            float rightZHoldAmount = CalculateHoldAmount(RightZHoldPct);

            jointPhysicsHandler.rotationSpring.Left.SetClothingAdjustment(leftYXHoldAmount);
            jointPhysicsHandler.rotationSpring.Right.SetClothingAdjustment(rightYXHoldAmount);
            jointPhysicsHandler.rotationDamper.Left.SetClothingAdjustment(0.01f * LeftYXDampeningPct);
            jointPhysicsHandler.rotationDamper.Right.SetClothingAdjustment(0.01f * RightYXDampeningPct);
            jointPhysicsHandler.inOutSpring.Left.SetClothingAdjustment(CalculateHoldAmount(LeftZHoldPct));
            jointPhysicsHandler.inOutSpring.Right.SetClothingAdjustment(CalculateHoldAmount(RightZHoldPct));
            jointPhysicsHandler.inOutDamper.Left.SetClothingAdjustment(0.01f * LeftZDampeningPct);
            jointPhysicsHandler.inOutDamper.Right.SetClothingAdjustment(0.01f * RightZDampeningPct);

            float collidersMassMultiplierLeft = 1 - (0.6f * leftYXHoldAmount + 0.28f * leftZHoldAmount);
            float collidersMassMultiplierRight = 1 - (0.6f * rightYXHoldAmount + 0.28f * rightZHoldAmount);
            Script.hardColliderHandler.UpdateClothingAdjustmentAmounts(collidersMassMultiplierLeft, collidersMassMultiplierRight);
            Script.hardColliderHandler.SyncCollidersMass();
        }

        protected static float CalculateHoldAmount(float percentage) =>
            Calc.Exponential1(0.01f * percentage, 0.50f, 3.40f, 1.20f);

        void SyncJointAngleOffsets()
        {
            if(!Script.enabled)
            {
                return;
            }

            var leftMin = Vector3.zero;
            var leftMax = Vector3.zero;
            var leftSum = Vector3.zero;
            var rightMin = Vector3.zero;
            var rightMax = Vector3.zero;
            var rightSum = Vector3.zero;

            foreach(var profile in activeProfiles)
            {
                if(profile.LeftActive.Value)
                {
                    var left = new Vector3(profile.GetLeftXMorphsOffset(), profile.GetLeftYMorphsOffset(), profile.GetLeftZMorphsOffset());
                    leftMin = Vector3.Min(left, leftMin);
                    leftMax = Vector3.Max(left, leftMax);
                    leftSum += left;
                }

                if(profile.RightActive.Value)
                {
                    var right = new Vector3(-profile.GetRightXMorphsOffset(), profile.GetRightYMorphsOffset(), profile.GetRightZMorphsOffset());
                    rightMin = Vector3.Min(right, rightMin);
                    rightMax = Vector3.Max(right, rightMax);
                    rightSum += right;
                }
            }

            LeftMorphsOffsets = ClampVector(leftSum, leftMin, leftMax);
            RightMorphsOffsets = ClampVector(rightSum, rightMin, rightMax);
        }

        internal void SyncMorphsOffsetsCompensation()
        {
            // must be synced before syncing offsets compensation since the compensation value depends on morph values being up to date
            foreach(var pair in morphConfigs)
            {
                SyncMorph(pair.Key, pair.Value);
            }

            Vector3 left = Vector3.zero, right = Vector3.zero;

            foreach(var pair in morphConfigs)
            {
                var config = pair.Value;
                left += config.GetLeftMorphsOffsetsCompensation();
                right += config.GetRightMorphsOffsetsCompensation();
            }

            LeftMorphsOffsetsCompensation = left;
            RightMorphsOffsetsCompensation = right;
        }

        void SyncSoftPhysicsAdjustments()
        {
            if(!Script.enabled || !_softPhysicsHandler.softPhysicsOn)
            {
                return;
            }

            foreach(var pair in SoftPhysicsJsfs)
            {
                SyncSoftPhysicsParam(pair.Key);
            }

            SetUseDefaultSoftColliderFriction(Side.LEFT);
            SetUseDefaultSoftColliderFriction(Side.RIGHT);
        }

        void SyncSoftPhysicsParam(string paramName)
        {
            if(!Script.enabled || !_softPhysicsHandler.softPhysicsOn)
            {
                return;
            }

            SetParamClothingAdjustment(Side.LEFT, paramName);
            SetParamClothingAdjustment(Side.RIGHT, paramName);
            CalculateAverageClothingAdjustments(paramName);
        }

        void SetParamClothingAdjustment(string side, string paramName)
        {
            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            foreach(var softJoint in context.allSoftJoints)
            {
                float min = 0f;
                float max = 0f;
                float sum = 0f;
                foreach(var profile in activeProfiles)
                {
                    if(profile.activeStorableBools[side].Value && profile.coveredSoftJoints[side].Contains(softJoint))
                    {
                        float current = profile.GetSoftPhysicsParamValue(side, paramName);
                        min = Mathf.Min(min, current);
                        max = Mathf.Max(max, current);
                        sum += current;
                    }
                }

                float adjustmentPct = sum > 0 ? Mathf.Min(sum, max) : Mathf.Max(sum, min);
                softJoint.SetClothingAdjustment(paramName, 0.01f * adjustmentPct);
            }
        }

        void CalculateAverageClothingAdjustments(string paramName)
        {
            var group = _softPhysicsHandler.softParamPairs.First(x => x.name == paramName);
            group.leftClothingAdjustmentAverage = CalculateAverageClothingAdjustment(Side.LEFT, paramName);
            group.rightClothingAdjustmentAverage = CalculateAverageClothingAdjustment(Side.RIGHT, paramName);
        }

        float CalculateAverageClothingAdjustment(string side, string paramName)
        {
            var set = new HashSet<SoftJoint>(activeProfiles.SelectMany(profile => profile.coveredSoftJoints[side]));
            return set.Select(softJoint => softJoint.GetClothingAdjustmentValue(paramName))
                .DefaultIfEmpty(0)
                .Average();
        }

        void SetUseDefaultSoftColliderFriction(string side)
        {
            var coveredJoints = activeProfiles
                .Where(profile => profile.activeStorableBools[side].Value)
                .SelectMany(profile => profile.coveredSoftJoints[side])
                .ToList();

            foreach(var softJoint in coveredJoints)
            {
                softJoint.UseDefaultFriction = true;
            }

            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            foreach(var softJoint in context.allSoftJoints.Except(coveredJoints))
            {
                softJoint.UseDefaultFriction = false;
            }
        }

        internal void AutoLoadDefaultProfiles()
        {
            foreach(var item in _itemsToAutoLoadProfile)
            {
                // Debug.Log($"AutoLoadDefaultProfiles: Loading default profile for {item.label}");
                presetHandler.UpdatePresetDir(item.ProfilesDir);
                bool defaultLoaded = presetHandler.LoadDefaultPreset(item, true);
                if(!defaultLoaded)
                {
                    AutoHoldAndDampening(item, GetProfile(item));
                }
            }

            _itemsToAutoLoadProfile.Clear();
        }

        internal void AutoLoadProfiles(IEnumerable<JSONNode> activeProfiles, string packageId)
        {
            if(Script.firstCalibrationDone)
            {
                UpdateItemsToAutoLoadProfile();
            }

            foreach(var jsonNode in activeProfiles)
            {
                string path = ParsePresetPath(jsonNode, packageId);
                if(!string.IsNullOrEmpty(path) && FileUtils.FileExists(path))
                {
                    presetHandler.LoadEnabledItemProfileFromPath(path);
                }
            }
        }

        internal override void ResetAll()
        {
            ResetMorphs();
            foreach(var param in jointPhysicsHandler.paramPairs)
            {
                param.ResetClothingAdjustment();
            }
        }

        internal void OnItemToggled(ClothingItem item)
        {
            /* Prevent null exception when exiting scene */
            if(!Script)
            {
                return;
            }

            if(_waitForWrapCoroutines.ContainsKey(item.label))
            {
                var coroutine = _waitForWrapCoroutines[item.label];
                if(coroutine != null)
                {
                    Script.StopCoroutine(_waitForWrapCoroutines[item.label]);
                }

                _waitForWrapCoroutines.Remove(item.label);
            }

            if(item.enabledJsb.val)
            {
                if(IsValidItem(item))
                {
                    if(!item.wrap)
                    {
                        _waitForWrapCoroutines[item.label] = Script.StartCoroutine(GetSkinWrapCo(item));
                    }

                    RebuildClothingOptions();
                    var profile = GetProfile(item);
                    if(profile == null)
                    {
                        InitializeProfile(item);
                    }

                    ClothingItemJssc.val = item.dazItem.uid;
                }
            }
            else
            {
                bool isClothingTabActive = Script.uiHandler?.IsClothingTabActive() ?? false;
                if(!isClothingTabActive && ClothingItemJssc.val == item.dazItem.uid)
                {
                    SelectFirstActiveItem();
                }
            }

            SyncActiveProfilesAdjustmentsAndMorphAngles(item);

            if(item == SelectedItem)
            {
                ItemEnabledJsb.valNoCallback = item.enabledJsb.val;
            }
        }

        internal void OnItemAllowDetachToggled(ClothingItem item, bool value)
        {
            if(!item.SimControl || !item.SimControl.simEnabledJSON.val)
            {
                return;
            }

            var profile = GetProfile(item);
            if(profile == null)
            {
                return;
            }

            if(profile.offWhenUndressing.Value && value)
            {
                profile.activeStorableBools[Side.LEFT].SetVal(false);
                if(!profile.symmetric.Value)
                {
                    profile.activeStorableBools[Side.RIGHT].SetVal(false);
                }
            }
        }

        void SyncActiveProfilesAdjustmentsAndMorphAngles(ClothingItem item)
        {
            SyncActiveProfiles(item);
            SyncAdjustments();
            Script.SyncMorphsAngleOffsets(MorphingTransitionJsf.val);
        }

        void SyncActiveProfiles(ClothingItem item)
        {
            if(item == null)
            {
                return;
            }

            var profile = GetProfile(item);
            if(IsValidItem(item) && item.enabledJsb.val && profile.IsActive)
            {
                activeProfiles.Add(profile);
            }
            else
            {
                activeProfiles.Remove(profile);
            }
        }

        static IEnumerator GetSkinWrapCo(ClothingItem item)
        {
            float timePassed = 0f;
            int count = 0;
            float interval = 0.1f;

            while(item.dazItem.GetComponentInChildren<DAZSkinWrap>() == null)
            {
                yield return new WaitForSeconds(interval);
                timePassed += interval;
                count++;
                interval *= 2;
                // Debug.Log($"Item {item.label} DAZSkinWrap not found in {timePassed} sec");
                if(count > 10)
                {
                    Debug.Log($"Item {item.label} DAZSkinWrap not found in {timePassed} sec");
                    break;
                }
            }

            item.wrap = item.dazItem.GetComponentInChildren<DAZSkinWrap>();
        }

        internal void AutoSelectItem()
        {
            if(ClothingItemJssc.val == Constant.SELECT)
            {
                SelectFirstActiveItem();
            }
        }

        void SelectFirstActiveItem()
        {
            var firstActiveItem = _store.enabledItems.FirstOrDefault(item => IsValidItem(item) && GetProfile(item).IsActive);
            ClothingItemJssc.val = firstActiveItem?.dazItem.uid ?? Constant.SELECT;
        }

        bool IsValidItem(ClothingItem item) => CheckTags(item.dazItem.tagsArray) && !_ignorePathHashes.Contains(item.ProfilesDirHash);

        bool CheckTags(string[] tags)
        {
            if(tags == null)
            {
                return true;
            }

            return tags.Any(tag => validRegionTags.Contains(tag)) ||
                !tags.Any(tag => invalidRegionTags.Contains(tag));
        }

        internal void SyncAdjustments()
        {
            if(!Script.enabled)
            {
                return;
            }

            SyncJointPhysicsAdjustments();
            SyncJointAngleOffsets();
            SyncMorphsOffsetsCompensation();
            SyncSoftPhysicsAdjustments();
        }

        void InitializeProfile(ClothingItem item)
        {
            // Debug.Log($"{Script.DomainName}: InitializeProfile for {item.label}");
            var profile = CreateNewProfile(item);
            SetProfile(item, profile);
            presetHandler.UpdatePresetDir(item.ProfilesDir);
            // during plugin initialization, default presets are loaded separately with AutoLoadDefaultProfiles
            bool defaultLoaded = initialized && presetHandler.LoadDefaultPreset(item, true);
            profile.SaveValues();
            if(initialized && !defaultLoaded)
            {
                AutoHoldAndDampening(item, profile);
            }
        }

        ClothingAdjustmentProfile CreateNewProfile(ClothingItem item)
        {
            // Debug.Log($"{Script.DomainName}: CreateNewProfile for {item.label}");
            string prefix = $"{item.label}:";
            var profile = new ClothingAdjustmentProfile(
                item,
                prefix,
                new StorableBool($"{prefix}{SymmetricJsb.name}", true)
                {
                    Temporary = true,
                },
                new StorableBool($"{prefix}{LeftActiveJsb.name}", true)
                {
                    Temporary = true,
                },
                new StorableBool($"{prefix}{RightActiveJsb.name}", true)
                {
                    Temporary = true,
                },
                StorableFloat.Copy(ProfileTransitionJsf, prefix, "", true),
                new StorableBool($"{prefix}{ProfileOffWhenUndressingJsb.name}", false)
                {
                    Temporary = true,
                }
            );

            if(!Person.IsFemale)
            {
                profile.isCoverageMapped = true;
                profile.exposedSoftJoints[Side.LEFT] = _softPhysicsHandler.leftContext.allSoftJoints.ToList();
                profile.exposedSoftJoints[Side.RIGHT] = _softPhysicsHandler.rightContext.allSoftJoints.ToList();
            }

            profile.symmetric.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(value)
                {
                    RightActiveJsb.valNoCallback = LeftActiveJsb.val;
                }

                profile.SyncSymmetric(value);
                if(profile == SelectedProfile)
                {
                    SelectLeft();
                }

                SyncActiveProfilesAdjustmentsAndMorphAngles(item);
            };
            profile.LeftActive.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(profile.symmetric.Value)
                {
                    profile.RightActive.SetValNoCallback(value);
                }

                if(profile == SelectedProfile)
                {
                    LeftActiveJsb.valNoCallback = value;
                    if(profile.symmetric.Value)
                    {
                        RightActiveJsb.valNoCallback = value;
                    }
                }

                SyncActiveProfilesAdjustmentsAndMorphAngles(item);
            };
            profile.RightActive.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(profile.symmetric.Value)
                {
                    profile.LeftActive.SetValNoCallback(value);
                }

                if(profile == SelectedProfile)
                {
                    RightActiveJsb.valNoCallback = value;
                    if(profile.symmetric.Value)
                    {
                        LeftActiveJsb.valNoCallback = value;
                    }
                }

                SyncActiveProfilesAdjustmentsAndMorphAngles(item);
            };
            profile.Transition.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                profile.SetTransitionValue();
                if(profile == SelectedProfile)
                {
                    ProfileTransitionJsf.valNoCallback = value;
                }

                SyncAdjustments();
                Script.SyncMorphsAngleOffsets();
            };
            profile.offWhenUndressing.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(profile == SelectedProfile)
                {
                    ProfileOffWhenUndressingJsb.valNoCallback = value;
                }
            };

            profile.@params[Side.LEFT] = NewClothingAdjustmentParams(prefix, false);
            profile.rightCachedParams = NewClothingAdjustmentParams(prefix, true);
            profile.@params[Side.RIGHT] = profile.@params[Side.LEFT];

            if(Script.initialized != true)
            {
                _profilesToPreRegister.Add(profile);
            }
            else
            {
                RegisterProfileParams(profile);
            }

            return profile;
        }

        ClothingAdjustmentParams NewClothingAdjustmentParams(string prefix, bool right, bool setCallbacks = true)
        {
            string suffix = right ? " (Right)" : "";
            var @params = new ClothingAdjustmentParams
            (
                YForceJsf,
                XForceJsf,
                ZForceJsf,
                YXHoldPctJsf,
                ZHoldPctJsf,
                YXDampeningPctJsf,
                ZDampeningPctJsf,
                YMorphsOffsetJsf,
                XMorphsOffsetJsf,
                ZMorphsOffsetJsf,
                SoftPhysicsJsfs,
                morphConfigs,
                prefix,
                suffix
            );
            if(setCallbacks)
            {
                SetParamsCallbackFunctions(@params);
            }

            return @params;
        }

        void SetParamsCallbackFunctions(ClothingAdjustmentParams @params)
        {
            @params.yForce.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetYForce(value, upDownMassMultiplier);
                if(@params == _selectedParams)
                {
                    YForceJsf.valNoCallback = value;
                }
            };
            @params.xForce.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetXForce(value, sidewaysMassMultiplier);
                if(@params == _selectedParams)
                {
                    XForceJsf.valNoCallback = value;
                }
            };
            @params.zForce.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetZForce(value, inOutMassMultiplier, !Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    ZForceJsf.valNoCallback = value;
                }
            };
            @params.yxHoldPct.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.yxHoldPct);
                if(@params == _selectedParams)
                {
                    YXHoldPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.zHoldPct.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.zHoldPct);
                if(@params == _selectedParams)
                {
                    ZHoldPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.yxDampeningPct.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.yxDampeningPct);
                if(@params == _selectedParams)
                {
                    YXDampeningPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.zDampeningPct.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.zDampeningPct);
                if(@params == _selectedParams)
                {
                    ZDampeningPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.yMorphsOffset.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetYMorphsOffset(value, Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    YMorphsOffsetJsf.valNoCallback = value;
                }

                SyncJointAngleOffsets();
                Script.SyncMorphsAngleOffsets();
            };
            @params.xMorphsOffset.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetXMorphsOffset(value, Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    XMorphsOffsetJsf.valNoCallback = value;
                }

                SyncJointAngleOffsets();
                Script.SyncMorphsAngleOffsets();
            };
            @params.zMorphsOffset.storableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetZMorphsOffset(value, Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    ZMorphsOffsetJsf.valNoCallback = value;
                }

                SyncJointAngleOffsets();
                Script.SyncMorphsAngleOffsets();
            };

            foreach(var pair in @params.softPhysicsParams)
            {
                pair.Value.storableParam.setCallbackFunction = value =>
                {
                    if(Script.IsInsideRestore)
                    {
                        return;
                    }

                    ClothingAdjustmentParams.SetPercentageValue(pair.Value);
                    if(@params == _selectedParams)
                    {
                        SoftPhysicsJsfs[pair.Key].valNoCallback = value;
                    }

                    SyncSoftPhysicsParam(pair.Key);
                };
            }

            foreach(var pair in @params.morphParams)
            {
                pair.Value.storableParam.setCallbackFunction += value =>
                {
                    if(Script.IsInsideRestore)
                    {
                        return;
                    }

                    if(@params == _selectedParams)
                    {
                        morphConfigs[pair.Key].storableFloat.valNoCallback = value;
                    }

                    SyncMorphsOffsetsCompensation(); // sync all morphs at once to ensure correct morphs offsets compensation
                    Script.SyncMorphsAngleOffsets();
                };
            }
        }

        void AutoHoldAndDampening(ClothingItem item, ClothingAdjustmentProfile profile)
        {
            var control = item.DazItemControl;
            if(!control || !control.GetBoolParamValue("enableJointSpringAndDamperAdjust"))
            {
                return;
            }

            bool enableJointAdjust = Script.IsBootyMagic
                ? control.GetBoolParamValue("enableGluteJointAdjust")
                : control.GetBoolParamValue("enableBreastJointAdjust");
            if(!enableJointAdjust)
            {
                return;
            }

            float multiplier = Script.IsBootyMagic
                ? control.GetFloatParamValue("gluteJointSpringAndDamperMultiplier")
                : control.GetFloatParamValue("breastJointSpringAndDamperMultiplier");
            if(Math.Abs(multiplier - 1.000f) < 0.100f)
            {
                return;
            }

            /* Reduce amount of hold and dampening compared to vanilla VAM multiplier */
            multiplier = 1.000f + 0.50f * (multiplier - 1.000f);

            var leftSpringParam = jointPhysicsHandler.rotationSpring.Left;
            float springValue = leftSpringParam.Value;
            float targetSpring = multiplier * springValue;
            float prevSpringDiff = ROTATION_SPRING.MAX;
            float selectedHoldPct = 0f;

            var leftDamperParam = jointPhysicsHandler.rotationDamper.Left;
            float damperValue = leftDamperParam.Value;
            float targetDamper = multiplier * damperValue;
            float prevDamperDiff = ROTATION_DAMPER.MAX;
            float selectedDampeningPct = 0f;

            for(int i = 0; i <= 100; i += 5)
            {
                float holdAmount = CalculateHoldAmount(i);
                float holdAdjustedSpring = Mathf.Lerp(springValue, ROTATION_SPRING.MAX, holdAmount);
                float springDiff = Mathf.Abs(holdAdjustedSpring - targetSpring);
                if(springDiff < prevSpringDiff)
                {
                    prevSpringDiff = springDiff;
                    selectedHoldPct = i;
                }

                float dampeningAmount = 0.01f * i;
                float dampeningAdjustedDamper = Mathf.Lerp(damperValue, ROTATION_DAMPER.MAX, dampeningAmount);
                float damperDiff = Mathf.Abs(dampeningAdjustedDamper - targetDamper);
                if(damperDiff < prevDamperDiff)
                {
                    prevDamperDiff = damperDiff;
                    selectedDampeningPct = i;
                }
            }

            var @params = profile.@params[Side.LEFT];
            @params.yxHoldPct.SetVal(selectedHoldPct);
            @params.yxDampeningPct.SetVal(selectedDampeningPct);
            @params.zHoldPct.SetVal(selectedHoldPct);
            @params.zDampeningPct.SetVal(selectedDampeningPct);
        }

        internal void OnItemsReloaded()
        {
            InitializeProfiles();
            RebuildClothingOptions(true);
            InitActiveProfiles();
            ClothingItemJssc.val = Constant.SELECT;
        }

        void RebuildClothingOptions(bool removeDisabled = false)
        {
            ClothingItemJssc.choices.Remove(Constant.SELECT);
            ClothingItemJssc.displayChoices.Remove(Constant.SELECT);

            var choices = removeDisabled ? new List<string>() : new List<string>(ClothingItemJssc.choices);
            var displayChoices = removeDisabled ? new List<string>() : new List<string>(ClothingItemJssc.displayChoices);

            foreach(var item in _store.enabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                string uid = item.dazItem.uid;
                string label = item.label;
                if(!choices.Contains(uid))
                {
                    choices.Add(uid);
                    displayChoices.Add(label);
                }
            }

            ClothingItemJssc.choices = choices;
            ClothingItemJssc.displayChoices = displayChoices;
        }

        void InitActiveProfiles()
        {
            activeProfiles.Clear();
            foreach(var item in _store.enabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                var profile = GetProfile(item);
                if(profile.IsActive)
                {
                    activeProfiles.Add(profile);
                }
            }
        }

        internal void SelectLeft()
        {
            if(SelectedProfile == null)
            {
                return;
            }

            SyncSelectedProfile(Side.LEFT);
        }

        internal void SelectRight()
        {
            if(SelectedProfile == null)
            {
                return;
            }

            SyncSelectedProfile(Side.RIGHT);
        }

        void OnItemSelected(string option)
        {
            if(option == Constant.SELECT)
            {
                ItemEnabledJsb.SetDefaultValNoCallback();
                LeftActiveJsb.SetDefaultValNoCallback();
                RightActiveJsb.SetDefaultValNoCallback();
                SymmetricJsb.SetDefaultValNoCallback();
                ProfileTransitionJsf.SetDefaultValNoCallback();
                YForceJsf.SetDefaultValNoCallback();
                XForceJsf.SetDefaultValNoCallback();
                ZForceJsf.SetDefaultValNoCallback();
                YXHoldPctJsf.SetDefaultValNoCallback();
                ZHoldPctJsf.SetDefaultValNoCallback();
                YXDampeningPctJsf.SetDefaultValNoCallback();
                ZDampeningPctJsf.SetDefaultValNoCallback();
                YMorphsOffsetJsf.SetDefaultValNoCallback();
                XMorphsOffsetJsf.SetDefaultValNoCallback();
                ZMorphsOffsetJsf.SetDefaultValNoCallback();

                foreach(var pair in SoftPhysicsJsfs)
                {
                    pair.Value.SetDefaultValNoCallback();
                }

                foreach(var pair in morphConfigs)
                {
                    pair.Value.storableFloat.SetDefaultValNoCallback();
                }

                VisualizeClothingCoverageJssc.val = "None";
                SelectedItem = null;
                SelectedProfile = null;
                _selectedParams = null;
                Script.uiHandler.clothingWindow.Refresh();
                return;
            }

            // Sanity checks
            if(!_store.items.ContainsKey(option))
            {
                Script.Loggr.LogError($"Clothing item with key {option} not found");
                return;
            }

            SelectedItem = _store.items[option];
            if(SelectedItem == null)
            {
                Script.Loggr.LogError("Selected item is null");
                return;
            }

            if(SelectedItem.enabledJsb == null)
            {
                Script.Loggr.LogError($"Selected item {SelectedItem.label} enabled toggle is null");
                return;
            }

            ItemEnabledJsb.valNoCallback = SelectedItem.enabledJsb.val;
            SelectedProfile = GetProfile(SelectedItem);

            if(SelectedProfile == null)
            {
                Script.Loggr.LogError($"Selected item {SelectedItem.label} profile is null");
                return;
            }

            SyncSelectedProfile(SelectedProfile.isRightSelected ? Side.RIGHT : Side.LEFT);
        }

        internal void FindCoveredSoftJoints()
        {
            if(!Script.enabled)
            {
                Script.Loggr.LogMessage("Enable the plugin first.");
                return;
            }

            if(SelectedItem == null)
            {
                Script.Loggr.LogMessage("Select a clothing item first.");
                return;
            }

            if(!SelectedItem.enabledJsb.val)
            {
                Script.Loggr.LogMessage("Item must be enabled in order to finding its covered soft joints.");
                return;
            }

            _softPhysicsHandler.ResetSoftJointsClothingAdjustments();

            var leftJoints = new List<SoftJoint>();
            var rightJoints = new List<SoftJoint>();

            if(SelectedItem.wrap)
            {
                SelectedItem.wrap.UpdateVerts();
                var vertices = SelectedItem.wrap.Mesh.vertices;

                foreach(var vertex in vertices)
                {
                    leftJoints.AddRange(NearbyJoints(vertex, Side.LEFT));
                    rightJoints.AddRange(NearbyJoints(vertex, Side.RIGHT));
                }
            }
            else
            {
                Script.Loggr.LogMessage($"Item {SelectedItem.label} does not support finding covered soft joints automatically.");
            }

            AssignJoints(leftJoints.Distinct().ToList(), SelectedProfile, Side.LEFT);
            AssignJoints(rightJoints.Distinct().ToList(), SelectedProfile, Side.RIGHT);
            SelectedProfile.isCoverageMapped = true;

            VisualizeClothingCoverageJssc.val = Constant.COVERAGE;

            if(!SelectedProfile.isCoverageMapped)
            {
                foreach(var pair in SoftPhysicsJsfs)
                {
                    pair.Value.valNoCallback = 0f;
                }
            }

            SyncSoftPhysicsAdjustments();
            Script.uiHandler.clothingWindow.Refresh();
        }

        IEnumerable<SoftJoint> NearbyJoints(Vector3 vertex, string side)
        {
            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            foreach(var softJoint in context.allSoftJoints)
            {
                var position = softJoint.jointRb.position;
                float radius = SoftJointSearchRadiusJsf.val;
                float distance = Vector3.Distance(position, vertex);
                if(distance < radius)
                {
                    yield return softJoint;
                }
            }
        }

        void AssignJoints(
            List<SoftJoint> coveredJoints,
            ClothingAdjustmentProfile profile,
            string side
        )
        {
            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            profile.coveredSoftJoints[side] = coveredJoints;
            profile.exposedSoftJoints[side] = context.allSoftJoints.Except(coveredJoints).ToList();
        }

        internal void SetAllJointsCovered(string side)
        {
            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            SelectedProfile.coveredSoftJoints[side] = context.allSoftJoints.ToList();
            SelectedProfile.exposedSoftJoints[side].Clear();
            VisualizeClothingCoverageJssc.val = Constant.COVERAGE;
        }

        internal void SetAllJointsExposed(string side)
        {
            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            SelectedProfile.coveredSoftJoints[side].Clear();
            SelectedProfile.exposedSoftJoints[side] = context.allSoftJoints.ToList();
            VisualizeClothingCoverageJssc.val = Constant.COVERAGE;
        }

        internal void MirrorCoverageJoints(string fromSide, string toSide)
        {
            var toContext = toSide == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            var joints = SelectedProfile.coveredSoftJoints[fromSide];
            var mirrorJoints = SoftPhysicsHandler.FindMirrorJoints(toContext, joints).ToList();
            SelectedProfile.coveredSoftJoints[toSide] = mirrorJoints;
            SelectedProfile.exposedSoftJoints[toSide] = toContext.allSoftJoints.Except(mirrorJoints).ToList();
        }

        const float HOVER_RADIUS = 80f;
        bool _jointToggled;
        float _prevDist;
        string _hoveredOverSide;

        internal void HandleToggleCoverageJoint()
        {
            /* Prevent toggling again while key is down */
            if(_jointToggled)
            {
                if(!Input.GetKey(KeyCode.LeftControl))
                {
                    _jointToggled = false;
                }

                return;
            }

            SelectedProfile.hoveredOverJoint = null;
            _hoveredOverSide = null;
            _prevDist = HOVER_RADIUS;
            HighlightHoveredOverJoint(Side.LEFT, Script.bootstrap.Camera);
            HighlightHoveredOverJoint(Side.RIGHT, Script.bootstrap.Camera);

            if(SelectedProfile.hoveredOverJoint != null && Input.GetKey(KeyCode.LeftControl))
            {
                _jointToggled = true;
                ToggleSoftJointSelection(_hoveredOverSide, SelectedProfile.hoveredOverJoint);
                SyncSoftPhysicsAdjustments();
            }
        }

        void HighlightHoveredOverJoint(string side, Camera camera)
        {
            var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
            foreach(var softJoint in context.allSoftJoints)
            {
                var targetPos = camera.WorldToScreenPoint(softJoint.jointRb.position);
                float dist = Vector2.Distance(Input.mousePosition, targetPos);
                if(dist < HOVER_RADIUS && dist < _prevDist)
                {
                    SelectedProfile.hoveredOverJoint = softJoint;
                    _hoveredOverSide = side;
                    _prevDist = dist;
                }
            }
        }

        void ToggleSoftJointSelection(string side, SoftJoint softJoint)
        {
            if(SelectedProfile.coveredSoftJoints[side].Contains(softJoint))
            {
                SelectedProfile.coveredSoftJoints[side].Remove(softJoint);
                if(!SelectedProfile.exposedSoftJoints[side].Contains(softJoint))
                {
                    SelectedProfile.exposedSoftJoints[side].Add(softJoint);
                }
            }
            else if(SelectedProfile.exposedSoftJoints[side].Contains(softJoint))
            {
                SelectedProfile.exposedSoftJoints[side].Remove(softJoint);
                if(!SelectedProfile.coveredSoftJoints[side].Contains(softJoint))
                {
                    SelectedProfile.coveredSoftJoints[side].Add(softJoint);
                }
            }
        }

        internal void SaveProfile()
        {
            if(SelectedItem == null)
            {
                return;
            }

            presetHandler.UpdatePresetDir(SelectedItem.ProfilesDir);
            presetHandler.savePresetJsu.FileBrowse();
        }

        internal void LoadProfile(bool isSettingTarget = false)
        {
            if(SelectedItem == null)
            {
                return;
            }

            presetHandler.UpdatePresetDir(SelectedItem.ProfilesDir);
            presetHandler.isSettingTarget = isSettingTarget;
            presetHandler.loadPresetJsu.FileBrowse();
        }

        internal void SaveDefault()
        {
            if(SelectedItem == null)
            {
                return;
            }

            presetHandler.UpdatePresetDir(SelectedItem.ProfilesDir);
            presetHandler.SaveDefaultPreset();
        }

        internal void LoadDefault()
        {
            if(SelectedItem == null)
            {
                return;
            }

            presetHandler.UpdatePresetDir(SelectedItem.ProfilesDir);
            presetHandler.LoadDefaultPreset(SelectedItem);
        }

        internal void ResetProfile()
        {
            if(SelectedItem == null)
            {
                return;
            }

            var profile = GetProfile(SelectedItem);
            profile.Reset(Person.IsFemale);
            profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            SyncActiveProfilesAdjustmentsAndMorphAngles(SelectedItem);
            SyncSelectedProfile(Side.LEFT);
        }

        internal void ResetTargetProfile()
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                ProfileTransitionJsf.val = 0f;
                return;
            }

            var profile = GetProfile(SelectedItem);
            profile.Transition.storableParam.val = 0f;
            profile.ResetTarget();
        }

        internal void SetProfileOffAsTarget(ClothingAdjustmentProfile profile)
        {
            if(profile == null)
            {
                return;
            }

            var offParams = NewClothingAdjustmentParams(profile.paramNamePrefix, false, false);
            profile.SetProfileOffAsTarget(offParams);
        }

        internal void SetCurrentTransitionStateAsCurrentProfile()
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                return;
            }

            var profile = GetProfile(SelectedItem);
            if(!profile.HasTarget || profile.Transition.storableParam.val == 0f)
            {
                return;
            }

            profile.SetValuesFromCurrentTransitionState();
            profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            SyncAdjustments();
            Script.SyncMorphsAngleOffsets();
            SyncSelectedProfile(Side.LEFT);
        }

        protected abstract ClothingAdjustmentProfile GetProfile(ClothingItem item);
        protected abstract void SetProfile(ClothingItem item, ClothingAdjustmentProfile profile);

        void SyncSelectedProfile(string side)
        {
            bool isRight = side == Side.RIGHT;
            if(isRight && SelectedProfile.symmetric.Value)
            {
                Script.Loggr.LogError("SyncSelectedProfile: Cannot sync right params to storables if profile is symmetric");
                return;
            }

            SymmetricJsb.valNoCallback = SelectedProfile.symmetric.Value;
            LeftActiveJsb.valNoCallback = SelectedProfile.LeftActive.Value;
            RightActiveJsb.valNoCallback = SelectedProfile.RightActive.Value;
            ProfileTransitionJsf.valNoCallback = SelectedProfile.Transition.storableParam.val;
            _selectedParams = SelectedProfile.@params[side];
            YForceJsf.valNoCallback = _selectedParams.yForce.storableParam.val;
            XForceJsf.valNoCallback = _selectedParams.xForce.storableParam.val;
            ZForceJsf.valNoCallback = _selectedParams.zForce.storableParam.val;
            YXHoldPctJsf.valNoCallback = _selectedParams.yxHoldPct.storableParam.val;
            ZHoldPctJsf.valNoCallback = _selectedParams.zHoldPct.storableParam.val;
            YXDampeningPctJsf.valNoCallback = _selectedParams.yxDampeningPct.storableParam.val;
            ZDampeningPctJsf.valNoCallback = _selectedParams.zDampeningPct.storableParam.val;
            YMorphsOffsetJsf.valNoCallback = _selectedParams.yMorphsOffset.storableParam.val;
            XMorphsOffsetJsf.valNoCallback = _selectedParams.xMorphsOffset.storableParam.val;
            ZMorphsOffsetJsf.valNoCallback = _selectedParams.zMorphsOffset.storableParam.val;

            foreach(var pair in SoftPhysicsJsfs)
            {
                pair.Value.valNoCallback = _selectedParams.softPhysicsParams[pair.Key].storableParam.val;
            }

            foreach(var pair in morphConfigs)
            {
                pair.Value.storableFloat.valNoCallback = _selectedParams.morphParams[pair.Key].Value;
            }

            SelectedProfile.isRightSelected = isRight;

            if(!SelectedProfile.HasSoftJoints())
            {
                VisualizeClothingCoverageJssc.val = Person.IsFemale ? "None" : Constant.COVERAGE;
            }

            Script.uiHandler.clothingWindow.Refresh();
        }

        float _relativeXGravityLeft, _relativeXGravityRight;
        float _relativeYGravityLeft, _relativeYGravityRight;
        float _relativeZGravityLeft, _relativeZGravityRight;

        internal override void PreRegisterParams()
        {
            Script.PreRegisterClothingParam(MorphingTransitionJsf);
            _profilesToPreRegister.ForEach(PreRegisterProfileParams);
        }

        void PreRegisterProfileParams(ClothingAdjustmentProfile profile)
        {
            Script.PreRegisterClothingParam(profile.symmetric.storableParam);
            Script.PreRegisterClothingParam(profile.LeftActive.storableParam);
            Script.PreRegisterClothingParam(profile.RightActive.storableParam);
            Script.PreRegisterClothingParam(profile.Transition.storableParam);
            PreRegisterAdjustmentParams(profile.@params[Side.LEFT]);
            PreRegisterAdjustmentParams(profile.rightCachedParams);
        }

        void PreRegisterAdjustmentParams(ClothingAdjustmentParams @params)
        {
            foreach(var pair in @params.allParams)
            {
                Script.PreRegisterClothingParam(pair.Value.storableParam);
            }

            foreach(var pair in @params.morphParams)
            {
                Script.PreRegisterClothingParam(pair.Value.storableParam);
            }
        }

        void RegisterProfileParams(ClothingAdjustmentProfile profile)
        {
            Script.RegisterBool(profile.symmetric.storableParam);
            Script.RegisterBool(profile.LeftActive.storableParam);
            Script.RegisterBool(profile.RightActive.storableParam);
            Script.RegisterFloat(profile.Transition.storableParam);
            RegisterAdjustmentParams(profile.@params[Side.LEFT]);
            RegisterAdjustmentParams(profile.rightCachedParams);
        }

        void RegisterAdjustmentParams(ClothingAdjustmentParams @params)
        {
            foreach(var pair in @params.allParams)
            {
                Script.RegisterFloat(pair.Value.storableParam);
            }

            foreach(var pair in @params.morphParams)
            {
                Script.RegisterFloat(pair.Value.storableParam);
            }
        }

        internal override void OnFixedUpdate()
        {
            UpdateForces();
            foreach(var pair in morphConfigs)
            {
                SyncMorph(pair.Key, pair.Value);
            }
        }

        void UpdateForces()
        {
            int count = activeProfiles.Count;
            if(count == 0)
            {
                return;
            }

            _relativeXGravityLeft = _gravityPhysicsHandler.CustomGravityLeft.RelativeTo(_leftContext.rigidbodyT, Vector3.right);
            _relativeXGravityRight = _gravityPhysicsHandler.CustomGravityRight.RelativeTo(_rightContext.rigidbodyT, Vector3.right);
            _relativeYGravityLeft = _gravityPhysicsHandler.CustomGravityLeft.RelativeTo(_leftContext.rigidbodyT, Vector3.up);
            _relativeYGravityRight = _gravityPhysicsHandler.CustomGravityRight.RelativeTo(_rightContext.rigidbodyT, Vector3.up);
            _relativeZGravityLeft = _gravityPhysicsHandler.CustomGravityLeft.RelativeTo(_leftContext.rigidbodyT, Vector3.forward);
            _relativeZGravityRight = _gravityPhysicsHandler.CustomGravityRight.RelativeTo(_rightContext.rigidbodyT, Vector3.forward);

            var leftForce = Vector3.zero;
            var rightForce = Vector3.zero;

            if(count == 1)
            {
                var profile = activeProfiles.First();
                if(profile.LeftActive.Value)
                {
                    leftForce = new Vector3(
                        CalculateXForceLeft(profile),
                        CalculateYForceLeft(profile),
                        CalculateZForceLeft(profile)
                    );
                }

                if(profile.RightActive.Value)
                {
                    rightForce = new Vector3(
                        CalculateXForceRight(profile),
                        CalculateYForceRight(profile),
                        CalculateZForceRight(profile)
                    );
                }
            }
            else
            {
                Vector3 leftMin = Vector3.zero, rightMin = Vector3.zero;
                Vector3 leftMax = Vector3.zero, rightMax = Vector3.zero;
                Vector3 leftSum = Vector3.zero, rightSum = Vector3.zero;

                foreach(var profile in activeProfiles)
                {
                    if(profile.LeftActive.Value)
                    {
                        var left = new Vector3(CalculateXForceLeft(profile), CalculateYForceLeft(profile), CalculateZForceLeft(profile));
                        leftMin = Vector3.Min(left, leftMin);
                        leftMax = Vector3.Max(left, leftMax);
                        leftSum += left;
                    }

                    if(profile.RightActive.Value)
                    {
                        var right = new Vector3(CalculateXForceRight(profile), CalculateYForceRight(profile), CalculateZForceRight(profile));
                        rightMin = Vector3.Min(right, rightMin);
                        rightMax = Vector3.Max(right, rightMax);
                        rightSum += right;
                    }
                }

                leftForce = ClampVector(leftSum, leftMin, leftMax);
                rightForce = ClampVector(rightSum, rightMin, rightMax);
            }

            LeftForce = leftForce;
            RightForce = rightForce;

            _leftContext.rigidbody.AddForce(_leftContext.rigidbodyT.TransformDirection(leftForce), ForceMode.Acceleration);
            _rightContext.rigidbody.AddForce(_rightContext.rigidbodyT.TransformDirection(rightForce), ForceMode.Acceleration);
        }

        static Vector3 ClampVector(Vector3 vector, Vector3 min, Vector3 max) => new Vector3(
            vector.x > 0 ? Mathf.Min(vector.x, max.x) : Mathf.Max(vector.x, min.x),
            vector.y > 0 ? Mathf.Min(vector.y, max.y) : Mathf.Max(vector.y, min.y),
            vector.z > 0 ? Mathf.Min(vector.z, max.z) : Mathf.Max(vector.z, min.z)
        );

        float CalculateXForceLeft(ClothingAdjustmentProfile profile)
        {
            float force = profile.GetLeftXForce();
            float target = force - _relativeXGravityLeft;

            /* Squeeze */
            if(force > 0)
            {
                return _relativeXGravityLeft > 0
                    ? Mathf.Max(0, target) /* Rotate right */
                    : force; /* Rotate left */
            }

            /* Separate */
            return _relativeXGravityLeft > 0
                ? force /* Rotate right */
                : Mathf.Min(0, target); /* Rotate left */
        }

        float CalculateXForceRight(ClothingAdjustmentProfile profile)
        {
            float force = profile.GetRightXForce();
            float target = force + _relativeXGravityRight;

            /* Squeeze */
            if(force > 0)
            {
                return _relativeXGravityRight > 0
                    ? -force /* Rotate right */
                    : -Mathf.Max(0, target); /* Rotate left */
            }

            /* Separate */
            return _relativeXGravityRight > 0
                ? -Mathf.Min(0, target) /* Rotate right */
                : -force; /* Rotate left */
        }

        float CalculateYForceLeft(ClothingAdjustmentProfile profile) => CalculateYForce(
            _relativeYGravityLeft,
            _gravityPhysicsHandler.RefDownGravityLeft.y,
            profile.GetLeftYForce()
        );

        float CalculateYForceRight(ClothingAdjustmentProfile profile) => CalculateYForce(
            _relativeYGravityRight,
            _gravityPhysicsHandler.RefDownGravityRight.y,
            profile.GetRightYForce()
        );

        static float CalculateYForce(float relativeGravity, float refDown, float gForce)
        {
            float target = gForce - (relativeGravity - refDown);
            return gForce > 0
                ? Mathf.Max(0, target) /* Lift up */
                : gForce; /* Press down */
        }

        float CalculateZForceLeft(ClothingAdjustmentProfile profile) => CalculateZForce(_relativeZGravityLeft, profile.GetLeftZForce());
        float CalculateZForceRight(ClothingAdjustmentProfile profile) => CalculateZForce(_relativeZGravityRight, profile.GetRightZForce());

        static float CalculateZForce(float relativeGravity, float gForce)
        {
            float target = gForce - relativeGravity;

            /* Pull */
            if(gForce > 0)
            {
                return relativeGravity > 0
                    ? Mathf.Max(0, target) /* Rotate forward */
                    : gForce; /* Rotate back */
            }

            /* Push */
            return relativeGravity > 0
                ? gForce /* Rotate back */
                : Mathf.Min(0, target); /* Rotate forward */
        }

        void SyncMorph(string id, ClothingMorphConfig config)
        {
            float leftValue = 0f, rightValue = 0f;

            foreach(var profile in activeProfiles)
            {
                if(profile.LeftActive.Value)
                {
                    leftValue = Mathf.Max(leftValue, profile.GetLeftMorphParamValue(id));
                }

                if(profile.RightActive.Value)
                {
                    rightValue = Mathf.Max(rightValue, profile.GetRightMorphParamValue(id));
                }
            }

            config.left.SetValue(leftValue);
            config.right.SetValue(rightValue);
        }

        internal void ResetMorphs()
        {
            foreach(var pair in morphConfigs)
            {
                var config = pair.Value;
                config.left.Reset();
                config.right.Reset();
            }
        }

        internal bool LoadSelectedItemProfile(string fullPath, JSONClass jc, bool isSettingTarget = false)
        {
            // Debug.Log($"{Script.DomainName}: LoadSelectedItemProfile fullPath {fullPath}, isSettingTarget {isSettingTarget}");
            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                Script.Loggr.LogError($"LoadSelectedItemProfile: Invalid profile path {fullPath}", false);
                return false;
            }

            return FinishLoadProfile(SelectedItem, parsedPath, jc, isSettingTarget);
        }

        internal bool LoadItemByUidProfile(string uid, string fullPath, JSONClass jc)
        {
            // Debug.Log($"{Script.DomainName}: LoadItemByUidProfile uid {uid}, fullPath {fullPath}");
            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                Script.Loggr.LogError($"LoadItemByUidProfile: Invalid profile path {fullPath}", false);
                return false;
            }

            var item = _store.items[uid];
            return FinishLoadProfile(item, parsedPath, jc);
        }

        internal void LoadEnabledItemProfileByPath(string fullPath, JSONClass jc, bool isSettingTarget = false)
        {
            // Debug.Log($"{Script.DomainName}: LoadEnabledItemProfileByPath fullPath {fullPath}, isSettingTarget {isSettingTarget}");
            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                if(Script.firstCalibrationDone)
                {
                    Script.Loggr.LogError($"LoadEnabledItemByPathProfile: Invalid profile path {fullPath}", false);
                }

                return;
            }

            var firstItem = _store.enabledItems.FirstOrDefault(item => IsValidItem(item) && item.ProfilesDir == parsedPath.itemSubDir);
            if(firstItem == null)
            {
                if(Script.firstCalibrationDone)
                {
                    Script.Loggr.LogError($"No enabled clothing item found for profile '{parsedPath.fileName}' (Full path: {fullPath})", false);
                }

                return;
            }

            FinishLoadProfile(firstItem, parsedPath, jc, isSettingTarget);
        }

        bool FinishLoadProfile(ClothingItem item, ParsedPath parsedPath, JSONClass jc, bool isSettingTarget = false)
        {
            // Debug.Log($"{Script.DomainName}: FinishLoadProfile {item.label}, path {parsedPath.fullPath}, isSettingTarget {isSettingTarget}");
            var profile = GetProfile(item);
            bool isSameItem = DetermineIsSameItem(item, parsedPath);

            if(isSettingTarget)
            {
                RestoreTargetProfileFromJSON(item, profile, jc);
                profile.targetName = BuildProfileName(parsedPath, isSameItem);
                profile.targetFullPath.SetVal(parsedPath.fullPath);
            }
            else
            {
                profile.loadedName = BuildProfileName(parsedPath, isSameItem);
                profile.loadedFullPath = parsedPath.fullPath;
                RestoreProfileFromJSON(item, profile, jc, parsedPath.packageId, isSameItem);
                profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);

                if(Script.firstCalibrationDone)
                {
                    SyncActiveProfilesAdjustmentsAndMorphAngles(item);
                }

                if(profile == SelectedProfile)
                {
                    SyncSelectedProfile(profile.isRightSelected ? Side.RIGHT : Side.LEFT);
                }
            }

            return true;
        }

        static bool DetermineIsSameItem(ClothingItem item, ParsedPath parsedPath)
        {
            bool isSameItem = item.ProfilesDir == parsedPath.itemSubDir;
            if(item.isBuiltIn && !isSameItem)
            {
                bool isSameDefaultBuiltInItem =
                    item.dazItem.creatorName == parsedPath.itemSubDir &&
                    item.dazItem.displayName == parsedPath.fileName.Replace(".json", "");
                isSameItem = isSameDefaultBuiltInItem;
            }

            return isSameItem;
        }

        ParsedPath ParseProfilePath(string fullPath) => new ParsedPath(fullPath, $"{FileUtils.savesDir}/{ClothingProfilesDir}");

        internal void SetSelectedProfileLoadedInfo(string fullPath)
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                Script.Loggr.LogError("SetSelectedProfileLoadedInfo: Selected item or profile was null");
                return;
            }

            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                Script.Loggr.LogError($"SetSelectedProfileLoadedInfo: Invalid profile path {fullPath}", false);
                return;
            }

            SelectedProfile.loadedName = BuildProfileName(parsedPath, true);
            SelectedProfile.loadedFullPath = parsedPath.fullPath;
            SelectedProfile.SaveValues();
        }

        static string BuildProfileName(ParsedPath parsedPath, bool isSameItem)
        {
            string name = parsedPath.fileName.Replace(".json", "");
            if(!string.IsNullOrEmpty(parsedPath.packageId))
            {
                name += $" ({parsedPath.packageId}.var)";
            }
            else if(!isSameItem)
            {
                name += $" ({parsedPath.itemSubDir})";
            }

            return name;
        }

        void RestoreProfileFromJSON(ClothingItem item, ClothingAdjustmentProfile profile, JSONClass jc, string packageId, bool isSameItem)
        {
            if(item == null)
            {
                return;
            }

            string prefix = $"{item.label}:";
            foreach(string key in jc.Keys.ToList())
            {
                JSONUtils.RenameKey(jc, key, $"{prefix}{key}");
            }

            profile.RestoreFromJSON(jc);
            AssignCoveredJointsFromJSON(jc, $"{prefix}Covered Joints", profile, isSameItem);
            SetTargetProfileFromJSON(jc, $"{prefix}Target Profile", profile, packageId);
            profile.SaveValues();
        }

        void AssignCoveredJointsFromJSON(JSONClass jc, string key, ClothingAdjustmentProfile profile, bool isSameItem)
        {
            if(!jc.HasKey(key))
            {
                return;
            }

            var coveredJointsJSON = jc[key].AsObject;
            AssignJointsOnSide(Side.LEFT, coveredJointsJSON, profile, isSameItem);
            AssignJointsOnSide(Side.RIGHT, coveredJointsJSON, profile, isSameItem);
        }

        void AssignJointsOnSide(
            string side,
            JSONClass coveredJoints,
            ClothingAdjustmentProfile profile,
            bool isSameItem
        )
        {
            var coveredJointsIndices = JSONUtils
                .GetJSONArrayEnumerable(coveredJoints, side)
                .Select(node => node.Value)
                .ToList();
            if(coveredJointsIndices.Count == 0)
            {
                return;
            }

            var joints = new List<SoftJoint>();
            if(isSameItem)
            {
                var context = side == Side.LEFT ? _softPhysicsHandler.leftContext : _softPhysicsHandler.rightContext; // TODO
                joints.AddRange(
                    context.allSoftJoints
                        .Where(joint => coveredJointsIndices.Contains($"{joint.softVerticesSet.targetVertex}"))
                        .Distinct()
                );
                AssignJoints(joints, profile, side);
            }

            // set true even if no joints are assigned because soft physics percentage floats are still loaded from json
            profile.isCoverageMapped = true;
        }

        void SetTargetProfileFromJSON(JSONClass jc, string key, ClothingAdjustmentProfile profile, string packageId)
        {
            if(!jc.HasKey(key))
            {
                profile.ResetTarget();
                return;
            }

            string targetFullPath = ParsePresetPath(jc[key], packageId);
            if(targetFullPath == "Profile Off")
            {
                SetProfileOffAsTarget(profile);
            }
            else if(!string.IsNullOrEmpty(targetFullPath) && FileUtils.FileExists(targetFullPath))
            {
                presetHandler.LoadEnabledItemProfileFromPath(targetFullPath, true);
            }
            else
            {
                profile.ResetTarget();
            }
        }

        void RestoreTargetProfileFromJSON(ClothingItem item, ClothingAdjustmentProfile profile, JSONClass jc)
        {
            if(item == null)
            {
                return;
            }

            if(!profile.HasTarget)
            {
                /* Assumes ProfileTransitionJsf doesn't change profile Transition.StorableFloat.val until a target has been selected */
                ProfileTransitionJsf.SetDefaultValNoCallback();
            }

            string prefix = $"{item.label}:";
            foreach(string key in jc.Keys.ToList())
            {
                JSONUtils.RenameKey(jc, key, $"{prefix}{key}");
            }

            profile.targetParams[Side.LEFT] = NewClothingAdjustmentParams(prefix, false, false);
            bool isSymmetric = jc[$"{prefix}{SymmetricJsb.name}"].AsBool;
            profile.targetParams[Side.LEFT].RestoreFromJSON(jc);

            if(isSymmetric)
            {
                profile.targetParams[Side.RIGHT] = profile.targetParams[Side.LEFT];
            }
            else
            {
                profile.targetParams[Side.RIGHT] = NewClothingAdjustmentParams(prefix, true, false);
                profile.targetParams[Side.RIGHT].RestoreFromJSON(jc);
            }

            if(Script.firstCalibrationDone)
            {
                profile.SetTargetAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            }
        }

        string ParsePresetPath(JSONNode jsonNode, string packageId)
        {
            try
            {
                var jsonClass = jsonNode.AsObject;
                if(jsonClass == null)
                {
                    // 1. returns "Profile Off" in the case of target presets
                    // 2. automatically works with pre-v1.0.0 presets and scene files, since in those the path is not wrapped in a JSONClass
                    return jsonNode.Value;
                }

                if(!jsonClass.HasKey("presetName"))
                {
                    return null;
                }

                string value = jsonClass["presetName"].Value;
                if(!string.IsNullOrEmpty(packageId))
                {
                    value = value.Replace("SELF:/", $"{packageId}:/");
                }

                // Debug.Log($"parsed path is {value}");
                return value;
            }
            catch(Exception e)
            {
                Script.Loggr.LogError($"ParsePresetPath: {e}");
                return null;
            }
        }

        internal JSONClass GetSelectedProfileJSON()
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                return null;
            }

            var jc = new JSONClass();
            var scriptJSON = Script.GetFullBaseJSON();
            if(!scriptJSON.HasKey(SelectedItem.label))
            {
                Script.Loggr.LogError($"GetSelectedProfileJSON: No JSON for {SelectedItem.label}");
                return null;
            }

            var profileJSON = scriptJSON[SelectedItem.label].AsObject;
            bool isSymmetric = profileJSON[SymmetricJsb.name].AsBool;
            foreach(string key in profileJSON.Keys)
            {
                if(isSymmetric && key.EndsWith(" (Right)"))
                {
                    continue;
                }

                jc[key] = profileJSON[key];
            }

            if(SelectedProfile.isCoverageMapped && SelectedProfile.HasSoftJoints())
            {
                jc["Covered Joints"] = new JSONClass();
                jc["Covered Joints"][Side.LEFT] = JointTargetVertexIndicesArray(SelectedProfile.coveredSoftJoints[Side.LEFT]);
                jc["Covered Joints"][Side.RIGHT] = JointTargetVertexIndicesArray(SelectedProfile.coveredSoftJoints[Side.RIGHT]);
            }

            if(SelectedProfile.HasTarget && SelectedProfile.targetFullPath.Value != SelectedProfile.loadedFullPath)
            {
                string path = SelectedProfile.targetFullPath.Value;
                if(path == "Profile Off")
                {
                    jc["Target Profile"] = path;
                }
                else
                {
                    jc["Target Profile"] = new JSONClass
                    {
                        // Hacky way to ensure that path is updated to SELF:/ by package manager when packaging the preset with its target preset in the same var
                        ["presetName"] = SelectedProfile.targetFullPath.Value,
                    };
                }
            }

            return jc;
        }

        static JSONArray JointTargetVertexIndicesArray(List<SoftJoint> softJoints)
        {
            var indicesArray = new JSONArray();
            foreach(var joint in softJoints)
            {
                indicesArray.Add(joint.softVerticesSet.targetVertex.ToString());
            }

            return indicesArray;
        }

        internal override void Disable()
        {
            ResetMorphs();
        }

        readonly HashSet<int> _ignorePathHashes = new HashSet<int>
        {
            1525714317, -275405496, 405930046, 468031865, -1726696007, 1646305290, 2039848008, 811889571, -1332967113, -576944167,
            1062655717, -106591142, 1266449527, 1512403363, -1431230130, -1012047721, -2007659129, 947173709, 1149493943, 90397658,
            1934750945, -1623821700, -1875532686, -1732806512, -28391486, 1469299812, -1936219866, 860641266, -1464953570, 1914121434,
            -1822097589, 1913868785, 1227944410, -919481284, 1572369843, 854262714, -30555405, 391029629, 1619955054, -1193687179,
            -495713474, 1471237650, -1403530000, -1548001400, -1306456402, 982758894, -754159802, 517907309, -664712225, -2018327746,
            -758913537, 614364960, -629971554, 404812007, 1251412286, -703061444, 695639964, 246938701, 824144434, 940816876,
            4463534, 851223646, 322157345, 697858755, 1779799571, -583507452, -611846028, 2002127299, -1187678316, -1695622598,
            -159581992, 1703939606, -1544358234, -1081928087, -424195579, -1339422173, -896307972, -1302288204, -1480081492, -839145292,
            914953347, -777363431, -838071991, -1624385022, -658484308, 733676756, -52113837, 192248105, 1251226558, 192544134,
            1443049068, 1649246538, 1358410009, -169110952, 1364886718, 1350305345, -1740077159, 381459904, 1184746512, -130142091,
            90491077, -1949728823, 512258255, -1910482249, -1985455815, -95788827, -583337659, -1169756183, 1742857976, -602180875,
            -816665451, -1694938206, 1719266879, 2101501027, 845641795, -1657536136, -1380743766, 2105635640, 1365711870, -692173599,
            975454884, 2060439769, 846718419, 271369704, 268004276, -1536762050, 1801280420, -1487307976, 799108764, -119605518,
            -1237674018, 117876702, 863050907, -1842447546, -309388879, -1483644999, 1490277736, 1704700366, -1985836145, 1547812819,
            -719754935, 273191626, -404411661, -602431528, -247962987, 1468930874, 1275041178, -491863435, -967096225, 681347020,
            1037185964, -281943241, -1108808554, 977620604, -1808023813, 1723906054, 382957122, -1556153398, -1715009146, -1247258718,
            1666653320, 1603661558, 1288719058, 1168299986, -846976199, -1499353110, -1875603226, 1603714246, 1771642666, 1940795477,
            1332252181, -283969470, -1481790638, -577283859, -74262699, -2073413591, 731929686, -1812078897, -1683634075, -1633781823,
            517975108, -1891209501, -1987583426, 1010376772, 1428478711, -2027945852, 1723502102, 1471062941, 1927097029, 2091662022,
            39778995, 643600941, -488067418, 83777600, -1098415549, -1422797459, -189276142, -1956032323, 938915011, 406321226,
            1152380449, -381099123, 1415859110, 107448166, 276810786, -149606198, -227966906, -618855555, -1126262471, -663236378,
            1806006006, 986341888, 812605246, -1932317241, -946726367, -1301418859, 230290245, 714598022, 1316669877, 855715995,
            -1993129010, 1042032342, 2079341534, 438075354, 784460955, -1065874946, -536614840, 71022180, 1100521073, -1831164960,
            -548205518, 810313189, 1445161960, -77165721, -73073973, 73240951, 417426356, -492100581, 993097730, 2025204634,
            -414409902, -481512780, 1790640906, -2135575149, -1145186464, -609898060, 355383997, 1270064901, 874831623, 1681840325,
            -788839637, 1404639067, -1744731387, 1390217085, 1667518427, -1488306248, 801500534, -2092976812, -1792816991, 1229229639,
            1509416555, 1361771351, -857888152, -1225905105, -1962153109, -1668239481, 1796767516, 1840094545, -1315121568, -1158634520,
            2119140995, 1579085552, -1554200466, -71355764, 37281262, -1544886474, 1335800853, 714619740, 926696272, 503993994,
            71696327, -1057739019, 935368376, 278127644, -1159266955, 1743757633, 872236884, -589045162, -1036661643, -1776809949,
            -629790130, 511952897, -670394491, 593064755, -1789147649, -1056218820, 692906416, -604167855, 1520127037, -146544919,
            1677454903, -1564248756, 1128417519, -361650072, 1219322606, -1295305383, -1189481895, -1229292081, 1569808027, 1197901606,
            354036637, 1670513490, 1960791127, -1703535475, -416898324, 1805402575, 1626913328, 1717067667, -568021668, -783664804,
            -309484386, -387906235, -1340107238, 630785433, 639830115, 587902509, -899099820, -558960662, 1466091640, -196867136,
            -969581833, -512535190, -1996132821, 1880285177, -1055188357, 1029076457, -40314301, 906314505, 1986350661, 639515532,
            757782003, 1695603451, 868354106, 1581532800, -239203897, -209003018, -1747001659, 681435463, 672636512, 232079863,
            1028626559, -749174799, -1850333298, 709321843, 1357142955, -1852711962, -1424212836, -1511233763, 1546117429, 1168420882,
            -869172988, -1068423235, 2042718377, -7744705, 324911963, 593518934, 798077098, -1243012991, -2036140591, -1851628359,
            121302511, 1921339777, 1540182185, 1758902579, 227714335, 368279809, -1395802329, -895410536, -1016413786, 1982676944,
            1082470298, 53199296, -957513338, 832852910, -1358783697, 869334440, 1857922834, 1104257087, -1697994491, -1428322722,
            1763765964, -611210808, 1579010654, 1046849357, -750764732, -1940562685, -1656845384, 964209896, -1572654385, 448010928,
            1251905126, 320889260, -1347511957, 1134131923, 1123303077, -1955982934, 1524978065, 1930556824, -1596182159, -253297123,
            426671363, -1950244943, 299115434, 451937442, 1296083005, 1563243354, 2123054982, 1021831228, -1176314150, -317262596,
            -1998254263, -532472987, 1288970972, 1096943052, -2116747254, 12833779, 1545381274, 540341047, 2097661152, -1856030534,
            -564576806, 2145338038, 2025027653, 1609983717, -173747012, -32850538, -615226217, 657535502, 442759250, 54176895,
            -1799992502, -2087686986, -870984831, -406466872, 753683206, 1418219021, -1909869650, -1063577647, -1959772933, -688717649,
            227743769, 2095839658, -1567080616, 696228970, -1946438442, 1886123357, 2012434838, -2041039869, 1950800964, 2029493524,
            -1123731423, -2146711430, -625189733, 278217170, -1126465309, 2058683739, 849286194, 326942488, 2122379014, -1345393263,
            -1392654224, -1444070572, 2046113519, -540590709, 1200353575, -1283906034, 940042840, 378471359, -632963355, 1618840296,
            1082743114, -402976968, -1583123439, 1686465369, 1480037140, -807045938, 154394889, -186133805, -2037811172, 1694820823,
            574333121, -1567610442, -404948947, 703365027, -428234583, 1518881553, 2090479830, -380735489, 519965761, -984366323,
            964758104, -807558433, -1654972418, -141023366, 1087944009, -1470602109, 1823019688, -1715621706, -309343250, 1924679937,
            -330753127, -1035813409, -1110584775, -1424198734, -1474450146, 1940648247, 257092796, 1056489549, -184487992, -1098395454,
            -828081942, -504133411, 2048862251, 714448403, 266840380, -812241719, -1451530780, 2060278214, -247804378, 1762524073,
            1990884596, 52198470, 1251318700, 822391874, 1627811469, -1470853746, -1348619274, 1736472504, -1649508572, -656285942,
            1314944349, 587691043, -2001190325, -2070247408, 1872181453, 233639202, -1734288595, -1399630287, 1787703613, -926505380,
            1253080885, -1217882515, -1238236639, -1549841273, -2061122216, 1101952855, -836454598, 2087064624, 1421436724, -2030122340,
            -889369296, 882763660, -964993754, -1919284939, -1091202887, -1435503092, -2081835581, -1639428444, -845897168, -406855331,
            -578193046, -1346545009, 1453097569, 773741904, 2107239701, 1282241274, 1092366201, 2121136110, -1280608386, -233485046,
            568217844, 1894127513, -419451412, -1724466317, -1747322704, 834151340, -1239817395, 334981262, 1360134307, -1898014272,
            1938952310, 1214036950, -1381686264, 1818088824, -1861079735, -1509875895, 218186490, 2144100592, 972571429, 2081193294,
            -886868547, 509708964, 1370767252, 518167538, 1037165500, 1127931619, -554661036, 1722525834, 1351928221, -2025634543,
            -146289640, 598653267, 1706960111, 1143167307, 695578157, 2048597400, 468681699, -82603107, -709318591, -460876114,
            1942295948, -40391357, -586800969, 392495911, -79119727, 398406623, 1299686438, 1211814815, 303549200, 1908531635,
            1993949119, -41291978, 587586827, -1644109475, 1989676001, 1171755632, 1338071615, 1969567644, -258406065, -1581012898,
            186962180, 696197964, 1353023595, 22104450, -724643389, 810585466, -513208824, 947978520, 1476892342, -693302172,
            -1982246485, 951664161, -1012440055, -725394659, 1495192312, -1701244092, 653569509, -1741365280, 222527917, 49365284,
            -1651801508, -476395996, 651982699, 549005894, -830066202, 53557357, 419382512, 1662631741, 1278411738, 625704982,
            1199044237, 1023635617, 1758980822, 783153497, -350311008, -220268357, 1921224257, 171176663, -1539462173, -799052359,
            532835184, 443369491, 1202217356, -448276583, -1192758044, 1509903983, -1571147217, 760733176, 350702488, -1939371638,
            -1597178696, 368132273, -932912283, -1447476354, -735267974, -1972838111, -1828123695, -672037600, -1953237502, -1409587456,
            -22130628, -515183307, 1777259385, 980513130, 953831079, 1629508217, 522848654, -1989715269, 2140846269, 1738901045,
            1996883074, -2068579167, 1522224870, 663479786, 2038987763, 1696009048, -724860693, 128561233, 1912890658, -1536418396,
            -1134105541, 1078676412, 42680388, -566536775, 1188214359, 610808781, 1672011358, 1288443314, -525155584, -1160288275,
            792719878, 586872663, -2132964877, 158524387, -1531538822, -690926196, 891630852, -1053867812, 1722627087, 458199734,
            -2129984627, 1912093493, -856867810, 1130632373, -1194703043, -939918963, 1112559679, -1805688874, -1525460542, -1822439828,
            1477668939, -1530157184, -392528798, 934448046, -343347205, -1666654002, -230144634, 1987203345, -2011547092, -2032104138,
            -1914965744, 138414652, -1965405102, 1659729004, -1758573112, 772500200, -1954036351, 5370858, 1548827353, 1774927775,
            462308203, -2091053917, 1416854752, 1446299301, 85076520, -1609457892, 2036364705, 695307146, -17506629, -1847377791,
            300075343, -1427213129, 2066818768, 1843333108, -1058930253, -107121420, -711802, 1326559092, -298435163, 212962312,
            1382252469, 491577690, 1431608819, -1611686186, 1208919761, -2010549842, 972241523, -558712004, -573807704, -1876703887,
            406476893, -132226968, 1796566730, -849009760, 646142434, 2130607453, -57542636, 1070390018, 1112067124, 1725412223,
            -1105429919, -145629608, 1645539130, 50306575, 1052709199, -1767122927, -474144920, 473441917, -249250374, 1010226213,
            1586203552, 1986343049, 1544396782, 206488812, -1392061980, 1520542167, -1935525046, 586135259, -1111593488, -1721854614,
            -1609621604, -2026374948, 226937271, 1653415023, 1363660444, 1123892124, 1557539208, -1833416494, -1491460497, -148536188,
            -34650102, -1975419074, -234124821, -2136860874, 2111191137, -231189813, 1374682055, -673175675, 1348143403, 1460557196,
            -1725831132, -1047448848, 991821358, -1560534336, 1123785092, 1814552472, 1785812021, 1025072014, 1392812817, -1037184232,
            -450407194, 2008455552, -789143855, 460245958, -1221764058, 31286842, -490965286, -1890620737, 1737343505, 442102400,
            -431576363, -1448992479, -895515461, -1599524101, 1102457916, 1563454680, -1804106677, 1133889932, -1116517998, 1265251326,
            687351593, -862750001, -493482261, -1421472999, -483720371, -1365264104, 1295499213, -1429821077, 1673087928, 368420749,
            -368583450, 613349391, 1569627580, -1437512277, -1537056559, 45118814, 1590569497, 1806455926, 202753781, -1077704743,
            -1323806443, 2083426676, -619372595, -1241799496, -1028111294, 1979303857, 1146421647, -1250154507, 2136144751, -865442699,
            1588010290, 2105023835, -877453995, -1797091252, 31355581, 2041701937, -860443818, -980355506, -289717170, -990762445,
            1403472579, 1037382289, 787070969, -575817726, 229240418, 1334012286, 195886845, 996240915, 1983904244, -125091300,
            410894762, 1895679359, 795660101, 1551312842, -1219921072, 933906888, 259065822, 1104149795, -1458784969, 792027582,
            1381475959, 134019434, 1882794656, 1586807957, 1907933806, -588341104, 128391928, 1873009787, -709665372, 2099263637,
            -119482213, 231655738, -1453991691, 1383460554, 1350698521, 801351776, 1266127469, -1485279621, 979763313, -487939994,
            521401686, -41606793, -2108276185, -1323451440, 900409737, 1999073838, -1276470203, 310950641, -677687913, 811524135,
            -1648100927, -949209466, 2086432887, -635909467, -1052635926, 1234006807, -202697730, -776799735, 1269114210, 243140872,
            1780966897, 1261426770, 1447062013, 946227673, -1772034030, 1242648331, 928478126, -2087449152, -1258704942, 921587388,
            -431018214, -65845044, 689659717, -348975151, 452949053, -2087467061, 1763806773, -716607790, -1221990628, 566461039,
            1388288727, 1913285963, -1984028433, -1417538844, 397764496, 1045015451, -628379941, -1336694094, 86520562, -230698568,
            14738052, -1616464762, -820386035, -61710076, -1338145627, -686968936, 584094984, 1512094871, 1423608844, 1977872573,
            -541511149, -439731592, -317338896, -1078216509, 1599672670, -1378814438, 1552692945, 1752253903, 1785293855, 825522919,
            -869830350, -675974275, 1182861438, 1550461466, 102277276, -253185999, 683911080, -1531791385, -697055469,
        };
    }
}
