using NAT.Common.Models;
using NAT.Common.Storables;
using Naturalis.Common.Handlers.Configs;
using SimpleJSON;
using System.Collections.Generic;

namespace NAT.Common.Handlers.Configs
{
    class ClothingAdjustmentParams
    {
        internal readonly Dictionary<string, AdjustmentFloatParam> allParams = new Dictionary<string, AdjustmentFloatParam>();
        internal readonly Dictionary<string, AdjustmentFloatParam> softPhysicsParams = new Dictionary<string, AdjustmentFloatParam>();
        internal readonly Dictionary<string, AdjustmentFloatParam> morphParams = new Dictionary<string, AdjustmentFloatParam>();

        internal readonly AdjustmentFloatParam yForce;
        internal readonly AdjustmentFloatParam xForce;
        internal readonly AdjustmentFloatParam zForce;
        internal readonly AdjustmentFloatParam yxHoldPct;
        internal readonly AdjustmentFloatParam zHoldPct;
        internal readonly AdjustmentFloatParam yxDampeningPct;
        internal readonly AdjustmentFloatParam zDampeningPct;
        internal readonly AdjustmentFloatParam yMorphsOffset;
        internal readonly AdjustmentFloatParam xMorphsOffset;
        internal readonly AdjustmentFloatParam zMorphsOffset;

        internal ClothingAdjustmentParams(
            JSONStorableFloat yForceJsf,
            JSONStorableFloat xForceJsf,
            JSONStorableFloat zForceJsf,
            JSONStorableFloat yxHoldPctJsf,
            JSONStorableFloat zHoldPctJsf,
            JSONStorableFloat yxDampeningPctJsf,
            JSONStorableFloat zDampeningPctJsf,
            JSONStorableFloat yMorphsOffsetJsf,
            JSONStorableFloat xMorphsOffsetJsf,
            JSONStorableFloat zMorphsOffsetJsf,
            Dictionary<string, JSONStorableFloat> softPhysicsJsfs,
            Dictionary<string, ClothingMorphConfig> morphConfigs,
            string prefix,
            string suffix
        )
        {
            yForce = new AdjustmentFloatParam(StorableFloat.Copy(yForceJsf, prefix, suffix, true));
            xForce = new AdjustmentFloatParam(StorableFloat.Copy(xForceJsf, prefix, suffix, true));
            zForce = new AdjustmentFloatParam(StorableFloat.Copy(zForceJsf, prefix, suffix, true));
            yxHoldPct = new AdjustmentFloatParam(StorableFloat.Copy(yxHoldPctJsf, prefix, suffix, true));
            zHoldPct = new AdjustmentFloatParam(StorableFloat.Copy(zHoldPctJsf, prefix, suffix, true));
            yxDampeningPct = new AdjustmentFloatParam(StorableFloat.Copy(yxDampeningPctJsf, prefix, suffix, true));
            zDampeningPct = new AdjustmentFloatParam(StorableFloat.Copy(zDampeningPctJsf, prefix, suffix, true));
            yMorphsOffset = new AdjustmentFloatParam(StorableFloat.Copy(yMorphsOffsetJsf, prefix, suffix, true));
            xMorphsOffset = new AdjustmentFloatParam(StorableFloat.Copy(xMorphsOffsetJsf, prefix, suffix, true));
            zMorphsOffset = new AdjustmentFloatParam(StorableFloat.Copy(zMorphsOffsetJsf, prefix, suffix, true));

            allParams.Add(yForceJsf.name, yForce);
            allParams.Add(xForceJsf.name, xForce);
            allParams.Add(zForceJsf.name, zForce);
            allParams.Add(yxHoldPctJsf.name, yxHoldPct);
            allParams.Add(zHoldPctJsf.name, zHoldPct);
            allParams.Add(yxDampeningPctJsf.name, yxDampeningPct);
            allParams.Add(zDampeningPctJsf.name, zDampeningPct);
            allParams.Add(yMorphsOffsetJsf.name, yMorphsOffset);
            allParams.Add(xMorphsOffsetJsf.name, xMorphsOffset);
            allParams.Add(zMorphsOffsetJsf.name, zMorphsOffset);

            if(softPhysicsJsfs != null)
            {
                foreach(var pair in softPhysicsJsfs)
                {
                    var adjustmentFloatParam = new AdjustmentFloatParam(StorableFloat.Copy(pair.Value, prefix, suffix, true));
                    softPhysicsParams[pair.Key] = adjustmentFloatParam;
                    allParams.Add(pair.Value.name, adjustmentFloatParam);
                }
            }

            foreach(var pair in morphConfigs)
            {
                morphParams[pair.Key] = new AdjustmentFloatParam(StorableFloat.Copy(pair.Value.storableFloat, prefix, suffix, true));
            }
        }

        internal void SetAllAdjustmentValues(
            float upDownMassMultiplier,
            float sidewaysMassMultiplier,
            float inOutMassMultiplier,
            bool isBootyMagic
        )
        {
            SetYForce(yForce.storableParam.val, upDownMassMultiplier);
            SetXForce(xForce.storableParam.val, sidewaysMassMultiplier);
            SetZForce(zForce.storableParam.val, inOutMassMultiplier, !isBootyMagic);
            SetPercentageValue(yxHoldPct);
            SetPercentageValue(zHoldPct);
            SetPercentageValue(yxDampeningPct);
            SetPercentageValue(zDampeningPct);
            SetYMorphsOffset(yMorphsOffset.storableParam.val, isBootyMagic);
            SetXMorphsOffset(xMorphsOffset.storableParam.val, isBootyMagic);
            SetZMorphsOffset(zMorphsOffset.storableParam.val, isBootyMagic);
            foreach(var pair in softPhysicsParams)
            {
                SetPercentageValue(pair.Value);
            }
        }

        internal void SetYForce(float value, float multiplier) => yForce.adjustedValue = CalculateForceValue(value, multiplier);
        internal void SetXForce(float value, float multiplier) => xForce.adjustedValue = CalculateForceValue(value, multiplier);
        internal void SetZForce(float value, float multiplier, bool invert) => zForce.adjustedValue = CalculateForceValue(value, multiplier, invert);

        static float CalculateForceValue(float value, float multiplier, bool invert = false)
        {
            float force = value * Constant.G * multiplier;
            if(invert)
            {
                force = -force;
            }

            return force;
        }

        internal static void SetPercentageValue(AdjustmentFloatParam param) => param.adjustedValue = Calc.PercentageValue(param.storableParam.val);
        internal void SetYMorphsOffset(float value, bool invert) => yMorphsOffset.adjustedValue = CalculateMorphsOffsetValue(value, invert);
        internal void SetXMorphsOffset(float value, bool invert) => xMorphsOffset.adjustedValue = CalculateMorphsOffsetValue(value, invert);
        internal void SetZMorphsOffset(float value, bool invert) => zMorphsOffset.adjustedValue = CalculateMorphsOffsetValue(value, invert, 0.001f);
        static float CalculateMorphsOffsetValue(float value, bool invert, float multiplier = 1) => multiplier * (invert ? -value : value);

        internal void CopyFrom(ClothingAdjustmentParams profile)
        {
            foreach(var pair in allParams)
            {
                pair.Value.CopyFrom(profile.allParams[pair.Key]);
            }

            foreach(var pair in profile.morphParams)
            {
                morphParams[pair.Key].SetVal(pair.Value.Value);
            }
        }

        internal void RestoreFromJSON(JSONClass jc)
        {
            foreach(var pair in allParams)
            {
                pair.Value.SetValFromJSON(jc);
            }

            foreach(var pair in morphParams)
            {
                var param = pair.Value;
                param.SetValNoCallback(jc[param.Name].AsFloat);
            }
        }

        internal void Reset()
        {
            foreach(var pair in allParams)
            {
                pair.Value.SetDefaultValNoCallback();
            }

            foreach(var pair in morphParams)
            {
                pair.Value.SetDefaultValNoCallback();
            }
        }

        internal bool IsSaved()
        {
            foreach(var pair in allParams)
            {
                if(!pair.Value.IsSaved)
                {
                    return false;
                }
            }

            foreach(var pair in morphParams)
            {
                if(!pair.Value.IsSaved)
                {
                    return false;
                }
            }

            return true;
        }

        internal void SaveValues()
        {
            foreach(var pair in allParams)
            {
                pair.Value.SaveValue();
            }

            foreach(var pair in morphParams)
            {
                pair.Value.SaveValue();
            }
        }
    }
}
