using NAT.Common.Storables;
using NAT.Common.Models;
using Naturalis.Common.Handlers.Configs;
using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Handlers.Configs
{
    sealed class ClothingAdjustmentProfile
    {
        readonly ClothingItem _item;
        internal readonly string paramNamePrefix;
        internal string loadedName = "None";
        internal string targetName = "";
        internal string loadedFullPath;
        internal readonly AdjustmentStringParam targetFullPath;
        internal readonly AdjustmentBoolParam symmetric;
        internal readonly AdjustmentBoolParam offWhenUndressing;

        internal readonly Dictionary<string, AdjustmentBoolParam> activeStorableBools = new Dictionary<string, AdjustmentBoolParam>();
        internal AdjustmentBoolParam LeftActive => activeStorableBools[Side.LEFT];
        internal AdjustmentBoolParam RightActive => activeStorableBools[Side.RIGHT];

        internal readonly Dictionary<string, ClothingAdjustmentParams> @params = new Dictionary<string, ClothingAdjustmentParams>();
        ClothingAdjustmentParams LeftParams => @params[Side.LEFT];
        ClothingAdjustmentParams RightParams => @params[Side.RIGHT];
        internal ClothingAdjustmentParams rightCachedParams;

        internal readonly Dictionary<string, ClothingAdjustmentParams> targetParams = new Dictionary<string, ClothingAdjustmentParams>
        {
            { Side.LEFT, null },
            { Side.RIGHT, null },
        };

        ClothingAdjustmentParams LeftTargetParams => targetParams[Side.LEFT];
        ClothingAdjustmentParams RightTargetParams => targetParams[Side.RIGHT];
        internal AdjustmentFloatParam Transition { get; }

        internal readonly Dictionary<string, List<SoftJoint>> coveredSoftJoints = new Dictionary<string, List<SoftJoint>>
        {
            { Side.LEFT, new List<SoftJoint>() },
            { Side.RIGHT, new List<SoftJoint>() },
        };

        internal readonly Dictionary<string, List<SoftJoint>> exposedSoftJoints = new Dictionary<string, List<SoftJoint>>
        {
            { Side.LEFT, new List<SoftJoint>() },
            { Side.RIGHT, new List<SoftJoint>() },
        };

        readonly Dictionary<string, List<SoftJoint>> _savedCoveredSoftJoints = new Dictionary<string, List<SoftJoint>>
        {
            { Side.LEFT, new List<SoftJoint>() },
            { Side.RIGHT, new List<SoftJoint>() },
        };

        internal SoftJoint hoveredOverJoint;

        internal bool isRightSelected;
        internal bool IsActive => LeftActive.Value || RightActive.Value;
        internal bool isCoverageMapped;
        internal bool HasTarget => targetParams[Side.LEFT] != null || targetParams[Side.RIGHT] != null;
        bool _isSymmetricUnset;

        internal ClothingAdjustmentProfile(
            ClothingItem item,
            string paramNamePrefix,
            StorableBool symmetricJsb,
            StorableBool leftActiveJsb,
            StorableBool rightActiveJsb,
            StorableFloat transition,
            StorableBool profileOffWhenUndressingJsb
        )
        {
            _item = item;
            this.paramNamePrefix = paramNamePrefix;
            symmetric = new AdjustmentBoolParam(symmetricJsb);
            activeStorableBools[Side.LEFT] = new AdjustmentBoolParam(leftActiveJsb);
            activeStorableBools[Side.RIGHT] = new AdjustmentBoolParam(rightActiveJsb);
            Transition = new AdjustmentFloatParam(transition);
            offWhenUndressing = new AdjustmentBoolParam(profileOffWhenUndressingJsb);
            targetFullPath = new AdjustmentStringParam(new StorableString("Target Full Path", ""));
        }

        internal string GetItemName() => _item.dazItem.displayName;

        internal float GetLeftXForce() => GetParamValue(LeftParams.xForce, LeftTargetParams?.xForce);
        internal float GetRightXForce() => GetParamValue(RightParams.xForce, RightTargetParams?.xForce);
        internal float GetLeftYForce() => GetParamValue(LeftParams.yForce, LeftTargetParams?.yForce);
        internal float GetRightYForce() => GetParamValue(RightParams.yForce, RightTargetParams?.yForce);
        internal float GetLeftZForce() => GetParamValue(LeftParams.zForce, LeftTargetParams?.zForce);
        internal float GetRightZForce() => GetParamValue(RightParams.zForce, RightTargetParams?.zForce);
        internal float GetLeftYXHoldPct() => GetParamValue(LeftParams.yxHoldPct, LeftTargetParams?.yxHoldPct);
        internal float GetRightYXHoldPct() => GetParamValue(RightParams.yxHoldPct, RightTargetParams?.yxHoldPct);
        internal float GetLeftZHoldPct() => GetParamValue(LeftParams.zHoldPct, LeftTargetParams?.zHoldPct);
        internal float GetRightZHoldPct() => GetParamValue(RightParams.zHoldPct, RightTargetParams?.zHoldPct);
        internal float GetLeftYXDampeningPct() => GetParamValue(LeftParams.yxDampeningPct, LeftTargetParams?.yxDampeningPct);
        internal float GetRightYXDampeningPct() => GetParamValue(RightParams.yxDampeningPct, RightTargetParams?.yxDampeningPct);
        internal float GetLeftZDampeningPct() => GetParamValue(LeftParams.zDampeningPct, LeftTargetParams?.zDampeningPct);
        internal float GetRightZDampeningPct() => GetParamValue(RightParams.zDampeningPct, RightTargetParams?.zDampeningPct);
        internal float GetLeftYMorphsOffset() => GetParamValue(LeftParams.yMorphsOffset, LeftTargetParams?.yMorphsOffset);
        internal float GetRightYMorphsOffset() => GetParamValue(RightParams.yMorphsOffset, RightTargetParams?.yMorphsOffset);
        internal float GetLeftXMorphsOffset() => GetParamValue(LeftParams.xMorphsOffset, LeftTargetParams?.xMorphsOffset);
        internal float GetRightXMorphsOffset() => GetParamValue(RightParams.xMorphsOffset, RightTargetParams?.xMorphsOffset);
        internal float GetLeftZMorphsOffset() => GetParamValue(LeftParams.zMorphsOffset, LeftTargetParams?.zMorphsOffset);
        internal float GetRightZMorphsOffset() => GetParamValue(RightParams.zMorphsOffset, RightTargetParams?.zMorphsOffset);

        float GetParamValue(AdjustmentFloatParam param, AdjustmentFloatParam transitionParam) => transitionParam == null
            ? param.adjustedValue
            : Mathf.Lerp(param.adjustedValue, transitionParam.adjustedValue, Transition.adjustedValue);

        internal float GetLeftMorphParamValue(string id) => LeftTargetParams != null
            ? Mathf.Lerp(LeftParams.morphParams[id].Value, LeftTargetParams.morphParams[id].Value, Transition.adjustedValue)
            : LeftParams.morphParams[id].Value;

        internal float GetRightMorphParamValue(string id) => RightTargetParams != null
            ? Mathf.Lerp(RightParams.morphParams[id].Value, RightTargetParams.morphParams[id].Value, Transition.adjustedValue)
            : RightParams.morphParams[id].Value;

        internal float GetSoftPhysicsParamValue(string side, string paramName)
        {
            float value = @params[side].softPhysicsParams[paramName].adjustedValue;
            return targetParams[side] != null
                ? Mathf.Lerp(value, targetParams[side].softPhysicsParams[paramName].adjustedValue, Transition.adjustedValue)
                : value;
        }

        internal void SetAllAdjustmentValues(
            float upDownMassMultiplier,
            float sidewaysMassMultiplier,
            float inOutMassMultiplier,
            bool isBootyMagic
        )
        {
            SetTransitionValue();
            LeftParams.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            if(!symmetric.storableParam.val)
            {
                RightParams.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            }

            SetTargetAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
        }

        internal void SetTargetAdjustmentValues(
            float upDownMassMultiplier,
            float sidewaysMassMultiplier,
            float inOutMassMultiplier,
            bool isBootyMagic
        )
        {
            LeftTargetParams?.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            if(!symmetric.storableParam.val)
            {
                RightTargetParams?.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, isBootyMagic);
            }
        }

        internal void SetTransitionValue() => Transition.adjustedValue = 0.01f * Transition.storableParam.val;

        internal void SyncSymmetric(bool value)
        {
            if(value)
            {
                rightCachedParams = @params[Side.RIGHT];
                @params[Side.RIGHT] = @params[Side.LEFT];
            }
            else
            {
                @params[Side.RIGHT] = rightCachedParams;
                if(!_isSymmetricUnset)
                {
                    @params[Side.RIGHT].CopyFrom(@params[Side.LEFT]);
                    _isSymmetricUnset = true;
                }
            }
        }

        internal bool HasSoftJoints()
        {
            foreach(var pair in coveredSoftJoints)
            {
                if(pair.Value.Count > 0)
                {
                    return true;
                }
            }

            return false;
        }

        internal void RestoreFromJSON(JSONClass jc)
        {
            symmetric.SetValNoCallback(jc[symmetric.Name].AsBool);
            LeftActive.SetValNoCallback(jc[LeftActive.Name].AsBool);
            Transition.SetValNoCallback(jc[Transition.storableParam.name].AsFloat);
            LeftParams.RestoreFromJSON(jc);
            if(symmetric.Value)
            {
                @params[Side.RIGHT] = LeftParams;
                RightActive.SetValNoCallback(LeftActive.Value);
                _isSymmetricUnset = false;
            }
            else
            {
                @params[Side.RIGHT] = rightCachedParams;
                RightParams.RestoreFromJSON(jc);
                RightActive.SetValNoCallback(jc[RightActive.Name].AsBool);
                _isSymmetricUnset = true;
            }
        }

        internal void SaveValues()
        {
            symmetric.SaveValue();
            LeftActive.SaveValue();
            RightActive.SaveValue();
            Transition.SaveValue();
            _savedCoveredSoftJoints[Side.LEFT] = coveredSoftJoints[Side.LEFT].ToList();
            _savedCoveredSoftJoints[Side.RIGHT] = coveredSoftJoints[Side.RIGHT].ToList();
            LeftParams.SaveValues();
            targetFullPath.SaveValue();
            if(!symmetric.Value)
            {
                RightParams.SaveValues();
            }
        }

        internal void Reset(bool isFemale)
        {
            symmetric.SetDefaultValNoCallback();
            LeftActive.SetDefaultValNoCallback();
            RightActive.SetDefaultValNoCallback();
            Transition.SetDefaultValNoCallback();
            LeftParams.Reset();
            if(symmetric.Value)
            {
                @params[Side.RIGHT] = LeftParams;
                RightActive.SetValNoCallback(LeftActive.Value);
                _isSymmetricUnset = false;
            }
            else
            {
                RightParams.Reset();
                _isSymmetricUnset = true;
            }

            coveredSoftJoints[Side.LEFT].Clear();
            coveredSoftJoints[Side.RIGHT].Clear();
            exposedSoftJoints[Side.LEFT].Clear();
            exposedSoftJoints[Side.RIGHT].Clear();
            if(isFemale)
            {
                isCoverageMapped = false;
            }

            loadedName = "None";
            loadedFullPath = null;
            ResetTarget();
            SaveValues();
        }

        internal void ResetTarget()
        {
            targetParams[Side.LEFT] = null;
            targetParams[Side.RIGHT] = null;
            targetName = "";
            targetFullPath.SetDefaultValNoCallback();
        }

        internal void SetProfileOffAsTarget(ClothingAdjustmentParams offParams)
        {
            targetParams[Side.LEFT] = offParams;
            targetParams[Side.RIGHT] = offParams;
            targetName = "Profile Off";
            targetFullPath.SetVal("Profile Off");
        }

        internal void SetValuesFromCurrentTransitionState()
        {
            if(LeftTargetParams != RightTargetParams)
            {
                symmetric.SetVal(false);
            }

            SetParamsOnSideFromCurrentTransitionState(Side.LEFT);
            if(!symmetric.Value)
            {
                SetParamsOnSideFromCurrentTransitionState(Side.RIGHT);
            }

            Transition.storableParam.valNoCallback = 0f;
        }

        void SetParamsOnSideFromCurrentTransitionState(string side)
        {
            var sideParams = targetParams[side];
            foreach(var pair in @params[side].allParams)
            {
                var param = pair.Value;
                var targetParam = sideParams.allParams[pair.Key];
                param.SetValNoCallback(Mathf.Lerp(param.Value, targetParam.Value, Transition.adjustedValue));
            }

            foreach(var pair in @params[side].morphParams)
            {
                var param = pair.Value;
                var targetParam = sideParams.morphParams[pair.Key];
                param.SetValNoCallback(Mathf.Lerp(param.Value, targetParam.Value, Transition.adjustedValue));
            }
        }

        internal bool IsSaved =>
            symmetric.IsSaved &&
            LeftActive.IsSaved &&
            RightActive.IsSaved &&
            CoveredJointsIsSaved(Side.LEFT) &&
            CoveredJointsIsSaved(Side.RIGHT) &&
            Transition.IsSaved &&
            targetFullPath.IsSaved &&
            (symmetric.Value ? LeftParams.IsSaved() : LeftParams.IsSaved() && RightParams.IsSaved());

        bool CoveredJointsIsSaved(string side) => coveredSoftJoints[side]
            .OrderBy(joint => joint.softVerticesSet.targetVertex)
            .SequenceEqual(_savedCoveredSoftJoints[side].OrderBy(joint => joint.softVerticesSet.targetVertex));
    }
}
