using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;

namespace NAT.Common.Handlers
{
    abstract class FrictionHandler : Handler
    {
        internal StorableFloat FrictionOffsetJsf { get; private set; }
        internal JSONStorableFloat FrictionValueJsf { get; private set; }
        internal StorableBool AdaptiveFrictionJsb { get; private set; }
        internal StorableFloat DrySkinFrictionJsf { get; private set; }

        JSONStorable _skinMaterialsStorable;
        protected JSONStorableFloat glossJsf;
        protected JSONStorableFloat specularBumpinessJsf;

        internal bool IsEnabled { get; private set; }

        protected FrictionHandler(ScriptBase script) : base(script)
        {
        }

        protected void SetupStorables()
        {
            AdaptiveFrictionJsb = new StorableBool("Use Adaptive Friction", false)
            {
                setCallbackFunction = value =>
                {
                    Script.uiModManager.drySkinFrictionSlider.SetActiveStyle(value, true);
                    CalculateFriction();
                },
            };
            FrictionOffsetJsf = new StorableFloat("Friction Offset", 0, -0.5f, 0.5f)
            {
                setCallbackFunction = CalculateFrictionCallback,
            };
            FrictionValueJsf = new JSONStorableFloat("Friction Value", 1.00f, 0.00f, 1.00f, false)
            {
                setCallbackFunction = UpdateSoftColliders,
            };
            DrySkinFrictionJsf = new StorableFloat("Dry Skin Friction", 1.000f, 0.000f, 1.000f)
            {
                setCallbackFunction = CalculateFrictionCallback,
            };

            _skinMaterialsStorable = Script.containingAtom.GetStorableByID("skin");
            if(_skinMaterialsStorable)
            {
                SetGlossAndSpecularStorables();
                IsEnabled = true;
            }
            else
            {
                IsEnabled = false;
            }
        }

        internal void Refresh()
        {
            _skinMaterialsStorable = Script.containingAtom.GetStorableByID("skin");
            IsEnabled = _skinMaterialsStorable;
            if(IsEnabled)
            {
                SetGlossAndSpecularStorables();
                LoadSettings();
                CalculateFriction();
            }
        }

        protected static PhysicsCalculator FatColliderRadiusCalculator() => new PhysicsCalculator
        {
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = 0.0090f,
                multiplierB = 0,
                curveA = x => 1.75f * Calc.InverseSmoothStep(x, 0.4f, 0.7f),
            },
        };

        protected abstract void UpdateSoftColliders(float friction);

        void SetGlossAndSpecularStorables()
        {
            glossJsf = _skinMaterialsStorable.GetFloatJSONParam("Gloss");
            glossJsf.setCallbackFunction += CalculateFrictionCallback;

            specularBumpinessJsf = _skinMaterialsStorable.GetFloatJSONParam("Specular Bumpiness");
            specularBumpinessJsf.setCallbackFunction += CalculateFrictionCallback;
        }

        void CalculateFrictionCallback(float _) => CalculateFriction();

        internal override void PreRegisterParams()
        {
            Script.PreRegisterFrictionParam(AdaptiveFrictionJsb);
            Script.PreRegisterFrictionParam(FrictionOffsetJsf);
            Script.PreRegisterFrictionParam(DrySkinFrictionJsf);
        }

        internal abstract void CalculateFriction();

        internal override void Destroy()
        {
            if(glossJsf != null)
            {
                glossJsf.setCallbackFunction -= CalculateFrictionCallback;
            }

            if(specularBumpinessJsf != null)
            {
                specularBumpinessJsf.setCallbackFunction -= CalculateFrictionCallback;
            }
        }
    }
}
