using NAT.Common.Contexts;
using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using NAT.Store;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using static NAT.Direction;

namespace NAT.Common.Handlers
{
    abstract class GravityPhysicsHandler : Handler
    {
        readonly AngleCalculator _angleCalculator;
        protected readonly JointPhysicsHandler jointPhysicsHandler;
        internal StorableFloat twistAngleTargetJsf;
        readonly ClothingStore _clothingStore;
        readonly HairStore _hairStore;

        protected readonly GravityPhysicsContext leftContext = new GravityPhysicsContext(LEFT);
        protected readonly GravityPhysicsContext rightContext = new GravityPhysicsContext(RIGHT);

        readonly Rigidbody[] _gravityRigidbodies;

        internal StorableBool LinkDirectionsJsb { get; }
        internal StorableFloat UpGravityJsf { get; }
        internal StorableFloat DownGravityJsf { get; }
        internal StorableFloat BackGravityJsf { get; }
        internal StorableFloat ForwardGravityJsf { get; }
        internal StorableFloat SidewaysGravityJsf { get; }
        internal StorableFloat[] UserMultipliers { get; }

        float _transition;
        Vector3 _prevSimulationForce = Vector3.zero;
        Vector3 _prevDefaultGravityForce = Vector3.zero;

        protected float upGravityMultiplier,
            downGravityMultiplier,
            forwardGravityMultiplier,
            backGravityMultiplier,
            sidewaysOutGravityMultiplier,
            sidewaysInGravityMultiplier;

        float _upGravityLeft,
            _upGravityRight,
            _downGravityLeft,
            _downGravityRight,
            _backGravityLeft,
            _backGravityRight,
            _forwardGravityLeft,
            _forwardGravityRight,
            _sidewaysInGravityLeft,
            _sidewaysInGravityRight,
            _sidewaysOutGravityLeft,
            _sidewaysOutGravityRight;

        internal float ExtraDownGravityLeft { get; private set; }
        internal float ExtraDownGravityRight { get; private set; }

        internal Vector3 RefDownGravityLeft { get; private set; }
        internal Vector3 RefDownGravityRight { get; private set; }
        internal Vector3 RefUpGravityLeft { get; private set; }
        internal Vector3 RefUpGravityRight { get; private set; }
        internal Vector3 RefLeftGravityLeft { get; private set; }
        internal Vector3 RefLeftGravityRight { get; private set; }
        internal Vector3 RefRightGravityLeft { get; private set; }
        internal Vector3 RefRightGravityRight { get; private set; }
        internal Vector3 RefBackGravityLeft { get; private set; }
        internal Vector3 RefBackGravityRight { get; private set; }
        internal Vector3 RefForwardGravityLeft { get; private set; }
        internal Vector3 RefForwardGravityRight { get; private set; }

        internal Vector3 CustomGravityLeft { get; private set; }
        internal Vector3 CustomGravityRight { get; private set; }

        internal readonly Dictionary<string, StorableAction> simulateGravityActions = new Dictionary<string, StorableAction>();
        internal StorableAction StopGravitySimulationAction { get; }
        internal bool IsSimulatingGravity { get; private set; }
        internal string GravitySimulationDirection { get; private set; }

        protected GravityPhysicsHandler(ScriptBase script) : base(script)
        {
            _angleCalculator = Script.angleCalculator;
            jointPhysicsHandler = Script.jointPhysicsHandler;
            _clothingStore = Script.bootstrap.ClothingStore;
            _hairStore = Script.bootstrap.HairStore;

            if(Script.IsBootyMagic)
            {
                leftContext.rigidbody = Person.LeftGlute.rigidbody;
                rightContext.rigidbody = Person.RightGlute.rigidbody;
            }
            else if(Script.IsTittyMagic)
            {
                leftContext.rigidbody = Person.LeftBreast.rigidbody;
                rightContext.rigidbody = Person.RightBreast.rigidbody;
            }

            _gravityRigidbodies = Person.atom.realRigidbodies
                .Where(rigidbody => rigidbody.useGravity)
                .Where(rigidbody => !rigidbody.name.EndsWith("Link"))
                .Where(rigidbody => !rigidbody.isKinematic)
                .Where(rigidbody => rigidbody.gameObject.activeSelf)
                .Where(rigidbody => rigidbody != Person.LeftGlute?.rigidbody && rigidbody != Person.RightGlute?.rigidbody)
                .Where(rigidbody => rigidbody != Person.LeftBreast.rigidbody && rigidbody != Person.RightBreast.rigidbody)
                .ToArray();

            float max = Script.IsBootyMagic ? 3.00f : 2.00f;
            LinkDirectionsJsb = new StorableBool("Link Directions", true)
            {
                setCallbackFunction = OnToggleLinkMultipliers,
            };
            UpGravityJsf = new StorableFloat("Up Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            DownGravityJsf = new StorableFloat("Down Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            BackGravityJsf = new StorableFloat("Back Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            ForwardGravityJsf = new StorableFloat("Forward Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };
            SidewaysGravityJsf = new StorableFloat("Sideways Gravity", 1.00f, 0.00f, max)
            {
                setJSONCallbackFunction = OnSetGravityMultiplier,
            };

            UserMultipliers = new[] { UpGravityJsf, DownGravityJsf, BackGravityJsf, ForwardGravityJsf, SidewaysGravityJsf };

            string[] gravityDirections =
            {
                UP,
                DOWN,
                BACK,
                FORWARD,
                LEFT,
                RIGHT,
            };

            foreach(string direction in gravityDirections)
            {
                simulateGravityActions[direction] = new StorableAction($"Simulate {direction} Gravity", () => SetGravityDirection(direction));
            }

            StopGravitySimulationAction = new StorableAction("Stop Gravity Simulation", UnsetGravityDirection);
        }

        void OnToggleLinkMultipliers(bool value)
        {
            if(Script.restoringFromJSON)
            {
                return;
            }

            if(value)
            {
                SyncMultipliersToJsf(UpGravityJsf);
            }
        }

        void OnSetGravityMultiplier(JSONStorableFloat jsf)
        {
            if(LinkDirectionsJsb.val)
            {
                SyncMultipliersToJsf(jsf);
            }

            FinishSetGravityMultiplier(jsf);
            SetGravityValues();
        }

        void SyncMultipliersToJsf(JSONStorableFloat reference)
        {
            float multiplier = reference.val;
            foreach(var jsf in UserMultipliers)
            {
                if(jsf == reference)
                {
                    continue;
                }

                jsf.valNoCallback = multiplier;
                FinishSetGravityMultiplier(jsf);
            }
        }

        void FinishSetGravityMultiplier(JSONStorableFloat jsf)
        {
            var storable = (StorableFloat) jsf;
            Script.SetCalibrationPending(storable);
        }

        internal override void PreRegisterParams()
        {
            foreach(var pair in simulateGravityActions)
            {
                Script.PreRegisterGeneralParam(pair.Value);
            }

            Script.PreRegisterGeneralParam(StopGravitySimulationAction);
        }

        internal void AddGravitySimulationForce(bool fullBody)
        {
            var simulatedGravityLeft = Utils.DirectionVector(leftContext.rigidbody, GravitySimulationDirection) * Physics.gravity.magnitude;
            var simulatedGravityRight = Utils.DirectionVector(rightContext.rigidbody, GravitySimulationDirection) * Physics.gravity.magnitude;
            var leftForce = CalculateCustomGravityLeft(simulatedGravityLeft);
            var rightForce = CalculateCustomGravityRight(simulatedGravityRight);
            var defaultGravityForce = (leftForce + rightForce).normalized * Physics.gravity.magnitude;

            CustomGravityLeft = 0.5f * (leftForce + rightForce);
            CustomGravityRight = CustomGravityLeft;
            if(_transition < 1f)
            {
                _transition += 0.020f / Utils.PhysicsRateMultiplier();
                CustomGravityLeft = Vector3.Lerp(_prevSimulationForce, CustomGravityLeft, _transition);
                CustomGravityRight = CustomGravityLeft;
                defaultGravityForce = Vector3.Lerp(_prevDefaultGravityForce, defaultGravityForce, _transition);
                _prevSimulationForce = CustomGravityLeft;
                _prevDefaultGravityForce = defaultGravityForce;
            }

            leftContext.rigidbody.AddForce(CustomGravityLeft, ForceMode.Acceleration);
            rightContext.rigidbody.AddForce(CustomGravityRight, ForceMode.Acceleration);

            if(fullBody)
            {
                foreach(var rigidbody in _gravityRigidbodies)
                {
                    rigidbody.AddForce(defaultGravityForce, ForceMode.Acceleration);
                }

                if(_transition <= 1f)
                {
                    _clothingStore.enabledItems.ForEach(item => item.SetSimForce(defaultGravityForce));
                    _hairStore.enabledItems.ForEach(item => item.SetSimForce(defaultGravityForce));
                }
            }
        }

        void SetGravityDirection(string direction)
        {
            if(GravitySimulationDirection == direction || Script.calibrationStateHandler.CalibratingJsb.val)
            {
                return;
            }

            if(Script.calibrationStateHandler.EnableCalibrationStateJsb.val)
            {
                Loggr.Message("Exit the calibration state to simulate gravity.");
                return;
            }

            EnableGravitySimulation(direction);
            Script.uiHandler.SetButtonColors(direction);
        }

        internal void EnableGravitySimulation(string direction)
        {
            GravitySimulationDirection = direction;
            IsSimulatingGravity = true;
            _transition = 0f;
            foreach(var rigidbody in _gravityRigidbodies)
            {
                rigidbody.useGravity = false;
            }
        }

        void UnsetGravityDirection()
        {
            if(Script.calibrationStateHandler.CalibratingJsb.val)
            {
                return;
            }

            StopGravitySimulation();
            Script.uiHandler.ResetButtonColors();
        }

        // TODO store useGravity values, restore original values on stop
        // TODO unnecessary foreaches post-calibration
        internal void StopGravitySimulation()
        {
            foreach(var rigidbody in _gravityRigidbodies)
            {
                rigidbody.useGravity = rigidbody != leftContext.rigidbody && rigidbody != rightContext.rigidbody;
            }

            _clothingStore.enabledItems.ForEach(item => item.UnsetSimForce());
            _hairStore.enabledItems.ForEach(item => item.UnsetSimForce());

            IsSimulatingGravity = false;
            GravitySimulationDirection = null;
            _prevSimulationForce = Vector3.zero;
        }

        readonly Quaternion _upright = Quaternion.Euler(Vector3.zero);
        readonly Quaternion _upsideDown = Quaternion.Euler(new Vector3(180, 0, 0));
        readonly Quaternion _left = Quaternion.Euler(new Vector3(0, -90, 0));
        readonly Quaternion _right = Quaternion.Euler(new Vector3(0, 90, 0));
        readonly Quaternion _forward = Quaternion.Euler(new Vector3(90, 0, 0));
        readonly Quaternion _backward = Quaternion.Euler(new Vector3(-90, 0, 0));

        internal void SetGravityValues()
        {
            _upGravityLeft = UpGravityJsf.val * upGravityMultiplier;
            _downGravityLeft = DownGravityJsf.val * downGravityMultiplier;
            _forwardGravityLeft = ForwardGravityJsf.val * forwardGravityMultiplier;
            _backGravityLeft = BackGravityJsf.val * backGravityMultiplier;
            _sidewaysInGravityLeft = SidewaysGravityJsf.val * sidewaysInGravityMultiplier;
            _sidewaysOutGravityLeft = SidewaysGravityJsf.val * sidewaysOutGravityMultiplier;

            _upGravityRight = _upGravityLeft;
            _downGravityRight = _downGravityLeft;
            _forwardGravityRight = _forwardGravityLeft;
            _backGravityRight = _backGravityLeft;
            _sidewaysInGravityRight = _sidewaysInGravityLeft;
            _sidewaysOutGravityRight = _sidewaysOutGravityLeft;

            ExtraDownGravityLeft = Mathf.Clamp(_downGravityLeft - 1f, 0, _downGravityLeft - 1f);
            ExtraDownGravityRight = Mathf.Clamp(_downGravityRight - 1f, 0, _downGravityRight - 1f);
            // TODO callbacks for actual gravity being manipulated at runtime

            RefDownGravityLeft = CalculateReferenceGravity(_upright, false);
            RefDownGravityRight = CalculateReferenceGravity(_upright, true);
            RefUpGravityLeft = CalculateReferenceGravity(_upsideDown, false);
            RefUpGravityRight = CalculateReferenceGravity(_upsideDown, true);
            RefLeftGravityLeft = CalculateReferenceGravity(_left, false);
            RefLeftGravityRight = CalculateReferenceGravity(_left, true);
            RefRightGravityLeft = CalculateReferenceGravity(_right, false);
            RefRightGravityRight = CalculateReferenceGravity(_right, true);
            RefBackGravityLeft = CalculateReferenceGravity(_forward, false);
            RefBackGravityRight = CalculateReferenceGravity(_forward, true);
            RefForwardGravityLeft = CalculateReferenceGravity(_backward, false);
            RefForwardGravityRight = CalculateReferenceGravity(_backward, true);
        }

        protected bool isCalibrating;
        protected float mass;
        protected float softness;
        float _leftRoll, _rightRoll;
        float _leftPitch, _rightPitch;

        internal void FixedUpdate(bool isCalibrating = false)
        {
            this.isCalibrating = isCalibrating;
            mass = jointPhysicsHandler.normalizedMass;
            softness = Script.softness;
            _leftRoll = _angleCalculator.SmoothLeftRoll;
            _rightRoll = _angleCalculator.SmoothRightRoll;
            _leftPitch = _angleCalculator.Smooth2XLeftPitch;
            _rightPitch = _angleCalculator.Smooth2XRightPitch;

            if(!IsSimulatingGravity)
            {
                ApplyRealGravity();
            }

            AdjustSidewaysPhysics();
            AdjustVerticalPhysics();
            AdjustDepthPhysics();
        }

        void AdjustSidewaysPhysics()
        {
            float leftEffect = Mathf.Abs(_leftRoll);
            if(_leftRoll > 0)
            {
                ResetLeft(SIDEWAYS_IN);
                UpdateLeft(SIDEWAYS_OUT, leftEffect);
            }
            else
            {
                ResetLeft(SIDEWAYS_OUT);
                UpdateLeft(SIDEWAYS_IN, leftEffect);
            }

            float rightEffect = Mathf.Abs(_rightRoll);
            if(_rightRoll > 0)
            {
                ResetRight(SIDEWAYS_OUT);
                UpdateRight(SIDEWAYS_IN, rightEffect);
            }
            else
            {
                ResetRight(SIDEWAYS_IN);
                UpdateRight(SIDEWAYS_OUT, rightEffect);
            }
        }

        void AdjustVerticalPhysics()
        {
            float leftEffect = AngleCalculator.UpDownAdjustByAngle(_leftPitch) * AngleCalculator.RollMultiplier(_leftRoll);
            if(_leftPitch > 1 || _leftPitch < -1)
            {
                // upside down, leaning forward or back
                ResetLeft(DOWN);
                UpdateLeft(UP, leftEffect);
            }
            else
            {
                ResetLeft(UP);
                UpdateLeft(DOWN, leftEffect);
            }

            float rightEffect = AngleCalculator.UpDownAdjustByAngle(_rightPitch) * AngleCalculator.RollMultiplier(_rightRoll);
            if(_rightPitch > 1 || _rightPitch < -1)
            {
                // upside down, leaning forward or back
                ResetRight(DOWN);
                UpdateRight(UP, rightEffect);
            }
            else
            {
                ResetRight(UP);
                UpdateRight(DOWN, rightEffect);
            }
        }

        void AdjustDepthPhysics()
        {
            float leftEffect = AngleCalculator.DepthAdjustByAngle(_leftPitch) * AngleCalculator.RollMultiplier(_leftRoll);
            if(_leftPitch > 0)
            {
                // leaning forward, upright or upside down
                ResetLeft(BACK);
                UpdateLeft(FORWARD, leftEffect);
            }
            else
            {
                // leaning back
                ResetLeft(FORWARD);
                UpdateLeft(BACK, leftEffect);
            }

            float rightEffect = AngleCalculator.DepthAdjustByAngle(_rightPitch) * AngleCalculator.RollMultiplier(_rightRoll);
            if(_rightPitch > 0)
            {
                // leaning forward, upright or upside down
                ResetRight(BACK);
                UpdateRight(FORWARD, rightEffect);
            }
            else
            {
                // leaning back
                UpdateRight(BACK, rightEffect);
                ResetRight(FORWARD);
            }
        }

        internal override void ResetAll()
        {
            foreach(string direction in new[]
            {
                SIDEWAYS_IN,
                SIDEWAYS_OUT,
                UP,
                DOWN,
                FORWARD,
                BACK,
            })
            {
                ResetLeft(direction);
                ResetRight(direction);
            }
        }

        void UpdateLeft(string direction, float effect) => leftContext.updateHandlers(direction, effect);
        void UpdateRight(string direction, float effect) => rightContext.updateHandlers(direction, effect);
        void ResetLeft(string direction) => leftContext.resetHandlers(direction);
        void ResetRight(string direction) => rightContext.resetHandlers(direction);

        void ApplyRealGravity()
        {
            var leftForce = CalculateCustomGravityLeft(Physics.gravity);
            var rightForce = CalculateCustomGravityRight(Physics.gravity);
            CustomGravityLeft = 0.5f * (leftForce + rightForce);
            CustomGravityRight = CustomGravityLeft;
            leftContext.rigidbody.AddForce(CustomGravityLeft, ForceMode.Acceleration);
            rightContext.rigidbody.AddForce(CustomGravityRight, ForceMode.Acceleration);
        }

        Vector3 CalculateCustomGravityLeft(Vector3 gravity)
        {
            var localGravity = leftContext.rigidbody.transform.InverseTransformDirection(gravity);
            var customGravity = GravityVectorLeft(localGravity);
            return leftContext.rigidbody.transform.TransformDirection(customGravity);
        }

        Vector3 CalculateCustomGravityRight(Vector3 gravity)
        {
            var localGravity = rightContext.rigidbody.transform.InverseTransformDirection(gravity);
            var customGravity = GravityVectorRight(localGravity);
            return rightContext.rigidbody.transform.TransformDirection(customGravity);
        }

        Vector3 CalculateReferenceGravity(Quaternion relativeTo, bool right)
        {
            var localGravity = Quaternion.Inverse(relativeTo) * Physics.gravity;
            var customGravity = right ? GravityVectorRight(localGravity) : GravityVectorLeft(localGravity);
            return relativeTo * customGravity;
        }

        Vector3 GravityVectorLeft(Vector3 localGravity) =>
            new Vector3(
                localGravity.x * (localGravity.x > 0 ? _sidewaysInGravityLeft : _sidewaysOutGravityLeft),
                localGravity.y * (localGravity.y > 0 ? _upGravityLeft : _downGravityLeft),
                localGravity.z * (localGravity.z > 0 ? _forwardGravityLeft : _backGravityLeft)
            );

        Vector3 GravityVectorRight(Vector3 localGravity) =>
            new Vector3(
                localGravity.x * (localGravity.x > 0 ? _sidewaysOutGravityRight : _sidewaysInGravityRight),
                localGravity.y * (localGravity.y > 0 ? _upGravityRight : _downGravityRight),
                localGravity.z * (localGravity.z > 0 ? _forwardGravityRight : _backGravityRight)
            );

        protected static PhysicsCalculator DefaultPhysicsCalculator() => new PhysicsCalculator
        {
            primaryConfig = new MultiplierConfig(),
            secondaryConfig = new MultiplierConfig(),
        };

        protected static StorableFloat NewGravityMultiplierJsf(string paramName, string direction) =>
            new StorableFloat($"{paramName} {direction} Gravity", 1, 0, 3);

        internal override void SaveCalibrated()
        {
            foreach(var jsf in UserMultipliers)
            {
                jsf.SaveCalibrated();
            }
        }

        internal bool IsGravityModified =>
            UpGravityJsf.IsModified ||
            DownGravityJsf.IsModified ||
            ForwardGravityJsf.IsModified ||
            BackGravityJsf.IsModified ||
            SidewaysGravityJsf.IsModified;
    }
}
