using ColliderEditor;
using NAT.Common.Contexts;
using NAT.Common.Diagnostics;
using NAT.Common.Models;
using NAT.Common.Storables;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Handlers
{
    sealed class HardColliderHandler : Handler
    {
        readonly HardCollidersContext _leftContext = new HardCollidersContext(Side.LEFT);
        readonly HardCollidersContext _rightContext = new HardCollidersContext(Side.RIGHT);

        readonly List<AutoColliderPair> _autoColliderPairs = new List<AutoColliderPair>();
        internal readonly List<HardColliderPair> hardColliderPairs = new List<HardColliderPair>();
        internal HardColliderPair selectedHardColliderPair;

        internal ColliderVisualizer colliderVisualizer;
        internal ColliderDrawService colliderDrawService;
        internal JSONStorableStringChooser colliderGroupsJssc;
        internal StorableFloat baseMassMultiplierJsf;
        internal StorableFloat baseSpringMultiplierJsf;
        internal StorableFloat baseDamperMultiplierJsf;
        internal StorableFloat baseRadiusOffsetJsf;
        internal StorableFloat baseDepthJsf;
        internal JSONStorableBool highlightAllJsb;

        internal float collidersMassAmount;
        internal float maxHardColliderFriction;

        float _backgroundRadiusOffset;
        float _backgroundDepth;
        float _solverIterationsMultiplier;

        internal HardColliderHandler(ScriptBase script) : base(script)
        {
            InitAutoColliders();
            LoadSettings();
            InitColliderVisualizer();
            colliderDrawService = Script.gameObject.AddComponent<ColliderDrawService>().Init(Script, _autoColliderPairs);
            SetupStorables();
        }

        void SetupStorables()
        {
            var options = new List<string>
            {
                Constant.SELECT,
            };
            var displayOptions = new List<string>(options);
            options.AddRange(hardColliderPairs.Select(pair => pair.left.collider.name));
            displayOptions.AddRange(hardColliderPairs.Select(pair => pair.id));
            colliderGroupsJssc = new JSONStorableStringChooser(
                "colliderGroup",
                options,
                displayOptions,
                "",
                "Collider"
            )
            {
                setCallbackFunction = value =>
                {
                    selectedHardColliderPair = hardColliderPairs.FirstOrDefault(pair => pair.left.collider.name == value);
                    colliderVisualizer.selectablesJSON.val = value;
                },
                val = options[0],
            };

            colliderVisualizer.groupsJSON.setCallbackFunction += value =>
            {
                bool enabled = value != "Off";
                colliderVisualizer.enabled = enabled;
                if(enabled)
                {
                    colliderVisualizer.selectablesJSON.val = colliderGroupsJssc.val;
                }
            };

            baseMassMultiplierJsf = new StorableFloat(
                $"Hard Colliders Base {COLLIDER_MASS_MULTIPLIER.NAME}",
                Person.IsFemale ? 1.00f : 0.50f,
                COLLIDER_MASS_MULTIPLIER.MIN,
                COLLIDER_MASS_MULTIPLIER.MAX
            )
            {
                setCallbackFunction = _ =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    if(Script.IsBootyMagic)
                    {
                        SyncGluteCollidersMass();
                    }
                    else if(Script.IsTittyMagic)
                    {
                        SyncBreastCollidersMass();
                    }
                },
            };

            baseSpringMultiplierJsf = new StorableFloat(
                $"Hard Colliders Base {COLLIDER_SPRING_MULTIPLIER.NAME}",
                1.0f,
                COLLIDER_SPRING_MULTIPLIER.MIN,
                COLLIDER_SPRING_MULTIPLIER.MAX
            )
            {
                setCallbackFunction = _ =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    SyncSpringMultipliers();
                },
            };

            baseDamperMultiplierJsf = new StorableFloat(
                $"Hard Colliders Base {COLLIDER_DAMPER_MULTIPLIER.NAME}",
                1.0f,
                COLLIDER_DAMPER_MULTIPLIER.MIN,
                COLLIDER_DAMPER_MULTIPLIER.MAX
            )
            {
                setCallbackFunction = _ =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    SyncDamperMultipliers();
                },
            };

            baseRadiusOffsetJsf = new StorableFloat($"Hard Colliders Base {COLLIDER_RADIUS_OFFSET.NAME}", 0.00f, COLLIDER_RADIUS_OFFSET.MIN, COLLIDER_RADIUS_OFFSET.MAX)
            {
                setCallbackFunction = _ =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    SyncRadiusOffsets();
                },
            };
            baseDepthJsf = new StorableFloat($"Hard Colliders Base {COLLIDER_DEPTH_OFFSET.NAME}", 0.0000f, 0.0000f, COLLIDER_DEPTH_OFFSET.MAX)
            {
                setCallbackFunction = _ =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    foreach(var pair in hardColliderPairs)
                    {
                        OnDepthChanged(pair.left);
                        OnDepthChanged(pair.right);
                    }
                },
            };

            highlightAllJsb = new JSONStorableBool(
                "Highlight All Colliders",
                false,
                value => colliderVisualizer.previewOpacityJSON.val = value ? 1.00f : Constant.COLLIDER_PREVIEW_OPACITY
            );
        }

        void InitColliderVisualizer() => colliderVisualizer = Script.gameObject.AddComponent<ColliderVisualizer>(c =>
        {
            var groups = new List<Group>
            {
                new Group("Off", "$off"), // match nothing
            };

            if(Script.IsBootyMagic)
            {
                groups.Add(new Group("Left glute", "^((AutoCollider)?FemaleAutoColliders)?LGlute1"));
                groups.Add(new Group("Right glute", "^((AutoCollider)?FemaleAutoColliders)?RGlute1"));
                groups.Add(new Group("Both glutes", "^((AutoCollider)?FemaleAutoColliders)?[LR]Glute1"));
            }
            else if(Script.IsTittyMagic)
            {
                groups.Add(new Group("Left breast", @"lPectoral\d"));
                groups.Add(new Group("Right breast", @"rPectoral\d"));
                groups.Add(new Group("Both breasts", @"[lr](Pectoral\d)"));
            }

            c.Init(
                hardColliderPairs.Select(group => group.left.collider),
                hardColliderPairs.Select(group => group.right.collider),
                groups
            );
            c.previewOpacityJSON.val = Constant.COLLIDER_PREVIEW_OPACITY;
            c.previewOpacityJSON.defaultVal = Constant.COLLIDER_PREVIEW_OPACITY;
            c.selectedPreviewOpacityJSON.val = 1;
            c.selectedPreviewOpacityJSON.defaultVal = 1;

            c.groupsJSON.label = "Show Colliders";
            if(Script.IsBootyMagic)
            {
                c.groupsJSON.val = "Both glutes";
                c.groupsJSON.defaultVal = "Both glutes";
            }
            else if(Script.IsTittyMagic)
            {
                c.groupsJSON.val = "Both breasts";
                c.groupsJSON.defaultVal = "Both breasts";
            }

            c.highlightMirrorJSON.val = true;
            c.showPreviewsJSON.val = true;

            foreach(string option in new[] { Constant.SELECT, "Other", "All" })
            {
                c.groupsJSON.choices.Remove(option);
            }
        });

        internal void InitAutoColliders()
        {
            var leftAutoColliders = new List<AutoCollider>();
            var rightAutoColliders = new List<AutoCollider>();

            if(Script.IsBootyMagic)
            {
                AddGluteAutoColliders(leftAutoColliders, "L");
                AddGluteAutoColliders(rightAutoColliders, "R");
                IgnoreSoftColliders(leftAutoColliders);
                IgnoreSoftColliders(rightAutoColliders);
            }
            else if(Script.IsTittyMagic)
            {
                AddBreastAutoColliders(leftAutoColliders, "l");
                AddBreastAutoColliders(rightAutoColliders, "r");
            }

            for(int i = 0; i < leftAutoColliders.Count; i++)
            {
                _autoColliderPairs.Add(new AutoColliderPair { left = leftAutoColliders[i], right = rightAutoColliders[i] });
            }
        }

        void AddGluteAutoColliders(List<AutoCollider> listToAdd, string side)
        {
            string name = $"FemaleAutoColliders{side}Glute";
            var autoColliderGroup = Person.autoColliderGroups.First(group => group.name == name);
            var autoColliderGroupAlt = Person.autoColliderGroups.First(group => group.name == $"{name}Alt");
            autoColliderGroup.enabled = false;
            autoColliderGroupAlt.enabled = true;
            listToAdd.AddRange(autoColliderGroupAlt.GetAutoColliders());
        }

        void AddBreastAutoColliders(List<AutoCollider> listToAdd, string side)
        {
            string name = $"FemaleAutoColliders{side}Pectoral";
            var autoColliderGroup = Person.autoColliderGroups.First(group => group.name == name);
            listToAdd.AddRange(autoColliderGroup.GetAutoColliders());
        }

        void IgnoreSoftColliders(List<AutoCollider> autoColliders)
        {
            foreach(var softVerticesGroup in Person.GlutePhysicsMesh.softVerticesGroups)
            {
                foreach(var softVerticesSet in softVerticesGroup.softVerticesSets)
                {
                    foreach(var ac in autoColliders)
                    {
                        Physics.IgnoreCollision(softVerticesSet.jointCollider, ac.jointCollider);
                        if(softVerticesSet.jointCollider2)
                        {
                            Physics.IgnoreCollision(softVerticesSet.jointCollider2, ac.jointCollider);
                        }
                    }
                }
            }
        }

        protected override void LoadSettings()
        {
            const AutoCollider.ColliderOrient orient = AutoCollider.ColliderOrient.Up;

            if(Script.IsBootyMagic)
            {
                _leftContext.autoColliderParams = new Dictionary<string, AutoColliderParams>
                {
                    ["Glute1"] = new AutoColliderParams(orient, 287, 8731, 277),
                    ["Glute1 (1)"] = new AutoColliderParams(orient, 4037, 8731, 277),
                    ["Glute1 (2)"] = new AutoColliderParams(orient, 295, 8743, 7877),
                    ["Glute1 (3)"] = new AutoColliderParams(orient, 282, 8743, 7877),
                    ["Glute1 (4)"] = new AutoColliderParams(orient, 7878, 8743, 7877),
                    ["Glute1 (5)"] = new AutoColliderParams(orient, 8760, 8744, 7866),
                    ["Glute1 (6)"] = new AutoColliderParams(orient, 7836, 8744, 7866),
                    ["Glute1 (7)"] = new AutoColliderParams(orient, 337, 8744, 7866),
                };

                _rightContext.autoColliderParams = new Dictionary<string, AutoColliderParams>
                {
                    ["Glute1"] = new AutoColliderParams(orient, 11215, 19402, 11205),
                    ["Glute1 (1)"] = new AutoColliderParams(orient, 14813, 19402, 11205),
                    ["Glute1 (2)"] = new AutoColliderParams(orient, 11223, 19414, 18557),
                    ["Glute1 (3)"] = new AutoColliderParams(orient, 11210, 19414, 18557),
                    ["Glute1 (4)"] = new AutoColliderParams(orient, 18558, 19414, 18557),
                    ["Glute1 (5)"] = new AutoColliderParams(orient, 19431, 19415, 18546),
                    ["Glute1 (6)"] = new AutoColliderParams(orient, 18516, 19415, 18546),
                    ["Glute1 (7)"] = new AutoColliderParams(orient, 11265, 19415, 18546),
                };

                _leftContext.protrusionIndices = new[]
                {
                    265, 272, 287, 288, 294, 2388, 2391, 4036, 4040, 7172, 7833,
                    7838, 7839, 7844, 7845, 7862, 7864, 7865, 7866, 7872, 7875,
                    7876, 7877, 7878, 7881, 8212, 8247, 8249, 8250, 21567, 21712,
                };

                _rightContext.protrusionIndices = new[]
                {
                    11193, 11200, 11215, 11216, 11222, 13218, 13221, 14812, 14816, 17879, 18513,
                    18518, 18519, 18524, 18525, 18542, 18544, 18545, 18546, 18552, 18555,
                    18556, 18557, 18558, 18561, 18883, 18918, 18920, 18921, 21602, 22309,
                };
            }
            else if(Script.IsTittyMagic)
            {
                _leftContext.autoColliderParams = new Dictionary<string, AutoColliderParams>
                {
                    ["Pectoral1"] = new AutoColliderParams(orient, 8935, 7246, 2592),
                    ["Pectoral2"] = new AutoColliderParams(orient, 8952, 8952, 136),
                    ["Pectoral3"] = new AutoColliderParams(orient, 8963, 8951, 7233),
                    ["Pectoral4"] = new AutoColliderParams(orient, 8958, 7230, 136),
                    ["Pectoral5"] = new AutoColliderParams(orient, 8955, 8952, 136),
                };

                _rightContext.autoColliderParams = new Dictionary<string, AutoColliderParams>
                {
                    ["Pectoral1"] = new AutoColliderParams(orient, 19601, 17952, 13410),
                    ["Pectoral2"] = new AutoColliderParams(orient, 19618, 19618, 11064),
                    ["Pectoral3"] = new AutoColliderParams(orient, 19629, 19617, 17939),
                    ["Pectoral4"] = new AutoColliderParams(orient, 19624, 17936, 11064),
                    ["Pectoral5"] = new AutoColliderParams(orient, 19621, 19618, 11064),
                };

                _leftContext.protrusionIndices = new[]
                {
                    0, 8, 136, 2402, 2403, 2407, 2412, 2417, 2418, 2590, 2594, 7230, 7232, 7234,
                    7236, 7238, 7240, 7241, 7945, 7946, 7947, 7950, 7958, 7966, 8849, 8852,
                    8926, 8927, 8930, 8932, 8934, 8937, 8939, 8940, 8943, 8944, 8946, 8948,
                };

                _rightContext.protrusionIndices = new[]
                {
                    10928, 10936, 11064, 13232, 13233, 13237, 13242, 13247, 13248, 13408, 13411, 17936, 17938,
                    17940, 17942, 17944, 17946, 17947, 18623, 18624, 18625, 18628, 18636, 18644, 19520, 19523,
                    19592, 19593, 19596, 19598, 19600, 19603, 19605, 19606, 19609, 19610, 19612, 19614,
                };
            }

            foreach(var pair in _autoColliderPairs)
            {
                hardColliderPairs.Add(NewHardColliderPair(pair.left, pair.right));
            }
        }

        HardColliderPair NewHardColliderPair(AutoCollider left, AutoCollider right)
        {
            string rightId = "";
            string id = "";
            if(Script.IsBootyMagic)
            {
                rightId = right.name.Replace("AutoColliderFemaleAutoColliders", ""); // -> RGlute1, RGlute1 (1), RGlute1 (2) etc.
                id = rightId.Substring(1, rightId.Length - 1); // -> Glute1, Glute1 (1), Glute1 (2) etc.
            }
            else if(Script.IsTittyMagic)
            {
                rightId = right.name.Replace("AutoColliderFemaleAutoColliders", ""); // -> rPectoral1, rPectoral2, etc.
                id = rightId.Substring(1, rightId.Length - 1); // -> Pectoral1, Pectoral2, etc.
            }

            var leftCollider = HardCollider.New(id, left, _leftContext.autoColliderParams[id]);
            var rightCollider = HardCollider.New(rightId, right, _rightContext.autoColliderParams[id]);

            var radiusMultipliers = new Dictionary<string, float>
            {
                { "Glute1", 0.40f },
                { "Glute1 (1)", 0.44f },
                { "Glute1 (2)", 0.34f },
                { "Glute1 (3)", 0.30f },
                { "Glute1 (4)", 0.26f },
                { "Glute1 (5)", 0.33f },
                { "Glute1 (6)", 0.38f },
                { "Glute1 (7)", 0.33f },
                { "Pectoral1", 0.43f },
                { "Pectoral2", 0.32f },
                { "Pectoral3", 0.38f },
                { "Pectoral4", 0.37f },
                { "Pectoral5", 0.32f },
            };

            var pair = new HardColliderPair(
                id,
                leftCollider,
                rightCollider,
                radiusMultipliers[id]
            );

            SetCallbackFunctions(leftCollider, rightCollider);
            return pair;
        }

        void SetCallbackFunctions(HardCollider left, HardCollider right)
        {
            if(Script.IsBootyMagic)
            {
                left.massMultiplierJsf.setCallbackFunction = value =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    float result = _solverIterationsMultiplier * value * MassPerCollider;
                    left.SetMass(result);
                    right.SetMass(result);
                };
            }
            else if(Script.IsTittyMagic)
            {
                left.massMultiplierJsf.setCallbackFunction = value =>
                {
                    if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                    {
                        return;
                    }

                    float result = CalculateBreastColliderMass(value);
                    left.SetMass(result);
                    right.SetMass(result);
                };
            }

            left.springMultiplierJsf.setCallbackFunction = value =>
            {
                if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                {
                    return;
                }

                OnSpringMultiplierChanged(left);
                right.springMultiplierJsf.val = value;
            };

            left.damperMultiplierJsf.setCallbackFunction = value =>
            {
                if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                {
                    return;
                }

                OnDamperMultiplierChanged(left);
                right.damperMultiplierJsf.val = value;
            };

            left.radiusOffsetJsf.setCallbackFunction = value =>
            {
                if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                {
                    return;
                }

                OnRadiusOffsetChanged(left);
                right.radiusOffsetJsf.val = value;
            };

            left.depthJsf.setCallbackFunction = value =>
            {
                if(!Script.firstCalibrationDone || !Script.IsEnabledAndActive())
                {
                    return;
                }

                OnDepthChanged(left);
                right.depthJsf.val = value;
            };

            right.springMultiplierJsf.setCallbackFunction = _ => OnSpringMultiplierChanged(right);
            right.damperMultiplierJsf.setCallbackFunction = _ => OnDamperMultiplierChanged(right);
            right.radiusOffsetJsf.setCallbackFunction = _ => OnRadiusOffsetChanged(right);
            right.depthJsf.setCallbackFunction = _ => OnDepthChanged(right);
        }

        void OnSpringMultiplierChanged(HardCollider hardCollider)
        {
            float multiplier = _springMultiplier * baseSpringMultiplierJsf.val * hardCollider.springMultiplierJsf.val;
            hardCollider.SetJointSpring(Person.oneOverScale * multiplier);
        }

        void OnDamperMultiplierChanged(HardCollider hardCollider)
        {
            float multiplier = _damperMultiplier * baseDamperMultiplierJsf.val * hardCollider.damperMultiplierJsf.val;
            hardCollider.SetJointDamper(Person.oneOverScale * multiplier);
        }

        float CalculateBreastColliderMass(float value) =>
            _solverIterationsMultiplier *
            baseMassMultiplierJsf.val *
            value *
            _massSoftnessMultiplier *
            _massExtraMultiplier *
            0.025f;

        void OnRadiusOffsetChanged(HardCollider hardCollider)
        {
            float result = _backgroundRadiusOffset + baseRadiusOffsetJsf.val + hardCollider.radiusOffsetJsf.val;
            if(-result > hardCollider.baseRadius)
            {
                hardCollider.radiusOffsetJsf.valNoCallback = -hardCollider.baseRadius;
            }

            hardCollider.SyncAutoSize(result);
            OnDepthChanged(hardCollider);
        }

        void OnDepthChanged(HardCollider hardCollider)
        {
            float depth = _backgroundDepth + baseDepthJsf.val + hardCollider.depthJsf.val;
            hardCollider.SyncDepth(Script.person.Skin.rawSkinnedVerts, depth);
        }

        float _massSoftnessMultiplier;
        float _massExtraMultiplier;
        float _springMultiplier;
        float _damperMultiplier;

        internal override void SetMultipliers()
        {
            float softness = Script.softness;
            float mass = Script.jointPhysicsHandler.normalizedRealMass;
            bool softPhysicsOn = Script.softPhysicsHandler.softPhysicsOn;

            /* Radius */
            _backgroundRadiusOffset = softPhysicsOn ? Mathf.Lerp(0.000f, -0.010f, softness) : 0.00f;
            if(softPhysicsOn)
            {
                float backGroundDepthMax = Mathf.Lerp(0.0050f, 0.0220f, mass);
                _backgroundDepth = Mathf.Lerp(0.0000f, backGroundDepthMax, softness);
            }
            else
            {
                _backgroundDepth = 0.00f;
            }

            /* Depth */
            baseDepthJsf.min = -_backgroundDepth;

            /* Softness */
            if(softness < 0.60f)
            {
                float normalized = Mathf.InverseLerp(0.00f, 0.60f, softness);
                _massSoftnessMultiplier = Mathf.Lerp(1.80f, 1.00f, normalized);
            }
            else if(softness < 0.80f)
            {
                float normalized = Mathf.InverseLerp(0.60f, 0.80f, softness);
                _massSoftnessMultiplier = Mathf.SmoothStep(1.00f, 0.80f, normalized);
            }
            else
            {
                float normalized = Mathf.InverseLerp(0.80f, 1.00f, softness);
                _massSoftnessMultiplier = Mathf.Lerp(0.80f, 1.00f, normalized);
            }

            /* Extra */
            if(softPhysicsOn)
            {
                float ldpMulti = Script.IsLdpUpdateOn ? 1.25f : 1.00f;
                _massExtraMultiplier = ldpMulti * Mathf.Lerp(0.64f, 0.36f, Calc.InverseSmoothStep(mass, 0.33f, 0.70f));
            }
            else
            {
                _massExtraMultiplier = Mathf.Lerp(0.32f, 0.55f, Calc.InverseSmoothStep(mass, 0.40f, 0.84f));
            }

            _springMultiplier = Mathf.Lerp(9.00f, 3.00f, Calc.InverseSmoothStep(softness, 0.60f, 0.40f));
            _damperMultiplier = Mathf.Lerp(3.33f, 10.00f, Calc.InverseSmoothStep(softness, 0.60f, 0.40f));
        }

        internal void UpdateClothingAdjustmentAmounts(float massMultiplierLeft, float massMultiplierRight)
        {
            foreach(var pair in hardColliderPairs)
            {
                pair.SetClothingMassMultipliers(massMultiplierLeft, massMultiplierRight);
            }
        }

        Action _updateColliderFriction;

        internal void OnSoftPhysicsEnabled(IEnumerable<SoftJoint> leftSoftJoints, IEnumerable<SoftJoint> rightSoftJoints)
        {
            RepopulateSoftJointRigidbodies(_leftContext, leftSoftJoints);
            RepopulateSoftJointRigidbodies(_rightContext, rightSoftJoints);
            _updateColliderFriction = UpdateSoftJointBasedFriction;
        }

        static void RepopulateSoftJointRigidbodies(HardCollidersContext context, IEnumerable<SoftJoint> softJoints)
        {
            context.softJointRigidbodies.Clear();
            foreach(var softJoint in softJoints)
            {
                if(context.protrusionIndices.Contains(softJoint.softVerticesSet.targetVertex))
                {
                    context.softJointRigidbodies.Add(softJoint.jointRb);
                }
            }
        }

        internal void OnSoftPhysicsDisabled() => _updateColliderFriction = UpdateSkinVertexBasedFriction;

        void UpdateSoftJointBasedFriction()
        {
            foreach(var pair in hardColliderPairs)
            {
                pair.UpdateFriction(maxHardColliderFriction, _leftContext.softJointRigidbodies, _rightContext.softJointRigidbodies);
            }
        }

        void UpdateSkinVertexBasedFriction()
        {
            var skinVerts = Script.person.Skin.rawSkinnedVerts;
            foreach(var pair in hardColliderPairs)
            {
                pair.UpdateFriction(maxHardColliderFriction, skinVerts, _leftContext.protrusionIndices, _rightContext.protrusionIndices);
            }
        }

        readonly CustomFixedUpdate _frictionFixedUpdate = new CustomFixedUpdate(1 / 20f);

        internal override void OnFixedUpdate()
        {
            _frictionFixedUpdate.Run(_updateColliderFriction);
        }

        internal void SyncAll()
        {
            SyncCollidersMass();
            SyncSpringMultipliers();
            SyncDamperMultipliers();
            CalculateBaseRadiuses();
            SyncRadiusOffsets();
        }

        internal void SyncCollidersMass()
        {
            _solverIterationsMultiplier = 1.00f / (UserPreferences.singleton.physicsHighQuality ? 5 : Script.bootstrap.SolverIterationsMultiplier);
            if(Script.IsBootyMagic)
            {
                SyncGluteCollidersMass();
            }
            else if(Script.IsTittyMagic)
            {
                SyncBreastCollidersMass();
            }
        }

        void SyncGluteCollidersMass()
        {
            float massPerCollider = MassPerCollider;
            foreach(var pair in hardColliderPairs)
            {
                float result = _solverIterationsMultiplier * pair.left.massMultiplierJsf.val * massPerCollider;
                pair.left.SetMass(result);
                pair.right.SetMass(result);
            }
        }

        float MassPerCollider => baseMassMultiplierJsf.val * collidersMassAmount / hardColliderPairs.Count;

        void SyncBreastCollidersMass()
        {
            foreach(var pair in hardColliderPairs)
            {
                float result = CalculateBreastColliderMass(pair.left.massMultiplierJsf.val);
                pair.left.SetMass(result);
                pair.right.SetMass(result);
            }
        }

        void SyncSpringMultipliers()
        {
            foreach(var pair in hardColliderPairs)
            {
                OnSpringMultiplierChanged(pair.left);
                OnSpringMultiplierChanged(pair.right);
            }
        }

        void SyncDamperMultipliers()
        {
            foreach(var pair in hardColliderPairs)
            {
                OnDamperMultiplierChanged(pair.left);
                OnDamperMultiplierChanged(pair.right);
            }
        }

        void CalculateBaseRadiuses()
        {
            var skinVerts = Script.person.Skin.rawSkinnedVerts;
            foreach(var pair in hardColliderPairs)
            {
                pair.CalculateBaseRadius(skinVerts);
            }
        }

        void SyncRadiusOffsets()
        {
            foreach(var pair in hardColliderPairs)
            {
                OnRadiusOffsetChanged(pair.left);
                OnRadiusOffsetChanged(pair.right);
            }
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterHardCollidersParam(baseMassMultiplierJsf);
            Script.PreRegisterHardCollidersParam(baseSpringMultiplierJsf);
            Script.PreRegisterHardCollidersParam(baseDamperMultiplierJsf);
            Script.PreRegisterHardCollidersParam(baseRadiusOffsetJsf);
            Script.PreRegisterHardCollidersParam(baseDepthJsf);
            foreach(var pair in hardColliderPairs)
            {
                pair.PreRegisterParams(Script);
            }
        }

        internal override void Enable()
        {
            foreach(var pair in hardColliderPairs)
            {
                pair.ConfigureAutoColliders();
            }
        }

        internal override void Disable()
        {
            /* Required for restoring default collider mass. Enabled in case disabled
             * programmatically, and not yet restored back on by PersonMonitor.
             */
            if(Script.IsTittyMagic)
            {
                Person.SetupUseColliders();
            }

            colliderVisualizer.SetEnabledNullSafe(false);
            colliderDrawService.SetEnabledNullSafe(false);
            RestoreOriginalValues();
        }

        void RestoreOriginalValues()
        {
            foreach(var pair in hardColliderPairs)
            {
                pair.RestoreOriginalValues();
            }
        }

        internal bool RigidbodiesFound()
        {
            foreach(var pair in hardColliderPairs)
            {
                if(!pair.HasRigidbodies())
                {
                    return false;
                }
            }

            return true;
        }

        internal override void Destroy()
        {
            UnityEngine.Object.Destroy(colliderVisualizer);
            colliderVisualizer = null;
            UnityEngine.Object.Destroy(colliderDrawService);
            colliderDrawService = null;
        }
    }
}
