using NAT.Common.Contexts;
using NAT.Common.Models;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class JointPhysicsHandler : Handler
    {
        // AdjustJoints.joint1.slerpDrive.maximumForce value
        const float DEFAULT_SLERP_MAX_FORCE = 500; // TODO check

        protected AdjustJoints control;
        internal readonly JointPhysicsContext leftContext = new JointPhysicsContext(Side.LEFT);
        internal readonly JointPhysicsContext rightContext = new JointPhysicsContext(Side.RIGHT);

        // TODO move to ScriptBase for easier referencing
        internal float calibratedMass;
        internal float normalizedMass;
        internal float normalizedRealMass;

        internal readonly List<JointPhysicsParamPair> paramPairs = new List<JointPhysicsParamPair>();
        internal JointPhysicsParamPair jointMass;
        internal JointPhysicsParamPair centerOfGravity;
        internal JointPhysicsParamPair rotationSpring;
        internal JointPhysicsParamPair rotationDamper;

        // internal JointPhysicsParamPair AngularSpring { get; private set; }
        // internal JointPhysicsParamPair AngularDamper { get; private set; }
        // internal JointPhysicsParamPair UpDownSpring { get; private set; }
        // internal JointPhysicsParamPair UpDownDamper { get; private set; }

        internal JointPhysicsParamPair rightLeftSpring;
        internal JointPhysicsParamPair rightLeftDamper;
        internal JointPhysicsParamPair inOutSpring;
        internal JointPhysicsParamPair inOutDamper;
        internal JointPhysicsParamPair upDownBaseAngle;
        internal JointPhysicsParamPair upAngleLimit;
        internal JointPhysicsParamPair downAngleLimit;
        internal JointPhysicsParamPair upDownAngleTarget;
        internal JointPhysicsParamPair rightLeftBaseAngle;
        internal JointPhysicsParamPair rightLeftAngleLimit;
        internal JointPhysicsParamPair rightLeftAngleTarget;
        internal JointPhysicsParamPair twistAngleTarget;

        internal bool UpDownAngleIsModified => upDownBaseAngle.IsModified || upDownAngleTarget.IsModified || upAngleLimit.IsModified || downAngleLimit.IsModified;
        internal bool RightLeftAngleIsModified => rightLeftBaseAngle.IsModified || rightLeftAngleTarget.IsModified || rightLeftAngleLimit.IsModified;

        internal float twistAngleTargetValue;

        protected JointPhysicsHandler(ScriptBase script) : base(script)
        {
        }

        internal abstract void CalculateMass();

        internal abstract void UpdateMassValueAndAmounts(float mass);

        internal void UpdateBaseValues(bool updateOffsetMinMax = true)
        {
            foreach(var pair in paramPairs)
            {
                pair.UpdateBaseValue(
                    pair.usesRealMass ? normalizedRealMass : normalizedMass,
                    Script.softness,
                    Script.quickness
                );
                if(updateOffsetMinMax)
                {
                    pair.UpdateOffsetMinMax();
                }
            }
        }

        internal void SetGravitySimulationDamper()
        {
            float interpolant = Calc.InverseSmoothStep(normalizedMass, 0.50f, 0.10f);
            float max = Script.IsBootyMagic ? 5.0f : 10.0f;
            rotationDamper.OverrideValue(Mathf.Lerp(2.5f, max, interpolant));
        }

        internal void UnsetGravitySimulationDamper() =>
            rotationDamper.UpdateBaseValue(
                normalizedMass,
                Script.softness,
                Script.quickness
            );

        internal override void OnFixedUpdate()
        {
            foreach(var pair in paramPairs)
            {
                pair.DoSync();
            }
        }

        protected void SyncMass(Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            if(value.IsEqual(rb.mass))
            {
                return;
            }

            rb.mass = value;
            rb.WakeUp();
        }

        protected void SyncCenterOfGravity(Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var newCenterOfMass = Vector3.Lerp(control.lowCenterOfGravity, control.highCenterOfGravity, value);
            if(newCenterOfMass.IsEqual(rb.centerOfMass, 0.01f))
            {
                return;
            }

            rb.centerOfMass = newCenterOfMass;
            rb.WakeUp();
        }

        protected void SyncRotationSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            // TODO set multiplier
            float scalePow = Mathf.Pow(1.7f, control.scale - 1f);
            float scaledSpring = value * scalePow;
            if(scaledSpring.IsEqual(joint.slerpDrive.positionSpring, 1f))
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionSpring = scaledSpring;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow; // TODO set once
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionSpring = scaledSpring;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionSpring = 2 * scaledSpring;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.spring = 2 * scaledSpring * control.limitSpringMultiplier; // 1.5f, // TODO check
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.spring = 2 * scaledSpring * control.limitSpringMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        protected void SyncRotationDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            // TODO set multiplier
            float scalePow = Mathf.Pow(1.7f, control.scale - 1f);

            float scaledDamper = value * scalePow;
            if(scaledDamper.IsEqual(joint.slerpDrive.positionDamper, 0.01f))
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionDamper = scaledDamper;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionDamper = 2 * scaledDamper;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionDamper = 2 * scaledDamper;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.damper = 2 * scaledDamper * control.limitDamperMultiplier; // 0
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.damper = 2 * scaledDamper * control.limitDamperMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        // protected void SyncUpDownSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     var drive = joint.yDrive;
        //     if(value.IsEqual(drive.positionSpring, 1f))
        //     {
        //         return;
        //     }
        //
        //     drive.positionSpring = value;
        //     joint.yDrive = drive;
        //     rb.WakeUp();
        // }

        // protected void SyncUpDownDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        // {
        //     if(!Script.enabled)
        //     {
        //         return;
        //     }
        //
        //     var drive = joint.yDrive;
        //     if(value.IsEqual(drive.positionDamper, 0.01f))
        //     {
        //         return;
        //     }
        //
        //     drive.positionDamper = value;
        //     joint.yDrive = drive;
        //     rb.WakeUp();
        // }

        protected void SyncRightLeftSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.xDrive;
            if(value.IsEqual(drive.positionSpring, 1f))
            {
                return;
            }

            drive.positionSpring = value;
            joint.xDrive = drive;
            rb.WakeUp();
        }

        protected void SyncRightLeftDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.xDrive;
            if(value.IsEqual(drive.positionDamper, 0.01f))
            {
                return;
            }

            drive.positionDamper = value;
            joint.xDrive = drive;
            rb.WakeUp();
        }

        protected void SyncInOutSpring(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.zDrive;
            if(value.IsEqual(drive.positionSpring, 1f))
            {
                return;
            }

            drive.positionSpring = value;
            joint.zDrive = drive;
            rb.WakeUp();
        }

        protected void SyncInOutDamper(ConfigurableJoint joint, Rigidbody rb, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var drive = joint.zDrive;
            if(value.IsEqual(drive.positionDamper, 0.01f))
            {
                return;
            }

            drive.positionDamper = value;
            joint.zDrive = drive;
            rb.WakeUp();
        }

        protected void SyncUpAngleLimit(ConfigurableJoint joint, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var jointLimit = joint.highAngularXLimit;
            if(value.IsEqual(jointLimit.limit, 0.01f))
            {
                return;
            }

            jointLimit.limit = value;
            joint.highAngularXLimit = jointLimit;
        }

        protected void SyncDownAngleLimit(ConfigurableJoint joint, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var jointLimit = joint.lowAngularXLimit;
            if(value.IsEqual(jointLimit.limit, 0.01f))
            {
                return;
            }

            jointLimit.limit = value;
            joint.lowAngularXLimit = jointLimit;
        }

        protected void SyncRightLeftAngleLimit(ConfigurableJoint joint, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            var yLimit = joint.angularYLimit;
            if(value.IsEqual(yLimit.limit, 0.01f))
            {
                return;
            }

            yLimit.limit = value;
            joint.angularYLimit = yLimit;
        }

        internal override void PreRegisterParams()
        {
            var gravityPhysicsHandler = Script.gravityPhysicsHandler;
            Script.PreRegisterJointPhysicsParam(jointMass.Left.offsetJsf);
            Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.LinkDirectionsJsb);
            Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.UpGravityJsf);
            Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.DownGravityJsf);
            Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.BackGravityJsf);
            Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.ForwardGravityJsf);
            Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.SidewaysGravityJsf);

            foreach(var paramPair in paramPairs)
            {
                paramPair.PreRegisterParam();
            }

            if(gravityPhysicsHandler.twistAngleTargetJsf != null)
            {
                Script.PreRegisterJointPhysicsParam(gravityPhysicsHandler.twistAngleTargetJsf);
            }
        }

        internal override void SaveCalibrated()
        {
            jointMass.SaveCalibrated();
            foreach(var pair in paramPairs)
            {
                pair.SaveCalibrated();
            }
        }

        protected void RestoreOriginalPhysics()
        {
            foreach(string paramName in control.GetFloatParamNames())
            {
                var param = control.GetFloatJSONParam(paramName);
                param.setCallbackFunction(param.val);
            }
        }
    }
}
