using MVR.FileManagementSecure;
using NAT.Common.Storables;
using System.Collections.Generic;

namespace NAT.Common.Handlers
{
    abstract class PresetHandler
    {
        protected readonly ScriptBase script;
        protected string presetDir;
        protected string lastDir;
        internal readonly JSONStorableUrl loadPresetJsu;
        internal readonly JSONStorableUrl savePresetJsu;

        const string PRESET_EXT = "json";

        protected PresetHandler(ScriptBase script, string subDir)
        {
            this.script = script;
            presetDir = $"{FileUtils.savesDir}/{subDir}";
            LastDirFallback();

            loadPresetJsu = new JSONStorableUrl("Load", string.Empty, PRESET_EXT, lastDir)
            {
                allowFullComputerBrowse = true,
                allowBrowseAboveSuggestedPath = true,
                hideExtension = true,
                showDirs = true,
                beginBrowseWithObjectCallback = BeginLoadBrowse,
                endBrowseWithObjectCallback = jsu => OnLoadPathSelected(jsu.val),
            };

            savePresetJsu = new JSONStorableUrl("Save", string.Empty, PRESET_EXT, lastDir)
            {
                allowFullComputerBrowse = true,
                allowBrowseAboveSuggestedPath = true,
                hideExtension = true,
                showDirs = true,
                beginBrowseWithObjectCallback = BeginSaveBrowse,
                endBrowseWithObjectCallback = jsu => OnSavePathSelected(jsu.val),
                shortCuts = new List<ShortCut>(), // clear shortcuts
            };
        }

        protected StorableActionPresetFilePath NewLoadWithPathAction(string name)
        {
            var loadWithPathJsu = new JSONStorableUrl($"_{name}", string.Empty, PRESET_EXT, lastDir)
            {
                allowFullComputerBrowse = true,
                allowBrowseAboveSuggestedPath = true,
                hideExtension = true,
                showDirs = true,
                beginBrowseWithObjectCallback = BeginLoadBrowse,
            };

            return new StorableActionPresetFilePath(name, null, loadWithPathJsu);
        }

        // see e.g. DAZCharacterTextureControl.BeginBrowse
        void BeginLoadBrowse(JSONStorableUrl jsu)
        {
            FileUtils.EnsureDirExists(presetDir);
            LastDirFallback();
            jsu.shortCuts = GetShortCutsForDirectory();
        }

        protected virtual List<ShortCut> GetShortCutsForDirectory() => FileUtils.GetShortCutsForDirectory(presetDir);

        void BeginSaveBrowse(JSONStorableUrl jsu)
        {
            FileUtils.EnsureDirExists(presetDir);
            LastDirFallback();
            script.DeferSetSaveFileBrowser(PRESET_EXT); // JSONStorableUrl calls BeginSaveBrowse before creating the file browser
        }

        void OnLoadPathSelected(string path)
        {
            if(string.IsNullOrEmpty(path))
            {
                return;
            }

            OnPathSelected(ref path);
            loadPresetJsu.suggestedPath = lastDir;
            LoadAction(path);
            loadPresetJsu.val = ""; // ensure empty value when next browse cancelled
        }

        protected abstract void LoadAction(string path);

        void OnSavePathSelected(string path)
        {
            if(string.IsNullOrEmpty(path))
            {
                return;
            }

            OnPathSelected(ref path);
            loadPresetJsu.suggestedPath = lastDir;
            if(!path.ToLower().EndsWith(PRESET_EXT.ToLower()))
            {
                path += "." + PRESET_EXT;
            }

            SaveAction(path);
            savePresetJsu.val = ""; // ensure empty value when next browse cancelled
        }

        protected abstract void SaveAction(string path);

        protected void OnPathSelected(ref string path)
        {
            path = SuperController.singleton.NormalizePath(path);
            string containingDir = GetContainingDir(path);
            if(FileUtils.DirectoryExists(containingDir))
            {
                lastDir = containingDir + @"\";
            }
        }

        static string GetContainingDir(string path)
        {
            string containingDir = path.Substring(0, path.LastIndexOfAny(new[] { '/', '\\' }));
            FileUtils.EnsureDirExists(containingDir);
            return containingDir;
        }

        void LastDirFallback()
        {
            if(!FileUtils.DirectoryExists(lastDir))
            {
                lastDir = presetDir;
            }
        }
    }
}
