using ColliderEditor;
using NAT.Common.Contexts;
using NAT.Common.Diagnostics;
using NAT.Common.Models;
using NAT.Common.Storables;
using SimpleJSON;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class SoftPhysicsHandler : Handler
    {
        readonly AngleCalculator _angleCalculator;
        internal ColliderVisualizer colliderVisualizer;
        internal SoftJointDrawService softJointDrawService;

        protected DAZPhysicsMesh physicsMesh;
        internal SoftPhysicsContext leftContext;
        internal SoftPhysicsContext rightContext;

        internal readonly List<SoftPhysicsParamPair> softParamPairs = new List<SoftPhysicsParamPair>();
        internal readonly List<JointPhysicsParamPair> paramPairs = new List<JointPhysicsParamPair>();

        internal SoftPhysicsParamPair fatSpring;
        internal SoftPhysicsParamPair fatDamper;
        internal SoftPhysicsParamPair fatMass;
        internal SoftPhysicsParamPair fatColliderRadius;
        internal SoftPhysicsParamPair fatColliderDepth;
        internal SoftPhysicsParamPair fatColliderFriction;
        internal SoftPhysicsParamPair fatDistanceLimit;
        internal JointPhysicsParamPair fatBackForce;
        internal JointPhysicsParamPair fatBackForceMaxForce;
        internal JointPhysicsParamPair fatBackForceThreshold;

        internal bool BackForceIsModified => fatBackForce.IsModified || fatBackForceMaxForce.IsModified || fatBackForceThreshold.IsModified;

        protected CustomFixedUpdate softJointsFixedUpdate;

        internal StorableBool softPhysicsOnJsb;
        internal StorableBool ldpEnabledJsb;
        internal JSONStorableBool showColliderPreviewsJsb;
        internal StorableBool softPhysicsSelfCollideJsb;

        internal bool softPhysicsSynced;
        internal bool softPhysicsOn;
        internal bool ldpOn;

        protected SoftPhysicsHandler(ScriptBase script) : base(script)
        {
            _angleCalculator = Script.angleCalculator;
        }

        protected SoftJointDrawService AddDrawService() => Script.gameObject.AddComponent<SoftJointDrawService>(service =>
        {
            service.Init(leftContext.allSoftJoints, rightContext.allSoftJoints, Script.bootstrap.Camera);
            if(ldpEnabledJsb.val)
            {
                service.visualizeLdpJsb.val = true;
            }
        });

        internal abstract Rigidbody[] GetLeftPositionalRigidbodies(string group);
        internal abstract Rigidbody[] GetRightPositionalRigidbodies(string group);

        internal void CalibrateSoftJoints(string direction)
        {
            foreach(var softJoint in leftContext.allSoftJoints)
            {
                softJoint.CalibrateDistortion(direction);
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                softJoint.CalibrateDistortion(direction);
            }
        }

        internal void UpdateSoftJoints(bool force = false)
        {
            if(!Script.IsLdpUpdateOn)
            {
                return;
            }

            UpdateSoftJoints(leftContext, force);
            UpdateSoftJoints(rightContext, force);
        }

        void UpdateSoftJoints(SoftPhysicsContext context, bool force = false)
        {
            string side = context.id;
            foreach(var softJoint in context.allSoftJoints)
            {
                if(softJoint.UpdateDistortion(_angleCalculator.CenterRoll, _angleCalculator.CenterPitch2X, force))
                {
                    softJoint.SyncSpringAndDamperDistortion(side);
                    softJoint.SyncMassDistortion(side);
                    softJoint.SyncColliderRadiusAndDepthDistortion(side);
                    softJoint.SyncColliderFrictionDistortion(side);
                    softJoint.SyncDistanceLimitDistortion(side);
                }
            }
        }

        internal override void ResetAll()
        {
            foreach(var softJoint in leftContext.allSoftJoints)
            {
                softJoint.ResetDistortion();
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                softJoint.ResetDistortion();
            }

            ResetSoftJointsClothingAdjustments();
        }

        internal void EnableMultiplyFriction()
        {
            foreach(var softJoint in leftContext.allSoftJoints)
            {
                softJoint.EnableMultiplyFriction();
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                softJoint.EnableMultiplyFriction();
            }
        }

        /* Update value if external value changed */
        internal abstract void ReverseSyncSoftPhysicsOn();

        /* Update value if external value changed */
        internal virtual void ReverseSyncAllowSelfCollision()
        {
        }

        internal abstract void SyncSoftPhysicsOn();

        internal void SetShowColliderPreviews(bool hasSoftColliderVisualization) =>
            colliderVisualizer.SetEnabledNullSafe(Script.enabled && showColliderPreviewsJsb.val && hasSoftColliderVisualization);

        internal static IEnumerable<SoftJoint> FindMirrorJoints(SoftPhysicsContext context, List<SoftJoint> softJoints)
        {
            foreach(var fromJoint in softJoints)
            {
                int index = fromJoint.softVerticesSet.targetVertex;
                yield return context.FindSoftJointFromMirror(index);
            }
        }

        internal void UpdatePhysics(bool updateOffsetMinMax = true)
        {
            if(!softPhysicsOn)
            {
                return;
            }

            foreach(var group in leftContext.gradientGroups)
            {
                group.CalculateDistances();
            }

            foreach(var group in rightContext.gradientGroups)
            {
                group.CalculateDistances();
            }

            bool isLdpUpdateOn = Script.IsLdpUpdateOn;
            foreach(var pair in softParamPairs)
            {
                float mass = pair.usesRealMass ? Script.jointPhysicsHandler.normalizedRealMass : Script.jointPhysicsHandler.normalizedMass;
                pair.UpdateBaseValue(mass, Script.softness, Script.quickness, isLdpUpdateOn);
                if(updateOffsetMinMax)
                {
                    pair.UpdateOffsetMinMax();
                }
            }

            foreach(var pair in paramPairs)
            {
                pair.UpdateBaseValue(Script.jointPhysicsHandler.normalizedMass, Script.softness, Script.quickness, isLdpUpdateOn);
                if(updateOffsetMinMax)
                {
                    pair.UpdateOffsetMinMax();
                }
            }
        }

        internal override void OnFixedUpdate()
        {
            if(!softPhysicsOn)
            {
                return;
            }

            softJointsFixedUpdate.Run();
        }

        internal bool AnySoftJointMoved()
        {
            foreach(var x in leftContext.allSoftJoints)
            {
                if(x.Moved(0.008f))
                {
                    return true;
                }
            }

            foreach(var x in rightContext.allSoftJoints)
            {
                if(x.Moved(0.008f))
                {
                    return true;
                }
            }

            return false;
        }

        internal override void PreRegisterParams()
        {
            foreach(var pair in softParamPairs)
            {
                var left = pair.Left;
                if(pair != fatColliderFriction)
                {
                    Script.PreRegisterSoftPhysicsParam(left.offsetJsf);
                }

                Script.PreRegisterSoftPhysicsParam(left.gradientConfig.middleShiftJsf);
                Script.PreRegisterSoftPhysicsParam(left.gradientConfig.outerMultiplierJsf);
                Script.PreRegisterSoftPhysicsParam(left.gradientConfig.centerMultiplierJsf);
                if(left.gradientConfig.nippleMultiplierJsf != null)
                {
                    Script.PreRegisterSoftPhysicsParam(left.gradientConfig.nippleMultiplierJsf);
                }

                Script.PreRegisterSoftPhysicsParam(left.distortionConfig.compressionEffectJsf);
                Script.PreRegisterSoftPhysicsParam(left.distortionConfig.compressionSensitivityJsf);
                Script.PreRegisterSoftPhysicsParam(left.distortionConfig.stretchingEffectJsf);
                Script.PreRegisterSoftPhysicsParam(left.distortionConfig.stretchingSensitivityJsf);
            }

            foreach(var pair in paramPairs)
            {
                var left = pair.Left;
                Script.PreRegisterSoftPhysicsParam(left.offsetJsf);
            }

            Script.PreRegisterGeneralParam(softPhysicsOnJsb);
            Script.PreRegisterGeneralParam(ldpEnabledJsb);
            if(softPhysicsSelfCollideJsb != null)
            {
                Script.PreRegisterGeneralParam(softPhysicsSelfCollideJsb);
            }
        }

        internal override void SaveCalibrated() => ldpEnabledJsb.SaveCalibrated();

        internal JSONClass SerializeCalibrationData()
        {
            var jc = new JSONClass();
            if(!Script.IsLdpUpdateOn)
            {
                return jc;
            }

            foreach(var softJoint in leftContext.allSoftJoints)
            {
                jc[softJoint.id] = softJoint.SerializeCalibrationData();
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                jc[softJoint.id] = softJoint.SerializeCalibrationData();
            }

            return jc;
        }

        internal void CalibrateFromJSON(JSONClass jc)
        {
            if(!Script.IsLdpUpdateOn)
            {
                return;
            }

            foreach(var softJoint in leftContext.allSoftJoints)
            {
                softJoint.CalibrateFromJSON(jc[softJoint.id].AsObject);
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                softJoint.CalibrateFromJSON(jc[softJoint.id].AsObject);
            }
        }

        internal void ResetSoftJointsClothingAdjustments()
        {
            foreach(var softJoint in leftContext.allSoftJoints)
            {
                softJoint.ResetClothingAdjustments();
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                softJoint.ResetClothingAdjustments();
            }
        }

        protected void RestoreOriginalFriction()
        {
            foreach(var softJoint in leftContext.allSoftJoints)
            {
                softJoint.RestoreOriginalFriction();
            }

            foreach(var softJoint in rightContext.allSoftJoints)
            {
                softJoint.RestoreOriginalFriction();
            }
        }

        protected void RestoreOriginalPhysics()
        {
            var json = physicsMesh.GetJSON(true, true, true);
            physicsMesh.RestoreAllFromDefaults();
            physicsMesh.RestoreFromJSON(json);
        }

        internal override void Destroy()
        {
            Object.Destroy(colliderVisualizer);
            colliderVisualizer = null;
            Object.Destroy(softJointDrawService);
            softJointDrawService = null;
        }
    }
}
