using System;
using UnityEngine;

namespace NAT.Common.Models
{
    class AutoColliderParams
    {
        // target and anchor
        readonly AutoCollider.ColliderOrient _colliderOrient;
        readonly int _targetVertex;
        readonly int _anchorVertex1;
        readonly int _anchorVertex2;
        // joint
        readonly float _mass;
        readonly float _springLook;
        readonly float _springUp;
        readonly float _springRight;
        readonly float _damperLook;
        readonly float _damperUp;
        readonly float _damperRight;
        readonly float _jointBackForce;
        // autocollider
        readonly bool _useAutoLength;
        readonly bool _useAutoRadius;
        readonly float _autoRadiusBuffer;
        readonly float _autoLengthBuffer;
        readonly bool _centerJoint;
        readonly float _colliderRightOffset;
        readonly float _colliderUpOffset;
        readonly float _colliderLookOffset;
        // friction
        readonly PhysicMaterialCombine _frictionCombine;
        readonly float _dynamicFriction;
        readonly float _staticFriction;

        readonly bool _isFullParams;

        internal AutoColliderParams(
            // target and anchor
            AutoCollider.ColliderOrient colliderOrient,
            int targetVertex,
            int anchorVertex1,
            int anchorVertex2,
            // joint
            float mass,
            float springLook,
            float springUp,
            float springRight,
            float damperLook,
            float damperUp,
            float damperRight,
            float jointBackForce,
            // autocollider
            bool useAutoLength,
            bool useAutoRadius,
            float autoRadiusBuffer,
            float autoLengthBuffer,
            bool centerJoint,
            float colliderRightOffset,
            float colliderUpOffset,
            float colliderLookOffset,
            // friction
            PhysicMaterialCombine frictionCombine,
            float dynamicFriction,
            float staticFriction
        )
        {
            // target and anchor
            _colliderOrient = colliderOrient;
            _targetVertex = targetVertex;
            _anchorVertex1 = anchorVertex1;
            _anchorVertex2 = anchorVertex2;
            // joint
            _mass = mass;
            _springLook = springLook;
            _springUp = springUp;
            _springRight = springRight;
            _damperLook = damperLook;
            _damperUp = damperUp;
            _damperRight = damperRight;
            _jointBackForce = jointBackForce;
            // autocollider
            _useAutoLength = useAutoLength;
            _useAutoRadius = useAutoRadius;
            _autoRadiusBuffer = autoRadiusBuffer;
            _autoLengthBuffer = autoLengthBuffer;
            _centerJoint = centerJoint;
            _colliderRightOffset = colliderRightOffset;
            _colliderUpOffset = colliderUpOffset;
            _colliderLookOffset = colliderLookOffset;
            // friction
            _frictionCombine = frictionCombine;
            _dynamicFriction = dynamicFriction;
            _staticFriction = staticFriction;

            _isFullParams = true;
        }

        internal AutoColliderParams(
            AutoCollider.ColliderOrient colliderOrient,
            int targetVertex,
            int anchorVertex1,
            int anchorVertex2
        )
        {
            _colliderOrient = colliderOrient;
            _targetVertex = targetVertex;
            _anchorVertex1 = anchorVertex1;
            _anchorVertex2 = anchorVertex2;
            _useAutoLength = false;
            _useAutoRadius = false;
            _centerJoint = true;
            _frictionCombine = PhysicMaterialCombine.Multiply;
        }

        internal void Restore(AutoCollider autoCollider, Collider collider)
        {
            try
            {
                if(_isFullParams)
                {
                    // target and anchor
                    autoCollider.colliderOrient = _colliderOrient;
                    autoCollider.targetVertex = _targetVertex;
                    autoCollider.anchorVertex1 = _anchorVertex1;
                    autoCollider.anchorVertex2 = _anchorVertex2;
                    // joint
                    if(collider.attachedRigidbody)
                    {
                        collider.attachedRigidbody.mass = _mass;
                    }

                    autoCollider.jointSpringLook = _springLook;
                    autoCollider.jointSpringUp = _springUp;
                    autoCollider.jointSpringRight = _springRight;
                    autoCollider.jointDamperLook = _damperLook;
                    autoCollider.jointDamperUp = _damperUp;
                    autoCollider.jointDamperRight = _damperRight;
                    autoCollider.jointBackForce = _jointBackForce;
                    // autocollider
                    autoCollider.useAutoLength = _useAutoLength;
                    autoCollider.useAutoRadius = _useAutoRadius;
                    autoCollider.autoRadiusBuffer = _autoRadiusBuffer;
                    autoCollider.autoLengthBuffer = _autoLengthBuffer;
                    autoCollider.centerJoint = _centerJoint;
                    autoCollider.colliderRightOffset = _colliderRightOffset;
                    autoCollider.colliderUpOffset = _colliderUpOffset;
                    autoCollider.colliderLookOffset = _colliderLookOffset;
                    // friction
                    var material = collider.material;
                    material.frictionCombine = _frictionCombine;
                    material.dynamicFriction = _dynamicFriction;
                    material.staticFriction = _staticFriction;
                }
                else
                {
                    // target and anchor
                    autoCollider.colliderOrient = _colliderOrient;
                    autoCollider.targetVertex = _targetVertex;
                    autoCollider.anchorVertex1 = _anchorVertex1;
                    autoCollider.anchorVertex2 = _anchorVertex2;
                    // autocollider
                    autoCollider.useAutoLength = _useAutoLength;
                    autoCollider.useAutoRadius = _useAutoRadius;
                    autoCollider.centerJoint = _centerJoint;
                    // friction
                    var material = collider.material;
                    material.frictionCombine = _frictionCombine;
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"HardColliderConfig.Restore: {e}");
            }
        }
    }
}
