using SimpleJSON;
using System;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class CalibrationTarget
    {
        readonly Rigidbody _originRb;
        readonly Rigidbody _targetRb;
        readonly Person _person;

        internal int[] centerVertexIndices;

        float _prevAngleX;
        float _prevAngleY;

        internal float angleX;
        internal float angleY;
        internal float depthDiff;

        Vector3 _calibratedOffsetFromOrigin;
        Quaternion _calibratedOriginRotation;
        Vector3 _simulatedJointPosition;

        Vector3 _skinCenterRelativeToTarget;
        Vector3 _positionRelativeToOrigin;
        Vector2 _angleXBaseVector;
        Vector2 _angleYBaseVector;

        float _depthBase;

        Func<Vector3> _getPositionRelativeToOrigin;
        Func<float> _getRelativeDepth;

        internal CalibrationTarget(Rigidbody originRb, Rigidbody targetRb, Person person)
        {
            _originRb = originRb;
            _targetRb = targetRb;
            _person = person;
        }

        [SuppressMessage("ReSharper", "ParameterTypeCanBeEnumerable.Global")]
        internal void SetBreastFunctionsWhenSoftPhysicsOn(
            Person person,
            Rigidbody[] outerRigidbodies,
            Rigidbody[] mainRigidbodies,
            Rigidbody[] centerRigidbodies
        )
        {
            var combinedRigidbodies = outerRigidbodies.Concat(mainRigidbodies).ToArray();
            _getPositionRelativeToOrigin = () => Calc.RelativePosition(_originRb, Calc.AverageRigidbodiesPosition(combinedRigidbodies));
            _getRelativeDepth = () =>
            {
                float jointDepth = JointRelativeDepth();
                float softMeshDepth = SoftMeshRelativeDepth(centerRigidbodies, outerRigidbodies);
                return 0.5f * (jointDepth + softMeshDepth) * person.oneOverScale;
            };
        }

        [SuppressMessage("ReSharper", "ParameterTypeCanBeEnumerable.Global")]
        internal void SetGluteFunctionsWhenSoftPhysicsOn(
            Person person,
            Rigidbody[] outerRigidbodies,
            Rigidbody[] mainRigidbodies,
            Rigidbody[] centerRigidbodies
        )
        {
            var combinedRigidbodies = outerRigidbodies
                .Concat(mainRigidbodies)
                .Concat(centerRigidbodies)
                .ToArray();
            _getPositionRelativeToOrigin = () => Calc.RelativePosition(
                _simulatedJointPosition,
                _originRb.transform,
                _targetRb.transform,
                Calc.AverageRigidbodiesPosition(combinedRigidbodies)
            );
            _getRelativeDepth = () =>
            {
                float jointDepth = AltJointRelativeDepth();
                float softMeshDepth = SoftMeshRelativeDepth(centerRigidbodies, outerRigidbodies);
                return 0.5f * (jointDepth + softMeshDepth) * person.oneOverScale;
            };
        }

        internal void SetBreastFunctionsWhenSoftPhysicsOff(Rigidbody nippleRb)
        {
            if(_person.IsFemale)
            {
                _getPositionRelativeToOrigin = () => Calc.RelativePosition(_originRb, nippleRb.position);
            }
            else
            {
                _getPositionRelativeToOrigin = () => Calc.RelativePosition(_originRb, _person.SkinVertsAveragePosition(centerVertexIndices));
            }

            _getRelativeDepth = () => JointRelativeDepth() * _person.oneOverScale;
        }

        internal void SetGluteFunctionsWhenSoftPhysicsOff()
        {
            _getPositionRelativeToOrigin = () => Calc.RelativePosition(
                _simulatedJointPosition,
                _originRb.transform,
                _person.SkinVertsAveragePosition(centerVertexIndices)
            );
            _getRelativeDepth = () => AltJointRelativeDepth() * _person.oneOverScale;
        }

        internal bool Moved(float tolerance)
        {
            float xAngle = RelativeXAngle();
            float yAngle = RelativeYAngle();
            bool moved = !xAngle.IsEqual(_prevAngleX, tolerance) || !yAngle.IsEqual(_prevAngleY, tolerance);
            _prevAngleX = xAngle;
            _prevAngleY = yAngle;
            return moved;
        }

        float RelativeXAngle() => Vector2.SignedAngle(
            _angleXBaseVector,
            new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x)
        );

        float RelativeYAngle() => Vector2.SignedAngle(
            _angleYBaseVector,
            new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y)
        );

        internal void Calibrate()
        {
            _calibratedOffsetFromOrigin = _targetRb.position - _originRb.position;
            _calibratedOriginRotation = _originRb.rotation;
            _simulatedJointPosition = SimulatedJointPosition();
            _positionRelativeToOrigin = _getPositionRelativeToOrigin();
            _skinCenterRelativeToTarget = Calc.RelativePositionAlt(_targetRb.transform, _person.SkinVertsAveragePosition(centerVertexIndices));
            _angleXBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.x);
            _angleYBaseVector = new Vector2(_positionRelativeToOrigin.z, _positionRelativeToOrigin.y);
            _depthBase = _getRelativeDepth();
        }

        internal void Measure(Vector3 offset)
        {
            _simulatedJointPosition = SimulatedJointPosition();
            _positionRelativeToOrigin = _getPositionRelativeToOrigin();
            angleX = RelativeXAngle() + offset.x;
            angleY = RelativeYAngle() + offset.y;
            depthDiff = _depthBase - _getRelativeDepth() + offset.z;
        }

        Vector3 SimulatedJointPosition()
        {
            var originRotationDiff = _originRb.rotation * Quaternion.Inverse(_calibratedOriginRotation);
            return _originRb.position + originRotationDiff * _calibratedOffsetFromOrigin;
        }

        float SoftMeshRelativeDepth(Rigidbody[] softJointCenterRbs, Rigidbody[] softJointOuterRbs)
        {
            var centerToRb = Calc.RelativePosition(
                _simulatedJointPosition,
                _targetRb.transform,
                _originRb.transform,
                Calc.AverageRigidbodiesPosition(softJointCenterRbs)
            );
            var outerToRb = Calc.RelativePosition(
                _simulatedJointPosition,
                _targetRb.transform,
                _originRb.transform,
                Calc.AverageRigidbodiesPosition(softJointOuterRbs)
            );
            return (centerToRb - outerToRb).z;
        }

        float JointRelativeDepth()
        {
            var position = Calc.RelativePosition(
                _simulatedJointPosition,
                _targetRb.transform,
                _originRb.transform,
                _targetRb.position
            );
            return position.z;
        }

        float AltJointRelativeDepth()
        {
            var position = Calc.RelativePosition(
                _simulatedJointPosition,
                _originRb.transform,
                _targetRb.transform,
                _originRb.position
            );
            return position.z;
        }

        internal void Reset()
        {
            angleY = 0;
            angleX = 0;
            depthDiff = 0;
        }

        internal JSONClass Serialize() => new JSONClass
        {
            ["offset_from_origin"] = JSONUtils.SerializeVector3(_calibratedOffsetFromOrigin),
            ["origin_rotation"] = JSONUtils.SerializeQuaternion(_calibratedOriginRotation),
            ["angle_x_base_vector"] = JSONUtils.SerializeVector2(_angleXBaseVector),
            ["angle_y_base_vector"] = JSONUtils.SerializeVector2(_angleYBaseVector),
            ["center_relative_to_target"] = JSONUtils.SerializeVector3(_skinCenterRelativeToTarget),
            ["depth_base"] = { AsFloat = _depthBase },
        };

        internal void RestoreFromJSON(JSONClass jc)
        {
            _calibratedOffsetFromOrigin = JSONUtils.DeserializeVector3(jc["offset_from_origin"]);
            _calibratedOriginRotation = JSONUtils.DeserializeQuaternion(jc["origin_rotation"]);
            _angleXBaseVector = JSONUtils.DeserializeVector2(jc["angle_x_base_vector"]);
            _angleYBaseVector = JSONUtils.DeserializeVector2(jc["angle_y_base_vector"]);
            if(jc.HasKey("center_relative_to_target"))
            {
                _skinCenterRelativeToTarget = JSONUtils.DeserializeVector3(jc["center_relative_to_target"]);
            }

            _depthBase = jc["depth_base"].AsFloat;
        }
    }
}
