﻿using NAT.Common.Storables;
using System;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Models
{
    abstract class HardCollider
    {
        const float DEFAULT_MASS = 0.04f;
        const float DEFAULT_SPRING = 200f;
        const float DEFAULT_DAMPER = 10f;

        internal static HardCollider New(
            string id,
            AutoCollider autoCollider,
            AutoColliderParams autoColliderParams
        )
        {
            HardCollider hardCollider;
            if(autoCollider.jointCollider is SphereCollider)
            {
                hardCollider = new SphereHardCollider(id, autoCollider, autoColliderParams);
            }
            else if(autoCollider.jointCollider is CapsuleCollider)
            {
                hardCollider = new CapsuleHardCollider(id, autoCollider, autoColliderParams);
            }
            else
            {
                throw new InvalidOperationException($"Invalid collider type: {autoCollider.jointCollider.GetType()}");
            }

            return hardCollider;
        }

        // ReSharper disable once NotAccessedField.Local
        readonly string _id;
        internal readonly AutoCollider autoCollider;
        internal readonly Collider collider;
        readonly PhysicMaterial _colliderMaterial;
        readonly ConfigurableJoint _joint;

        internal readonly StorableFloat massMultiplierJsf;
        internal readonly StorableFloat springMultiplierJsf;
        internal readonly StorableFloat damperMultiplierJsf;
        internal readonly StorableFloat radiusOffsetJsf;
        internal readonly StorableFloat depthJsf;

        readonly AutoColliderParams _autoColliderParams;
        AutoColliderParams _originalParams;

        internal float baseRadius;
        protected abstract Vector3 Center { get; set; }
        internal Vector3 CenterWorldPosition => collider.transform.TransformPoint(Center);

        internal float depth;

        float _clothingMassMultiplier;

        protected HardCollider(
            string id,
            AutoCollider autoCollider,
            AutoColliderParams autoColliderParams
        )
        {
            _id = id;
            this.autoCollider = autoCollider;
            _autoColliderParams = autoColliderParams;
            autoCollider.resizeTrigger = AutoCollider.ResizeTrigger.None;
            collider = autoCollider.jointCollider;
            collider.enabled = true;
            _colliderMaterial = collider.material;
            _joint = autoCollider.joint;

            massMultiplierJsf = new StorableFloat($"{id} {COLLIDER_MASS_MULTIPLIER.NAME}", 1.00f, COLLIDER_MASS_MULTIPLIER.MIN, COLLIDER_MASS_MULTIPLIER.MAX);
            springMultiplierJsf = new StorableFloat($"{id} {COLLIDER_SPRING_MULTIPLIER.NAME}", 1.00f, COLLIDER_SPRING_MULTIPLIER.MIN, COLLIDER_SPRING_MULTIPLIER.MAX);
            damperMultiplierJsf = new StorableFloat($"{id} {COLLIDER_DAMPER_MULTIPLIER.NAME}", 1.00f, COLLIDER_DAMPER_MULTIPLIER.MIN, COLLIDER_DAMPER_MULTIPLIER.MAX);
            radiusOffsetJsf = new StorableFloat($"{id} {COLLIDER_RADIUS_OFFSET.NAME}", 0.000f, COLLIDER_RADIUS_OFFSET.MIN, COLLIDER_RADIUS_OFFSET.MAX);
            depthJsf = new StorableFloat($"{id} {COLLIDER_DEPTH_OFFSET.NAME}", 0.0000f, COLLIDER_DEPTH_OFFSET.MIN, COLLIDER_DEPTH_OFFSET.MAX);
        }

        internal void SetMass(float value)
        {
            if(!autoCollider.jointRB)
            {
                return;
            }

            autoCollider.jointRB.mass = _clothingMassMultiplier * value;
        }

        internal float GetMass() => autoCollider.jointRB ? autoCollider.jointRB.mass : 0f;

        internal void SetJointSpring(float multiplier)
        {
            float value = multiplier * DEFAULT_SPRING;
            _joint.xDrive = UpdateDriveSpring(_joint.xDrive, value);
            _joint.yDrive = UpdateDriveSpring(_joint.yDrive, value);
            _joint.zDrive = UpdateDriveSpring(_joint.zDrive, value);
        }

        static JointDrive UpdateDriveSpring(JointDrive drive, float spring) => new JointDrive
        {
            positionSpring = spring,
            positionDamper = drive.positionDamper,
            maximumForce = drive.maximumForce,
        };

        internal float GetSpring() => _joint.xDrive.positionSpring;

        internal void SetJointDamper(float multiplier)
        {
            float value = multiplier * DEFAULT_DAMPER;
            _joint.xDrive = UpdateDriveDamper(_joint.xDrive, value);
            _joint.yDrive = UpdateDriveDamper(_joint.yDrive, value);
            _joint.zDrive = UpdateDriveDamper(_joint.zDrive, value);
        }

        static JointDrive UpdateDriveDamper(JointDrive drive, float damper) => new JointDrive
        {
            positionSpring = drive.positionSpring,
            positionDamper = damper,
            maximumForce = drive.maximumForce,
        };

        internal float GetDamper() => _joint.xDrive.positionDamper;

        internal void UpdateFriction(float max, IEnumerable<Rigidbody> rigidbodies)
        {
            var position = CenterWorldPosition;
            float radius = autoCollider.colliderRadius;

            float sum = 0f;
            int count = 0;
            foreach(var rb in rigidbodies)
            {
                float distance = (rb.position - position).magnitude - radius;
                if(distance < 0)
                {
                    sum -= distance;
                    count++;
                }
            }

            ApplyFriction(max, 0.25f * radius, sum, count);
        }

        internal void UpdateFriction(float max, Vector3[] skinVerts, IEnumerable<int> protrusionIndices)
        {
            var position = CenterWorldPosition;
            float radius = autoCollider.colliderRadius;

            float sum = 0f;
            int count = 0;
            foreach(int i in protrusionIndices)
            {
                float distance = (skinVerts[i] - position).magnitude - radius;
                if(distance < 0)
                {
                    sum -= distance;
                    count++;
                }
            }

            ApplyFriction(max, 0.5f * radius, sum, count);
        }

        void ApplyFriction(float max, float maxDistance, float negativeDistancesSum, float negativeDistancesCount)
        {
            float averageNegativeDistance = negativeDistancesCount > 0 ? negativeDistancesSum / negativeDistancesCount : 0;
            float protrusion = Mathf.InverseLerp(0, maxDistance, averageNegativeDistance);
            float friction = Mathf.Lerp(0, max, 1 - protrusion);
            _colliderMaterial.dynamicFriction = friction;
            _colliderMaterial.staticFriction = friction;
        }

        internal void SetClothingMassMultiplier(float massMultiplier) => _clothingMassMultiplier = massMultiplier;

        void AutoColliderSizeSet()
        {
            autoCollider.resizeTrigger = AutoCollider.ResizeTrigger.Always;
            autoCollider.AutoColliderSizeSet();
            autoCollider.resizeTrigger = AutoCollider.ResizeTrigger.None;
        }

        internal void CalculateBaseRadius(Vector3[] skinVerts, float autoRadiusMultiplier)
        {
            var anchorVertex1Position = skinVerts[autoCollider.anchorVertex1];
            var anchorVertex2Position = skinVerts[autoCollider.anchorVertex2];
            baseRadius = autoRadiusMultiplier * (anchorVertex2Position - anchorVertex1Position).magnitude / autoCollider.scale;
        }

        internal void SyncAutoSize(float radiusOffset)
        {
            autoCollider.colliderRadius = baseRadius + radiusOffset;
            autoCollider.colliderLength = 2.0f * autoCollider.colliderRadius;
        }

        internal void SyncDepth(Vector3[] skinVerts, float depth)
        {
            this.depth = depth;
            var target = skinVerts[autoCollider.targetVertex];
            var direction = (CenterWorldPosition - target).normalized;
            Center = collider.transform.InverseTransformDirection(direction) * depth;
        }

        internal void ConfigureAutoColliders()
        {
            SaveOriginalValues();
            _autoColliderParams.Restore(autoCollider, collider);
        }

        void SaveOriginalValues() => _originalParams = new AutoColliderParams
        (
            // target and anchor
            autoCollider.colliderOrient,
            autoCollider.targetVertex,
            autoCollider.anchorVertex1,
            autoCollider.anchorVertex2,
            // joint
            DEFAULT_MASS,
            autoCollider.jointSpringLook,
            autoCollider.jointSpringUp,
            autoCollider.jointSpringRight,
            autoCollider.jointDamperLook,
            autoCollider.jointDamperUp,
            autoCollider.jointDamperRight,
            autoCollider.jointBackForce,
            // autocollider
            autoCollider.useAutoLength,
            autoCollider.useAutoRadius,
            autoCollider.autoRadiusBuffer,
            autoCollider.autoLengthBuffer,
            autoCollider.centerJoint,
            autoCollider.colliderRightOffset,
            autoCollider.colliderUpOffset,
            autoCollider.colliderLookOffset,
            // friction
            _colliderMaterial.frictionCombine,
            _colliderMaterial.dynamicFriction,
            _colliderMaterial.staticFriction
        );

        internal void RestoreOriginalValues()
        {
            _originalParams?.Restore(autoCollider, collider);
            AutoColliderSizeSet();
        }
    }
}
