using NAT.Common.Handlers.Configs;
using System;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class JointPhysicsParam : PhysicsParam
    {
        internal float Value { get; private set; }

        readonly Dictionary<string, float> _extraAdjustments = new Dictionary<string, float>();

        // TODO readonly
        internal Dictionary<string, PhysicsCalculator> gravityCalculators = new Dictionary<string, PhysicsCalculator>();
        readonly Dictionary<string, float> _additiveGravityAdjustments = new Dictionary<string, float>();

        // TODO readonly
        internal Dictionary<string, PhysicsCalculator> forceCalculators = new Dictionary<string, PhysicsCalculator>();
        readonly Dictionary<string, float> _additiveForceAdjustments = new Dictionary<string, float>();

        float _clothingAdjustment;
        // internal float ClothingAdjustmentValue { get; private set; }

        readonly Action<float> _sync;

        internal JointPhysicsParam(string side, string name, float min, float max, Action<float> sync = null)
            : base(side, name, min, max)
        {
            _sync = sync ?? (v => { });
        }

        internal bool HasAdjustableGravityPhysics()
        {
            if(gravityCalculators == null || gravityCalculators.Count == 0)
            {
                return false;
            }

            foreach(var pair in gravityCalculators)
            {
                if(pair.Value.multiplierJsf == null)
                {
                    return false;
                }
            }

            return true;
        }

        internal override void DoSync()
        {
            float adjustments = 0f;
            foreach(var pair in _extraAdjustments)
            {
                adjustments += pair.Value;
            }

            foreach(var pair in _additiveGravityAdjustments)
            {
                adjustments += pair.Value;
            }

            foreach(var pair in _additiveForceAdjustments)
            {
                adjustments += pair.Value;
            }

            float value = offsetJsf.val + baseValueJsf.val + adjustments;
            float clamped = Mathf.Clamp(value, baseValueJsf.min, baseValueJsf.max);
            Value = Mathf.Lerp(clamped, baseValueJsf.max, _clothingAdjustment);
            _sync.Invoke(Value);
        }

        internal void UpdateMassValue(float mass)
        {
            baseValueJsf.val = mass;
            DoSync();
        }

        // For temp use, call UpdateValue after
        internal void OverrideValue(float value)
        {
            Value = value;
            _sync.Invoke(value);
        }

        internal void UpdateGravityValue(string direction, float effect, float mass, float softness)
        {
            if(!gravityCalculators.ContainsKey(direction))
            {
                return;
            }

            float gravityValue = gravityCalculators[direction].CalculateDynamic(effect, mass, softness, true);
            _additiveGravityAdjustments[direction] = gravityValue;
        }

        internal void ResetGravityValue(string direction)
        {
            if(!gravityCalculators.ContainsKey(direction))
            {
                return;
            }

            _additiveGravityAdjustments[direction] = 0;
        }

        internal void UpdateForceValue(string direction, float effect, float mass, float softness)
        {
            if(!forceCalculators.ContainsKey(direction))
            {
                return;
            }

            float forceValue = forceCalculators[direction].CalculateDynamic(effect, mass, softness, true);
            _additiveForceAdjustments[direction] = forceValue;
        }

        internal void ResetForceValue(string direction)
        {
            if(!forceCalculators.ContainsKey(direction))
            {
                return;
            }

            _additiveForceAdjustments[direction] = 0;
        }

        internal void SetClothingAdjustment(float value) => _clothingAdjustment = value;
        internal void SetExtraAdjustment(string key, float value) => _extraAdjustments[key] = value;

        internal void SaveCalibrated()
        {
            offsetJsf.SaveCalibrated();
            if(HasAdjustableGravityPhysics())
            {
                foreach(var pair in gravityCalculators)
                {
                    pair.Value.multiplierJsf.SaveCalibrated();
                }
            }
        }
    }
}
