using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using System.Collections.Generic;

namespace NAT.Common.Models
{
    class JointPhysicsParamPair : PhysicsParamPair
    {
        internal string needsCalibration;
        internal bool rightInverted;

        JointPhysicsParam _left, _right;

        internal void DoSync()
        {
            leftParam.DoSync();
            rightParam.DoSync();
        }

        internal JointPhysicsParam Left
        {
            get { return _left; }
            set
            {
                _left = value;
                leftParam = value;
            }
        }

        internal JointPhysicsParam Right
        {
            get { return _right; }
            set
            {
                _right = value;
                rightParam = value;
            }
        }

        internal JointPhysicsParamPair(ScriptBase script, string name, string buttonLabel = null) : base(script, name, buttonLabel)
        {
        }

        internal void PreRegisterParam()
        {
            if(name == TWIST_ANGLE_TARGET.NAME)
            {
                return;
            }

            script.PreRegisterJointPhysicsParam(leftParam.offsetJsf);
            foreach(var pair in Left.gravityCalculators)
            {
                script.PreRegisterJointPhysicsParam(pair.Value.multiplierJsf);
            }
        }

        internal void UpdateMassValue(float mass)
        {
            Left.UpdateMassValue(mass);
            Right.UpdateMassValue(mass);
        }

        internal void OverrideValue(float value)
        {
            Left.OverrideValue(value);
            Right.OverrideValue(value);
        }

        internal void SetForceCalculators(Dictionary<string, Dictionary<string, PhysicsCalculator>> configs)
        {
            Left.forceCalculators = configs[Side.LEFT];
            Right.forceCalculators = configs[Side.RIGHT];
        }

        internal void SetCallbackFunctions()
        {
            leftParam.offsetJsf.setJSONCallbackFunction = jsf =>
            {
                var leftStorable = (StorableFloat) jsf;
                leftParam.UpdateOffsetValue(leftStorable.val);
                rightParam.offsetJsf.val = rightInverted ? -leftStorable.val : leftStorable.val;
                SetCalibrationPendingIfNeeded(leftStorable);
                script.uiHandler.physicsWindow.OnMainParamAdjusted(this);
            };
            rightParam.offsetJsf.setCallbackFunction = value => rightParam.UpdateOffsetValue(value);
        }

        internal void SetGravityMultiplierCallbacks()
        {
            foreach(var pair in Left.gravityCalculators)
            {
                var config = pair.Value;
                config.multiplierJsf.setJSONCallbackFunction = jsf =>
                {
                    var leftStorable = (StorableFloat) jsf;
                    var rightStorable = Right.gravityCalculators[pair.Key].multiplierJsf;
                    rightStorable.val = leftStorable.val;
                    SetCalibrationPendingIfNeeded(leftStorable);
                    script.uiHandler.physicsWindow.OnMainParamAdjusted(this);
                };
            }
        }

        void SetCalibrationPendingIfNeeded(StorableFloat leftStorable)
        {
            if(
                needsCalibration == NeedsCalibration.ALWAYS ||
                (needsCalibration == NeedsCalibration.IF_LDP_ENABLED && script.IsLdpUpdateOn)
            )
            {
                script.SetCalibrationPending(leftStorable);
            }
        }

        internal void SetExtraAdjustment(string key, float value)
        {
            Left.SetExtraAdjustment(key, value);
            Right.SetExtraAdjustment(key, value);
        }

        internal void ResetClothingAdjustment()
        {
            Left.SetClothingAdjustment(0);
            Right.SetClothingAdjustment(0);
        }

        internal override bool IsModified => base.IsModified || IsGravityPhysicsModified();

        bool IsGravityPhysicsModified()
        {
            if(!Left.HasAdjustableGravityPhysics())
            {
                return false;
            }

            foreach(var pair in Left.gravityCalculators)
            {
                if(pair.Value.multiplierJsf.IsModified)
                {
                    return true;
                }
            }

            return false;
        }

        internal void SaveCalibrated()
        {
            if(needsCalibration != NeedsCalibration.ALWAYS)
            {
                return;
            }

            Left.SaveCalibrated();
            Right.SaveCalibrated();
        }
    }
}
