using System;
using SimpleJSON;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class SoftJoint
    {
        internal readonly string id;
        internal readonly DAZPhysicsMeshSoftVerticesSet softVerticesSet;
        readonly SphereCollider _jointCollider;
        readonly SphereCollider _jointCollider2;
        readonly bool _useSecondCollider;
        readonly ConfigurableJoint _joint;
        internal readonly Rigidbody jointRb;
        readonly bool _enableSyncLinkParams;

        const float JOINT_SPRING_MAX_FORCE = 1E+07f;
        const float DEFAULT_FRICTION = 0.60f;
        internal float baseSpringMultiplier = 1;

        readonly SoftJointCalibration _distortionCalibration = new SoftJointCalibration();

        SoftPhysicsParamPair _fatMass, _fatSpring, _fatDamper, _fatColliderRadius, _fatColliderDepth, _fatColliderFriction, _fatDistanceLimit;

        internal SoftJoint[] adjacentJoints;
        internal SoftJoint[] nearbyJoints;

        internal SoftJoint(string id, DAZPhysicsMeshSoftVerticesSet softVerticesSet)
        {
            this.id = id;
            this.softVerticesSet = softVerticesSet;
            _jointCollider = (SphereCollider) softVerticesSet.jointCollider;
            _jointCollider2 = (SphereCollider) softVerticesSet.jointCollider2;
            _useSecondCollider = _jointCollider2 != null;
            _joint = this.softVerticesSet.joint;
            _enableSyncLinkParams = this.softVerticesSet.linkJoints != null;
            jointRb = this.softVerticesSet.jointRB;
        }

        internal void SetPhysicsParams(
            SoftPhysicsParamPair fatMass,
            SoftPhysicsParamPair fatSpring,
            SoftPhysicsParamPair fatDamper,
            SoftPhysicsParamPair fatColliderRadius,
            SoftPhysicsParamPair fatColliderDepth,
            SoftPhysicsParamPair fatColliderFriction,
            SoftPhysicsParamPair fatDistanceLimit
        )
        {
            _fatMass = fatMass;
            _fatSpring = fatSpring;
            _fatDamper = fatDamper;
            _fatColliderRadius = fatColliderRadius;
            _fatColliderDepth = fatColliderDepth;
            _fatColliderFriction = fatColliderFriction;
            _fatDistanceLimit = fatDistanceLimit;
        }

        internal IEnumerable<Collider> GetColliders()
        {
            yield return _jointCollider;
            if(_useSecondCollider)
            {
                yield return _jointCollider2;
            }
        }

        // values with user offset
        float _unmultipliedBaseSpring;
        float _baseSpring;
        float _baseDamper;
        float _baseMass;
        float _baseColliderRadius;
        float _baseColliderDepth;
        float _baseColliderFriction;
        float _baseDistanceLimit;

        internal float GetBaseValue(string paramName, bool unmultiplied = true)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return unmultiplied ? _unmultipliedBaseSpring : _baseSpring;
                case FAT_DAMPER.NAME: return _baseDamper;
                case FAT_MASS.NAME: return _baseMass;
                case FAT_COLLIDER_RADIUS.NAME: return _baseColliderRadius;
                case FAT_COLLIDER_DEPTH.NAME: return _baseColliderDepth;
                case FAT_COLLIDER_FRICTION.NAME: return _baseColliderFriction;
                case FAT_DISTANCE_LIMIT.NAME: return _baseDistanceLimit;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        internal void SetBaseValue(string paramName, float value)
        {
            if(float.IsNaN(value))
            {
                Debug.LogError($"{paramName} value is NaN");
                return;
            }

            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _unmultipliedBaseSpring = value;
                    _baseSpring = baseSpringMultiplier * _unmultipliedBaseSpring;
                    break;
                case FAT_DAMPER.NAME:
                    _baseDamper = value;
                    break;
                case FAT_MASS.NAME:
                    _baseMass = value;
                    break;
                case FAT_COLLIDER_RADIUS.NAME:
                    _baseColliderRadius = value;
                    break;
                case FAT_COLLIDER_DEPTH.NAME:
                    _baseColliderDepth = value;
                    break;
                case FAT_COLLIDER_FRICTION.NAME:
                    _baseColliderFriction = value;
                    break;
                case FAT_DISTANCE_LIMIT.NAME:
                    _baseDistanceLimit = value;
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // offsets consisting of any additive gravity physics, force physics, friction based or other adjustments
        float _springOffset;
        float _damperOffset;
        float _massOffset;
        float _colliderRadiusOffset;
        float _colliderDepthOffset;
        float _colliderFrictionOffset;
        float _distanceLimitOffset;

        // ReSharper disable once UnusedMember.Global
        internal float GetOffsetValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _springOffset;
                case FAT_DAMPER.NAME: return _damperOffset;
                case FAT_MASS.NAME: return _massOffset;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadiusOffset;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepthOffset;
                case FAT_COLLIDER_FRICTION.NAME: return _colliderFrictionOffset;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimitOffset;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        internal void SetAdditionalOffset(string paramName, float value)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _springOffset = value;
                    SetSpring();
                    break;
                case FAT_DAMPER.NAME:
                    _damperOffset = value;
                    SetDamper();
                    break;
                case FAT_MASS.NAME:
                    _massOffset = value;
                    SetMass();
                    break;
                case FAT_COLLIDER_RADIUS.NAME:
                    _colliderRadiusOffset = value;
                    SetColliderRadius();
                    break;
                case FAT_COLLIDER_DEPTH.NAME:
                    _colliderDepthOffset = value;
                    SetColliderDepth();
                    break;
                case FAT_COLLIDER_FRICTION.NAME:
                    _colliderFrictionOffset = value;
                    SetColliderFriction();
                    break;
                case FAT_DISTANCE_LIMIT.NAME:
                    _distanceLimitOffset = value;
                    SetJointLimits();
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // additional multipliers (used for nipple erection physics)
        float _additionalSpringMulti = 1f;
        float _additionalDamperMulti = 1f;

        internal void SetAdditionalBaseMultiplier(string paramName, float value)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _additionalSpringMulti = value;
                    SetSpring();
                    break;
                case FAT_DAMPER.NAME:
                    _additionalDamperMulti = value;
                    SetDamper();
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        float _springClothingAdjustmentValue;
        float _damperClothingAdjustmentValue;
        float _massClothingAdjustmentValue;
        float _colliderRadiusClothingAdjustmentValue;
        float _colliderDepthClothingAdjustmentValue;
        float _distanceLimitClothingAdjustmentValue;

        internal float GetClothingAdjustmentValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _springClothingAdjustmentValue;
                case FAT_DAMPER.NAME: return _damperClothingAdjustmentValue;
                case FAT_MASS.NAME: return _massClothingAdjustmentValue;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadiusClothingAdjustmentValue;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepthClothingAdjustmentValue;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimitClothingAdjustmentValue;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        float _springClothingAdjustment;
        float _damperClothingAdjustment;
        float _massClothingAdjustment;
        float _colliderRadiusClothingAdjustment;
        float _colliderDepthClothingAdjustment;
        float _distanceLimitClothingAdjustment;

        internal void SetClothingAdjustment(string paramName, float value)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME:
                    _springClothingAdjustment = value;
                    SetSpring();
                    break;
                case FAT_DAMPER.NAME:
                    _damperClothingAdjustment = value;
                    SetDamper();
                    break;
                case FAT_MASS.NAME:
                    _massClothingAdjustment = value;
                    SetMass();
                    break;
                case FAT_COLLIDER_RADIUS.NAME:
                    _colliderRadiusClothingAdjustment = value;
                    SetColliderRadius();
                    break;
                case FAT_COLLIDER_DEPTH.NAME:
                    _colliderDepthClothingAdjustment = value;
                    SetColliderDepth();
                    break;
                case FAT_DISTANCE_LIMIT.NAME:
                    _distanceLimitClothingAdjustment = value;
                    SetJointLimits();
                    break;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        internal void ResetClothingAdjustments()
        {
            _springClothingAdjustment = 0f;
            _damperClothingAdjustment = 0f;
            _massClothingAdjustment = 0f;
            _colliderRadiusClothingAdjustment = 0f;
            _colliderDepthClothingAdjustment = 0f;
            _distanceLimitClothingAdjustment = 0f;
        }

        // calculated values (not including real time distortion adjustments)
        float _spring;
        float _damper;
        float _mass;
        float _colliderRadius;
        float _colliderDepth;
        float _colliderFriction;
        float _distanceLimit;

        // ReSharper disable once UnusedMember.Global
        internal float GetCalculatedValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _spring;
                case FAT_DAMPER.NAME: return _damper;
                case FAT_MASS.NAME: return _mass;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadius;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepth;
                case FAT_COLLIDER_FRICTION.NAME: return _colliderFriction;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimit;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        // real-time values
        float _realSpring;
        float _realDamper;
        float _realMass;
        float _realColliderRadius;
        float _realColliderDepth;
        float _realColliderFriction;
        float _realDistanceLimit;

        internal float GetRealValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _realSpring;
                case FAT_DAMPER.NAME: return _realDamper;
                case FAT_MASS.NAME: return _realMass;
                case FAT_COLLIDER_RADIUS.NAME: return _realColliderRadius;
                case FAT_COLLIDER_DEPTH.NAME: return _realColliderDepth;
                case FAT_COLLIDER_FRICTION.NAME: return _realColliderFriction;
                case FAT_DISTANCE_LIMIT.NAME: return _realDistanceLimit;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        void SetSpring()
        {
            float clamped = Mathf.Clamp(
                _additionalSpringMulti * _baseSpring + _springOffset,
                FAT_SPRING.MIN,
                FAT_SPRING.GRADIENT_MAX
            );
            _spring = Mathf.Lerp(clamped, FAT_SPRING.GRADIENT_MAX, _springClothingAdjustment);
            _springClothingAdjustmentValue = _spring - clamped;
            _realSpring = _spring;
            float damper = _realDamper == 0 ? 1 : _realDamper;
            SetJointDrive(_realSpring, damper);

            if(_enableSyncLinkParams)
            {
                SyncLinkParams(_realSpring, damper);
            }
        }

        void SetDamper()
        {
            float clamped = Mathf.Clamp(
                _additionalDamperMulti * _baseDamper + _damperOffset,
                FAT_DAMPER.MIN,
                FAT_DAMPER.GRADIENT_MAX
            );
            _damper = Mathf.Lerp(clamped, FAT_DAMPER.GRADIENT_MAX, _damperClothingAdjustment);
            _damperClothingAdjustmentValue = _damper - clamped;
            _realDamper = _damper;
            float spring = _realSpring == 0 ? 100 : _realSpring;
            SetJointDrive(spring, _realDamper);

            if(_enableSyncLinkParams)
            {
                SyncLinkParams(spring, _realDamper);
            }
        }

        void SetJointDrive(float spring, float damper)
        {
            _joint.xDrive = new JointDrive
            {
                positionSpring = spring,
                positionDamper = damper,
                maximumForce = JOINT_SPRING_MAX_FORCE,
            };
            _joint.yDrive = new JointDrive
            {
                positionSpring = spring,
                positionDamper = damper,
                maximumForce = JOINT_SPRING_MAX_FORCE,
            };
            _joint.zDrive = new JointDrive
            {
                positionSpring = spring,
                positionDamper = damper,
                maximumForce = JOINT_SPRING_MAX_FORCE,
            };
        }

        // Not sure what the link joints are, exactly
        void SyncLinkParams(float spring, float damper)
        {
            foreach(var linkJoint in softVerticesSet.linkJoints)
            {
                if(linkJoint)
                {
                    linkJoint.spring = spring;
                    linkJoint.damper = damper;
                }
            }
        }

        void SetMass()
        {
            float clamped = Mathf.Clamp(_baseMass + _massOffset, FAT_MASS.MIN, FAT_MASS.GRADIENT_MAX);
            _mass = _massClothingAdjustment > 0
                ? Mathf.Lerp(clamped, FAT_MASS.GRADIENT_MAX, _massClothingAdjustment)
                : Mathf.Lerp(clamped, FAT_MASS.MIN, -_massClothingAdjustment);
            _massClothingAdjustmentValue = _mass - clamped;
            _realMass = _mass;
            jointRb.mass = _realMass;
        }

        void SetColliderRadius()
        {
            float clamped = Mathf.Clamp(_baseColliderRadius + _colliderRadiusOffset, FAT_COLLIDER_RADIUS.MIN, FAT_COLLIDER_RADIUS.GRADIENT_MAX);
            _colliderRadius = _colliderRadiusClothingAdjustment > 0
                ? Mathf.Lerp(clamped, FAT_COLLIDER_RADIUS.GRADIENT_MAX, _colliderRadiusClothingAdjustment)
                : Mathf.Lerp(clamped, FAT_COLLIDER_RADIUS.MIN, -_colliderRadiusClothingAdjustment);
            _colliderRadiusClothingAdjustmentValue = _colliderRadius - clamped;
            _realColliderRadius = _colliderRadius;
            SetColliders();
        }

        void SetColliderDepth()
        {
            float clamped = Mathf.Clamp(_baseColliderDepth + _colliderDepthOffset, FAT_COLLIDER_DEPTH.GRADIENT_MIN, FAT_COLLIDER_DEPTH.GRADIENT_MAX);
            _colliderDepth = _colliderDepthClothingAdjustment > 0
                ? Mathf.Lerp(clamped, FAT_COLLIDER_DEPTH.GRADIENT_MAX, _colliderDepthClothingAdjustment)
                : Mathf.Lerp(clamped, FAT_COLLIDER_DEPTH.GRADIENT_MIN, -_colliderDepthClothingAdjustment);
            _colliderDepthClothingAdjustmentValue = _colliderDepth - clamped;
            _realColliderDepth = _colliderDepth;
            SetColliders();
        }

        bool _useDefaultFriction;

        internal bool UseDefaultFriction
        {
            private get { return _useDefaultFriction; }
            set
            {
                _useDefaultFriction = value;
                SetColliderFriction();
            }
        }

        void SetColliderFriction()
        {
            _colliderFriction = UseDefaultFriction
                ? DEFAULT_FRICTION
                : Mathf.Clamp(_baseColliderFriction + _colliderFrictionOffset, FAT_COLLIDER_FRICTION.MIN, FAT_COLLIDER_FRICTION.MAX);
            _realColliderFriction = _colliderFriction;
            var material = _jointCollider.material;
            material.dynamicFriction = _realColliderFriction;
            material.staticFriction = _realColliderFriction;
            if(_useSecondCollider)
            {
                material = _jointCollider2.material;
                material.dynamicFriction = _realColliderFriction;
                material.staticFriction = _realColliderFriction;
            }
        }

        // Ignores second collider of nipple joint
        void SetColliders()
        {
            Vector3 colliderCenter;

            // lookAtOption != DAZPhysicsMeshSoftVerticesGroup.LookAtOption.Anchor
            colliderCenter.x = 0;
            colliderCenter.y = 0;
            colliderCenter.z = _realColliderRadius + _realColliderDepth;

            _jointCollider.radius = _realColliderRadius * softVerticesSet.sizeMultiplier;
            _jointCollider.center = colliderCenter * softVerticesSet.sizeMultiplier;

            if(_useSecondCollider)
            {
                _jointCollider2.radius = _jointCollider.radius;
                _jointCollider2.center = _jointCollider.center;
            }
        }

        void SetJointLimits()
        {
            float clamped = Mathf.Clamp(_baseDistanceLimit + _distanceLimitOffset, FAT_DISTANCE_LIMIT.MIN, FAT_DISTANCE_LIMIT.GRADIENT_MAX);
            _distanceLimit = _distanceLimitClothingAdjustment > 0
                ? Mathf.Lerp(clamped, FAT_DISTANCE_LIMIT.GRADIENT_MAX, _distanceLimitClothingAdjustment)
                : Mathf.Lerp(clamped, FAT_DISTANCE_LIMIT.MIN, -_distanceLimitClothingAdjustment);
            _distanceLimitClothingAdjustmentValue = _distanceLimit - clamped;
            _realDistanceLimit = _distanceLimit;
            var linearLimit = _joint.linearLimit;
            linearLimit.limit = _realDistanceLimit;
            _joint.linearLimit = linearLimit;
        }

        Vector3 _prevPosition;

        internal bool Moved(float tolerance)
        {
            var position = jointRb.position;
            bool moved = (position - _prevPosition).magnitude > tolerance;
            _prevPosition = position;
            return moved;
        }

        /* A combination of the average distance of this joint from its nearby joints
         * and the distance of this joint from the furthest of its nearby joints.
         */
        internal void CalibrateDistortion(string direction)
        {
            _distortionCalibration.SetBaseValue(direction, CalculateDistances());
        }

        float _springDistortion;
        float _damperDistortion;
        float _massDistortion;
        float _colliderRadiusDistortion;
        float _colliderDepthDistortion;
        float _colliderFrictionDistortion;
        float _distanceLimitDistortion;

        internal float GetDistortionValue(string paramName)
        {
            switch(paramName)
            {
                case FAT_SPRING.NAME: return _springDistortion;
                case FAT_DAMPER.NAME: return _damperDistortion;
                case FAT_MASS.NAME: return _massDistortion;
                case FAT_COLLIDER_RADIUS.NAME: return _colliderRadiusDistortion;
                case FAT_COLLIDER_DEPTH.NAME: return _colliderDepthDistortion;
                case FAT_COLLIDER_FRICTION.NAME: return _colliderFrictionDistortion;
                case FAT_DISTANCE_LIMIT.NAME: return _distanceLimitDistortion;
                default: throw new Exception($"Invalid parameter name {paramName}");
            }
        }

        float _distortion;
        float _prevAbsDistortion;

        internal bool UpdateDistortion(float centerRoll, float centerPitch2X, bool force = false)
        {
            _distortionCalibration.CalculateInterpBaseValue(centerRoll, centerPitch2X);
            _distortion = CalculateDistances() - _distortionCalibration.BaseValue;
            float absDistortion = Mathf.Abs(_distortion);
            if(force || !absDistortion.IsEqual(_prevAbsDistortion, 0.01f)) // more than 0.1 mm
            {
                _prevAbsDistortion = absDistortion;
                return true;
            }

            return false;
        }

        internal void SyncSpringAndDamperDistortion(string side)
        {
            _springDistortion = Mathf.Clamp(
                _fatSpring.CalculateDistortionOffset(
                    side,
                    _distortionCalibration.BaseValue,
                    _distortion
                ),
                FAT_SPRING.MIN - _spring,
                FAT_SPRING.GRADIENT_MAX - _spring
            );

            _damperDistortion = Mathf.Clamp(
                _fatDamper.CalculateDistortionOffset(
                    side,
                    _distortionCalibration.BaseValue,
                    _distortion
                ),
                FAT_DAMPER.MIN - _damper,
                FAT_DAMPER.GRADIENT_MAX - _damper
            );

            _realSpring = _spring + _springDistortion;
            _realDamper = _damper + _damperDistortion;
            SetJointDrive(_realSpring, _realDamper);
            if(_enableSyncLinkParams)
            {
                SyncLinkParams(_realSpring, _realDamper);
            }
        }

        internal void SyncMassDistortion(string side)
        {
            _massDistortion = Mathf.Clamp(
                _fatMass.CalculateDistortionOffset(
                    side,
                    _distortionCalibration.BaseValue,
                    _distortion
                ),
                FAT_MASS.MIN - _mass,
                FAT_MASS.GRADIENT_MAX - _mass
            );

            _realMass = _mass + _massDistortion;
            jointRb.mass = _realMass;
        }

        internal void SyncColliderRadiusAndDepthDistortion(string side)
        {
            _colliderRadiusDistortion = Mathf.Clamp(
                _fatColliderRadius.CalculateDistortionOffset(side, _distortionCalibration.BaseValue, _distortion),
                FAT_COLLIDER_RADIUS.MIN - _colliderRadius,
                FAT_COLLIDER_RADIUS.GRADIENT_MAX - _colliderRadius
            );

            _colliderDepthDistortion = Mathf.Clamp(
                _fatColliderDepth.CalculateDistortionOffset(side, _distortionCalibration.BaseValue, _distortion),
                FAT_COLLIDER_DEPTH.GRADIENT_MIN - _colliderDepth,
                FAT_COLLIDER_DEPTH.GRADIENT_MAX - _colliderDepth
            );

            _realColliderRadius = _colliderRadius + _colliderRadiusDistortion;
            _realColliderDepth = _colliderDepth + _colliderDepthDistortion;
            SetColliders();
        }

        internal void SyncColliderFrictionDistortion(string side)
        {
            _colliderFrictionDistortion = Mathf.Clamp(
                _fatColliderFriction.CalculateDistortionOffset(side, _distortionCalibration.BaseValue, _distortion),
                FAT_COLLIDER_FRICTION.MIN - _colliderFriction,
                FAT_COLLIDER_FRICTION.MAX - _colliderFriction
            );

            _realColliderFriction = _colliderFriction + _colliderFrictionDistortion;
            var material = softVerticesSet.jointCollider.material;
            material.dynamicFriction = _realColliderFriction;
            material.staticFriction = _realColliderFriction;
        }

        internal void SyncDistanceLimitDistortion(string side)
        {
            _distanceLimitDistortion = Mathf.Clamp(
                _fatDistanceLimit.CalculateDistortionOffset(
                    side,
                    _distortionCalibration.BaseValue,
                    _distortion
                ),
                FAT_DISTANCE_LIMIT.MIN - _distanceLimit,
                FAT_DISTANCE_LIMIT.GRADIENT_MAX - _distanceLimit
            );

            _realDistanceLimit = _distanceLimit + _distanceLimitDistortion;
            var linearLimit = _joint.linearLimit;
            linearLimit.limit = _realDistanceLimit;
            _joint.linearLimit = linearLimit;
        }

        internal void ResetDistortion() => _distortion = 0;

        /* Average of the maximum distance to a nearby joint and the average of distances to every nearby joint. Result in centimeters. */
        float CalculateDistances()
        {
            float maxDistance = 0;
            float sumDistances = 0;
            foreach(var softJoint in nearbyJoints)
            {
                float current = 100f * Vector3.Distance(jointRb.position, softJoint.jointRb.position);
                if(current - maxDistance > 0.01f)
                {
                    maxDistance = current;
                }

                sumDistances += current;
            }

            return (maxDistance + sumDistances / nearbyJoints.Length) / 2;
        }

        internal void EnableMultiplyFriction()
        {
            var material = softVerticesSet.jointCollider.material;
            material.frictionCombine = PhysicMaterialCombine.Multiply;
        }

        internal JSONClass SerializeCalibrationData() => _distortionCalibration.SerializeCalibrationData();

        internal void CalibrateFromJSON(JSONClass jc) => _distortionCalibration.CalibrateFromJSON(jc["distortion"].AsObject);

        internal void RestoreOriginalFriction()
        {
            var material = softVerticesSet.jointCollider.material;
            material.dynamicFriction = DEFAULT_FRICTION;
            material.staticFriction = DEFAULT_FRICTION;
            material.frictionCombine = PhysicMaterialCombine.Average;
        }
    }
}
