using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;

namespace NAT.Common.Models
{
    sealed class SoftPhysicsParamPair : PhysicsParamPair
    {
        readonly SoftPhysicsHandler _softPhysicsHandler;

        internal bool hasSoftColliderVisualization;
        internal float visualizationLimit;
        internal float clothingAdjustmentVisualizationLimit;
        internal float distortionLimit;
        internal bool visualizeRelativeChange;

        internal float leftClothingAdjustmentAverage;
        internal float rightClothingAdjustmentAverage;

        SoftPhysicsParam _left, _right;

        internal SoftPhysicsParamPair(ScriptBase script, SoftPhysicsHandler softPhysicsHandler, string name, string buttonLabel = null) : base(script, name, buttonLabel)
        {
            _softPhysicsHandler = softPhysicsHandler;
        }

        internal SoftPhysicsParam Left
        {
            get { return _left; }
            set
            {
                _left = value;
                leftParam = value;
            }
        }

        internal SoftPhysicsParam Right
        {
            get { return _right; }
            set
            {
                _right = value;
                rightParam = value;
            }
        }

        internal void SetCallbackFunctions()
        {
            leftParam.offsetJsf.setCallbackFunction = value =>
            {
                leftParam.UpdateOffsetValue(value);
                rightParam.offsetJsf.val = value;
                _softPhysicsHandler.UpdateSoftJoints(true);
                script.uiHandler.physicsWindow?.OnSoftParamAdjusted(this);
            };
            rightParam.offsetJsf.setCallbackFunction = value => rightParam.UpdateOffsetValue(value);

            /* Gradient config storable callbacks */
            {
                var lGradient = Left.gradientConfig;
                var rGradient = Right.gradientConfig;
                lGradient.outerMultiplierJsf.setCallbackFunction = value => UpdateGradient(value, rGradient.outerMultiplierJsf);
                lGradient.centerMultiplierJsf.setCallbackFunction = value => UpdateGradient(value, rGradient.centerMultiplierJsf);
                if(script.IsTittyMagic)
                {
                    lGradient.nippleMultiplierJsf.setCallbackFunction = value => UpdateGradient(value, rGradient.nippleMultiplierJsf);
                }

                lGradient.middleShiftJsf.setCallbackFunction = value => UpdateGradient(value, rGradient.middleShiftJsf);
            }

            /* Distortion config storable callbacks */
            {
                var config = Left.distortionConfig;
                config.compressionEffectJsf.setCallbackFunction = value =>
                {
                    SetCompressionMaxAndMultiplier(value);
                    _softPhysicsHandler.UpdateSoftJoints(true);
                    script.uiHandler.physicsWindow?.OnSoftParamAdjusted(this);
                };
                config.compressionSensitivityJsf.setCallbackFunction = value =>
                {
                    SetCompressionSlope(value);
                    _softPhysicsHandler.UpdateSoftJoints(true);
                    script.uiHandler.physicsWindow?.OnSoftParamAdjusted(this);
                };
                config.stretchingEffectJsf.setCallbackFunction = value =>
                {
                    SetStretchingMaxAndMultiplier(value);
                    _softPhysicsHandler.UpdateSoftJoints(true);
                    script.uiHandler.physicsWindow?.OnSoftParamAdjusted(this);
                };
                config.stretchingSensitivityJsf.setCallbackFunction = value =>
                {
                    SetStretchingSlope(value);
                    _softPhysicsHandler.UpdateSoftJoints(true);
                    script.uiHandler.physicsWindow?.OnSoftParamAdjusted(this);
                };
                SetCompressionMaxAndMultiplier(config.compressionEffectJsf.val);
                SetCompressionSlope(config.compressionSensitivityJsf.val);
                SetStretchingMaxAndMultiplier(config.stretchingEffectJsf.val);
                SetStretchingSlope(config.stretchingSensitivityJsf.val);
            }
        }

        void UpdateGradient(float value, JSONStorableFloat rightJsf)
        {
            rightJsf.val = value;
            leftParam.DoSync();
            rightParam.DoSync();
            script.uiHandler.physicsWindow?.OnSoftParamAdjusted(this);
        }

        void SetCompressionMaxAndMultiplier(float value)
        {
            _left.distortionConfig.SetCompressionMaxAndMultiplier(value);
            _right.distortionConfig.SetCompressionMaxAndMultiplier(value);
        }

        void SetCompressionSlope(float value)
        {
            _left.distortionConfig.SetCompressionSlope(value);
            _right.distortionConfig.SetCompressionSlope(value);
        }

        void SetStretchingMaxAndMultiplier(float value)
        {
            _left.distortionConfig.SetStretchingMaxAndMultiplier(value);
            _right.distortionConfig.SetStretchingMaxAndMultiplier(value);
        }

        void SetStretchingSlope(float value)
        {
            _left.distortionConfig.SetStretchingSlope(value);
            _right.distortionConfig.SetStretchingSlope(value);
        }

        internal void SetFrictionCalculators(PhysicsCalculator leftCalculator, PhysicsCalculator rightCalculator)
        {
            _left.inverseFrictionCalculator = leftCalculator;
            _right.inverseFrictionCalculator = rightCalculator;
        }

        internal void UpdateInverseFrictionValue(float friction, float mass, float softness)
        {
            _left.UpdateInverseFrictionValue(friction, mass, softness);
            _right.UpdateInverseFrictionValue(friction, mass, softness);
        }

        internal void ResetInverseFrictionValue()
        {
            _left.ResetInverseFrictionValue();
            _right.ResetInverseFrictionValue();
        }

        internal float CalculateDistortionOffset(string side, float distortionBase, float distortion) => side == Side.LEFT
            ? _left.distortionConfig.CalculateOffset(distortionLimit, distortionBase, distortion)
            : _right.distortionConfig.CalculateOffset(distortionLimit, distortionBase, distortion);

        internal override bool IsModified =>
            base.IsModified ||
            _left.gradientConfig.IsModified ||
            _left.distortionConfig.IsModified;
    }
}
