//
using NAT.Common.Components;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.Common.UI;
using NAT.TM.Handlers;
using System;
using System.Collections;
using SimpleJSON;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.Common
{
    class ScriptBase : MVRScript, IModParent
    {
        internal string DomainName { get; private set; }
        internal Color DomainColor { get; private set; }
        internal Loggr Loggr { get; private set; }
        internal Bootstrap bootstrap;
        internal bool? initialized;
        internal bool restoringFromJSON;
        internal bool restoredFromJSON;
        protected string calibrationDataVersion;
        internal Person person;

        protected readonly List<Handler> handlers = new List<Handler>();
        internal PluginPresetHandler presetHandler;
        internal AngleCalculator angleCalculator;
        internal JointPhysicsHandler jointPhysicsHandler;
        internal SoftPhysicsHandler softPhysicsHandler;
        internal HardColliderHandler hardColliderHandler;
        internal GravityPhysicsHandler gravityPhysicsHandler;
        internal Handler forcePhysicsHandler;
        internal ForceMorphHandler forceMorphHandler;
        internal ClothingHandler clothingHandler;
        internal FrictionHandler frictionHandler;
        internal NippleErectionHandler nippleErectionHandler;
        protected MorphListener morphListener;

        internal string LoadedPresetName { get; private set; }

        internal StorableFloat softnessJsf;
        internal StorableFloat quicknessJsf;
        internal float softness;
        internal float quickness;
        internal StorableAction calibrationAction;
        internal StorableFloat leftRightAngleOffsetJsf;
        internal StorableFloat upDownAngleOffsetJsf;
        internal StorableFloat depthOffsetJsf;
        protected float calibratedScale;

        internal CalibrationTarget leftCalibrationTarget, rightCalibrationTarget;
        protected Vector3 leftMeasurementOffset, rightMeasurementOffset;

        UnityEventsListener _parentObjectEventsListener;
        UnityEventsListener _pluginUIEventsListener;
        internal Bindings bindings;
        internal CalibrationStateHandler calibrationStateHandler;
        internal UIModManagerBase uiModManager;
        internal UIHandler uiHandler;

        internal virtual bool IsLdpUpdateOn => false;
        internal virtual bool IsSoftPhysicsOn => false;
        protected bool isUpdatePaused;
        internal bool IsDevelopmentEnv => bootstrap.IsDevelopmentEnv;
        internal bool IsTittyMagic { get; private set; }
        internal bool IsBootyMagic { get; private set; }
        internal bool IsInsideRestore => insideRestore;
        public bool IsEnabledAndActive() => this.IsEnabledNullSafe() && gameObject.IsActiveInHierarchyNullSafe();

        List<StorableGroup> _storableDomains;
        Dictionary<string, StorableGroup> _storableDomainsByName;

        internal ScriptBase OtherPlugin => IsBootyMagic ? bootstrap.TittyMagic : bootstrap.BootyMagic;
        bool _uiOpened;

        UIDynamicTextField _postponedInfoField;
        Action _initPostponedHandlers;

        void Start()
        {
            _initPostponedHandlers?.Invoke();
            _initPostponedHandlers = null;

            if(_parentObjectEventsListener == null)
            {
                _parentObjectEventsListener = transform.parent.gameObject.AddComponent<UnityEventsListener>(listener =>
                {
                    listener.disabledHandlers += () =>
                    {
                        SetGrayBackground();
                        uiHandler?.HideTabButtons();
                        uiHandler?.GoToInactiveWindow();
                    };
                    listener.enabledHandlers += () =>
                    {
                        if(_uiOpened && initialized == true)
                        {
                            uiHandler?.ShowTabButtons();
                            uiHandler?.GoToMainWindow();
                        }
                    };
                });
            }
        }

        protected void StartOrPostponeCoroutine(IEnumerator coroutine, Action onPostpone = null)
        {
            if(gameObject.activeInHierarchy)
            {
                StartCoroutine(coroutine);
            }
            else
            {
                onPostpone?.Invoke();
                _initPostponedHandlers += () => StartCoroutine(coroutine);
            }
        }

        protected void SetupDomain(string domainName, Color domainColor)
        {
            DomainName = domainName;
            DomainColor = domainColor;
            IsBootyMagic = domainName == Constant.BOOTY_MAGIC;
            IsTittyMagic = domainName == Constant.TITTY_MAGIC;
            Loggr = new Loggr(DomainName);
            _storableDomains = new List<StorableGroup>
            {
                new StorableGroup(StorableGroupName.GENERAL,
                    "General plugin parameters."
                ),
                new StorableGroup(StorableGroupName.JOINT_PHYSICS,
                    "Joint physics parameters."
                ),
                new StorableGroup(StorableGroupName.SOFT_PHYSICS,
                    "Soft physics parameters."
                ),
                new StorableGroup(StorableGroupName.FORCE_PHYSICS,
                    "Force physics parameters. These are not configurable in the UI."
                ),
                new StorableGroup(StorableGroupName.MORPHS,
                    "Directional force morphing and other morphing related parameters. Clothing morph settings are at the end of the list!"
                ),
                new StorableGroup(StorableGroupName.HARD_COLLIDERS,
                    "Hard colliders parameters."
                ),
                new StorableGroup(StorableGroupName.FRICTION,
                    "Friction parameters."
                ),
                new StorableGroup(StorableGroupName.UI_ACTIONS,
                    "Actions for opening specific UI windows. These actions are also available via the Keybindings plugin."
                ),
                new StorableGroup(StorableGroupName.INTERNAL,
                    "Internal parameters the used for inter-plugin communication. Do not touch."
                ),
                new StorableGroup(StorableGroupName.CLOTHING,
                    "Clothing parameters. Item specific settings are not stored - use clothing profile presets to save your settings."
                ),
            };
            for(int i = 0; i < _storableDomains.Count; i++)
            {
                _storableDomains[i].PreRegisterHeader(i);
            }

            _storableDomainsByName = new Dictionary<string, StorableGroup>();
            foreach(var domain in _storableDomains)
            {
                _storableDomainsByName[domain.Name] = domain;
            }
        }

        // Prevent ScriptBase from showing up as a plugin in Plugins tab.
        public override bool ShouldIgnore() => true;

        public override void InitUI()
        {
            if(ShouldIgnore() || DomainName == null)
            {
                return;
            }

            base.InitUI();
            if(!UITransform)
            {
                return;
            }

            StartOrPostponeCoroutine(InitUICo(), () =>
            {
                SetGrayBackground();
                _postponedInfoField = CreateTextField(new JSONStorableString("info", "Enable the atom to initialize.".Bold()));
                _postponedInfoField.backgroundColor = Color.clear;
            });
        }

        protected virtual void OnUIEnabled()
        {
        }

        protected virtual void OnUIDisabled()
        {
        }

        IEnumerator InitUICo()
        {
            if(_postponedInfoField != null)
            {
                RemoveTextField(_postponedInfoField);
            }

            if(!_pluginUIEventsListener)
            {
                _pluginUIEventsListener = UITransform.gameObject.AddComponent<UnityEventsListener>(listener =>
                {
                    listener.enabledHandlers += () => this.StartCoroutineIfGameObjectActive(OnUIEnabledCo());
                    listener.disabledHandlers += () => this.StartCoroutineIfGameObjectActive(OnUIDisabledCo());
                });
            }

            yield return null;
        }

        bool _inEnabledCo;

        IEnumerator OnUIEnabledCo()
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures onEnable logic executes just once.
                 */
                yield break;
            }

            _inEnabledCo = true;

            if(!_uiOpened)
            {
                SetGrayBackground();
            }

            while(initialized == null)
            {
                yield return null;
            }

            if(initialized == false)
            {
                yield break;
            }

            if(!_uiOpened)
            {
                uiHandler.CreateTabButtons();
                if(gameObject.activeInHierarchy)
                {
                    uiHandler.GoToMainWindow();
                }
                else
                {
                    uiHandler.HideTabButtons();
                    uiHandler.GoToInactiveWindow();
                }

                _uiOpened = true;
            }

            yield return null;
            yield return null;
            yield return null;
            OnUIEnabled();

            _inEnabledCo = false;
        }

        IEnumerator OnUIDisabledCo()
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures only onEnable logic executes.
                 */
                yield break;
            }

            OnUIDisabled();
        }

        void SetGrayBackground()
        {
            var background = rightUIContent.parent.parent.parent.transform.GetComponent<Image>();
            background.color = Colors.backgroundGray;
        }

        public Transform GetParentUI() => bootstrap.GetParentUI();

        internal void ConfigureColliderFriction()
        {
            if(!IsEnabledAndActive())
            {
                Loggr.LogMessage("Enable the plugin to configure collider friction.");
                return;
            }

            bindings.actions[$"{Constant.OPEN_UI}_{Constant.FRICTION}"].actionCallback();
        }

        internal virtual void OnSoftPhysicsToggled(bool value)
        {
        }

        internal void OnPhysicsRateChanged()
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            if(!calibrationStateHandler.Pending && !calibrationStateHandler.calibrateOnNextPending)
            {
                Loggr.LogMessage("Physics rate changed - calibration set to pending.");
            }

            SetHighQualityPhysics();
            calibrationStateHandler.SetCalibrationPending(Constant.PHYSICS_RATE_CHANGED);
        }

        void SetHighQualityPhysics()
        {
            if(IsBootyMagic)
            {
                person.SetGlutesHighQualityPhysics(true);
            }
            else if(IsTittyMagic)
            {
                person.SetBreastsHighQualityPhysics(true);
            }
        }

        internal void OnHighQualityPhysicsChanged()
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            SetHighQualityPhysics();
            SetMultipliersAndUpdatePhysics();
            hardColliderHandler.SyncCollidersMass();
        }

        protected virtual void SetMultipliersAndUpdatePhysics()
        {
        }

        internal void OnCharacterChanged()
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            frictionHandler.Refresh();
            uiModManager.SetFrictionUIVisibility();
        }

        internal void DeferProcessMorphsForListener()
        {
            if(IsEnabledAndActive() && !morphListener.ProcessingDone)
            {
                StartCoroutine(ProcessMorphsForListenerCo());
            }
        }

        IEnumerator ProcessMorphsForListenerCo()
        {
            Loggr.LogMessage("Processing all morphs to be able to detect changes. This could take a while if you have lots of morphs.");
            yield return null;
            ProcessMorphsForListener();
        }

        protected void ProcessMorphsForListener()
        {
            morphListener.ProcessMorphs(person.Geometry.morphBank1);
            if(IsTittyMagic && !person.IsFemale)
            {
                morphListener.ProcessMorphs(person.Geometry.morphBank1OtherGender);
            }
        }

        internal void SetCalibrationPending(IStorableParam storableParam)
        {
            if(restoringFromJSON || !IsEnabledAndActive())
            {
                return;
            }

            calibrationStateHandler.SetCalibrationPending(storableParam.Name, !storableParam.IsCalibrated);
        }

        internal bool inCalibrationCo;
        internal bool firstCalibrationDone;
        protected Coroutine calibrationCo;

        internal void StartCalibration(bool forceUpdateMass = false, JSONClass calibrationData = null)
        {
            if(restoringFromJSON || inCalibrationCo)
            {
                return;
            }

            if(!IsEnabledAndActive())
            {
                Loggr.LogMessage($"Enable {DomainName} to calibrate.");
                return;
            }

            if(forceUpdateMass || !bootstrap.CheckForSizeChangesJsb.val)
            {
                calibrationStateHandler.updateMassWhenCalibrating = true;
            }

            calibrationCo = StartCoroutine(CalibrationCo(calibrationData));
        }

        protected virtual IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            yield break;
        }

        internal virtual void SyncMorphsAngleOffsets(float? timer = null)
        {
        }

        protected CalibrationStateHandler AddCalibrationStateHandler() => gameObject.AddComponent<CalibrationStateHandler>(handler =>
        {
            handler.Init(this, () => StartCalibration());
            PreRegisterInternalParam(handler.CalibratingJsb);
        });

        // https://github.com/vam-community/vam-plugins-interop-specs/blob/main/keybindings.md
        public void OnBindingsListRequested(List<object> bindingsList)
        {
            bindings = bindings ?? new Bindings(this); // Might already be setup in Init.
            bindingsList.Add(bindings.namespaceDict);
            bindingsList.AddRange(bindings.GetActionsList());
        }

        internal virtual void SetCalibrationTargetFunctions(bool softPhysicsOn)
        {
        }

        // adapted from Timeline v4.3.1 (c) acidbubbles
        IEnumerator SelectPluginUICo(Action postAction = null)
        {
            while(initialized == null)
            {
                yield return null;
            }

            if(initialized == false)
            {
                yield break;
            }

            if(UITransform && UITransform.gameObject.activeInHierarchy)
            {
                if(enabled)
                {
                    postAction?.Invoke();
                }

                yield break;
            }

            yield return SelectContainingAtomTabCo("Plugins");

            float timeout = Time.unscaledTime + 1;
            while(Time.unscaledTime < timeout)
            {
                yield return null;

                if(!UITransform)
                {
                    continue;
                }

                /* Close any currently open plugin UI before opening this plugin's UI */
                foreach(Transform scriptController in manager.pluginContainer)
                {
                    var mvrScript = scriptController.gameObject.GetComponent<MVRScript>();
                    if(mvrScript && mvrScript != this)
                    {
                        mvrScript.UITransform.gameObject.SetActive(false);
                    }
                }

                if(enabled)
                {
                    UITransform.gameObject.SetActive(true);
                    postAction?.Invoke();
                    yield break;
                }
            }
        }

        internal void DeferOpenUIMain() => this.StartCoroutineIfGameObjectActive(SelectPluginUICo(() => uiHandler.GoToMainWindow()));
        internal void DeferOpenUIPhysics() => this.StartCoroutineIfGameObjectActive(SelectPluginUICo(() => uiHandler.GoToPhysicsWindow()));
        internal void DeferOpenUIMorphs() => this.StartCoroutineIfGameObjectActive(SelectPluginUICo(() => uiHandler.GoToMorphsWindow()));
        internal void DeferOpenUIHardColliders() => this.StartCoroutineIfGameObjectActive(SelectPluginUICo(() => uiHandler.GoToHardCollidersWindow()));

        internal void DeferOpenUIFriction() => this.StartCoroutineIfGameObjectActive(enabled
            ? SelectContainingAtomTabCo("Skin Materials 2", uiModManager.SwitchToOwnFrictionUI)
            : SelectContainingAtomTabCo("Plugins")
        );

        internal void DeferOpenUIClothing() => this.StartCoroutineIfGameObjectActive(SelectPluginUICo(() => uiHandler.GoToClothingWindow()));
        internal void DeferOpenUIDev() => StartCoroutine(SelectPluginUICo(() => uiHandler.GoToDevWindow()));
        internal void DeferOpenUIDevMorph() => StartCoroutine(SelectPluginUICo(() => uiHandler.GoToDevMorphWindow()));
        internal void DeferOpenUIColliderDev() => StartCoroutine(SelectPluginUICo(() => uiHandler.GoToColliderDevWindow()));
        internal void DeferOpenUI() => this.StartCoroutineIfGameObjectActive(SelectPluginUICo());

        IEnumerator SelectContainingAtomTabCo(string tabName, Action postAction = null)
        {
            if(SuperController.singleton.gameMode != SuperController.GameMode.Edit)
            {
                SuperController.singleton.gameMode = SuperController.GameMode.Edit;
            }

            SuperController.singleton.SelectController(containingAtom.mainController, false, false);
            SuperController.singleton.ShowMainHUDAuto();

            float timeout = Time.unscaledTime + 1;
            while(Time.unscaledTime < timeout)
            {
                yield return null;
                var selector = containingAtom.gameObject.GetComponentInChildren<UITabSelector>();
                if(selector)
                {
                    selector.SetActiveTab(tabName);
                    postAction?.Invoke();
                    break;
                }
            }
        }

        internal void DeferSetSaveFileBrowser(string fileExt) => StartCoroutine(SetSaveFileBrowserCo(fileExt));

        static IEnumerator SetSaveFileBrowserCo(string fileExt)
        {
            yield return null;
            var browser = SuperController.singleton.mediaFileBrowserUI;
            browser.browseVarFilesAsDirectories = false;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = $"{((int) (DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString()}.{fileExt}";
            browser.ActivateFileNameField();
        }

        internal void PreRegisterGeneralParam(IStorableParam param) => PreRegisterParam(StorableGroupName.GENERAL, param);
        internal void PreRegisterJointPhysicsParam(IStorableParam param) => PreRegisterParam(StorableGroupName.JOINT_PHYSICS, param);
        internal void PreRegisterSoftPhysicsParam(IStorableParam param) => PreRegisterParam(StorableGroupName.SOFT_PHYSICS, param);
        internal void PreRegisterForcePhysicsParam(IStorableParam param) => PreRegisterParam(StorableGroupName.FORCE_PHYSICS, param);
        internal void PreRegisterHardCollidersParam(IStorableParam param) => PreRegisterParam(StorableGroupName.HARD_COLLIDERS, param);
        internal void PreRegisterMorphsParam(IStorableParam param) => PreRegisterParam(StorableGroupName.MORPHS, param);
        internal void PreRegisterFrictionParam(IStorableParam param) => PreRegisterParam(StorableGroupName.FRICTION, param);
        internal void PreRegisterUIActionsParam(IStorableParam param) => PreRegisterParam(StorableGroupName.UI_ACTIONS, param);

        internal void PreRegisterInternalParam(IStorableParam param, bool temporary = true)
        {
            param.Temporary = temporary;
            PreRegisterParam(StorableGroupName.INTERNAL, param);
        }

        internal void PreRegisterClothingParam(IStorableParam param) => PreRegisterParam(StorableGroupName.CLOTHING, param);

        void PreRegisterParam(string domainName, IStorableParam param)
        {
            if(param == null)
            {
                throw new ArgumentException($"Parameter null! Cannot pre-register to {domainName}");
            }

            _storableDomainsByName[domainName].PreRegister(param);
        }

        protected void RegisterParams()
        {
            foreach(var handler in handlers)
            {
                handler.PreRegisterParams();
            }

            foreach(var domain in _storableDomains)
            {
                domain.RegisterAllTo(this);
            }
        }

        protected void UpdateBase()
        {
            if(bootstrap.CheckForSizeChangesJsb.val)
            {
                if(morphListener.ChangeDetected())
                {
                    calibrationStateHandler.SetCalibrationPending(Constant.SIZE_CHANGED, morphListener.IsChanged);
                }
            }

            if(softPhysicsHandler.softJointDrawService.isDrawingClothingCoverage)
            {
                clothingHandler.HandleToggleCoverageJoint();
            }

            if(_pluginUIEventsListener.IsEnabled)
            {
                var activeWindow = uiHandler.GetActiveWindow();
                if(activeWindow != null)
                {
                    bool atInterval = Time.frameCount % 6 == 0;
                    if(atInterval)
                    {
                        uiHandler.SetInfoPanelText(activeWindow.GetInfo());
                    }

                    activeWindow.UpdateDiagnostics(atInterval);
                }
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            jc[Constant.NATURALIS_VERSION] = Bootstrap.NATURALIS_VERSION;
            foreach(var domain in _storableDomains)
            {
                domain.RemoveTemporaryParamsFrom(jc);
            }

            return jc;
        }

        internal JSONClass GetFullBaseJSON() => base.GetJSON(true, true, true);

        internal void SavePresetJSON(JSONClass jc, string fullPath)
        {
            jc.Remove("enabled");
            SaveJSON(jc, fullPath);
            UpdateLoadedPresetName(fullPath);
        }

        public void RestoreFromPresetJSON(JSONClass jc, string fullPath)
        {
            if(initialized != true || restoringFromJSON)
            {
                Loggr.LogMessage("Preset file must be loaded after initialization.");
                return;
            }

            var restoredFromVersion = ParseVersion(jc);

            bool versionMisMatch = restoredFromVersion != null && !restoredFromVersion.IsMajorMinorEqual(bootstrap.NaturalisVersion);
            string restoreInfo = null;
            if(versionMisMatch)
            {
                restoreInfo = $"Loaded preset saved with version {restoredFromVersion.versionStr}";
            }

            string migrationInfo = MigrateFromVersion(jc, restoredFromVersion);
            if(!string.IsNullOrEmpty(restoreInfo) && !string.IsNullOrEmpty(migrationInfo))
            {
                Loggr.LogMessage(restoreInfo + migrationInfo);
            }

            jc["enabled"].AsBool = enabled;
            base.RestoreFromJSON(jc);
            string packageId = FileUtils.ParsePackageIdFromPath(fullPath);
            clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), packageId);
            UpdateLoadedPresetName(fullPath, versionMisMatch ? restoredFromVersion : null);

            if(bootstrap.CalibrateOnPresetLoadedJsb.val)
            {
                StartCalibration(true);
            }
        }

        public void ResetPluginParams()
        {
            var json = new JSONClass
            {
                [Constant.NATURALIS_VERSION] = Bootstrap.NATURALIS_VERSION,
            };

            if(IsBootyMagic)
            {
                json[softPhysicsHandler.ldpEnabledJsb.name].AsBool = bootstrap.GlutesLdpDefaultEnabledJsb.val;
            }
            else if(IsTittyMagic)
            {
                json[softPhysicsHandler.ldpEnabledJsb.name].AsBool = bootstrap.BreastsLdpDefaultEnabledJsb.val;
            }

            RestoreFromPresetJSON(json, null);
        }

        public void InitializeFromPresetJSON(JSONClass jc, string fullPath)
        {
            var restoredFromVersion = ParseVersion(jc);

            bool versionMisMatch = restoredFromVersion != null && !restoredFromVersion.IsMajorMinorEqual(bootstrap.NaturalisVersion);
            string restoreInfo = null;
            if(versionMisMatch)
            {
                restoreInfo = $"Loaded default preset saved with version {restoredFromVersion.versionStr}";
            }

            string migrationInfo = MigrateFromVersion(jc, restoredFromVersion);
            if(!string.IsNullOrEmpty(restoreInfo) && !string.IsNullOrEmpty(migrationInfo))
            {
                Loggr.LogMessage(restoreInfo + migrationInfo);
            }

            jc["enabled"].AsBool = enabled;
            base.RestoreFromJSON(jc);
            UpdateLoadedPresetName(fullPath, versionMisMatch ? restoredFromVersion : null);
        }

        static SemVer ParseVersion(JSONClass jc)
        {
            if(!jc.HasKey(Constant.NATURALIS_VERSION))
            {
                return null;
            }

            string versionStr = jc[Constant.NATURALIS_VERSION].Value;
            /* Prevent overriding registered param */
            jc[Constant.NATURALIS_VERSION] = Bootstrap.NATURALIS_VERSION;
            return new SemVer(versionStr, true);
        }

        string MigrateFromVersion(JSONClass jc, SemVer version)
        {
            if(!IsDevelopmentEnv)
            {
                try
                {
                    return new Migrater(DomainName, version, jc).Migrate();
                }
                catch(Exception e)
                {
                    Loggr.Error($"Error during migration from {version?.versionStr ?? "unknown version"}: {e}");
                }
            }

            return "";
        }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            restoringFromJSON = true;

            /* Disable early to allow correct enabled value to be used during Init */
            if(jc.HasKey("enabled"))
            {
                enabledJSON.val = jc["enabled"].AsBool;
            }

            StartOrPostponeCoroutine(RestoreFromJSONCo(
                jc,
                restorePhysical,
                restoreAppearance,
                presetAtoms,
                setMissingToDefault
            ));
        }

        IEnumerator RestoreFromJSONCo(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            JSONArray presetAtoms,
            bool setMissingToDefault
        )
        {
            while(initialized == null)
            {
                yield return null;
            }

            if(initialized == false)
            {
                yield break;
            }

            restoringFromJSON = false;
            string restoreInfo = null;

            var restoredFromVersion = ParseVersion(jc);
            if(restoredFromVersion != null && !restoredFromVersion.IsMajorMinorEqual(bootstrap.NaturalisVersion))
            {
                restoreInfo = $"Restored parameters from version {restoredFromVersion.versionStr}.";
            }

            string migrationInfo = MigrateFromVersion(jc, restoredFromVersion);
            if(!string.IsNullOrEmpty(restoreInfo))
            {
                Loggr.LogMessage(restoreInfo + migrationInfo);
            }

            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            JSONClass calibrationData = null;
            string scenePackageId = null;
            if(jc.HasKey("calibration_data"))
            {
                #if VAM_GT_1_21
                {
                    scenePackageId = FileUtils.ParsePackageIdFromPath(SuperController.singleton.LoadedSceneName);
                }
                #endif
                var tmpData = jc["calibration_data"].AsObject;
                if(
                    tmpData.HasKey("version") &&
                    new SemVer(tmpData["version"].Value).IsMajorMinorEqual(calibrationDataVersion)
                )
                {
                    calibrationData = tmpData;
                }
            }

            // ReSharper disable once ExpressionIsAlwaysNull
            FinishRestoreFromJSON(jc, scenePackageId, calibrationData);
            restoredFromJSON = true;
        }

        protected virtual void FinishRestoreFromJSON(JSONClass jc, string scenePackageId, JSONClass calibrationData)
        {
        }

        void UpdateLoadedPresetName(string fullPath, SemVer restoredFromVersion = null)
        {
            string loadedPresetName = "None";
            if(!string.IsNullOrEmpty(fullPath))
            {
                var parsedPath = new ParsedPath(fullPath, $"{FileUtils.savesDir}/{DomainName}Presets");
                loadedPresetName = parsedPath.fileName.Replace(".json", "");
                if(!string.IsNullOrEmpty(parsedPath.packageId))
                {
                    loadedPresetName += $" ({parsedPath.packageId}.var)";
                }

                if(restoredFromVersion != null && !restoredFromVersion.IsEmpty)
                {
                    loadedPresetName += $" [v{restoredFromVersion.versionStr}]";
                }
            }

            LoadedPresetName = loadedPresetName;
            uiHandler.mainWindow?.Refresh();
        }

        protected void BaseOnEnable()
        {
            if(firstCalibrationDone)
            {
                bootstrap.PersonMonitor.Check(true);
            }

            DeferProcessMorphsForListener();
        }

        protected void BaseOnDestroy()
        {
            DestroyImmediate(_parentObjectEventsListener);
            _parentObjectEventsListener = null;

            DestroyImmediate(_pluginUIEventsListener);
            _pluginUIEventsListener = null;

            Destroy(calibrationStateHandler);
            calibrationStateHandler = null;

            uiModManager?.Destroy();
        }
    }
}
