//
using NAT.Common.Handlers;
using NAT.Common.UI.Windows;
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;

namespace NAT.Common.UI
{
    sealed class UIHandler
    {
        readonly ScriptBase _script;
        readonly GravityPhysicsHandler _gravityPhysicsHandler;

        const string PHYSICS = "Physics";
        const string MORPHS = "Morphs";
        const string HARD_COLLIDERS = "Hard Colliders";
        const string CLOTHING = "Clothing";

        readonly Dictionary<string, Window> _tabs = new Dictionary<string, Window>();
        readonly Dictionary<string, NavigationButton> _tabButtons = new Dictionary<string, NavigationButton>();

        Window _activeTabWindow;
        // ReSharper disable MemberCanBePrivate.Global
        internal readonly Window mainWindow;
        internal readonly Window physicsWindow;
        internal readonly Window morphsWindow;
        internal readonly Window hardCollidersWindow;
        internal readonly Window clothingWindow;
        internal Window inactiveWindow;
        // ReSharper restore MemberCanBePrivate.Global

        readonly Section _uiDevSection;
        internal readonly Section infoPanelSection;
        internal readonly Section gravityControlsSection;
        UIDynamicTextField _infoPanelTextField;

        // TODO nested window as active window
        internal Window GetActiveWindow() => GetActiveNestedWindow() ?? _activeTabWindow;
        internal Window GetActiveNestedWindow() => _activeTabWindow?.activeNestedWindow;

        // ReSharper disable UnusedMember.Global
        internal bool IsMainTabActive() => _activeTabWindow == mainWindow;
        internal bool IsPhysicsTabActive() => _activeTabWindow == physicsWindow;

        internal bool IsMorphsTabActive() => _activeTabWindow == morphsWindow;
        internal bool IsHardCollidersTabActive() => _activeTabWindow == hardCollidersWindow;
        internal bool IsClothingTabActive() => _activeTabWindow == clothingWindow;
        internal bool IsInactiveWindowActive() => _activeTabWindow == inactiveWindow;
        internal bool IsNestedWindowActive() => GetActiveNestedWindow() != null;
        // ReSharper restore UnusedMember.Global

        internal bool IsActive(Window window) =>
            window == _activeTabWindow ? !IsNestedWindowActive() : GetActiveNestedWindow() == window;

        internal UIHandler(ScriptBase script)
        {
            _script = script;
            _gravityPhysicsHandler = script.gravityPhysicsHandler;
            mainWindow = new MainWindow(script);
            physicsWindow = new PhysicsWindow(script);
            morphsWindow = new MorphsWindow(script);
            hardCollidersWindow = new HardCollidersWindow(script);
            clothingWindow = new ClothingWindow(script);

            _tabs[_script.DomainName] = mainWindow;
            _tabs[PHYSICS] = physicsWindow;
            _tabs[MORPHS] = morphsWindow;
            _tabs[HARD_COLLIDERS] = hardCollidersWindow;
            _tabs[CLOTHING] = clothingWindow;
            infoPanelSection = new Section(BuildInfoPanelSection);
            gravityControlsSection = new Section(BuildGravityControlsSection);

            #if ENV_DEVELOPMENT
            {
                _uiDevSection = new Section(BuildUIDevSection);
                _rectJssc = new JSONStorableStringChooser("Rect", new List<string>(), Constant.SELECT, "Rect");
                _posX = new JSONStorableFloat("Pos X", 0.00f, -20.00f, 20.00f);
                _posY = new JSONStorableFloat("Pos Y", 0.00f, -20.00f, 20.00f);
                _sizeX = new JSONStorableFloat("Size X", 0.00f, -15.00f, 15.00f);
                _sizeY = new JSONStorableFloat("Size Y", 0.00f, -15.00f, 15.00f);
            }
            #endif
        }

        const string TOGGLE_ADJUSTMENTS_INFO = "Toggle Adjustments Info";
        const string ADJUSTMENTS_INFO = "Adjustments Info";

        static string AdjustmentsInfoLabelOpen => Constant.IS_GLOBAL_UI_SIDE_LEFT
            ? $"{"<".Bold()}  {ADJUSTMENTS_INFO}"
            : $"{ADJUSTMENTS_INFO}  {">".Bold()}";

        static string OpenSectionLabel => Constant.IS_GLOBAL_UI_SIDE_LEFT ? ">".Bold() : "<".Bold();

        void BuildInfoPanelSection(Section section)
        {
            float posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? 1088 : -307;
            {
                var uiDynamic = CreateButton(AdjustmentsInfoLabelOpen, new Vector2(posX, -214), new Vector2(-781, 50));
                uiDynamic.textColor = Color.white;
                uiDynamic.buttonColor = Colors.transparentGray;
                uiDynamic.AddListener(ToggleInfoPanel);
                section.AddElement(TOGGLE_ADJUSTMENTS_INFO, new Element(uiDynamic));
            }
            {
                var uiDynamic = CreateTextField("", new Vector2(posX, -1058), new Vector2(-781, 845));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.textColor = Color.white;
                uiDynamic.backgroundColor = Colors.transparentGray;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _infoPanelTextField = uiDynamic;
                section.AddElement(ADJUSTMENTS_INFO, new Element(uiDynamic));
            }
        }

        void ToggleInfoPanel()
        {
            var button = infoPanelSection.GetElement(TOGGLE_ADJUSTMENTS_INFO);
            var textField = infoPanelSection.GetElement(ADJUSTMENTS_INFO);

            if(button == null || textField == null || button.isAnimating)
            {
                return;
            }

            const float shiftX = 240;
            IEnumerator animation;

            if(textField.IsVisible)
            {
                animation = AnimateRect(
                    button,
                    new Vector2(Constant.IS_GLOBAL_UI_SIDE_LEFT ? 0 : shiftX, 0),
                    new Vector2(-shiftX, 0),
                    () =>
                    {
                        button.SetLabel(OpenSectionLabel);
                        textField.Hide();
                    }
                );
            }
            else
            {
                animation = AnimateRect(
                    button,
                    new Vector2(Constant.IS_GLOBAL_UI_SIDE_LEFT ? 0 : -shiftX, 0),
                    new Vector2(shiftX, 0),
                    () => textField.Show(),
                    () => button.SetLabel(AdjustmentsInfoLabelOpen)
                );
            }

            _script.StartCoroutine(animation);
        }

        internal void SetInfoPanelText(string text)
        {
            if(_infoPanelTextField == null)
            {
                return;
            }

            _infoPanelTextField.text = text;
        }

        const string TOGGLE_GRAVITY_CONTROLS = "Toggle Gravity Controls";
        const string SIMULATE_GRAVITY = "Simulate Gravity";
        const string STOP_GRAVITY_SIMULATION = "Stop";

        static string GravityControlsLabelOpen => Constant.IS_GLOBAL_UI_SIDE_LEFT
            ? $"{"<".Bold()}  {SIMULATE_GRAVITY}"
            : $"{SIMULATE_GRAVITY}  {">".Bold()}";

        void BuildGravityControlsSection(Section section)
        {
            float posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? 1088 : -282;

            {
                var button = CreateButton(GravityControlsLabelOpen, new Vector2(posX, -1121), new Vector2(-806, 50));
                button.textColor = Color.white;
                button.buttonColor = Colors.transparentGray;
                button.AddListener(ToggleGravityControls);
                section.AddElement(TOGGLE_GRAVITY_CONTROLS, new Element(button));
            }

            float controlsPosX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? 1088 : -100;
            float controlsSpacing = Constant.IS_GLOBAL_UI_SIDE_LEFT ? 91 : -91;

            CreateGravityControlButton(Direction.UP, "↑", new Vector2(controlsPosX + controlsSpacing, -1180), section);
            CreateGravityControlButton(Direction.DOWN, "↓", new Vector2(controlsPosX + controlsSpacing, -1239), section);

            posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? controlsPosX : controlsPosX + 2 * controlsSpacing;
            CreateGravityControlButton(Direction.BACK, "BK", new Vector2(posX, -1180), section);

            posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? controlsPosX + 2 * controlsSpacing : controlsPosX;
            CreateGravityControlButton(Direction.FORWARD, "FW", new Vector2(posX, -1180), section);

            posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? controlsPosX : controlsPosX + 2 * controlsSpacing;
            CreateGravityControlButton(Direction.RIGHT, "←", new Vector2(posX, -1239), section);

            posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? controlsPosX + 2 * controlsSpacing : controlsPosX;
            CreateGravityControlButton(Direction.LEFT, "→", new Vector2(posX, -1239), section);

            {
                posX = Constant.IS_GLOBAL_UI_SIDE_LEFT ? controlsPosX : controlsPosX + 2 * controlsSpacing;
                var button = CreateButton(STOP_GRAVITY_SIMULATION.Bold(), new Vector2(posX, -1289), new Vector2(-806, 50));
                button.buttonText.fontSize = 28;
                button.buttonColor = Colors.transparentPink;
                _gravityPhysicsHandler.StopGravitySimulationAction.button = button.button;
                section.AddElement(STOP_GRAVITY_SIMULATION, new Element(button));
            }
        }

        void CreateGravityControlButton(string direction, string label, Vector2 position, Section section)
        {
            var size = new Vector2(-988, 60);
            var button = CreateButton(label, position, size);
            button.textColor = _gravityPhysicsHandler.GravitySimulationDirection == direction && !_script.calibrationStateHandler.CalibratingJsb.val
                ? Colors.funkyCyan
                : Color.white;
            button.buttonColor = Colors.transparentGray;
            button.buttonText.fontSize = direction == Direction.BACK || direction == Direction.FORWARD ? 28 : 36;
            _gravityPhysicsHandler.simulateGravityActions[direction].button = button.button;
            section.AddElement(direction, new Element(button));
        }

        void ToggleGravityControls()
        {
            var toggleButton = gravityControlsSection.GetElement(TOGGLE_GRAVITY_CONTROLS);
            var stopButton = gravityControlsSection.GetElement(STOP_GRAVITY_SIMULATION);

            if(toggleButton == null || stopButton == null || toggleButton.isAnimating)
            {
                return;
            }

            const float shiftX = 215;
            IEnumerator animation;

            if(stopButton.IsVisible)
            {
                animation = AnimateRect(
                    toggleButton,
                    new Vector2(Constant.IS_GLOBAL_UI_SIDE_LEFT ? 0 : shiftX, 0),
                    new Vector2(-shiftX, 0),
                    () =>
                    {
                        toggleButton.SetLabel(OpenSectionLabel);
                        foreach(var pair in _gravityPhysicsHandler.simulateGravityActions)
                        {
                            gravityControlsSection.GetElement(pair.Key).Hide();
                        }

                        stopButton.Hide();
                        _gravityPhysicsHandler.StopGravitySimulationAction.actionCallback();
                    }
                );
            }
            else
            {
                animation = AnimateRect(
                    toggleButton,
                    new Vector2(Constant.IS_GLOBAL_UI_SIDE_LEFT ? 0 : -shiftX, 0),
                    new Vector2(shiftX, 0),
                    () =>
                    {
                        foreach(var pair in _gravityPhysicsHandler.simulateGravityActions)
                        {
                            gravityControlsSection.GetElement(pair.Key).Show();
                        }

                        stopButton.Show();
                    },
                    () => toggleButton.SetLabel(GravityControlsLabelOpen)
                );
            }

            _script.StartCoroutine(animation);
        }

        internal void SetButtonColors(string direction)
        {
            if(gravityControlsSection.GetElement(TOGGLE_GRAVITY_CONTROLS) == null)
            {
                return;
            }

            var triggeredAction = _gravityPhysicsHandler.simulateGravityActions[direction];
            foreach(var pair in _gravityPhysicsHandler.simulateGravityActions)
            {
                var uiDynamic = gravityControlsSection.GetElement(pair.Key).GetUIDynamicAs<UIDynamicButton>();
                uiDynamic.textColor = pair.Value == triggeredAction ? Colors.funkyCyan : Color.white;
            }
        }

        internal void ResetButtonColors()
        {
            if(gravityControlsSection.GetElement(TOGGLE_GRAVITY_CONTROLS) == null)
            {
                return;
            }

            foreach(var pair in _gravityPhysicsHandler.simulateGravityActions)
            {
                var uiDynamic = gravityControlsSection.GetElement(pair.Key).GetUIDynamicAs<UIDynamicButton>();
                uiDynamic.textColor = Color.white;
            }
        }

        static IEnumerator AnimateRect(
            Element element,
            Vector2 pos,
            Vector2 size,
            Action startAction = null,
            Action endAction = null,
            float duration = 0.50f
        )
        {
            var rect = element.Rect;
            var startPosition = rect.anchoredPosition;
            var startSize = rect.sizeDelta;
            var targetPosition = startPosition + pos;
            var targetSize = startSize + size;
            float elapsedTime = 0f;

            element.isAnimating = true;
            startAction?.Invoke();
            while(elapsedTime < duration)
            {
                rect.anchoredPosition = new Vector2(
                    Mathf.SmoothStep(startPosition.x, targetPosition.x, elapsedTime / duration),
                    Mathf.SmoothStep(startPosition.y, targetPosition.y, elapsedTime / duration)
                );
                rect.sizeDelta = new Vector2(
                    Mathf.SmoothStep(startSize.x, targetSize.x, elapsedTime / duration),
                    Mathf.SmoothStep(startSize.y, targetSize.y, elapsedTime / duration)
                );
                elapsedTime += Time.deltaTime;
                yield return null;
            }

            endAction?.Invoke();
            element.isAnimating = false;
            rect.anchoredPosition = targetPosition;
            rect.sizeDelta = targetSize;
        }

        internal void CreateTabButtons()
        {
            const float spacing = 215;
            var keys = new List<string>(_tabs.Keys);
            for(int i = 0; i < keys.Count; i++)
            {
                string id = keys[i];
                CreateNavigationButton(id, 2f + i * spacing, () => NavigateToTab(id));
            }

            _script.CreateSpacer().height = 50f;
            _script.CreateSpacer(true).height = 50f;
        }

        void CreateNavigationButton(string id, float posX, UnityAction callback)
        {
            var element = new NavigationButton(
                CreateButton(id, new Vector2(posX, -72), new Vector2(-864, 60)),
                _script.DomainColor,
                Colors.darkerGray
            );
            element.AddListener(callback);
            element.SetInactive();
            _tabButtons[id] = element;
        }

        internal void HideTabButtons()
        {
            foreach(var pair in _tabButtons)
            {
                pair.Value.Hide();
            }
        }

        internal void ShowTabButtons()
        {
            foreach(var pair in _tabButtons)
            {
                pair.Value.Show();
            }
        }

        // ReSharper disable UnusedMethodReturnValue.Global
        internal Window GoToMainWindow() => NavigateToTab(_script.DomainName);
        internal Window GoToPhysicsWindow() => NavigateToTab(PHYSICS);
        internal Window GoToMorphsWindow() => NavigateToTab(MORPHS);
        internal Window GoToHardCollidersWindow() => NavigateToTab(HARD_COLLIDERS);
        internal Window GoToClothingWindow() => NavigateToTab(CLOTHING);

        internal Window GoToDevWindow() => GoToMainWindow()?.OpenNestedWindow(nameof(DevWindow));
        internal Window GoToDevMorphWindow() => GoToMainWindow()?.OpenNestedWindow(nameof(DevMorphWindow));
        internal Window GoToColliderDevWindow() => GoToHardCollidersWindow()?.OpenNestedWindow(nameof(ColliderDevWindow));
        // ReSharper restore UnusedMethodReturnValue.Global

        internal void GoToInactiveWindow()
        {
            if(inactiveWindow == null)
            {
                inactiveWindow = new InactiveWindow(_script);
            }

            _activeTabWindow?.Clear();
            _activeTabWindow = inactiveWindow;
            inactiveWindow.Show();
        }

        Window NavigateToTab(string id)
        {
            if(!_tabs.ContainsKey(id))
            {
                return null;
            }

            if(id == HARD_COLLIDERS && !_script.enabled)
            {
                _script.Loggr.LogMessage("Enable the plugin to configure hard colliders.");
                return null;
            }

            var window = _tabs[id];
            if(_activeTabWindow == mainWindow && IsNestedWindowActive())
            {
                _activeTabWindow.OnReturn();
            }

            _activeTabWindow?.Clear();
            _activeTabWindow = window;
            window.Show();

            var navigationButton = _tabButtons[id];
            navigationButton.SetActive();
            foreach(var pair in _tabButtons)
            {
                if(pair.Key != id)
                {
                    pair.Value.SetInactive();
                }
            }

            if(_script.IsDevelopmentEnv)
            {
                UpdateRectOptions(id);
            }

            return window;
        }

        internal void ClosePopups() => _activeTabWindow.ClosePopups();

        internal UIDynamicButton CreateButton(string label, Vector2 position, Vector2 size)
        {
            var t = _script.InstantiateButton(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicButton>();
            uiDynamic.label = label;
            return uiDynamic;
        }

        internal UIDynamicTextField CreateTextField(string text, Vector2 position, Vector2 size)
        {
            var t = _script.InstantiateTextField(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicTextField>();
            uiDynamic.text = text;
            uiDynamic.backgroundColor = Color.clear;
            return uiDynamic;
        }

        internal UIDynamicSlider CreateSlider(string label, string valueFormat, Vector2 position)
        {
            var t = _script.InstantiateSlider(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = new Vector2(-555, 120);
            var uiDynamic = t.GetComponent<UIDynamicSlider>();
            uiDynamic.label = label;
            uiDynamic.rangeAdjustEnabled = false;
            uiDynamic.valueFormat = valueFormat;
            ConfigureSlider(uiDynamic);
            return uiDynamic;
        }

        internal UIDynamicToggle CreateToggle(string label, Vector2 position, float sizeX)
        {
            var t = _script.InstantiateToggle(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = new Vector2(sizeX, 50);
            var uiDynamic = t.GetComponent<UIDynamicToggle>();
            uiDynamic.label = label;
            return uiDynamic;
        }

        internal UIDynamicPopup CreatePopup(string label, Vector2 position, Vector2 size)
        {
            var t = _script.InstantiatePopup(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;
            uiDynamic.popup.selectColor = Colors.lightBlue;
            return uiDynamic;
        }

        internal UIDynamicPopup CreateScrollablePopup(string label, Vector2 position, Vector2 size)
        {
            var t = _script.InstantiateScrollablePopup(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;
            uiDynamic.popup.selectColor = Colors.lightBlue;
            return uiDynamic;
        }

        internal UIDynamicPopup CreateFilterablePopup(string label, Vector2 position, Vector2 size)
        {
            var t = _script.InstantiateFilterablePopup(_script.UITransform);
            var rectT = t.GetComponent<RectTransform>();
            rectT.pivot = Vector2.zero;
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;
            uiDynamic.popup.selectColor = Colors.lightBlue;
            return uiDynamic;
        }

        static void ConfigureSlider(UIDynamicSlider uiDynamic)
        {
            var uiDynamicT = uiDynamic.transform;
            var rectT = uiDynamicT.GetComponent<RectTransform>();
            var position = rectT.anchoredPosition;
            rectT.anchoredPosition = new Vector2(position.x - 55, position.y);
            rectT.sizeDelta = new Vector2(-444, 118);

            // Debug.Log(Utils.ObjectHierarchyToString(slider.transform));
            // foreach(Transform t1 in uiDynamic.transform)
            // {
            //     Debug.Log($"{t1.name}:\n{Utils.EnumerableToString(Utils.GetChildComponentData(t1))}");
            // }

            var panelT = uiDynamicT.Find("Panel");
            var sliderT = uiDynamicT.Find("Slider");
            var textT = uiDynamicT.Find("Text");
            var valueInputFieldT = uiDynamicT.Find("ValueInputField");
            var defaultValueButtonT = uiDynamicT.Find("DefaultValueButton");
            var quickButtonsLeftT = uiDynamicT.Find("QuickButtonsGroup/QuickButtonsLeft");
            var quickButtonsRightT = uiDynamicT.Find("QuickButtonsGroup/QuickButtonsRight");
            var m1ButtonT = quickButtonsLeftT.Find("M1Button");
            var m2ButtonT = quickButtonsLeftT.Find("M2Button");
            var m3ButtonT = quickButtonsLeftT.Find("M3Button");
            var m4ButtonT = quickButtonsLeftT.Find("M4Button");
            var p1ButtonT = quickButtonsRightT.Find("P1Button");
            var p2ButtonT = quickButtonsRightT.Find("P2Button");
            var p3ButtonT = quickButtonsRightT.Find("P3Button");
            var p4ButtonT = quickButtonsRightT.Find("P4Button");

            textT.GetComponent<Text>().fontSize = 27;
            foreach(var childText in valueInputFieldT.GetComponentsInChildren<Text>())
            {
                childText.fontSize = 27;
            }

            m4ButtonT.gameObject.SetActive(false);
            p4ButtonT.gameObject.SetActive(false);

            var panelRectT = panelT.GetComponent<RectTransform>();
            var sliderRectT = sliderT.GetComponent<RectTransform>();
            var textRectT = textT.GetComponent<RectTransform>();
            var valueInputFieldRectT = valueInputFieldT.GetComponent<RectTransform>();
            var defaultValueButtonRectT = defaultValueButtonT.GetComponent<RectTransform>();
            var m1ButtonRectT = m1ButtonT.GetComponent<RectTransform>();
            var m2ButtonRectT = m2ButtonT.GetComponent<RectTransform>();
            var m3ButtonRectT = m3ButtonT.GetComponent<RectTransform>();
            var p1ButtonRectT = p1ButtonT.GetComponent<RectTransform>();
            var p2ButtonRectT = p2ButtonT.GetComponent<RectTransform>();
            var p3ButtonRectT = p3ButtonT.GetComponent<RectTransform>();

            panelRectT.sizeDelta = new Vector2(-114, 0);
            sliderRectT.sizeDelta = new Vector2(-125, 35);
            sliderRectT.anchoredPosition = new Vector2(0, 53);
            textRectT.anchoredPosition = new Vector2(60, -25);
            valueInputFieldRectT.anchoredPosition = new Vector2(-65, -25);

            defaultValueButtonRectT.anchoredPosition = new Vector2(1, 18);
            defaultValueButtonRectT.sizeDelta = new Vector2(100, 32);
            defaultValueButtonRectT.GetComponentInChildren<Text>().fontSize = 20;

            ConfigureQuickButton(m1ButtonRectT);
            ConfigureQuickButton(m2ButtonRectT);
            ConfigureQuickButton(m3ButtonRectT);
            ConfigureQuickButton(p1ButtonRectT, true);
            ConfigureQuickButton(p2ButtonRectT, true);
            ConfigureQuickButton(p3ButtonRectT, true);

            // AddRect(panelT.name, panelRectT);
            // AddRect(sliderT.name, sliderRectT);
            // AddRect(textT.name, textRectT);
            // AddRect(valueInputFieldT.name, valueInputFieldRectT);
            // AddRect(defaultValueButtonT.name, defaultValueButtonRectT);
            // AddRect(m1ButtonT.name, m1ButtonRectT);
            // AddRect(m2ButtonT.name, m2ButtonRectT);
            // AddRect(m3ButtonT.name, m3ButtonRectT);
            // AddRect(p1ButtonT.name, p1ButtonRectT);
            // AddRect(p2ButtonT.name, p2ButtonRectT);
            // AddRect(p3ButtonT.name, p3ButtonRectT);
        }

        static void ConfigureQuickButton(RectTransform rectT, bool right = false)
        {
            rectT.GetComponentInChildren<Text>().fontSize = 20;
            rectT.anchoredPosition = new Vector2(right ? 18 : -17, 18);
            rectT.sizeDelta = new Vector2(0, 32);
        }

        // ReSharper disable once MemberCanBeMadeStatic.Global
        internal void ConfigurePopup(UIDynamicPopup uiDynamic, float height, float offsetX = 0, bool upwards = false)
        {
            uiDynamic.popup.labelText.color = Color.black;

            if(height > 0f)
            {
                uiDynamic.popupPanelHeight = height;
            }

            float offsetY = upwards ? height + 60 : 0;
            uiDynamic.popup.popupPanel.offsetMin += new Vector2(offsetX, offsetY);
            uiDynamic.popup.popupPanel.offsetMax += new Vector2(offsetX, offsetY);
        }

        internal void AddPrevNextButtons(UIDynamicPopup uiDynamic, bool cycle = false)
        {
            var popup = uiDynamic.popup;
            popup.labelText.alignment = TextAnchor.UpperCenter;
            popup.labelText.GetComponent<RectTransform>().anchorMax = new Vector2(0, 0.89f);

            /* Prev */
            {
                var t = _script.InstantiateButton(uiDynamic.transform);
                var uiDynamicButton = t.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = "<";
                if(cycle)
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetPreviousOrLastValue();
                    });
                }
                else
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetPreviousValue();
                    });
                }

                var rectT = t.GetComponent<RectTransform>();
                rectT.pivot = Vector2.zero;
                rectT.anchoredPosition = new Vector2(10, 0);
                rectT.sizeDelta = Vector2.zero;
                rectT.offsetMin = new Vector2(5, 5f);
                rectT.offsetMax = new Vector2(80, 70);
                rectT.anchorMin = Vector2.zero;
                rectT.anchorMax = Vector2.zero;
            }

            /* Next */
            {
                var t = _script.InstantiateButton(uiDynamic.transform);
                var uiDynamicButton = t.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = ">";
                if(cycle)
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetNextOrFirstValue();
                    });
                }
                else
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetNextValue();
                    });
                }

                var rectT = t.GetComponent<RectTransform>();
                rectT.pivot = Vector2.zero;
                rectT.anchoredPosition = Vector2.zero;
                rectT.sizeDelta = Vector2.zero;
                rectT.offsetMin = new Vector2(82, 5);
                rectT.offsetMax = new Vector2(157, 70);
                rectT.anchorMin = Vector2.zero;
                rectT.anchorMax = Vector2.zero;
            }
        }

#region Development

        const float SCALE_MULT = 100f;
        RectTransform _selectedRect;
        readonly Dictionary<string, RectTransform> _rectTransforms = new Dictionary<string, RectTransform>();
        readonly JSONStorableStringChooser _rectJssc;
        readonly JSONStorableFloat _posX;
        readonly JSONStorableFloat _posY;
        readonly JSONStorableFloat _sizeX;
        readonly JSONStorableFloat _sizeY;

        internal void ToggleDevSection()
        {
            if(!_uiDevSection.IsShown)
            {
                _uiDevSection.Show();
                UpdateRectOptions(_activeTabWindow.activeNestedWindow ?? _activeTabWindow);
            }
            else
            {
                _uiDevSection.Hide();
            }
        }

        void UpdateRectOptions(string windowId)
        {
            if(!_tabs.ContainsKey(windowId))
            {
                return;
            }

            UpdateRectOptions(_tabs[windowId]);
        }

        internal void UpdateRectOptions(Window window)
        {
            if(!_script.IsDevelopmentEnv || !_uiDevSection.IsShown)
            {
                return;
            }

            _rectTransforms.Clear();
            foreach(var pair in window.GetVisibleElements())
            {
                _rectTransforms[pair.Key] = pair.Value.Rect;
            }

            var options = new List<string> { Constant.SELECT };
            options.AddRange(_rectTransforms.Keys);
            _rectJssc.choices = options;
        }

        void BuildUIDevSection(Section section)
        {
            var rectPopup = CreateFilterablePopup(_rectJssc.name, new Vector2(1475, -1050), new Vector2(-555, 120));
            ConfigurePopup(rectPopup, 800, 0, true);
            AddPrevNextButtons(rectPopup);
            _rectJssc.setCallbackFunction = SelectRect;
            _rectJssc.popup = rectPopup.popup;

            var posX = CreateSlider(_posX.name, "F2", new Vector2(1475, -1200));
            _posX.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.anchoredPosition = SCALE_MULT * new Vector2(value, _posY.val);
                }
            };
            _posX.slider = posX.slider;

            var posY = CreateSlider(_posY.name, "F2", new Vector2(1475, -1350));
            _posY.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.anchoredPosition = SCALE_MULT * new Vector2(_posX.val, value);
                }
            };
            _posY.slider = posY.slider;

            var sizeX = CreateSlider(_sizeX.name, "F2", new Vector2(1475, -1500));
            _sizeX.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.sizeDelta = SCALE_MULT * new Vector2(value, _sizeY.val);
                }
            };
            _sizeX.slider = sizeX.slider;

            var sizeY = CreateSlider(_sizeY.name, "F2", new Vector2(1475, -1650));
            _sizeY.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.sizeDelta = SCALE_MULT * new Vector2(_sizeX.val, value);
                }
            };
            _sizeY.slider = sizeY.slider;

            section.AddElement(_rectJssc.name, new Element(rectPopup));
            section.AddElement(_posX.name, new Element(posX));
            section.AddElement(_posY.name, new Element(posY));
            section.AddElement(_sizeX.name, new Element(sizeX));
            section.AddElement(_sizeY.name, new Element(sizeY));
        }

        void SelectRect(string key)
        {
            if(!_rectTransforms.ContainsKey(key))
            {
                _selectedRect = null;
                return;
            }

            _selectedRect = _rectTransforms[key];
            var anchoredPosition = _selectedRect.anchoredPosition;
            _posX.defaultVal = anchoredPosition.x / SCALE_MULT;
            _posY.defaultVal = anchoredPosition.y / SCALE_MULT;
            _posX.val = _posX.defaultVal;
            _posY.val = _posY.defaultVal;
            var sizeDelta = _selectedRect.sizeDelta;
            _sizeX.defaultVal = sizeDelta.x / SCALE_MULT;
            _sizeY.defaultVal = sizeDelta.y / SCALE_MULT;
            _sizeX.val = _sizeX.defaultVal;
            _sizeY.val = _sizeY.defaultVal;
        }

#endregion Development
    }
}
