using NAT.Common.Models;
using System;
using System.Collections.Generic;
using System.Text;
using UnityEngine;

namespace NAT.Common.UI
{
    abstract class Window
    {
        protected readonly ScriptBase script;

        readonly List<JSONStorableString> _stringStorables = new List<JSONStorableString>();
        // readonly List<JSONStorableAction> _actionStorables = new List<JSONStorableAction>();
        readonly List<JSONStorableFloat> _floatStorables = new List<JSONStorableFloat>();
        readonly List<JSONStorableBool> _boolStorables = new List<JSONStorableBool>();
        readonly List<JSONStorableStringChooser> _chooserStorables = new List<JSONStorableStringChooser>();
        readonly Dictionary<string, Element> _elements = new Dictionary<string, Element>();

        readonly string _id;

        protected Section currentSection;
        protected readonly Dictionary<string, Section> sections = new Dictionary<string, Section>();

        readonly Dictionary<string, Window> _nestedWindows = new Dictionary<string, Window>();
        protected Window parent;
        internal Window activeNestedWindow;
        protected bool isBuilt;

        protected bool showInfoPanel;
        protected readonly StringBuilder infoSb = new StringBuilder();

        protected Window(ScriptBase script, string id = null)
        {
            this.script = script;
            _id = id;
        }

        protected UIDynamicButton CreateButton(string text, Vector2 position, Vector2 size) => script.uiHandler.CreateButton(text, position, size);
        protected UIDynamicTextField CreateTextField(string text, Vector2 position, Vector2 size) => script.uiHandler.CreateTextField(text, position, size);
        protected UIDynamicSlider CreateSlider(string label, string valueFormat, Vector2 position) => script.uiHandler.CreateSlider(label, valueFormat, position);
        protected UIDynamicToggle CreateToggle(string label, Vector2 position, float sizeX) => script.uiHandler.CreateToggle(label, position, sizeX);
        protected UIDynamicPopup CreatePopup(string label, Vector2 position, Vector2 size) => script.uiHandler.CreatePopup(label, position, size);
        protected UIDynamicPopup CreateFilterablePopup(string label, Vector2 position, Vector2 size) => script.uiHandler.CreateFilterablePopup(label, position, size);
        protected UIDynamicPopup CreateScrollablePopup(string label, Vector2 position, Vector2 size) => script.uiHandler.CreateScrollablePopup(label, position, size);

        // ReSharper disable MemberCanBePrivate.Global UnusedMethodReturnValue.Global
        protected Element AddTextField(JSONStorableString jss, Func<string, UIDynamicTextField> createUIDynamic, Section section = null)
        {
            _stringStorables.Add(jss);
            return AddElement(jss.name, createUIDynamic, section);
        }

        protected Element AddSlider(JSONStorableFloat jsf, Func<string, UIDynamicSlider> createUIDynamic, Section section = null)
        {
            _floatStorables.Add(jsf);
            return AddElement(jsf.name, createUIDynamic, section);
        }

        protected Element AddToggle(JSONStorableBool jsb, Func<string, UIDynamicToggle> createUIDynamic, Section section = null)
        {
            _boolStorables.Add(jsb);
            return AddElement(jsb.name, createUIDynamic, section);
        }

        protected Element AddPopup(JSONStorableStringChooser jssc, Func<string, UIDynamicPopup> createUIDynamic, Section section = null)
        {
            _chooserStorables.Add(jssc);
            return AddElement(jssc.name, createUIDynamic, section);
        }

        protected Element AddElement<T>(string name, Func<string, T> createUIDynamic, Section section = null) where T : UIDynamic
        {
            var element = new Element(createUIDynamic(name));
            _elements.Add(name, element);
            if(section != null)
            {
                section.AddElement(name, element);
            }

            return element;
        }

        protected void AddSection(string id, Action<Section> buildAction) => sections.Add(id, new Section(buildAction));

        protected void AddSectionNavigationButton(string sectionId, Vector2 position, float sizeX, bool enabled = true) =>
            AddElement(sectionId, _ =>
            {
                var uiDynamic = CreateButton(sectionId, position, new Vector2(sizeX, 55));
                uiDynamic.buttonText.fontSize = 26;
                var navigationButton = new NavigationButton(uiDynamic, script.DomainColor, Color.gray);
                navigationButton.AddListener(() => ShowSection(sectionId));
                navigationButton.SetEnabled(enabled);
                sections[sectionId].navigationButton = navigationButton;
                return uiDynamic;
            });

        protected void AddNestedWindow(Window window)
        {
            window.parent = this;
            _nestedWindows.Add(window._id, window);
        }

        void AddReturnButton() => AddElement("<  Return", name =>
        {
            var uiDynamic = CreateButton(name, new Vector2(0, -140), new Vector2(-645, 60));
            uiDynamic.buttonText.fontSize = 28;
            uiDynamic.textColor = Color.white;
            var colors = uiDynamic.button.colors;
            colors.normalColor = Colors.sliderGray;
            colors.highlightedColor = Color.grey;
            colors.pressedColor = Color.grey;
            uiDynamic.button.colors = colors;
            uiDynamic.AddListener(parent.OnReturn);
            return uiDynamic;
        });

        internal void Show()
        {
            if(activeNestedWindow != null)
            {
                activeNestedWindow.Show();
                return;
            }

            ShowPrevElements();

            if(showInfoPanel)
            {
                script.uiHandler.infoPanelSection.Show(true);
            }

            if(parent != null && !isBuilt)
            {
                AddReturnButton();
            }

            if(!isBuilt)
            {
                Build();
                isBuilt = true;
            }

            SyncElementsToStorables();
            PostShow();
        }

        void ShowPrevElements()
        {
            foreach(var pair in _elements)
            {
                pair.Value.ShowPrev();
            }

            foreach(var pair in _elements)
            {
                pair.Value.ShowPrev();
            }
        }

        protected abstract void Build();

        void SyncElementsToStorables()
        {
            _stringStorables.ForEach(storable => storable.dynamicText = _elements[storable.name].GetUIDynamicAs<UIDynamicTextField>());
            // _actionStorables.ForEach(storable => storable.button = _elements[storable.name].GetUIDynamicAs<UIDynamicButton>().button);
            _floatStorables.ForEach(storable => storable.slider = _elements[storable.name].GetUIDynamicAs<UIDynamicSlider>().slider);
            _boolStorables.ForEach(storable => storable.toggle = _elements[storable.name].GetUIDynamicAs<UIDynamicToggle>().toggle);
            _chooserStorables.ForEach(storable => storable.popup = _elements[storable.name].GetUIDynamicAs<UIDynamicPopup>().popup);
        }

        protected virtual void PostShow()
        {
        }

        protected virtual void OnClose()
        {
        }

        internal void OnReturn()
        {
            activeNestedWindow.Clear();
            activeNestedWindow = null;
            Show();
            script.uiHandler.UpdateRectOptions(this);
        }

        protected Element GetElement(string name)
        {
            Element element;
            _elements.TryGetValue(name, out element);
            return element;
        }

        protected void ShowSection(string id)
        {
            foreach(var pair in sections)
            {
                if(pair.Key == id)
                {
                    currentSection = pair.Value;
                    currentSection.Show();
                }
                else
                {
                    pair.Value.Hide();
                }
            }

            SyncElementsToStorables();
            script.uiHandler.UpdateRectOptions(this);
            PostShowSection(currentSection);
        }

        protected virtual void PostShowSection(Section section)
        {
        }

        internal Window OpenNestedWindow(string id)
        {
            activeNestedWindow?.Clear();
            activeNestedWindow = null;
            Clear();

            if(!_nestedWindows.ContainsKey(id))
            {
                script.Loggr.LogError($"Nested window with id '{id}' not found");
                return null;
            }

            activeNestedWindow = _nestedWindows[id];
            activeNestedWindow.Show();
            script.uiHandler.UpdateRectOptions(activeNestedWindow);
            return activeNestedWindow;
        }

        internal virtual void OnMainParamAdjusted(PhysicsParamPair pair)
        {
        }

        internal virtual void OnSoftParamAdjusted(PhysicsParamPair pair)
        {
        }

        internal virtual void UpdateDiagnostics(bool atInterval)
        {
        }

        internal virtual string GetInfo() => "";

        // TODO execute on UI opened, prevent if UI not open
        internal virtual void Refresh()
        {
        }

        internal void ClosePopups()
        {
            if(activeNestedWindow != null)
            {
                activeNestedWindow.ClosePopups();
            }
            else
            {
                ClosePopupsSelf();
            }
        }

        internal virtual void Clear()
        {
            if(activeNestedWindow != null)
            {
                activeNestedWindow.Clear();
                // activeNestedWindow = null; // Do not set null - it gets rebuilt on return
            }
            else
            {
                ClearSelf();
            }

            OnClose();
        }

        void ClearSelf()
        {
            foreach(var pair in _elements)
            {
                pair.Value.Hide();
            }

            if(showInfoPanel)
            {
                script.uiHandler.infoPanelSection.Hide();
            }
        }

        void ClosePopupsSelf()
        {
            foreach(var pair in _elements)
            {
                pair.Value.ClosePopup();
            }
        }

        internal IEnumerable<KeyValuePair<string, Element>> GetVisibleElements()
        {
            foreach(var pair in _elements)
            {
                if(pair.Value.IsVisible)
                {
                    yield return pair;
                }
            }
        }
    }
}
