using NAT.Common.Diagnostics;
using NAT.Common.Handlers;
using System.Linq;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.Common.UI.Windows
{
    sealed class ClothingWindow : Window
    {
        readonly ClothingHandler _clothingHandler;
        readonly SoftPhysicsHandler _softPhysicsHandler;
        readonly SoftJointDrawService _softJointDrawService;
        UIDynamicTextField _currentPresetTextField;
        UIDynamicTextField _transitionInfoTextField;
        readonly StringBuilder _currentPresetSb = new StringBuilder();
        readonly StringBuilder _transitionInfoSb = new StringBuilder();

        internal ClothingWindow(ScriptBase script) : base(script)
        {
            _clothingHandler = script.clothingHandler;
            _softPhysicsHandler = script.softPhysicsHandler;
            _softJointDrawService = _softPhysicsHandler.softJointDrawService;
            AddNestedWindow(new AdditionalClothingSettingsWindow(script));
            AddSection(FORCES_HOLD_AND_DAMPENING, BuildForcesHoldAndDampeningSection);
            AddSection(SOFT_PHYSICS, BuildSoftPhysicsSection);
            AddSection(MORPH_ADJUSTMENTS, BuildMorphAdjustmentsSection);
            showInfoPanel = true;
        }

        const string FORCES_HOLD_AND_DAMPENING = "Forces, Hold & Dampening";
        const string SOFT_PHYSICS = "Soft Physics";
        const string MORPH_ADJUSTMENTS = "Morph Adjustments";
        const string LEFT_PROFILE = "Left Profile";
        const string RIGHT_PROFILE = "Right Profile";

        protected override void Build()
        {
            BuildClothingPopupRow();
            BuildPresetArea();
            BuildSettingsArea();
            ShowSection(FORCES_HOLD_AND_DAMPENING);
        }

        void BuildClothingPopupRow()
        {
            AddPopup(_clothingHandler.ClothingItemJssc, _ =>
            {
                var uiDynamic = CreatePopup(_clothingHandler.ClothingItemJssc.label, new Vector2(0, -190), new Vector2(-310, 110));
                script.uiHandler.ConfigurePopup(uiDynamic, 510, Utils.IsVR() ? 0 : Mathf.Lerp(0, 12, (float) Screen.width / Screen.height - 1));
                return uiDynamic;
            });
            AddToggle(_clothingHandler.ItemEnabledJsb, name => CreateToggle(name, new Vector2(763, -136), -890));
            AddElement("Customize", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(763, -185), new Vector2(-890, 50));
                uiDynamic.AddListener(() =>
                {
                    var selectedItem = _clothingHandler.SelectedItem;
                    if(selectedItem == null || !selectedItem.enabledJsb.val)
                    {
                        return;
                    }

                    selectedItem.dazItem.OpenUI();
                });
                return uiDynamic;
            });
            AddElement("Reload Items", _ =>
            {
                var uiDynamic = CreateButton("Reload\nItems", new Vector2(960, -185), new Vector2(-970, 100));
                uiDynamic.buttonText.fontSize = 26;
                uiDynamic.AddListener(() => script.bootstrap.ClothingStore.ReloadItems());
                return uiDynamic;
            });
        }

        void BuildPresetArea()
        {
            AddElement("Profile Preset Header", _ =>
            {
                var uiDynamic = CreateTextField("Profile\nPreset", new Vector2(0, -275), new Vector2(-930, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });
            AddElement("Save", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(160, -253), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.SaveProfile());
                return uiDynamic;
            });
            AddElement("Load", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(288, -253), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.LoadProfile());
                return uiDynamic;
            });
            AddElement("Reset", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(415, -253), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.ResetProfile());
                return uiDynamic;
            });
            AddElement("Set As Item Default", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(551, -253), new Vector2(-820, 50));
                uiDynamic.AddListener(() => _clothingHandler.SaveDefault());
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddElement("Load Item Default", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(810, -253), new Vector2(-820, 50));
                uiDynamic.AddListener(() => _clothingHandler.LoadDefault());
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddElement("Current Preset Info", _ =>
            {
                var uiDynamic = CreateTextField("Current: None", new Vector2(160, -289), new Vector2(-170, 32));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.MiddleLeft;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _currentPresetTextField = uiDynamic;
                return uiDynamic;
            });
        }

        void BuildSettingsArea()
        {
            AddElement("Profile Settings Header", _ =>
            {
                var uiDynamic = CreateTextField("Profile\nSettings", new Vector2(0, -386), new Vector2(-930, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });

            AddToggle(_clothingHandler.SymmetricJsb, name =>
            {
                var uiDynamic = CreateToggle(name, new Vector2(160, -373), -856);
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);
                return uiDynamic;
            });
            AddElement(LEFT_PROFILE, _ =>
            {
                var uiDynamic = CreateButton("Left", new Vector2(391, -347), new Vector2(-931, 50));
                uiDynamic.AddListener(() => _clothingHandler.SelectLeft());
                uiDynamic.SetActiveStyle(!(_clothingHandler.SelectedProfile?.symmetric.Value ?? true), true);
                return uiDynamic;
            });
            AddElement(RIGHT_PROFILE, _ =>
            {
                var uiDynamic = CreateButton("Right", new Vector2(539, -347), new Vector2(-931, 50));
                uiDynamic.AddListener(() => _clothingHandler.SelectRight());
                uiDynamic.SetActiveStyle(!(_clothingHandler.SelectedProfile?.symmetric.Value ?? true), true);
                return uiDynamic;
            });
            AddToggle(_clothingHandler.LeftActiveJsb, _ =>
            {
                var uiDynamic = CreateToggle("Active", new Vector2(391, -395), -931);
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);
                return uiDynamic;
            });
            AddToggle(_clothingHandler.RightActiveJsb, _ =>
            {
                var uiDynamic = CreateToggle("Active", new Vector2(539, -395), -931);
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);
                return uiDynamic;
            });

            AddElement("Additional Settings Header", _ =>
            {
                var uiDynamic = CreateTextField("Additional\nSettings", new Vector2(750, -356), new Vector2(-906, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });
            AddElement("Show Additional Settings", _ =>
            {
                var uiDynamic = CreateButton("Show...", new Vector2(920, -343), new Vector2(-931, 50));
                uiDynamic.AddListener(() => OpenNestedWindow(nameof(AdditionalClothingSettingsWindow)));
                return uiDynamic;
            });
            AddElement("Transition Info", _ =>
            {
                var uiDynamic = CreateTextField("", new Vector2(750, -395), new Vector2(-760, 32));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                _transitionInfoTextField = uiDynamic;
                return uiDynamic;
            });

            AddSectionNavigationButton(FORCES_HOLD_AND_DAMPENING, new Vector2(60, -468), -710);
            AddSectionNavigationButton(SOFT_PHYSICS, new Vector2(430, -468), -870, _softPhysicsHandler.softPhysicsOn);
            AddSectionNavigationButton(MORPH_ADJUSTMENTS, new Vector2(640, -468), -710);
        }

        void BuildForcesHoldAndDampeningSection(Section section)
        {
            AddElement($"{FORCES_HOLD_AND_DAMPENING} Header", _ =>
            {
                var uiDynamic = CreateTextField("Forces", new Vector2(12, -530), new Vector2(-558, 40));
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            }, section);

            AddElement($"{FORCES_HOLD_AND_DAMPENING} Info", _ =>
            {
                var sb = new StringBuilder();
                if(script.IsBootyMagic)
                {
                    sb.Append("Forces simulate the way clothing pushes glutes.");
                }
                else if(script.IsTittyMagic)
                {
                    sb.Append("Forces simulate the way clothing pushes breasts.");
                }

                string text =
                    sb.Append(" Forces automatically change the physical behavior and apply directional morphing.")
                        .AppendLine(" Gravity in the same direction can cancel out the force.")
                        .AppendLine()
                        .AppendLine("Hold limits the amount of movement by increasing joint springs. It also reduces hard collider mass.")
                        .AppendLine()
                        .Append("Dampening prevents jiggling by increasing joint damper parameters.")
                        .ToString();

                var uiDynamic = CreateTextField(text, new Vector2(547, -900), new Vector2(-558, 371));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            }, section);

            const float startYPos = -650;
            const float spacing = 127;

            AddSlider(_clothingHandler.YForceJsf, _ => CreateSlider("Y Force (Up/Down)", "F2", new Vector2(10, startYPos)), section);
            AddSlider(_clothingHandler.XForceJsf, _ => CreateSlider("X Force (Sideways)", "F2", new Vector2(10, startYPos - spacing)), section);
            AddSlider(_clothingHandler.ZForceJsf, _ => CreateSlider("Z Force (In/Out)", "F2", new Vector2(10, startYPos - 2 * spacing)), section);

            AddElement("Hold Header", _ =>
            {
                var uiDynamic = CreateTextField("Hold", new Vector2(12, -955), new Vector2(-558, 40));
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            }, section);

            AddSlider(_clothingHandler.YXHoldPctJsf, name => CreateSlider(name, "F1", new Vector2(10, startYPos - 45 - 3 * spacing)), section);
            AddSlider(_clothingHandler.ZHoldPctJsf, name => CreateSlider(name, "F1", new Vector2(10, startYPos - 45 - 4 * spacing)), section);

            AddElement("Dampening Header", _ =>
            {
                var uiDynamic = CreateTextField("Dampening", new Vector2(547, -955), new Vector2(-558, 40));
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            }, section);

            AddSlider(_clothingHandler.YXDampeningPctJsf, name => CreateSlider(name, "F1", new Vector2(545, startYPos - 45 - 3 * spacing)), section);
            AddSlider(_clothingHandler.ZDampeningPctJsf, name => CreateSlider(name, "F1", new Vector2(545, startYPos - 45 - 4 * spacing)), section);
        }

        void BuildSoftPhysicsSection(Section section)
        {
            AddElement("Soft Physics Info", _ =>
            {
                var sb = new StringBuilder();
                if(script.person.IsFemale)
                {
                    sb.Append("Find the area covered by the clothing item. To ensure full coverage, adjust the radius or");
                    sb.Append(" hover over a joint in desktop mode and press CTRL to toggle it manually.");
                }
                else
                {
                    sb.Append("Select covered joints with the quick buttons, and by hovering over a joint in desktop mode");
                    sb.Append(" and pressing CTRL to toggle. (Auto-detection is not supported on male geometry.)");
                }

                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(10, -855), new Vector2(-10, 371));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            }, section);

            if(script.person.IsFemale)
            {
                AddSlider(_clothingHandler.SoftJointSearchRadiusJsf, name => CreateSlider(name, "F4", new Vector2(10, -670)), section);
                AddElement("Find Covered Soft Joints", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(12, -730), new Vector2(-557, 50));
                    uiDynamic.AddListener(_clothingHandler.FindCoveredSoftJoints);
                    return uiDynamic;
                }, section);
            }
            else
            {
                const float sizeY = 55;
                var size = new Vector2(-817, sizeY);
                const float startYPos = -620;
                AddElement("Select All (Left)", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(12, startYPos), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsCovered(Side.LEFT));
                    return uiDynamic;
                }, section);
                AddElement("Select All (Right)", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(274, startYPos), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsCovered(Side.RIGHT));
                    return uiDynamic;
                }, section);
                AddElement("Select None (Left)", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(12, startYPos - sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsExposed(Side.LEFT));
                    return uiDynamic;
                }, section);
                AddElement("Select None (Right)", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(274, startYPos - sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.SetAllJointsExposed(Side.RIGHT));
                    return uiDynamic;
                }, section);
                AddElement("Mirror Right to Left", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(12, startYPos - 2 * sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.MirrorCoverageJoints(Side.RIGHT, Side.LEFT));
                    return uiDynamic;
                }, section);
                AddElement("Mirror Left to Right", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(274, startYPos - 2 * sizeY), size);
                    uiDynamic.buttonText.fontSize = 26;
                    uiDynamic.AddListener(() => _clothingHandler.MirrorCoverageJoints(Side.LEFT, Side.RIGHT));
                    return uiDynamic;
                }, section);
            }

            AddPopup(_clothingHandler.VisualizeClothingCoverageJssc, _ =>
            {
                var uiDynamic = CreateScrollablePopup(_clothingHandler.VisualizeClothingCoverageJssc.label, new Vector2(545, -670), new Vector2(-552, 118));
                script.uiHandler.ConfigurePopup(uiDynamic, 400);
                script.uiHandler.AddPrevNextButtons(uiDynamic);
                uiDynamic.SetActiveStyle(_clothingHandler.SelectedProfile?.isCoverageMapped ?? false, true);
                return uiDynamic;
            }, section);
            AddToggle(_clothingHandler.AutoSelectVisualizedParamJsb, name => CreateToggle(name, new Vector2(545, -730), -557), section);

            AddElement("Soft Physics Parameters Info", _ =>
            {
                string text = new StringBuilder()
                    .Append("Shift a parameter from its current value towards its maximum value (if positive %) or minimum value (if negative %).")
                    .Append(" Increase Fat Spring and Fat Damper to make the clothed area more rigid. The other parameters are provided for experimentation.")
                    .ToString();
                var uiDynamic = CreateTextField(text, new Vector2(10, -1120), new Vector2(-10, 371));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            }, section);

            const float sliderSpacing = 127;
            var list = _clothingHandler.SoftPhysicsJsfs.Values.ToList();
            for(int i = 0; i < list.Count; i++)
            {
                var parameterJsf = list[i];
                float posX = i < 3 ? 10 : 545;
                float currentSpacing = i < 3 ? i * sliderSpacing : (i - 3) * sliderSpacing;
                AddSlider(parameterJsf, name => CreateSlider(name, "F1", new Vector2(posX, -970 - currentSpacing)), section);
            }
        }

        void BuildMorphAdjustmentsSection(Section section)
        {
            const float startYPos = -730;
            const float spacing = 127;

            AddElement("Morphing Offsets Header", _ =>
            {
                var uiDynamic = CreateTextField("Morphing Offsets", new Vector2(272, -530), new Vector2(-558, 40));
                uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                return uiDynamic;
            }, section);

            AddElement("Morphing Offsets Info", _ =>
            {
                const string text = "Apply directional morphs independently of forces.";
                var uiDynamic = CreateTextField(text, new Vector2(272, -610), new Vector2(-558, 75));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            }, section);

            AddSlider(_clothingHandler.YMorphsOffsetJsf, _ =>
            {
                string label = new StringBuilder()
                    .AppendFormat("{0} (°)", _clothingHandler.YMorphsOffsetJsf.name)
                    .ToString();
                return CreateSlider(label, "F2", new Vector2(270, startYPos));
            }, section);

            AddSlider(_clothingHandler.XMorphsOffsetJsf, _ =>
            {
                string label = new StringBuilder()
                    .AppendFormat("{0} (°)", _clothingHandler.XMorphsOffsetJsf.name)
                    .ToString();
                return CreateSlider(label, "F2", new Vector2(270, startYPos - spacing));
            }, section);

            AddSlider(_clothingHandler.ZMorphsOffsetJsf, _ =>
            {
                string label = new StringBuilder()
                    .AppendFormat("{0} (mm)", _clothingHandler.ZMorphsOffsetJsf.name)
                    .ToString();
                return CreateSlider(label, "F1", new Vector2(270, startYPos - 2 * spacing));
            }, section);

            if(script.IsTittyMagic)
            {
                const float startYPos2 = -1050;
                AddElement("Extra Morphs Header", _ =>
                {
                    var uiDynamic = CreateTextField("Extra Morphs", new Vector2(272, startYPos2), new Vector2(-558, 40));
                    uiDynamic.UItext.alignment = TextAnchor.UpperCenter;
                    return uiDynamic;
                }, section);

                var configs = _clothingHandler.morphConfigs.Values.ToArray();
                for(int i = 0; i < configs.Length; i++)
                {
                    var config = configs[i];
                    float currentSpacing = (i + 1) * spacing;
                    AddSlider(
                        config.storableFloat,
                        name => CreateSlider(name, "F2", new Vector2(270, startYPos2 - currentSpacing)),
                        section
                    );
                }
            }
        }

        protected override void PostShow()
        {
            _clothingHandler.AutoSelectItem();
            _softJointDrawService.enabled = script.enabled;
            Refresh();
        }

        protected override void PostShowSection(Section section) =>
            section.SetActiveStyle(_clothingHandler.SelectedProfile != null, true);

        protected override void OnClose() =>
            _softJointDrawService.enabled = false;

        internal override void Refresh()
        {
            if(!isBuilt)
            {
                return;
            }

            if(activeNestedWindow != null)
            {
                activeNestedWindow.Refresh();
                return;
            }

            var selectedProfile = _clothingHandler.SelectedProfile;
            bool profileExists = selectedProfile != null;
            bool isSymmetric = !profileExists || selectedProfile.symmetric.Value;
            bool isCoverageMapped = profileExists && selectedProfile.isCoverageMapped;

            var leftButton = GetElement(LEFT_PROFILE);
            var rightButton = GetElement(RIGHT_PROFILE);
            var symmetricToggle = GetElement(_clothingHandler.SymmetricJsb.name);
            var leftActiveToggle = GetElement(_clothingHandler.LeftActiveJsb.name);
            var rightActiveToggle = GetElement(_clothingHandler.RightActiveJsb.name);
            var visualizeCoveragePopup = GetElement(_clothingHandler.VisualizeClothingCoverageJssc.name);

            leftButton.SetActiveStyle(!isSymmetric, true);
            rightButton.SetActiveStyle(!isSymmetric, true);
            symmetricToggle.SetActiveStyle(profileExists, true);
            leftActiveToggle.SetActiveStyle(profileExists, true);
            rightActiveToggle.SetActiveStyle(profileExists, true);
            visualizeCoveragePopup?.SetActiveStyle(isCoverageMapped, true);
            sections[SOFT_PHYSICS].navigationButton.SetEnabled(_softPhysicsHandler.softPhysicsOn);

            if(isSymmetric)
            {
                leftButton.SetLabel("Left");
                rightButton.SetLabel("Right");
            }
            else
            {
                if(selectedProfile.isRightSelected)
                {
                    leftButton.SetLabel("Left");
                    rightButton.SetLabel("Right".Bold());
                }
                else
                {
                    leftButton.SetLabel("Left".Bold());
                    rightButton.SetLabel("Right");
                }
            }

            if(!_softPhysicsHandler.softPhysicsOn && sections[SOFT_PHYSICS].IsShown)
            {
                ShowSection(FORCES_HOLD_AND_DAMPENING);
            }
            else
            {
                PostShowSection(currentSection);
            }
        }

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt)
            {
                return;
            }

            if(activeNestedWindow != null)
            {
                activeNestedWindow.UpdateDiagnostics(atInterval);
            }
            else
            {
                if(atInterval)
                {
                    // TODO on callback
                    _currentPresetTextField.text = ClothingUIHelpers.CurrentPreset(_currentPresetSb, _clothingHandler.SelectedProfile);
                    _transitionInfoTextField.text = ClothingUIHelpers.TransitionInfo(_transitionInfoSb, _clothingHandler.SelectedProfile);
                }
            }

            _softJointDrawService.ClearLines();
            UpdateSoftPhysicsVisualization();
        }

        void UpdateSoftPhysicsVisualization()
        {
            if(!sections[SOFT_PHYSICS].IsShown || _softJointDrawService == null)
            {
                return;
            }

            if(_clothingHandler.VisualizeClothingCoverageJssc.val == Constant.COVERAGE)
            {
                var profile = _clothingHandler.SelectedProfile;
                if(_clothingHandler.SelectedItem.enabledJsb.val &&
                    profile != null &&
                    profile.IsActive &&
                    profile.isCoverageMapped
                )
                {
                    _softJointDrawService.DrawClothingCoverage(profile);
                }
                else
                {
                    _softJointDrawService.isDrawingClothingCoverage = false;
                }
            }
            else
            {
                _softJointDrawService.DrawClothingAdjustments(_clothingHandler.VisualizedParamPair);
                _softJointDrawService.isDrawingClothingCoverage = false;
            }
        }

        internal override string GetInfo() => isBuilt
            ? ClothingUIHelpers.BuildAdjustmentsInfo(infoSb, _clothingHandler, _softPhysicsHandler, false)
            : "";
    }
}
