// ReSharper disable UnusedMember.Local
using NAT.Common.Handlers;
using NAT.Common.Models;
using UnityEngine;

namespace NAT.Common.UI.Windows
{
    sealed class DevWindow : Window
    {
        readonly AngleCalculator _angleCalculator;
        readonly JointPhysicsHandler _jointPhysicsHandler;
        readonly SoftPhysicsHandler _softPhysicsHandler;
        readonly GravityPhysicsHandler _gravityPhysicsHandler;
        readonly ForceMorphHandler _forceMorphHandler;
        readonly ClothingHandler _clothingHandler;

        UIDynamicTextField _leftDebugField;

        readonly BodyPart _leftPart, _rightPart;

        internal DevWindow(ScriptBase script) : base(script, nameof(DevWindow))
        {
            _angleCalculator = script.angleCalculator;
            _jointPhysicsHandler = script.jointPhysicsHandler;
            _softPhysicsHandler = script.softPhysicsHandler;
            _gravityPhysicsHandler = script.gravityPhysicsHandler;
            _forceMorphHandler = script.forceMorphHandler;
            _clothingHandler = script.clothingHandler;
            if(script.IsBootyMagic)
            {
                _leftPart = script.person.LeftGlute;
                _rightPart = script.person.RightGlute;
            }
            else if(script.IsTittyMagic)
            {
                _leftPart = script.person.LeftBreast;
                _rightPart = script.person.RightBreast;
            }
        }

        protected override void Build()
        {
            AddElement("leftDebug", _ =>
            {
                var uiDynamic = CreateTextField("", new Vector2(10, -1150), new Vector2(-555, 1000));
                uiDynamic.UItext.fontSize = 28;
                _leftDebugField = uiDynamic;
                return uiDynamic;
            });
        }

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt || !atInterval)
            {
                return;
            }

            // TODO optimize
            DebugMorphAngleLimits();
            // DebugGravityAndAngles();
            // DebugClothingMorphs();
        }

        void DebugMorphAngleLimits()
        {
            infoSb
                .Clear()
                .AppendFormat("NRealM: {0:0.000}\n", _jointPhysicsHandler.normalizedRealMass)
                .AppendFormat("UpLimit: {0:0.000}\n", _forceMorphHandler.upLimit)
                .AppendFormat("DepthOutLimit: {0:0.000}\n", _forceMorphHandler.depthOutLimit)
                .AppendFormat("SidewaysOutLimit: {0:0.000}\n", _forceMorphHandler.sidewaysOutLimit)
                .AppendFormat("SidewaysInLimit: {0:0.000}\n", _forceMorphHandler.sidewaysInLimit)
                .AppendLine()
                .AppendFormat("UpNormalizer: {0:0.000}\n", _forceMorphHandler.upDownNormalizer)
                .AppendFormat("DepthInNormalizer: {0:0.000}\n", _forceMorphHandler.depthInNormalizer)
                .AppendFormat("DepthOutNormalizer: {0:0.000}\n", _forceMorphHandler.depthOutNormalizer)
                .AppendFormat("SidewaysNormalizer: {0:0.000}\n", _forceMorphHandler.sidewaysNormalizer)
                .AppendLine()
                .AppendFormat("AngleY L: {0:0.000}\n", script.leftCalibrationTarget.angleY)
                .AppendFormat("AngleY R: {0:0.000}\n", script.rightCalibrationTarget.angleY)
                .AppendFormat("AngleX L: {0:0.000}\n", script.leftCalibrationTarget.angleX)
                .AppendFormat("AngleX R: {0:0.000}\n", script.rightCalibrationTarget.angleX)
                .AppendFormat("DepthDiff L: {0:0.000}\n", script.leftCalibrationTarget.depthDiff)
                .AppendFormat("DepthDiff R: {0:0.000}\n", script.rightCalibrationTarget.depthDiff);
            _leftDebugField.text = infoSb.ToString();
        }

        void DebugGravityAndAngles()
        {
            var leftT = _leftPart.rigidbody.transform;
            var rightT = _rightPart.rigidbody.transform;

            var gravity = _gravityPhysicsHandler.CustomGravityLeft;
            float relGLeftRight = gravity.RelativeTo(leftT, Vector3.right);
            float relGLeftUp = gravity.RelativeTo(leftT, Vector3.up);
            float relGLeftForward = gravity.RelativeTo(leftT, Vector3.forward);
            float relGRightRight = gravity.RelativeTo(rightT, Vector3.right);
            float relGRightUp = gravity.RelativeTo(rightT, Vector3.up);
            float relGRightForward = gravity.RelativeTo(rightT, Vector3.forward);

            infoSb
                .Clear()
                .AppendFormat("LeftPitch: {0:0.00}\n", _angleCalculator.LeftPitch)
                .AppendFormat("LeftRoll: {0:0.00}\n", _angleCalculator.LeftRoll)
                .AppendFormat("RightPitch: {0:0.00}\n", _angleCalculator.RightPitch)
                .AppendFormat("RightRoll: {0:0.00}\n", _angleCalculator.RightRoll)
                .AppendLine()
                .AppendFormat("leanback fix L: {0:0.00}\n", _forceMorphHandler.leftLeanBackFixer)
                .AppendFormat("leanback fix R: {0:0.00}\n", _forceMorphHandler.rightLeanBackFixer)
                // .AppendFormat("upsidedown fix L {0:0.00}\n", _forceMorphHandler.leftUpsideDownFixer)
                // .AppendFormat("upsidedown fix R {0:0.00}\n", _forceMorphHandler.rightUpsideDownFixer)
                .AppendFormat("roll fix L {0:0.00}\n", _forceMorphHandler.leftRollFixer)
                .AppendFormat("roll fix R {0:0.00}\n", _forceMorphHandler.rightRollFixer)
                .AppendLine()
                .AppendItalic("Breast joint gravity (m/s^2):\n")
                .AppendFormat(" x: {0:0.00}\n", gravity.x)
                .AppendFormat(" y: {0:0.00}\n", gravity.y)
                .AppendFormat(" z: {0:0.00}\n", gravity.z)
                .AppendLine()
                .AppendItalic("Rel G Left (m/s^2):\n")
                .AppendFormat(" x: {0:0.00}\n", relGLeftRight)
                .AppendFormat(" y: {0:0.00}\n", relGLeftUp)
                .AppendFormat(" z: {0:0.00}\n", relGLeftForward);
            _leftDebugField.text = infoSb.ToString();

            var refGLeftRight = _gravityPhysicsHandler.RefRightGravityLeft;
            var refGLeftUp = _gravityPhysicsHandler.RefUpGravityLeft;
            var refGLeftDown = _gravityPhysicsHandler.RefDownGravityLeft;
            var refGLeftForward = _gravityPhysicsHandler.RefForwardGravityLeft;
            var refGLeftBack = _gravityPhysicsHandler.RefBackGravityLeft;
            infoSb
                .Clear()
                .AppendItalic("Ref G Left Right (m/s^2):\n")
                .AppendFormat(" x: {0:0.00}\n", refGLeftRight.x)
                .AppendFormat(" y: {0:0.00}\n", refGLeftRight.y)
                .AppendFormat(" z: {0:0.00}\n", refGLeftRight.z)
                .AppendLine()
                .AppendItalic("Ref G Left Up (m/s^2):\n")
                .AppendFormat(" x: {0:0.00}\n", refGLeftUp.x)
                .AppendFormat(" y: {0:0.00}\n", refGLeftUp.y)
                .AppendFormat(" z: {0:0.00}\n", refGLeftUp.z)
                .AppendLine()
                .AppendItalic("Ref G Left Down (m/s^2):\n")
                .AppendFormat(" x: {0:0.00}\n", refGLeftDown.x)
                .AppendFormat(" y: {0:0.00}\n", refGLeftDown.y)
                .AppendFormat(" z: {0:0.00}\n", refGLeftDown.z)
                .AppendLine()
                .AppendItalic("Ref G Left Forward (m/s^2):\\n")
                .AppendFormat(" x: {0:0.00}\n", refGLeftForward.x)
                .AppendFormat(" y: {0:0.00}\n", refGLeftForward.y)
                .AppendFormat(" z: {0:0.00}\n", refGLeftForward.z)
                .AppendLine()
                .AppendItalic("Ref G Left Back (m/s^2):\n")
                .AppendFormat(" x: {0:0.00}\n", refGLeftBack.x)
                .AppendFormat(" y: {0:0.00}\n", refGLeftBack.y)
                .AppendFormat(" z: {0:0.00}\n", refGLeftBack.z);
            _leftDebugField.text = infoSb.ToString();
        }

        void DebugClothingMorphs()
        {
            infoSb
                .Clear()
                .AppendLine(_clothingHandler.LeftMorphsOffsets.ToPrettyString())
                .AppendLine(_clothingHandler.RightMorphsOffsets.ToPrettyString())
                .AppendLine(_clothingHandler.LeftMorphsOffsetsCompensation.ToPrettyString())
                .AppendLine(_clothingHandler.RightMorphsOffsetsCompensation.ToPrettyString())
                .AppendLine();
            foreach(var pair in _clothingHandler.morphConfigs)
            {
                var config = pair.Value;
                infoSb
                    .AppendFormat("{0} Left: {1}\n", pair.Key, config.left.GetValue())
                    .AppendFormat("{0} Right: {1}\n", pair.Key, config.right.GetValue());
            }

            _leftDebugField.text = infoSb.ToString();
        }
    }
}
