using ColliderEditor;
using NAT.Common.Diagnostics;
using NAT.Common.Handlers;
using NAT.Common.Models;
using System.Text;
using UnityEngine;

namespace NAT.Common.UI.Windows
{
    sealed class HardCollidersWindow : Window
    {
        readonly HardColliderHandler _hardColliderHandler;
        readonly ColliderDrawService _drawService;
        readonly ColliderVisualizer _visualizer;
        HardColliderPair _selectedPair;

        internal HardCollidersWindow(ScriptBase script) : base(script)
        {
            _hardColliderHandler = script.hardColliderHandler;
            _drawService = _hardColliderHandler.colliderDrawService;
            _visualizer = _hardColliderHandler.colliderVisualizer;
            if(script.IsDevelopmentEnv)
            {
                AddNestedWindow(new ColliderDevWindow(script));
            }

            showInfoPanel = true;
        }

        protected override void Build()
        {
            BuildMainArea();
            BuildColliderArea();
        }

        void BuildMainArea()
        {
            AddElement("Configure Hard Colliders", name =>
            {
                var uiDynamic = CreateTextField(name.Bold(), new Vector2(10, -165), new Vector2(-555, 60));
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddElement("Info1", _ =>
            {
                var sb = new StringBuilder();
                if(script.IsBootyMagic)
                {
                    sb.Append("Hard colliders apply a force back to the glute joint, making the glute move in response to collision.");
                }
                else if(script.IsTittyMagic)
                {
                    sb.Append("Hard colliders apply a force back to the pectoral joint, making the breast move in response to collision.");
                }

                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(10, -270), new Vector2(-555, 100));
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });

            AddPopup(_hardColliderHandler.colliderGroupsJssc, _ =>
            {
                var uiDynamic = CreateFilterablePopup(_hardColliderHandler.colliderGroupsJssc.label, new Vector2(545, -277), new Vector2(-555, 118));
                float popupPanelHeight = script.IsBootyMagic ? 580 : 400;
                script.uiHandler.ConfigurePopup(uiDynamic, popupPanelHeight);
                script.uiHandler.AddPrevNextButtons(uiDynamic, true);
                uiDynamic.popup.onValueChangeHandlers += OnColliderPopupValueChanged;
                return uiDynamic;
            });

            const float startYPos = -400;
            const float sliderSpacing = 130;

            AddSlider(_hardColliderHandler.baseMassMultiplierJsf, _ =>
            {
                var uiDynamic = CreateSlider($"Base {COLLIDER_MASS_MULTIPLIER.NAME}", COLLIDER_MASS_MULTIPLIER.FORMAT, new Vector2(10, startYPos));
                // if(script.IsTittyMagic)
                // {
                //     uiDynamic.AddListener(UpdateMassMultiplierSliderColor);
                // }

                return uiDynamic;
            });

            AddSlider(
                _hardColliderHandler.baseSpringMultiplierJsf,
                _ => CreateSlider($"Base {COLLIDER_SPRING_MULTIPLIER.NAME}", COLLIDER_SPRING_MULTIPLIER.FORMAT, new Vector2(10, startYPos - sliderSpacing))
            );

            AddSlider(
                _hardColliderHandler.baseDamperMultiplierJsf,
                _ => CreateSlider($"Base {COLLIDER_DAMPER_MULTIPLIER.NAME}", COLLIDER_DAMPER_MULTIPLIER.FORMAT, new Vector2(10, startYPos - 2 * sliderSpacing))
            );

            const float gap = 25f;
            AddSlider(
                _hardColliderHandler.baseRadiusOffsetJsf,
                _ => CreateSlider($"Base {COLLIDER_RADIUS_OFFSET.NAME}", COLLIDER_RADIUS_OFFSET.FORMAT, new Vector2(10, startYPos - gap - 3 * sliderSpacing))
            );

            AddSlider(
                _hardColliderHandler.baseDepthJsf,
                _ => CreateSlider($"Base {COLLIDER_DEPTH_OFFSET.NAME}", COLLIDER_DEPTH_OFFSET.FORMAT, new Vector2(10, startYPos - gap - 4 * sliderSpacing))
            );

            AddPopup(_visualizer.groupsJSON, _ =>
            {
                var uiDynamic = CreateScrollablePopup("Show Colliders", new Vector2(10, -1105), new Vector2(-555, 100));
                script.uiHandler.ConfigurePopup(uiDynamic, 280, 0, true);
                uiDynamic.popup.onValueChangeHandlers += OnShowPreviewsPopupValueChanged;
                return uiDynamic;
            });

            AddToggle(_visualizer.xRayPreviewsJSON, _ => CreateToggle("Use XRay Previews", new Vector2(10, -1165), -555));
            AddToggle(_hardColliderHandler.highlightAllJsb, name => CreateToggle(name, new Vector2(10, -1225), -555));

            AddElement("Configure Collider Friction", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(545, -1100), new Vector2(-555, 50));
                uiDynamic.AddListener(script.ConfigureColliderFriction);
                return uiDynamic;
            });

            AddElement("Friction Info", _ =>
            {
                const string text = "Collider friction controls how easily colliders stick to or slip away from hands or other colliding objects.";
                var uiDynamic = CreateTextField(text, new Vector2(545, -1210), new Vector2(-555, 100));
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });
        }

        void BuildColliderArea()
        {
            const float startYPos = -400;
            const float sliderSpacing = 130;

            AddElement(COLLIDER_MASS_MULTIPLIER.NAME, name =>
            {
                var uiDynamic = CreateSlider(name, COLLIDER_MASS_MULTIPLIER.FORMAT, new Vector2(545, startYPos));
                uiDynamic.slider.minValue = COLLIDER_MASS_MULTIPLIER.MIN;
                uiDynamic.slider.maxValue = COLLIDER_MASS_MULTIPLIER.MAX;
                return uiDynamic;
            });

            AddElement(COLLIDER_SPRING_MULTIPLIER.NAME, name =>
            {
                var uiDynamic = CreateSlider(name, COLLIDER_SPRING_MULTIPLIER.FORMAT, new Vector2(545, startYPos - sliderSpacing));
                uiDynamic.slider.minValue = COLLIDER_SPRING_MULTIPLIER.MIN;
                uiDynamic.slider.maxValue = COLLIDER_SPRING_MULTIPLIER.MAX;
                return uiDynamic;
            });

            AddElement(COLLIDER_DAMPER_MULTIPLIER.NAME, name =>
            {
                var uiDynamic = CreateSlider(name, COLLIDER_DAMPER_MULTIPLIER.FORMAT, new Vector2(545, startYPos - 2 * sliderSpacing));
                uiDynamic.slider.minValue = COLLIDER_DAMPER_MULTIPLIER.MIN;
                uiDynamic.slider.maxValue = COLLIDER_DAMPER_MULTIPLIER.MAX;
                return uiDynamic;
            });

            const float gap = 25f;

            AddElement(COLLIDER_RADIUS_OFFSET.NAME, name =>
            {
                var uiDynamic = CreateSlider(name, COLLIDER_RADIUS_OFFSET.FORMAT, new Vector2(545, startYPos - gap - 3 * sliderSpacing));
                uiDynamic.slider.minValue = COLLIDER_RADIUS_OFFSET.MIN;
                uiDynamic.slider.maxValue = COLLIDER_RADIUS_OFFSET.MAX;
                return uiDynamic;
            });

            AddElement(COLLIDER_DEPTH_OFFSET.NAME, name =>
            {
                var uiDynamic = CreateSlider(name, COLLIDER_DEPTH_OFFSET.FORMAT, new Vector2(545, startYPos - gap - 4 * sliderSpacing));
                uiDynamic.slider.minValue = COLLIDER_DEPTH_OFFSET.MIN;
                uiDynamic.slider.maxValue = COLLIDER_DEPTH_OFFSET.MAX;
                return uiDynamic;
            });
        }

        protected override void PostShow()
        {
            bool enabled = script.enabled;
            _visualizer.enabled = enabled;
            _drawService.enabled = enabled;
            OnColliderPopupValueChanged(_hardColliderHandler.colliderGroupsJssc.val);
            OnShowPreviewsPopupValueChanged(_visualizer.groupsJSON.val);
        }

        protected override void OnClose()
        {
            _visualizer.enabled = false;
            _drawService.enabled = false;
        }

        void OnShowPreviewsPopupValueChanged(string value)
        {
            if(!isBuilt)
            {
                return;
            }

            GetElement(_visualizer.xRayPreviewsJSON.name).SetActiveStyle(value != "Off");
            GetElement(_hardColliderHandler.highlightAllJsb.name).SetActiveStyle(value != "Off");
        }

        void OnColliderPopupValueChanged(string value)
        {
            if(!isBuilt)
            {
                return;
            }

            _selectedPair = null;
            foreach(var pair in _hardColliderHandler.hardColliderPairs)
            {
                if(pair.left.collider.name == value)
                {
                    _selectedPair = pair;
                }

                pair.left.massMultiplierJsf.slider = null;
                pair.left.springMultiplierJsf.slider = null;
                pair.left.damperMultiplierJsf.slider = null;
                pair.left.radiusOffsetJsf.slider = null;
                pair.left.depthJsf.slider = null;
            }

            var massMultiplierSlider = GetElement(COLLIDER_MASS_MULTIPLIER.NAME).GetUIDynamicAs<UIDynamicSlider>();
            var springMultiplierSlider = GetElement(COLLIDER_SPRING_MULTIPLIER.NAME).GetUIDynamicAs<UIDynamicSlider>();
            var damperMultiplierSlider = GetElement(COLLIDER_DAMPER_MULTIPLIER.NAME).GetUIDynamicAs<UIDynamicSlider>();
            var radiusOffsetSlider = GetElement(COLLIDER_RADIUS_OFFSET.NAME).GetUIDynamicAs<UIDynamicSlider>();
            var depthSlider = GetElement(COLLIDER_DEPTH_OFFSET.NAME).GetUIDynamicAs<UIDynamicSlider>();

            if(_selectedPair != null)
            {
                _selectedPair.left.massMultiplierJsf.slider = massMultiplierSlider.slider;
                _selectedPair.left.springMultiplierJsf.slider = springMultiplierSlider.slider;
                _selectedPair.left.damperMultiplierJsf.slider = damperMultiplierSlider.slider;
                _selectedPair.left.radiusOffsetJsf.slider = radiusOffsetSlider.slider;
                _selectedPair.left.depthJsf.slider = depthSlider.slider;

                massMultiplierSlider.label = $"{_selectedPair.id} {COLLIDER_MASS_MULTIPLIER.NAME}";
                springMultiplierSlider.label = $"{_selectedPair.id} {COLLIDER_SPRING_MULTIPLIER.NAME}";
                damperMultiplierSlider.label = $"{_selectedPair.id} {COLLIDER_DAMPER_MULTIPLIER.NAME}";
                radiusOffsetSlider.label = $"{_selectedPair.id} {COLLIDER_RADIUS_OFFSET.NAME}";
                depthSlider.label = $"{_selectedPair.id} {COLLIDER_DEPTH_OFFSET.NAME}";

                massMultiplierSlider.SetActiveStyle(true);
                springMultiplierSlider.SetActiveStyle(true);
                damperMultiplierSlider.SetActiveStyle(true);
                radiusOffsetSlider.SetActiveStyle(true);
                depthSlider.SetActiveStyle(true);
            }
            else
            {
                massMultiplierSlider.label = COLLIDER_MASS_MULTIPLIER.NAME;
                springMultiplierSlider.label = COLLIDER_SPRING_MULTIPLIER.NAME;
                damperMultiplierSlider.label = COLLIDER_DAMPER_MULTIPLIER.NAME;
                radiusOffsetSlider.label = COLLIDER_RADIUS_OFFSET.NAME;
                depthSlider.label = COLLIDER_DEPTH_OFFSET.NAME;

                massMultiplierSlider.slider.value = 0;
                springMultiplierSlider.slider.value = 0;
                damperMultiplierSlider.slider.value = 0;
                radiusOffsetSlider.slider.value = 0;
                radiusOffsetSlider.slider.minValue = COLLIDER_RADIUS_OFFSET.MIN;
                radiusOffsetSlider.slider.maxValue = COLLIDER_RADIUS_OFFSET.MAX;
                depthSlider.slider.value = 0;
                depthSlider.slider.minValue = COLLIDER_DEPTH_OFFSET.MIN;
                depthSlider.slider.maxValue = COLLIDER_DEPTH_OFFSET.MAX;

                massMultiplierSlider.SetActiveStyle(false, true);
                springMultiplierSlider.SetActiveStyle(false, true);
                damperMultiplierSlider.SetActiveStyle(false, true);
                radiusOffsetSlider.SetActiveStyle(false, true);
                depthSlider.SetActiveStyle(false, true);
            }
        }

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt)
            {
                return;
            }

            _drawService.ClearLines();
            string groupFilter = _hardColliderHandler.colliderVisualizer.groupsJSON.val;
            bool useXRayPreviews = _hardColliderHandler.colliderVisualizer.xRayPreviewsJSON.val;
            if(_hardColliderHandler.highlightAllJsb.val)
            {
                _drawService.DrawAllColliderPairsDiagnostics(_hardColliderHandler.hardColliderPairs, groupFilter, useXRayPreviews);
            }
            else if(_hardColliderHandler.selectedHardColliderPair != null)
            {
                _drawService.DrawColliderPairDiagnostics(_hardColliderHandler.selectedHardColliderPair, groupFilter, useXRayPreviews);
            }
        }

        readonly StringBuilder _collidersInfoSb = new StringBuilder();

        internal override string GetInfo()
        {
            float avgMass = 0f;
            float avgSpring = 0f;
            float avgDamper = 0f;
            float avgRadius = 0f;
            float avgDepth = 0f;

            _collidersInfoSb.Clear();

            foreach(var pair in _hardColliderHandler.hardColliderPairs)
            {
                var left = pair.left;
                float mass = 1000f * left.GetMass();
                float spring = left.GetSpring();
                float damper = left.GetDamper();
                float radius = left.autoCollider.colliderRadius;
                float depth = left.depth;

                avgMass += mass;
                avgSpring += spring;
                avgDamper += damper;
                avgRadius += radius;
                avgDepth += depth;

                _collidersInfoSb.AppendLine().AppendSize("\n", 8).AppendBold(pair.id);
                if(left.massMultiplierJsf.IsModified)
                {
                    _collidersInfoSb.AppendFormat("\n  Mass: {0:0.00} g", mass);
                }

                if(left.springMultiplierJsf.IsModified)
                {
                    _collidersInfoSb.AppendFormat("\n  Spring: {0:0}", spring);
                }

                if(left.damperMultiplierJsf.IsModified)
                {
                    _collidersInfoSb.AppendFormat("\n  Damper: {0:0.0}", damper);
                }

                if(left.radiusOffsetJsf.IsModified)
                {
                    _collidersInfoSb.AppendFormat("\n  Radius: {0:0.0000}", radius);
                }

                if(left.depthJsf.IsModified)
                {
                    _collidersInfoSb.AppendFormat("\n  Depth: {0:0.0000}", depth);
                }
            }

            int count = _hardColliderHandler.hardColliderPairs.Count;
            avgMass /= count;
            avgSpring /= count;
            avgDamper /= count;
            avgRadius /= count;
            avgDepth /= count;

            infoSb.Clear()
                .AppendSize("\n", 8)
                .AppendBold("Averages:")
                .AppendFormat("\n  Mass: {0:0.00} g", avgMass)
                .AppendFormat("\n  Spring: {0:0}", avgSpring)
                .AppendFormat("\n  Damper: {0:0.0}", avgDamper)
                .AppendFormat("\n  Radius: {0:0.0000}", avgRadius)
                .AppendFormat("\n  Depth: {0:0.0000}", avgDepth);

            infoSb.Append(_collidersInfoSb);

            return infoSb.ToString();
        }
    }
}
