﻿using NAT.Common.Models;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.UI.Windows
{
    sealed class JointAngleWindow : Window
    {
        readonly JointPhysicsParamPair _baseAngle;
        readonly JointPhysicsParamPair _angleTarget;
        readonly JointPhysicsParamPair _angleLimit1;
        readonly JointPhysicsParamPair _angleLimit2;
        readonly List<ParamValueInfo> _paramValueInfos = new List<ParamValueInfo>();

        internal JointAngleWindow(
            ScriptBase script,
            JointPhysicsParamPair baseAngle,
            JointPhysicsParamPair angleTarget,
            JointPhysicsParamPair angleLimit1,
            JointPhysicsParamPair angleLimit2 = null
        ) : base(script, baseAngle.name)
        {
            _baseAngle = baseAngle;
            _angleTarget = angleTarget;
            _angleLimit1 = angleLimit1;
            _angleLimit2 = angleLimit2;
            AddSection(_angleTarget.name, BuildAngleTargetSection);
            AddSection(_angleLimit2 == null ? _angleLimit1.name : UP_DOWN_ANGLE_LIMITS, BuildAngleLimitsSection);
        }

        const string UP_DOWN_ANGLE_LIMITS = "Up/Down Angle Limits";

        protected override void Build()
        {
            AddToggle(script.calibrationStateHandler.EnableCalibrationStateJsb, name => CreateToggle(name, new Vector2(545, -135), -555));
            BuildParamArea(_baseAngle);

            const float buttonYPos = -445;
            AddSectionNavigationButton(_angleTarget.name, new Vector2(60, buttonYPos), -590);
            AddSectionNavigationButton(_angleLimit2 == null ? _angleLimit1.name : UP_DOWN_ANGLE_LIMITS, new Vector2(550, buttonYPos), -590);
            ShowSection(_angleTarget.name);
        }

        void BuildParamArea(JointPhysicsParamPair paramPair, int yOffset = 0, bool buildGravityPhysicsArea = false, Section section = null)
        {
            var param = paramPair.Left;
            var paramValueInfo = new ParamValueInfo(paramPair);
            _paramValueInfos.Add(paramValueInfo);

            AddElement($"{paramPair.name} Info", _ =>
            {
                var uiDynamic = CreateTextField(paramPair.infoText, new Vector2(10, -435 + yOffset), new Vector2(-555, 280));
                uiDynamic.UItext.fontSize = 28;
                return uiDynamic;
            }, section);

            AddSlider(param.offsetJsf, _ => CreateSlider("Offset", paramPair.ValueFormat, new Vector2(545, -265 + yOffset)), section);

            AddElement($"{paramPair.name} Value", _ =>
            {
                var uiDynamic = CreateTextField("", new Vector2(545, -330 + yOffset), new Vector2(-555, 45));
                uiDynamic.UItext.fontSize = 28;
                paramValueInfo.textField = uiDynamic;
                return uiDynamic;
            }, section);

            if(buildGravityPhysicsArea && param.HasAdjustableGravityPhysics())
            {
                AddElement($"{paramPair.name} Gravity Physics", name =>
                {
                    var uiDynamic = CreateTextField("Gravity Physics".Bold(), new Vector2(10, -420 + yOffset), new Vector2(-555, 60));
                    uiDynamic.UItext.fontSize = 30;
                    uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                    return uiDynamic;
                }, section);

                AddElement("Gravity Targets Info", _ =>
                {
                    var uiDynamic = CreateTextField("", new Vector2(10, -470 + yOffset), new Vector2(-555, 50));
                    uiDynamic.UItext.fontSize = 28;
                    if(script.IsBootyMagic)
                    {
                        uiDynamic.text = "Left glute gravity target values:";
                    }
                    else if(script.IsTittyMagic)
                    {
                        uiDynamic.text = "Left breast gravity target values:";
                    }

                    return uiDynamic;
                }, section);

                AddElement($"{paramPair.name} Gravity Targets Values", _ =>
                {
                    var uiDynamic = CreateTextField("", new Vector2(10, -640 + yOffset), new Vector2(-555, 200));
                    uiDynamic.UItext.fontSize = 28;
                    paramValueInfo.gravityTargetsField = uiDynamic;
                    return uiDynamic;
                }, section);

                const float sliderSpacing = 130;
                var keys = param.gravityCalculators.Keys.ToList();

                for(int i = 0; i < keys.Count; i++)
                {
                    string key = keys[i];
                    var config = param.gravityCalculators[key];
                    float posY = -540 + yOffset - i * sliderSpacing;
                    AddSlider(config.multiplierJsf, _ => CreateSlider($"{key} Gravity", "F2", new Vector2(545, posY)), section);
                }
            }
        }

        void BuildAngleTargetSection(Section section) => BuildParamArea(_angleTarget, -325, true, section);

        void BuildAngleLimitsSection(Section section)
        {
            BuildParamArea(_angleLimit1, -325, true, section);
            if(_angleLimit2 != null)
            {
                BuildParamArea(_angleLimit2, -505, true, section);
            }
        }

        protected override void PostShow()
        {
            // _jointPhysicsHandler.jointDrawService.enabled = script.enabled;
        }

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt)
            {
                return;
            }

            if(atInterval)
            {
                foreach(var paramValueInfo in _paramValueInfos)
                {
                    paramValueInfo.Update(infoSb);
                    paramValueInfo.UpdateGravityTargets(infoSb, script.jointPhysicsHandler.normalizedMass, script.softness);
                }
            }
        }

        protected override void OnClose() => script.calibrationStateHandler.EnableCalibrationStateJsb.val = false;
    }
}
