using NAT.Common.Handlers;
using SimpleJSON;
using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.Common.UI.Windows
{
    sealed class MainWindow : Window
    {
        readonly SoftPhysicsHandler _softPhysicsHandler;
        ParamValueInfo _massValueInfo;
        UIDynamicTextField _currentPresetTextField;

        internal MainWindow(ScriptBase script) : base(script)
        {
            _softPhysicsHandler = script.softPhysicsHandler;

            AddNestedWindow(new SettingsWindow(script));
            if(script.IsDevelopmentEnv)
            {
                AddNestedWindow(new DevWindow(script));
                AddNestedWindow(new DevMorphWindow(script));
            }
        }

        protected override void Build()
        {
            BuildMainArea();
            BuildPresetArea();
            BuildLowerArea();
        }

        void BuildMainArea()
        {
            AddSlider(script.softnessJsf, name => CreateSlider(name, "F0", new Vector2(10, -225)));
            AddSlider(script.quicknessJsf, name => CreateSlider(name, "F2", new Vector2(545, -225)));

            AddElement("Info", _ =>
            {
                string text = new StringBuilder()
                    .Append("Softness, quickness and mass affect physics base values, the effect of gravity,")
                    .Append(" the amount of morphing, and hard collider properties. Use them to determine")
                    .Append(" the overall behavior, then fine tune physics, morphs and colliders as needed.")
                    .ToString();
                var uiDynamic = CreateTextField(text, new Vector2(10, -415), new Vector2(-555, 180));
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });

            var jointMassParamPair = script.jointPhysicsHandler.jointMass;
            _massValueInfo = new ParamValueInfo(script.jointPhysicsHandler.jointMass, jointMassParamPair.name, "kg");

            AddSlider(
                jointMassParamPair.Left.offsetJsf,
                name => CreateSlider(name, jointMassParamPair.ValueFormat, new Vector2(545, -355))
            );

            AddElement("Mass Value", _ =>
            {
                var uiDynamic = CreateTextField("", new Vector2(545, -415), new Vector2(-555, 50));
                uiDynamic.UItext.fontSize = 28;
                _massValueInfo.textField = uiDynamic;
                return uiDynamic;
            });

            AddToggle(_softPhysicsHandler.softPhysicsOnJsb, name => CreateToggle(name, new Vector2(10, -490), -555));
            AddToggle(_softPhysicsHandler.ldpEnabledJsb, name => CreateToggle(name, new Vector2(545, -490), -555));

            AddElement("Soft Physics Info", _ =>
            {
                var sb = new StringBuilder();
                if(script.IsBootyMagic)
                {
                    sb.AppendFormat("Simulate soft-body dynamics for glutes. Soft physics is required to get the most out of {0}.", script.DomainName);
                }
                else if(script.IsTittyMagic)
                {
                    sb.AppendFormat("Simulate soft-body dynamics for breasts. Soft physics is required to get the most out of {0}.", script.DomainName);
                }

                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(10, -600), new Vector2(-555, 100));
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });

            AddElement("LDP Info", _ =>
            {
                const string text = "Adjust soft physics parameters of individual soft joints dynamically with stretching and compression.";
                var uiDynamic = CreateTextField(text, new Vector2(545, -600), new Vector2(-555, 100));
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });

            if(script.IsTittyMagic)
            {
                AddToggle(_softPhysicsHandler.softPhysicsSelfCollideJsb, name => CreateToggle(name, new Vector2(10, -660), -555));
            }

            AddElement("Configure Collider Friction", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(545, -660), new Vector2(-555, 50));
                uiDynamic.AddListener(script.ConfigureColliderFriction);
                return uiDynamic;
            });

            AddElement("Friction Info", _ =>
            {
                const string text = "Collider friction controls how easily colliders stick to or slip away from hands or other colliding objects.";
                var uiDynamic = CreateTextField(text, new Vector2(545, -770), new Vector2(-555, 100));
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });
        }

        void BuildPresetArea()
        {
            AddElement("Preset Header", _ =>
            {
                var sb = new StringBuilder().AppendLine(script.DomainName).Append("Preset");
                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(5, -1140), new Vector2(-890, 70));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.UItext.alignment = TextAnchor.MiddleCenter;
                return uiDynamic;
            });
            AddElement("Save", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(200, -1118), new Vector2(-950, 50));
                uiDynamic.AddListener(() => script.presetHandler.savePresetJsu.FileBrowse());
                return uiDynamic;
            });
            AddElement("Load", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(328, -1118), new Vector2(-950, 50));
                uiDynamic.AddListener(() => script.presetHandler.loadPresetJsu.FileBrowse());
                return uiDynamic;
            });
            AddElement("Reset", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(455, -1118), new Vector2(-950, 50));
                uiDynamic.AddListener(script.ResetPluginParams);
                return uiDynamic;
            });
            AddElement("Set As Default", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(591, -1118), new Vector2(-840, 50));
                uiDynamic.AddListener(() => script.presetHandler.SaveDefaultPreset(script.person.IsFemale));
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddElement("Load Default", name =>
            {
                var uiDynamic = CreateButton(name, new Vector2(830, -1118), new Vector2(-840, 50));
                uiDynamic.AddListener(() => script.presetHandler.LoadDefaultPreset(script.person.IsFemale));
                uiDynamic.buttonText.fontSize = 26;
                return uiDynamic;
            });
            AddElement("Current Preset Info", _ =>
            {
                var uiDynamic = CreateTextField("Current: None", new Vector2(200, -1154), new Vector2(-210, 32));
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.MiddleLeft;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _currentPresetTextField = uiDynamic;
                return uiDynamic;
            });
        }

        void BuildLowerArea()
        {
            AddElement("User Preferences", _ =>
            {
                var sb = new StringBuilder().AppendFormat("{0} User Preferences", Constant.NATURALIS);
                var uiDynamic = CreateButton(sb.ToString(), new Vector2(360, -1224), new Vector2(-700, 50));
                uiDynamic.AddListener(() => OpenNestedWindow(nameof(SettingsWindow)));
                return uiDynamic;
            });

            AddElement("Version", _ =>
            {
                var sb = new StringBuilder().AppendFormat("v{0}", Bootstrap.NATURALIS_VERSION);
                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(841, -1230), new Vector2(-850, 42));
                uiDynamic.UItext.alignment = TextAnchor.LowerRight;
                uiDynamic.UItext.fontSize = 26;
                return uiDynamic;
            });
        }

        protected override void PostShow() => Refresh();

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt || !atInterval)
            {
                return;
            }

            _massValueInfo.Update(infoSb);
        }

        internal override void Refresh()
        {
            if(!isBuilt)
            {
                return;
            }

            var softPhysicsOnToggle = GetElement(_softPhysicsHandler.softPhysicsOnJsb.name);
            var ldpEnabledToggle = GetElement(_softPhysicsHandler.ldpEnabledJsb.name);

            var personMonitor = script.bootstrap.PersonMonitor;
            softPhysicsOnToggle.SetActiveStyle(
                script.IsBootyMagic ? personMonitor.IsGluteSoftPhysicsOn() : personMonitor.IsBreastSoftPhysicsOn(),
                false,
                _softPhysicsHandler.softPhysicsOnJsb.val && script.enabled
            );
            ldpEnabledToggle.SetActiveStyle(script.IsSoftPhysicsOn, false, script.enabled);

            _currentPresetTextField.text = new StringBuilder()
                .AppendFormat("Current: {0}", script.LoadedPresetName ?? "None")
                .ToString();
        }
    }
}
