using NAT.Common.Handlers;
using NAT.Common.Models;
using UnityEngine;

namespace NAT.Common.UI.Windows
{
    sealed class PhysicsWindow : Window
    {
        readonly JointPhysicsHandler _jointPhysicsHandler;
        readonly SoftPhysicsHandler _softPhysicsHandler;
        readonly GravityPhysicsHandler _gravityPhysicsHandler;

        internal PhysicsWindow(ScriptBase script) : base(script)
        {
            _jointPhysicsHandler = script.jointPhysicsHandler;
            _softPhysicsHandler = script.softPhysicsHandler;
            _gravityPhysicsHandler = script.gravityPhysicsHandler;

            AddNestedWindow(new WeightWindow(script, _jointPhysicsHandler.jointMass));
            AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.centerOfGravity));
            AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.rotationSpring));
            AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.rotationDamper));
            if(script.IsTittyMagic)
            {
                AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.rightLeftSpring));
                AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.rightLeftDamper));
            }

            AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.inOutSpring));
            AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.inOutDamper));

            AddNestedWindow(new JointAngleWindow(
                script,
                _jointPhysicsHandler.upDownBaseAngle,
                _jointPhysicsHandler.upDownAngleTarget,
                _jointPhysicsHandler.upAngleLimit,
                _jointPhysicsHandler.downAngleLimit
            ));
            AddNestedWindow(new JointAngleWindow(
                script,
                _jointPhysicsHandler.rightLeftBaseAngle,
                _jointPhysicsHandler.rightLeftAngleTarget,
                _jointPhysicsHandler.rightLeftAngleLimit
            ));

            if(script.IsTittyMagic)
            {
                AddNestedWindow(new ParameterWindow(script, _jointPhysicsHandler.twistAngleTarget));
            }

            foreach(var pair in _softPhysicsHandler.softParamPairs)
            {
                AddNestedWindow(new SoftParameterWindow(script, pair));
            }

            AddNestedWindow(new FatBackForceWindow(
                script,
                _softPhysicsHandler.fatBackForce,
                _softPhysicsHandler.fatBackForceMaxForce,
                _softPhysicsHandler.fatBackForceThreshold
            ));
        }

        /* Buttons are active if soft physics is currently enabled and in sync, and settingsMonitor agrees */

        bool IsSoftPhysicsOnAndSynced()
        {
            bool monitorSoftPhysicsOn = script.IsBootyMagic
                ? script.bootstrap.PersonMonitor.IsGluteSoftPhysicsOn()
                : script.bootstrap.PersonMonitor.IsBreastSoftPhysicsOn();
            return _softPhysicsHandler.softPhysicsOn && _softPhysicsHandler.softPhysicsSynced && monitorSoftPhysicsOn;
        }

        protected override void Build()
        {
            BuildJointPhysicsButtonsArea();
            BuildAnglesButtonsArea();
            BuildSoftPhysicsButtonsArea();
        }

        float _currentYPos;
        const float BUTTON_HEIGHT = 79;
        const float SPACING = BUTTON_HEIGHT + 10;

        void BuildJointPhysicsButtonsArea()
        {
            _currentYPos = -154;
            AddElement("Joint Physics", name =>
            {
                var uiDynamic = CreateTextField(name.Bold(), new Vector2(30, _currentYPos), new Vector2(-555, 50));
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddLeftSideButton(_jointPhysicsHandler.jointMass);
            AddLeftSideButton(_jointPhysicsHandler.centerOfGravity);
            AddLeftSideButtonPair(_jointPhysicsHandler.rotationSpring, _jointPhysicsHandler.rotationDamper);

            if(script.IsTittyMagic)
            {
                AddLeftSideButtonPair(_jointPhysicsHandler.rightLeftSpring, _jointPhysicsHandler.rightLeftDamper);
            }

            AddLeftSideButtonPair(_jointPhysicsHandler.inOutSpring, _jointPhysicsHandler.inOutDamper);
        }

        void BuildAnglesButtonsArea()
        {
            _currentYPos -= SPACING;
            if(script.IsBootyMagic)
            {
                _currentYPos -= SPACING;
            }

            AddElement("Joint Angles", name =>
            {
                var uiDynamic = CreateTextField(name.Bold(), new Vector2(30, _currentYPos), new Vector2(-555, 50));
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddLeftSideButton(_jointPhysicsHandler.upDownBaseAngle);
            AddLeftSideButton(_jointPhysicsHandler.rightLeftBaseAngle);

            if(script.IsTittyMagic)
            {
                AddLeftSideButton(_jointPhysicsHandler.twistAngleTarget);
            }
        }

        void BuildSoftPhysicsButtonsArea()
        {
            _currentYPos = -154;
            AddElement("Soft Physics", name =>
            {
                var uiDynamic = CreateTextField(name.Bold(), new Vector2(570, _currentYPos), new Vector2(-555, 50));
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddRightSideButton(_softPhysicsHandler.fatSpring);
            AddRightSideButton(_softPhysicsHandler.fatDamper);
            AddRightSideButton(_softPhysicsHandler.fatMass);
            AddRightSideButton(_softPhysicsHandler.fatDistanceLimit);
            AddRightSideButton(_softPhysicsHandler.fatBackForce);

            _currentYPos -= SPACING;

            AddElement("Fat Colliders", name =>
            {
                var uiDynamic = CreateTextField(name.Bold(), new Vector2(570, _currentYPos), new Vector2(-555, 50));
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddRightSideButton(_softPhysicsHandler.fatColliderRadius);
            AddRightSideButton(_softPhysicsHandler.fatColliderDepth);
            AddRightSideButton(_softPhysicsHandler.fatColliderFriction);
        }

        void AddLeftSideButton(PhysicsParamPair pair)
        {
            _currentYPos -= SPACING;
            AddParamButton(pair, new Vector2(10, _currentYPos), new Vector2(-515, BUTTON_HEIGHT));
        }

        void AddLeftSideButtonPair(PhysicsParamPair pairLeft, PhysicsParamPair pairRight)
        {
            _currentYPos -= SPACING;
            AddParamButton(pairLeft, new Vector2(10, _currentYPos), new Vector2(-797, BUTTON_HEIGHT));
            AddParamButton(pairRight, new Vector2(292, _currentYPos), new Vector2(-797, BUTTON_HEIGHT));
        }

        // ReSharper disable once UnusedMember.Local
        void AddLeftSideButtonTrio(PhysicsParamPair pairLeft, PhysicsParamPair pairMid, PhysicsParamPair pairRight)
        {
            _currentYPos -= SPACING;
            AddParamButton(pairLeft, new Vector2(10, _currentYPos), new Vector2(-891, BUTTON_HEIGHT));
            AddParamButton(pairMid, new Vector2(198, _currentYPos), new Vector2(-892, BUTTON_HEIGHT));
            AddParamButton(pairRight, new Vector2(385, _currentYPos), new Vector2(-892, BUTTON_HEIGHT));
        }

        void AddRightSideButton(PhysicsParamPair pair)
        {
            _currentYPos -= SPACING;
            AddParamButton(pair, new Vector2(599, _currentYPos), new Vector2(-610, BUTTON_HEIGHT));
        }

        void AddParamButton(PhysicsParamPair pair, Vector2 pos, Vector2 size) => AddElement(pair.name, name =>
        {
            var uiDynamic = CreateButton(pair.buttonLabel, pos, size);
            uiDynamic.buttonText.fontSize = 28;
            uiDynamic.AddListener(() => OpenNestedWindow(name));
            return uiDynamic;
        });

        protected override void PostShow()
        {
            UpdateMainParamsArea();
            UpdateSoftParamsArea(IsSoftPhysicsOnAndSynced());
        }

        void UpdateMainParamsArea()
        {
            SyncParamButton(_jointPhysicsHandler.jointMass, _jointPhysicsHandler.jointMass.IsModified || _gravityPhysicsHandler.IsGravityModified);

            foreach(var pair in _jointPhysicsHandler.paramPairs)
            {
                bool isModified = pair == _jointPhysicsHandler.twistAngleTarget
                    ? _gravityPhysicsHandler.twistAngleTargetJsf.val != 0
                    : pair.IsModified;
                if(pair == _jointPhysicsHandler.upDownBaseAngle)
                {
                    SyncParamButton(pair, _jointPhysicsHandler.UpDownAngleIsModified);
                }
                else if(pair == _jointPhysicsHandler.rightLeftBaseAngle)
                {
                    SyncParamButton(pair, _jointPhysicsHandler.RightLeftAngleIsModified);
                }
                else
                {
                    SyncParamButton(pair, isModified);
                }
            }
        }

        void UpdateSoftParamsArea(bool active)
        {
            var softPhysicsHeader = GetElement("Soft Physics").GetUIDynamicAs<UIDynamicTextField>();
            var fatCollidersHeader = GetElement("Fat Colliders").GetUIDynamicAs<UIDynamicTextField>();
            softPhysicsHeader.SetActiveStyle(active);
            fatCollidersHeader.SetActiveStyle(active);

            foreach(var pair in _softPhysicsHandler.softParamPairs)
            {
                SyncParamButton(pair, pair.IsModified, active);
            }

            SyncParamButton(_softPhysicsHandler.fatBackForce, _softPhysicsHandler.BackForceIsModified, active);
        }

        internal override void Refresh() => UpdateSoftParamsArea(_softPhysicsHandler.softPhysicsOn);

        internal override void OnMainParamAdjusted(PhysicsParamPair pair)
        {
            if(!script.uiHandler.IsActive(this))
            {
                return;
            }

            if(pair == _jointPhysicsHandler.jointMass)
            {
                SyncParamButton(pair, pair.IsModified || _gravityPhysicsHandler.IsGravityModified);
            }
            else if(
                pair == _jointPhysicsHandler.upDownBaseAngle ||
                pair == _jointPhysicsHandler.upDownAngleTarget ||
                pair == _jointPhysicsHandler.upAngleLimit ||
                pair == _jointPhysicsHandler.downAngleLimit
            )
            {
                SyncParamButton(_jointPhysicsHandler.upDownBaseAngle, _jointPhysicsHandler.UpDownAngleIsModified);
            }
            else if(
                pair == _jointPhysicsHandler.rightLeftBaseAngle ||
                pair == _jointPhysicsHandler.rightLeftAngleTarget ||
                pair == _jointPhysicsHandler.rightLeftAngleLimit
            )
            {
                SyncParamButton(_jointPhysicsHandler.rightLeftBaseAngle, _jointPhysicsHandler.RightLeftAngleIsModified);
            }
            else if(pair == _jointPhysicsHandler.twistAngleTarget)
            {
                SyncParamButton(pair, _gravityPhysicsHandler.twistAngleTargetJsf.val != 0);
            }
            else if(
                pair == _softPhysicsHandler.fatBackForce ||
                pair == _softPhysicsHandler.fatBackForceThreshold ||
                pair == _softPhysicsHandler.fatBackForceMaxForce
            )
            {
                bool active = IsSoftPhysicsOnAndSynced();
                SyncParamButton(_softPhysicsHandler.fatBackForce, _softPhysicsHandler.BackForceIsModified, active);
            }
            else
            {
                SyncParamButton(pair, pair.IsModified);
            }
        }

        internal override void OnSoftParamAdjusted(PhysicsParamPair pair)
        {
            if(!script.uiHandler.IsActive(this))
            {
                return;
            }

            SyncParamButton(pair, pair.IsModified, IsSoftPhysicsOnAndSynced());
        }

        void SyncParamButton(PhysicsParamPair pair, bool modified, bool active = true)
        {
            var button = GetElement(pair.name)?.GetUIDynamicAs<UIDynamicButton>();
            if(!button)
            {
                return;
            }

            string label = $"  {pair.buttonLabel}";
            if(modified && active)
            {
                label += " *".Bold();
            }

            button.label = label;
            button.SetActiveStyle(active, true);
        }
    }
}
