using ColliderEditor;
using NAT.Common.Diagnostics;
using NAT.Common.Handlers;
using NAT.Common.Models;
using System;
using System.Text;
using UnityEngine;

namespace NAT.Common.UI.Windows
{
    sealed class SoftParameterWindow : Window
    {
        readonly SoftPhysicsHandler _softPhysicsHandler;
        readonly ColliderVisualizer _colliderVisualizer;
        readonly SoftJointDrawService _drawService;
        readonly SoftPhysicsParamPair _paramPair;
        readonly SoftPhysicsParam _param;
        bool IsFrictionParam => _paramPair.name == FAT_COLLIDER_FRICTION.NAME;

        UIDynamicTextField _outerValueField;
        UIDynamicTextField _mainValueField;
        UIDynamicTextField _centerValueField;
        UIDynamicTextField _nippleValueField;

        internal SoftParameterWindow(ScriptBase script, SoftPhysicsParamPair paramPair) : base(script, paramPair.name)
        {
            _softPhysicsHandler = script.softPhysicsHandler;
            _colliderVisualizer = _softPhysicsHandler.colliderVisualizer;
            _drawService = _softPhysicsHandler.softJointDrawService;
            _paramPair = paramPair;
            _param = _paramPair.Left;
            AddSection(VALUE_GRADIENT, BuildValueGradientSection);
            AddSection(LOCAL_DISTORTION_PHYSICS, BuildLocalDistortionPhysicsSection);
        }

        const string VALUE_GRADIENT = "Value Gradient";
        const string LOCAL_DISTORTION_PHYSICS = "Local Distortion Physics";

        protected override void Build()
        {
            BuildMainArea();
            BuildVisualizationArea();
            ShowSection(VALUE_GRADIENT);
        }

        void BuildMainArea()
        {
            AddElement("Parameter Info", _ =>
            {
                float posY = IsFrictionParam ? -320 : -385;
                float sizeY = IsFrictionParam ? 160 : 225;
                var uiDynamic = CreateTextField(_paramPair.infoText, new Vector2(10, posY), new Vector2(-555, sizeY));
                uiDynamic.UItext.fontSize = 28;
                return uiDynamic;
            });

            if(IsFrictionParam)
            {
                AddElement("Configure Collider Friction", name =>
                {
                    var uiDynamic = CreateButton(name, new Vector2(10, -345), new Vector2(-555, 50));
                    uiDynamic.AddListener(script.ConfigureColliderFriction);
                    return uiDynamic;
                });
            }

            AddSlider(_param.offsetJsf, _ =>
            {
                string label = IsFrictionParam ? "Friction Value" : "Offset";
                var uiDynamic = CreateSlider(label, _paramPair.ValueFormat, new Vector2(545, -265));
                uiDynamic.SetActiveStyle(!IsFrictionParam, true);
                return uiDynamic;
            });

            _outerValueField = ValueInfoField(new Vector2(546, -320));
            AddElement("Outer Value", _ => _outerValueField);
            _centerValueField = ValueInfoField(new Vector2(808, -320));
            AddElement("Center Value", _ => _centerValueField);

            if(script.IsTittyMagic)
            {
                _mainValueField = ValueInfoField(new Vector2(546, -365));
                AddElement("Main Value", _ => _mainValueField);
                _nippleValueField = ValueInfoField(new Vector2(808, -365));
                AddElement("Nipple Value", _ => _nippleValueField);
            }

            AddSectionNavigationButton(VALUE_GRADIENT, new Vector2(60, -468), -590);
            AddSectionNavigationButton(LOCAL_DISTORTION_PHYSICS, new Vector2(550, -468), -590);
        }

        void BuildVisualizationArea()
        {
            if(!_paramPair.hasSoftColliderVisualization)
            {
                return;
            }

            AddPopup(_softPhysicsHandler.colliderVisualizer.groupsJSON, _ =>
            {
                var uiDynamic = CreateScrollablePopup(_softPhysicsHandler.colliderVisualizer.groupsJSON.label, new Vector2(10, -1220), new Vector2(-555, 100));
                script.uiHandler.ConfigurePopup(uiDynamic, 280, 0, true);
                return uiDynamic;
            });

            AddSlider(
                _softPhysicsHandler.colliderVisualizer.previewOpacityJSON,
                _ => CreateSlider("Collider Previews Opacity", "F2", new Vector2(545, -1215))
            );
        }

        void BuildValueGradientSection(Section section)
        {
            AddToggle(
                _drawService.visualizeGradientJsb,
                _ => CreateToggle("Visualize Gradient", new Vector2(10, -540), -555),
                section
            );

            var gradientConfig = _param.gradientConfig;
            const float startYPos = -670;
            const float sliderSpacing = 130;

            AddSlider(
                gradientConfig.middleShiftJsf,
                _ => CreateSlider("Middle Shift", "F2", new Vector2(10, startYPos)),
                section
            );

            AddSlider(
                gradientConfig.outerMultiplierJsf,
                _ => CreateSlider("Outer Multiplier", "F2", new Vector2(10, startYPos - sliderSpacing)),
                section
            );

            AddSlider(gradientConfig.centerMultiplierJsf, _ =>
            {
                string label = script.IsBootyMagic ? "Center Multiplier" : "Areola Multiplier";
                return CreateSlider(label, "F2", new Vector2(10, startYPos - 2 * sliderSpacing));
            }, section);

            AddElement("Gradient Info", _ =>
            {
                var sb = new StringBuilder();
                string center = script.IsBootyMagic ? "center" : "areola";
                sb.Append("Middle shift moves the gradient's mid point from the outer joints")
                    .AppendFormat(" at -1.00 to the {0} joints at 1.00.\n", center)
                    .Append("\n");

                if(script.IsBootyMagic)
                {
                    sb.Append("The outer multiplier affects joints around the edge of the glute.\n")
                        .Append("\n")
                        .AppendFormat("The {0} multiplier changes the values for joints at the center of the glute.\n", center)
                        .Append("\n")
                        .AppendFormat("Values between the outer and {0} joints are interpolated smoothly.\n", center);
                }
                else if(script.IsTittyMagic)
                {
                    sb.Append("The outer multiplier affects joints at the base of the breast.\n")
                        .Append("\n")
                        .AppendFormat("The {0} multiplier changes the values for joints at the center of the breast, excluding the nipple.\n", center)
                        .Append("\n")
                        .AppendFormat("Values between the outer and {0} joints are interpolated smoothly.\n", center)
                        .Append("\n")
                        .Append("The nipple multiplier affects only the single nipple joint.");
                }

                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(545, -1072), new Vector2(-555, 520));
                uiDynamic.UItext.fontSize = 28;
                return uiDynamic;
            }, section);

            if(script.IsTittyMagic)
            {
                AddSlider(
                    gradientConfig.nippleMultiplierJsf,
                    _ => CreateSlider("Nipple Multiplier", "F2", new Vector2(10, startYPos - 3 * sliderSpacing)),
                    section
                );
            }
        }

        void BuildLocalDistortionPhysicsSection(Section section)
        {
            AddToggle(
                _drawService.visualizeLdpJsb,
                _ => CreateToggle("Visualize LDP", new Vector2(10, -540), -555),
                section
            );

            var config = _param.distortionConfig;
            const float startYPos = -670;
            const float sliderSpacing = 130;

            AddSlider(
                config.compressionEffectJsf,
                _ => CreateSlider("Compression Effect", "F2", new Vector2(10, startYPos)),
                section
            );

            AddSlider(
                config.compressionSensitivityJsf,
                _ => CreateSlider("Compression Sensitivity", "F2", new Vector2(10, startYPos - sliderSpacing)),
                section
            );

            AddSlider(
                config.stretchingEffectJsf,
                _ => CreateSlider("Stretching Effect", "F2", new Vector2(10, startYPos - 2 * sliderSpacing)),
                section
            );

            AddSlider(
                config.stretchingSensitivityJsf,
                _ => CreateSlider("Stretching Sensitivity", "F2", new Vector2(10, startYPos - 3 * sliderSpacing)),
                section
            );

            AddElement("LDP Info", _ =>
            {
                string text = new StringBuilder()
                    .Append("Compression causes the parameter's value to change as joints move closer to each other.")
                    .AppendLine(" Stretching does the same as the joints move away from each other.")
                    .AppendLine()
                    .Append("The Effect sliders are multipliers on the amount of compression and stretching.")
                    .AppendLine(" A positive value causes the parameter's value to increase, and a negative value causes it to decrease.")
                    .AppendLine()
                    .Append("The Sensitivity sliders control how easily small distortions cause the value to change.")
                    .ToString();
                var uiDynamic = CreateTextField(text, new Vector2(545, -1060), new Vector2(-555, 508));
                uiDynamic.UItext.fontSize = 28;
                return uiDynamic;
            }, section);
        }

        UIDynamicTextField ValueInfoField(Vector2 position)
        {
            var uiDynamic = CreateTextField("", position, new Vector2(-820, 42));
            uiDynamic.UItext.fontSize = 26;
            uiDynamic.backgroundColor = Colors.defaultTextFieldBackgroundColor;
            return uiDynamic;
        }

        protected override void PostShow()
        {
            Refresh();
            _drawService.enabled = script.enabled;
            if(_paramPair.hasSoftColliderVisualization)
            {
                var uiDynamicPopup = GetElement(_softPhysicsHandler.colliderVisualizer.groupsJSON.name).GetUIDynamicAs<UIDynamicPopup>();
                var uiDynamicSlider = GetElement(_softPhysicsHandler.colliderVisualizer.previewOpacityJSON.name).GetUIDynamicAs<UIDynamicSlider>();
                _softPhysicsHandler.colliderVisualizer.groupsJSON.popup = uiDynamicPopup.popup;
                _softPhysicsHandler.colliderVisualizer.previewOpacityJSON.slider = uiDynamicSlider.slider;
            }

            SetShowColliderPreviews();
        }

        protected override void PostShowSection(Section section)
        {
            if(section == sections[LOCAL_DISTORTION_PHYSICS])
            {
                section.SetActiveStyle(script.IsLdpUpdateOn, true);
            }
        }

        internal override void Refresh()
        {
            if(!isBuilt)
            {
                return;
            }

            PostShowSection(currentSection);
        }

        internal void SetShowColliderPreviews() =>
            _softPhysicsHandler.SetShowColliderPreviews(_paramPair.hasSoftColliderVisualization);

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt)
            {
                return;
            }

            if(atInterval)
            {
                UpdateOuterValueText();
                UpdateCenterValueText();
                if(script.IsTittyMagic)
                {
                    UpdateMainValueText();
                    UpdateNippleValueText();
                }
            }

            _drawService.ClearLines();
            _drawService.DrawDiagnostics(_paramPair);
        }

        float _prevOuterValue = float.NaN;
        float _prevCenterValue = float.NaN;
        float _prevMainValue = float.NaN;
        float _prevNippleValue = float.NaN;

        void UpdateOuterValueText()
        {
            float averageRealValue = 0f;
            var softJoints = _softPhysicsHandler.leftContext.outerSoftJoints;
            foreach(var softJoint in softJoints)
            {
                averageRealValue += softJoint.GetRealValue(_paramPair.name);
            }

            float value = averageRealValue / softJoints.Length;
            if(value.IsEqual(_prevOuterValue, _paramPair.ValuePrecision))
            {
                return;
            }

            _prevOuterValue = value;
            infoSb.Clear()
                .AppendSize("\n", 5)
                .Append("Outer (avg): ")
                .AppendBold(value.ToString(_paramPair.ValueFormat));
            _outerValueField.text = infoSb.ToString();
        }

        void UpdateCenterValueText()
        {
            float averageRealValue = 0f;
            var softJoints = _softPhysicsHandler.leftContext.centerSoftJoints;
            foreach(var softJoint in softJoints)
            {
                averageRealValue += softJoint.GetRealValue(_paramPair.name);
            }

            float value = averageRealValue / softJoints.Length;
            if(value.IsEqual(_prevCenterValue, _paramPair.ValuePrecision))
            {
                return;
            }

            _prevCenterValue = value;
            infoSb.Clear()
                .AppendSize("\n", 5)
                .Append("Center (avg): ")
                .AppendBold(value.ToString(_paramPair.ValueFormat));
            _centerValueField.text = infoSb.ToString();
        }

        void UpdateMainValueText()
        {
            float averageRealValue = 0f;
            var softJoints = _softPhysicsHandler.leftContext.mainSoftJoints;
            foreach(var softJoint in softJoints)
            {
                averageRealValue += softJoint.GetRealValue(_paramPair.name);
            }

            float value = averageRealValue / softJoints.Length;
            if(value.IsEqual(_prevMainValue, _paramPair.ValuePrecision))
            {
                return;
            }

            _prevMainValue = value;
            infoSb.Clear()
                .AppendSize("\n", 5)
                .Append("Main (avg): ")
                .AppendBold(value.ToString(_paramPair.ValueFormat));
            _mainValueField.text = infoSb.ToString();
        }

        void UpdateNippleValueText()
        {
            float value = _softPhysicsHandler.leftContext.nippleSoftJoint.GetRealValue(_paramPair.name);
            if(value.IsEqual(_prevNippleValue, _paramPair.ValuePrecision))
            {
                return;
            }

            _prevNippleValue = value;

            infoSb.Clear()
                .AppendSize("\n", 5)
                .Append("Nipple: ")
                .AppendBold(value.ToString(_paramPair.ValueFormat));
            _nippleValueField.text = infoSb.ToString();
        }

        protected override void OnClose()
        {
            _colliderVisualizer.enabled = false;
            _drawService.enabled = false;
        }
    }
}
