using NAT.Common.Handlers;
using NAT.Common.Models;
using System.Text;
using UnityEngine;

namespace NAT.Common.UI.Windows
{
    sealed class WeightWindow : Window
    {
        readonly GravityPhysicsHandler _gravityPhysicsHandler;
        readonly JointPhysicsParamPair _paramPair;
        readonly JointPhysicsParam _param;
        readonly ParamValueInfo _paramValueInfo;
        UIDynamicTextField _gravityValuesField;

        internal WeightWindow(ScriptBase script, JointPhysicsParamPair paramPair) : base(script, paramPair.name)
        {
            _gravityPhysicsHandler = script.gravityPhysicsHandler;
            _paramPair = paramPair;
            _param = _paramPair.Left;
            _paramValueInfo = new ParamValueInfo(_paramPair);
        }

        protected override void Build()
        {
            AddElement("Parameter Info", _ =>
            {
                var uiDynamic = CreateTextField(_paramPair.infoText, new Vector2(10, -370), new Vector2(-555, 220));
                uiDynamic.UItext.fontSize = 28;
                return uiDynamic;
            });

            AddSlider(_param.offsetJsf, _ => CreateSlider("Offset", _paramPair.ValueFormat, new Vector2(545, -265)));

            AddElement("Value", _ =>
            {
                var uiDynamic = CreateTextField("", new Vector2(545, -330), new Vector2(-555, 45));
                uiDynamic.UItext.fontSize = 28;
                _paramValueInfo.textField = uiDynamic;
                return uiDynamic;
            });

            AddElement("Gravity Header", _ =>
            {
                string text = script.IsBootyMagic
                    ? "Glute Gravity"
                    : "Breast Gravity";
                var uiDynamic = CreateTextField(text.Bold(), new Vector2(10, -450), new Vector2(-555, 60));
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddElement("Gravity Info", _ =>
            {
                var sb = new StringBuilder();
                if(script.IsBootyMagic)
                {
                    sb.AppendSize("\n", 8)
                        .AppendLine("Adjust the force of gravity on the glute joint relative to different rotations of the joint.")
                        .AppendLine()
                        .Append("\"Down Gravity\" makes glutes hang down when standing up,")
                        .AppendLine(" while \"Back Gravity\" pulls glutes out when leaning back.");
                }
                else if(script.IsTittyMagic)
                {
                    sb.AppendSize("\n", 8)
                        .AppendLine("Adjust the force of gravity on the breast joint relative to different rotations of the joint.")
                        .AppendLine()
                        .Append("\"Down Gravity\" makes breasts hang down when standing up,")
                        .AppendLine(" while \"Back Gravity\" pushes breasts towards the chest when leaning back.");
                }

                sb.AppendLine()
                    .Append("The effective gravity is a result of scene gravity, background multipliers, and the multipliers on the right.");

                var uiDynamic = CreateTextField(sb.ToString(), new Vector2(10, -835), new Vector2(-555, 395));
                uiDynamic.UItext.fontSize = 28;
                return uiDynamic;
            });

            AddElement("Gravity Values Info", _ =>
            {
                var uiDynamic = CreateTextField("", new Vector2(10, -1200), new Vector2(-555, 350));
                uiDynamic.UItext.fontSize = 28;
                _gravityValuesField = uiDynamic;
                return uiDynamic;
            });

            AddToggle(_gravityPhysicsHandler.LinkDirectionsJsb, name => CreateToggle(name, new Vector2(545, -500), -555));

            const float sliderSpacing = 130;
            for(int i = 0; i < _gravityPhysicsHandler.UserMultipliers.Length; i++)
            {
                var multiplier = _gravityPhysicsHandler.UserMultipliers[i];
                float posY = -630 - i * sliderSpacing;
                AddSlider(multiplier, name => CreateSlider(name, "F2", new Vector2(545, posY)));
            }
        }

        Vector3 _prevSceneGravity = new Vector3(float.NaN, float.NaN, float.NaN);
        Vector3 _prevGravity = new Vector3(float.NaN, float.NaN, float.NaN);

        internal override void UpdateDiagnostics(bool atInterval)
        {
            if(!isBuilt || !atInterval)
            {
                return;
            }

            _paramValueInfo.Update(infoSb);

            bool doUpdateText = false;
            if(!Physics.gravity.IsEqual(_prevSceneGravity, 0.01f))
            {
                _prevSceneGravity = Physics.gravity;
                doUpdateText = true;
            }

            var gravity = _gravityPhysicsHandler.CustomGravityLeft;
            if(!gravity.IsEqual(_prevGravity, 0.01f))
            {
                _prevGravity = gravity;
                doUpdateText = true;
            }

            if(doUpdateText)
            {
                _gravityValuesField.text = infoSb
                    .Clear()
                    .AppendSize("\n", 8)
                    .AppendItalic("Scene gravity (m/s^2):")
                    .AppendLine()
                    .AppendSize("\n", 8)
                    .AppendFormat(" x: {0:0.00}", Physics.gravity.x)
                    .AppendLine()
                    .AppendFormat(" y: {0:0.00}", Physics.gravity.y)
                    .AppendLine()
                    .AppendFormat(" z: {0:0.00}", Physics.gravity.z)
                    .AppendLine()
                    .AppendLine()
                    .AppendItalic("Gravity on joint (m/s^2):")
                    .AppendLine()
                    .AppendSize("\n", 8)
                    .AppendFormat(" x: {0:0.00}", gravity.x)
                    .AppendLine()
                    .AppendFormat(" y: {0:0.00}", gravity.y)
                    .AppendLine()
                    .AppendFormat(" z: {0:0.00}", gravity.z)
                    .AppendLine()
                    .ToString();
            }
        }
    }
}
