// ReSharper disable UnusedMember.Global
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common
{
    static class Calc
    {
        // value returned is smoothed (for better animation) i.e. no longer maps linearly to the actual rotation angle
        internal static float SmoothStep(float val) => val > 0
            ? Mathf.SmoothStep(0, 1, val)
            : -Mathf.SmoothStep(0, 1, -val);

        internal static float RoundToDecimals(float value, float roundFactor = 1000f) =>
            Mathf.Round(value * roundFactor) / roundFactor;

        internal static Vector3 RelativePosition(Rigidbody origin, Vector3 position) =>
            RelativeDiff(position - origin.position, origin.transform);

        // TODO test with futa
        internal static Vector3 RelativePosition(Transform transform, Vector3 position) =>
            RelativeDiff(position - transform.position, transform);

        internal static Vector3 RelativePosition(Vector3 originPosition, Transform originT, Vector3 position) =>
            RelativeDiff(position - originPosition, originT);

        internal static Vector3 RelativePosition(Vector3 originPosition, Transform originT1, Transform originT2, Vector3 position)
        {
            var difference = position - originPosition;
            return new Vector3(
                Vector3.Dot(difference, (originT1.right + originT2.right) / 2),
                Vector3.Dot(difference, (originT1.up + originT2.up) / 2),
                Vector3.Dot(difference, (originT1.forward + originT2.forward) / 2)
            );
        }

        static Vector3 RelativeDiff(Vector3 difference, Transform originTransform) => new Vector3(
            Vector3.Dot(difference, originTransform.right),
            Vector3.Dot(difference, originTransform.up),
            Vector3.Dot(difference, originTransform.forward)
        );

        // TODO use?
        internal static Vector3 RelativePositionAlt(Transform t, Vector3 position) =>
            Quaternion.Inverse(t.rotation) * (position - t.position);

        internal static Vector3 AbsolutePosition(Transform t, Vector3 relativePosition) =>
            t.position + t.rotation * relativePosition;

        internal static Vector3 AverageVector3(IEnumerable<Vector3> vectors)
        {
            float sumX = 0;
            float sumY = 0;
            float sumZ = 0;
            int count = 0;

            foreach(var vector in vectors)
            {
                sumX += vector.x;
                sumY += vector.y;
                sumZ += vector.z;
                count++;
            }

            float averageX = sumX / count;
            float averageY = sumY / count;
            float averageZ = sumZ / count;

            return new Vector3(averageX, averageY, averageZ);
        }

        internal static Vector3 AverageRigidbodiesPosition(Rigidbody[] rigidbodies)
        {
            var sum = Vector3.zero;
            foreach(var rigidbody in rigidbodies)
            {
                sum += rigidbody.position;
            }

            return sum / rigidbodies.Length;
        }

        internal static bool DeviatesAtLeast(float v1, float v2, float percent)
        {
            bool result;
            if(v1 > v2)
            {
                result = (v1 - v2) / v1 > percent / 100;
            }
            else
            {
                result = (v2 - v1) / v2 > percent / 100;
            }

            return result;
        }

        internal static float[] ExponentialMovingAverage(float[] source, float k)
        {
            float[] result = new float[source.Length];
            result[source.Length - 1] = source[source.Length - 1];
            for(int i = source.Length - 2; i >= 0; i--)
            {
                result[i] = k * source[i] + (1 - k) * result[i + 1];
            }

            return result;
        }

        internal static Vector3[] ExponentialMovingAverage(Vector3[] source, float k)
        {
            var result = new Vector3[source.Length];
            result[source.Length - 1] = source[source.Length - 1];
            for(int i = source.Length - 2; i >= 0; i--)
            {
                result[i] = new Vector3(
                    k * source[i].x + (1 - k) * result[i + 1].x,
                    k * source[i].y + (1 - k) * result[i + 1].y,
                    k * source[i].z + (1 - k) * result[i + 1].z
                );
            }

            return result;
        }

        internal static float InverseSmoothStep(float value, float curvature, float midPoint)
        {
            if(value < 0f)
            {
                return 0f;
            }

            if(value > 1f)
            {
                return 1f;
            }

            float c = 2f / (1f - curvature) - midPoint;

            return value < midPoint
                ? F1(value, midPoint, c)
                : 1 - F1(1f - value, 1f - midPoint, c);
        }

        static float F1(float value, float midPoint, float exp) => Mathf.Pow(value, exp) / Mathf.Pow(midPoint, exp - 1);

        // https://www.desmos.com/calculator/ln9oecyv4y
        // dual: https://www.desmos.com/calculator/icmymkbfux
        internal static float InverseSmoothStepWithMax(float value, float max, float curvature, float midPoint)
        {
            if(value < 0f)
            {
                return 0f;
            }

            if(value > max)
            {
                return 1f;
            }

            float s = ClampSmoothStepCurvature(curvature);
            float p = midPoint * max;
            p = p < 0f ? 0f : p > max ? max : p;
            float c = 2f / (1f - s) - p / max;

            if(value < p)
            {
                return F1(value, max, p, c);
            }

            return 1f - F1(max - value, max, max - p, c);
        }

        static float F1(float value, float b, float n, float c) => Mathf.Pow(value, c) / (b * Mathf.Pow(n, c - 1));

        internal static float ClampSmoothStepCurvature(float curvature) => curvature < -2.99f ? -2.99f : curvature > 0.99f ? 0.99f : curvature;

        // https://www.desmos.com/calculator/7s7iak8rqb
        internal static float QuadraticRegression(float value) => -0.118f * value * value + 0.97f * value;

        // ReSharper disable once UnusedMember.Local
        static float Polynomial(
            float x,
            float a = 1,
            float b = 1,
            float c = 1,
            float p = 1,
            float q = 1
        ) => a * Mathf.Pow(x, p) + b * Mathf.Pow(x, q) + c * x;

        // https://www.desmos.com/calculator/okwggb6a3s
        // ReSharper disable once UnusedMember.Global
        internal static float QuadraticOneToZero(float x, float a) => a * Mathf.Pow(x - Mathf.Sqrt(1 / a), 2);
        internal static float SoftnessBaseCurve(float x) => Exponential1(x, 6.44f, 1.27f, 1.15f);
        internal static float ForcePhysicsSoftnessCurve(float x) => Exponential1(x, 1.60f, 2.29f, 1.00f);
        internal static float AngleTargetMassCurve(float x) => Exponential1(x, 0.30f, 1.75f, 1.00f);
        internal static float AngleTargetSoftnessCurve(float x) => Exponential1(x, 3.00f, 1.35f, 1.00f);
        internal static float InOutSpringSoftnessCurve(float x) => Exponential1(x, 2.11f, 1.21f, 0.88f);

        // https://www.desmos.com/calculator/ltyekxtlji
        internal static float Exponential1(float x, float b, float p, float q, float a = 1, float s = 0) =>
            (1 - b) * Mathf.Pow(a * x + s, p) + b * Mathf.Pow(a * x + s, q);

        // https://www.desmos.com/calculator/6pxtrvvyby
        internal static float Exponential2(float x, float c, float s = 0)
        {
            float baseValue = (2 - c) * (x + s) - 1 + c / 2;
            return c / (1 + c / 3) + baseValue * baseValue;
        }

        // https://www.desmos.com/calculator/mpmm8lzvs2
        // ReSharper disable once MemberCanBePrivate.Global
        internal static float Exponential3(float x, float c, float s, float l, float q, float p)
        {
            float xToLimitPow = Mathf.Pow(x / l, q);
            return c + s * Mathf.Pow(1 - xToLimitPow, p);
        }

        // https://www.desmos.com/calculator/732liojfhe
        internal static float LocalDip(float x, float p, float h, float q, float m)
        {
            float exponent = -p * Mathf.Pow(h * (x - m), q);
            float clampedExponent = Mathf.Min(exponent, 709); // 709 is the largest representable positive exponent in floating-point numbers
            return 1 - Mathf.Exp(clampedExponent);
        }

        internal static float PercentageValue(float value) => RoundToDecimals(value, 10);
    }
}
