// ReSharper disable MemberCanBePrivate.Global UnusedMember.Global
using System;
// ReSharper disable once RedundantUsingDirective
using UnityEngine;

namespace NAT.Common
{
    sealed class SemVer
    {
        readonly Version _version;
        internal readonly string versionStr;
        internal readonly string prereleaseTag = "";
        internal bool IsEmpty => string.IsNullOrEmpty(versionStr);

        internal SemVer(string versionStr, bool splitPrereleaseTag = false)
        {
            try
            {
                this.versionStr = versionStr;
                /* Without prereleaseTag set to true, a versionStr containing a prerelease tag will cause an error.
                 * The result is that _version remains null and all comparisons return false.
                 */
                if(splitPrereleaseTag)
                {
                    string[] split = versionStr.Split('-');
                    _version = new Version(split[0]);
                    if(split.Length > 1)
                    {
                        prereleaseTag = split[1];
                    }
                }
                else
                {
                    _version = new Version(versionStr);
                }
            }
            catch(Exception)
            {
                // Debug.Log("Invalid version string: " + versionStr);
            }
        }

        internal bool IsEqual(string other) => IsEqual(new SemVer(other));

        internal bool IsEqual(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version == other._version;
        }

        internal bool IsLt(string other) => IsLt(new SemVer(other));
        internal bool IsMajorLt(string other) => IsMajorLt(new SemVer(other));
        internal bool IsMajorMinorLt(string other) => IsMajorMinorLt(new SemVer(other));

        internal bool IsLt(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version < other._version;
        }

        internal bool IsMajorLt(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version.Major < other._version.Major;
        }

        internal bool IsMajorMinorLt(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            var majorMinorA = new Version(_version.Major, _version.Minor);
            var majorMinorB = new Version(other._version.Major, other._version.Minor);

            return majorMinorA < majorMinorB;
        }

        internal bool IsGte(string other) => IsGte(new SemVer(other));
        internal bool IsMajorGte(string other) => IsMajorGte(new SemVer(other));
        internal bool IsMajorMinorGte(string other) => IsMajorMinorGte(new SemVer(other));
        internal bool IsMajorMinorEqual(string other) => IsMajorMinorEqual(new SemVer(other));

        internal bool IsGte(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version >= other._version;
        }

        internal bool IsMajorGte(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version.Major >= other._version.Major;
        }

        internal bool IsMajorMinorGte(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            var majorMinorA = new Version(_version.Major, _version.Minor);
            var majorMinorB = new Version(other._version.Major, other._version.Minor);

            return majorMinorA >= majorMinorB;
        }

        internal bool IsMajorMinorEqual(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            var majorMinorA = new Version(_version.Major, _version.Minor);
            var majorMinorB = new Version(other._version.Major, other._version.Minor);

            return majorMinorA == majorMinorB;
        }
    }
}
