using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.Text.RegularExpressions;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Security.Cryptography;
using UnityEngine;

namespace NAT.Common
{
    [SuppressMessage("ReSharper", "UnusedMember.Global")]
    static class Utils
    {
        // ReSharper disable once UnusedMember.Global
        internal static MVRScript FindPluginOnAtom(Atom atom, string search)
        {
            string match = atom.GetStorableIDs().Find(id => id.Contains(search));
            return match == null ? null : atom.GetStorableByID(match) as MVRScript;
        }

        internal static IEnumerable<Atom> GetAtomsByType(string type)
        {
            foreach(var atom in SuperController.singleton.GetAtoms())
            {
                if(atom.type == type)
                {
                    yield return atom;
                }
            }
        }

        const float MAX_PHYSICS_RATE_MULTIPLIER = 2.00f;

        internal static float PhysicsRateMultiplier()
        {
            float multiplier = 0.01666667f / Time.fixedDeltaTime;
            return multiplier > MAX_PHYSICS_RATE_MULTIPLIER ? MAX_PHYSICS_RATE_MULTIPLIER : multiplier;
        }

        internal static string ParamName(string paramName, string side)
        {
            string result = paramName;
            if(side != Side.LEFT)
            {
                result += $" {side}";
            }

            return result;
        }

        // ReSharper disable once UnusedMember.Global
        internal static string ObjectPropertiesString(object obj)
        {
            string result = "";
            foreach(PropertyDescriptor descriptor in TypeDescriptor.GetProperties(obj))
            {
                result += $"{descriptor.Name} = {descriptor.GetValue(obj)}\n";
            }

            return result;
        }

        /// <summary>
        /// Get a string containing a visual representation of a Transform's children hierarchy.
        /// </summary>
        /// <param name="root">The parent Transform (GameObject)</param>
        /// <param name="maxDepth">Maximum depth to traverse the hierarchy.</param>
        /// <param name="propertyDel">A delegate function that takes a Transform and returns the string to print as an entry in the hierarchy.
        /// For example, to print a hierarchy of each transform's name: <code>thisTransform => thisTransform.name</code> A null value will print names.</param>
        /// <returns>A string containing a visual hierarchy of all child transforms, including the root</returns>
        // ReSharper disable once UnusedMember.Global
        internal static string ObjectHierarchyToString(Transform root, int? maxDepth = null, Func<Transform, string> propertyDel = null)
        {
            if(propertyDel == null)
            {
                propertyDel = t => t.name;
            }

            var builder = new StringBuilder();
            ObjectHierarchyToString(root, propertyDel, builder, maxDepth);

            if(builder.Length < 1024)
            {
                return builder.ToString();
            }

            return
                $"Output string length {builder.Length.ToString()} may be too large for viewing in VAM. " +
                $"See %userprofile%/AppData/LocalLow/MeshedVR/VaM/output_log.txt for the full output.\n{builder}";
        }

        static void ObjectHierarchyToString(
            Transform root,
            Func<Transform, string> propertyDel,
            StringBuilder builder,
            int? maxDepth,
            int currentDepth = 0
        )
        {
            if(currentDepth > maxDepth)
            {
                return;
            }

            for(int i = 0; i < currentDepth; i++)
            {
                builder.Append("|   ");
            }

            builder.Append(propertyDel(root) + "\n");
            foreach(Transform child in root)
            {
                ObjectHierarchyToString(child, propertyDel, builder, maxDepth, currentDepth + 1);
            }
        }

        internal static IEnumerable<string> GetChildComponentData(UnityEngine.Component component) =>
            GetComponentHierarchy(component.transform, 0);

        static IEnumerable<string> GetComponentHierarchy(Transform current, int indentLevel)
        {
            foreach(var childComponent in current.GetComponents<UnityEngine.Component>())
            {
                string indentation = "";
                for(int i = 0; i < indentLevel; i++)
                {
                    indentation += "> ";
                }

                string result = $"{indentation}{childComponent.GetType()}";
                if(!childComponent.gameObject.activeSelf)
                {
                    result += ", active: false";
                }

                yield return result;
            }

            foreach(Transform child in current)
            {
                foreach(string childData in GetComponentHierarchy(child, indentLevel + 1))
                {
                    yield return childData;
                }
            }
        }

        internal static string EnumerableToString<T>(IEnumerable<T> enumerable, string separator = "\n")
            => string.Join(separator, enumerable.Select(item => item.ToString()).ToArray());

        internal static bool GlobalAnimationFrozen()
        {
            bool mainToggleFrozen =
                SuperController.singleton.freezeAnimationToggle &&
                SuperController.singleton.freezeAnimationToggle.isOn;
            bool altToggleFrozen =
                SuperController.singleton.freezeAnimationToggleAlt &&
                SuperController.singleton.freezeAnimationToggleAlt.isOn;
            return mainToggleFrozen || altToggleFrozen;
        }

        internal static Rigidbody FindRigidbody(Atom atom, string name)
        {
            foreach(var rigidbody in atom.GetComponentsInChildren<Rigidbody>())
            {
                if(rigidbody.name == name)
                {
                    return rigidbody;
                }
            }

            throw new Exception($"Could not find Rigidbody {name} on atom {atom.name}");
        }

        internal static IEnumerator WaitAndRepeat(Action callback, int times, float intervalWait)
        {
            int count = 0;
            while(count < times)
            {
                count++;
                yield return new WaitForSeconds(intervalWait);
                callback.Invoke();
            }
        }

        internal static Regex NewRegex(string regexStr) => new Regex(regexStr, RegexOptions.Compiled);

        internal static string ToValueFormat(float roundFactor)
        {
            switch(Mathf.RoundToInt(roundFactor))
            {
                case 1: return "F0";
                case 10: return "F1";
                case 100: return "F2";
                case 1000: return "F3";
                case 10000: return "F4";
                case 100000: return "F5";
                default: throw new Exception($"Invalid roundFactor {roundFactor}");
            }
        }

        internal static float ValueFormatToTolerance(string valueFormat)
        {
            switch(valueFormat)
            {
                case "F0": return 0.1f;
                case "F1": return 0.01f;
                case "F2": return 0.001f;
                case "F3": return 0.0001f;
                case "F4": return 0.00001f;
                case "F5": return 0.000001f;
                default: throw new Exception($"Invalid valueFormat {valueFormat}");
            }
        }

        internal static Vector3 DirectionVector(Rigidbody rb, string direction)
        {
            switch(direction)
            {
                case Direction.UP: return rb.transform.up;
                case Direction.DOWN: return -rb.transform.up;
                case Direction.FORWARD: return rb.transform.forward;
                case Direction.BACK: return -rb.transform.forward;
                case Direction.LEFT: return -rb.transform.right;
                case Direction.RIGHT: return rb.transform.right;
                default: throw new ArgumentException($"Invalid direction {direction}");
            }
        }

        internal static Camera GetCamera() => GetVRCamera() ?? SuperController.singleton.MonitorCenterCamera;

        static Camera GetVRCamera()
        {
            if(SuperController.singleton.isOpenVR)
            {
                return SuperController.singleton.ViveCenterCamera;
            }

            return SuperController.singleton.isOVR ? SuperController.singleton.OVRCenterCamera : null;
        }

        internal static bool IsVR() => SuperController.singleton.isOpenVR || SuperController.singleton.isOVR;

        internal static string Sha256Hash(string value, StringBuilder stringBuilder)
        {
            stringBuilder.Clear();
            using(var hash = SHA256.Create())
            {
                var enc = Encoding.UTF8;
                byte[] result = hash.ComputeHash(enc.GetBytes(value));
                foreach(byte b in result)
                {
                    stringBuilder.Append(b.ToString("x2"));
                }
            }

            return stringBuilder.ToString().Substring(0, 6);
        }

        internal static int Sha256HashToInt(string value)
        {
            using(var hash = SHA256.Create())
            {
                byte[] bytes = hash.ComputeHash(Encoding.UTF8.GetBytes(value));
                return BitConverter.ToInt32(bytes, 0);
            }
        }
    }
}
