using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.TM.Handlers
{
    sealed class BreastGravityPhysicsHandler : GravityPhysicsHandler
    {
        internal BreastGravityPhysicsHandler(ScriptBase script) : base(script)
        {
            twistAngleTargetJsf = new StorableFloat(TWIST_ANGLE_TARGET.NAME, 0.00f, -30.00f, 30.00f);
            LoadSettings();
        }

        protected override void LoadSettings()
        {
            leftContext.calculators = Calculators(Side.LEFT);
            rightContext.calculators = Calculators(Side.RIGHT);

            leftContext.calculators.Add(
                jointPhysicsHandler.rightLeftAngleTarget,
                new Dictionary<string, PhysicsCalculator>
                {
                    {
                        SIDEWAYS_IN, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_IN),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = 0.00f,
                                multiplierB = 5.00f,
                                curveB = GravitySoftnessCurve,
                            },
                        }
                    },
                    {
                        SIDEWAYS_OUT, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.LEFT), SIDEWAYS_OUT),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = 0.00f,
                                multiplierB = -5.00f,
                                curveB = GravitySoftnessCurve,
                            },
                        }
                    },
                }
            );

            rightContext.calculators.Add(
                jointPhysicsHandler.rightLeftAngleTarget,
                new Dictionary<string, PhysicsCalculator>
                {
                    {
                        SIDEWAYS_IN, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_IN),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = 0.00f,
                                multiplierB = -5.00f,
                                curveB = GravitySoftnessCurve,
                            },
                        }
                    },
                    {
                        SIDEWAYS_OUT, new PhysicsCalculator
                        {
                            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_ANGLE_TARGET.NAME, Side.RIGHT), SIDEWAYS_OUT),
                            primaryConfig = new MultiplierConfig
                            {
                                baseValue = 0.00f,
                                multiplierB = 5.00f,
                                curveB = GravitySoftnessCurve,
                            },
                        }
                    },
                }
            );

            var twistAngleTargetCalculators = TwistAngleTargetCalculators();
            leftContext.calculators.Add(jointPhysicsHandler.twistAngleTarget, twistAngleTargetCalculators[Side.LEFT]);
            rightContext.calculators.Add(jointPhysicsHandler.twistAngleTarget, twistAngleTargetCalculators[Side.RIGHT]);

            foreach(var pair in leftContext.calculators)
            {
                var paramPair = pair.Key;
                var calculators = pair.Value;
                paramPair.Left.gravityCalculators = calculators;

                if(paramPair != jointPhysicsHandler.twistAngleTarget)
                {
                    paramPair.SetGravityMultiplierCallbacks();
                }

                if(paramPair == jointPhysicsHandler.rotationDamper)
                {
                    leftContext.updateHandlers += (direction, effect) =>
                    {
                        if(!isCalibrating)
                        {
                            paramPair.Left.UpdateGravityValue(direction, effect, mass, softness);
                        }
                    };
                    leftContext.resetHandlers += direction =>
                    {
                        if(!isCalibrating)
                        {
                            paramPair.Left.ResetGravityValue(direction);
                        }
                    };
                }
                else
                {
                    leftContext.updateHandlers += (direction, effect) => paramPair.Left.UpdateGravityValue(direction, effect, mass, softness);
                    leftContext.resetHandlers += direction => paramPair.Left.ResetGravityValue(direction);
                }
            }

            foreach(var pair in rightContext.calculators)
            {
                var paramPair = pair.Key;
                var calculators = pair.Value;
                paramPair.Right.gravityCalculators = calculators;

                if(paramPair != jointPhysicsHandler.twistAngleTarget)
                {
                    paramPair.SetGravityMultiplierCallbacks();
                }

                if(paramPair == jointPhysicsHandler.rotationDamper)
                {
                    rightContext.updateHandlers += (direction, effect) =>
                    {
                        if(!isCalibrating)
                        {
                            paramPair.Right.UpdateGravityValue(direction, effect, mass, softness);
                        }
                    };
                    rightContext.resetHandlers += direction =>
                    {
                        if(!isCalibrating)
                        {
                            paramPair.Right.ResetGravityValue(direction);
                        }
                    };
                }
                else
                {
                    rightContext.updateHandlers += (direction, effect) => paramPair.Right.UpdateGravityValue(direction, effect, mass, softness);
                    rightContext.resetHandlers += direction => paramPair.Right.ResetGravityValue(direction);
                }
            }
        }

        Dictionary<JointPhysicsParamPair, Dictionary<string, PhysicsCalculator>> Calculators(string side) =>
            new Dictionary<JointPhysicsParamPair, Dictionary<string, PhysicsCalculator>>
            {
                {
                    jointPhysicsHandler.rotationDamper, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpDamperCalculator(side, UP) },
                        { BACK, BackDamperCalculator(side, BACK) },
                        { FORWARD, ForwardDamperCalculator(side, FORWARD) },
                        { SIDEWAYS_IN, SidewaysInDamperCalculator(side, SIDEWAYS_IN) },
                    }
                },
                {
                    jointPhysicsHandler.rightLeftSpring, new Dictionary<string, PhysicsCalculator>
                    {
                        { SIDEWAYS_IN, RightLeftSpringCalculator(side, SIDEWAYS_IN) },
                        { SIDEWAYS_OUT, RightLeftSpringCalculator(side, SIDEWAYS_OUT) },
                    }
                },
                {
                    jointPhysicsHandler.inOutSpring, new Dictionary<string, PhysicsCalculator>
                    {
                        { BACK, InOutSpringCalculator(side, BACK) },
                        { FORWARD, InOutSpringCalculator(side, FORWARD) },
                    }
                },
                {
                    jointPhysicsHandler.upDownAngleTarget, new Dictionary<string, PhysicsCalculator>
                    {
                        { UP, UpAngleTargetCalculator(side) },
                        { DOWN, DownAngleTargetCalculator(side) },
                    }
                },
            };

        static PhysicsCalculator UpDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = 0.25f,
                curveA = UpGravityMassCurve,
            },
        };

        static PhysicsCalculator ForwardDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = 0.60f,
                multiplierB = 0.30f,
            },
        };

        static PhysicsCalculator BackDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = 0.40f,
                multiplierB = 0.20f,
            },
        };

        static PhysicsCalculator SidewaysInDamperCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(ROTATION_DAMPER.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0,
                multiplierA = 0.24f,
                multiplierB = 0.12f,
                curveA = x => 1 - x,
            },
        };

        static PhysicsCalculator RightLeftSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(RIGHT_LEFT_SPRING.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 1500f,
            },
        };

        static PhysicsCalculator InOutSpringCalculator(string side, string direction) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(IN_OUT_SPRING.NAME, side), direction),
            primaryConfig = new MultiplierConfig
            {
                baseValue = -50f,
                multiplierA = -67f,
                multiplierB = -83f,
                curveA = x => Calc.Exponential1(1 - x, 0.3f, 10f, 0.84f, 1.01f),
                curveB = Calc.InOutSpringSoftnessCurve,
            },
        };

        static PhysicsCalculator UpAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(BREAST_UP_DOWN_ANGLE_TARGET.NAME, side), UP),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0.00f,
                multiplierB = 6.00f,
                curveB = GravitySoftnessCurve,
            },
        };

        static PhysicsCalculator DownAngleTargetCalculator(string side) => new PhysicsCalculator
        {
            multiplierJsf = NewGravityMultiplierJsf(Utils.ParamName(BREAST_UP_DOWN_ANGLE_TARGET.NAME, side), DOWN),
            primaryConfig = new MultiplierConfig
            {
                baseValue = 0.00f,
                multiplierB = -6.00f,
                curveB = GravitySoftnessCurve,
            },
        };

        static Dictionary<string, PhysicsCalculator> TwistAngleTargetCalculator() => new Dictionary<string, PhysicsCalculator>
        {
            { UP, DefaultPhysicsCalculator() },
            { DOWN, DefaultPhysicsCalculator() },
        };

        Dictionary<string, Dictionary<string, PhysicsCalculator>> TwistAngleTargetCalculators()
        {
            var left = TwistAngleTargetCalculator();
            var right = TwistAngleTargetCalculator();

            twistAngleTargetJsf.setJSONCallbackFunction = jsf =>
            {
                var storable = (StorableFloat) jsf;
                float rounded = Calc.RoundToDecimals(storable.val);
                left[UP].primaryConfig.baseValue = -rounded;
                left[UP].secondaryConfig.baseValue = -rounded;
                left[DOWN].primaryConfig.baseValue = rounded;
                left[DOWN].secondaryConfig.baseValue = rounded;
                right[UP].primaryConfig.baseValue = rounded;
                right[UP].secondaryConfig.baseValue = rounded;
                right[DOWN].primaryConfig.baseValue = -rounded;
                right[DOWN].secondaryConfig.baseValue = -rounded;
                Script.SetCalibrationPending(storable);
                Script.jointPhysicsHandler.twistAngleTargetValue = rounded;
                Script.uiHandler.physicsWindow.OnMainParamAdjusted(jointPhysicsHandler.twistAngleTarget);
            };

            return new Dictionary<string, Dictionary<string, PhysicsCalculator>>
            {
                { Side.LEFT, left },
                { Side.RIGHT, right },
            };
        }

        internal override void SaveCalibrated()
        {
            base.SaveCalibrated();
            twistAngleTargetJsf.SaveCalibrated();
        }

        static float UpGravityMassCurve(float x) => Calc.InverseSmoothStep(1 - x, 0.67f, 0.98f);
        static float BackForwardGravityMassCurve(float x) => 1 - x;
        static float GravitySoftnessCurve(float x) => Calc.InverseSmoothStep(x, 0.00f, 0.15f);

        internal override void SetMultipliers()
        {
            mass = jointPhysicsHandler.normalizedRealMass;
            float baseMultiplier = Mathf.Lerp(Mathf.Lerp(0.20f, 0.75f, mass), 1.00f, GravitySoftnessCurve(Script.softness));
            backGravityMultiplier = baseMultiplier * Mathf.Lerp(1.00f, 1.73f, BackForwardGravityMassCurve(mass));
            forwardGravityMultiplier = baseMultiplier * Mathf.Lerp(0.52f, 2.28f, BackForwardGravityMassCurve(mass));
            upGravityMultiplier = baseMultiplier;
            downGravityMultiplier = Mathf.Lerp(
                Mathf.Lerp(0.28f, 1.05f, mass),
                Mathf.Lerp(1.20f, 1.68f, mass),
                GravitySoftnessCurve(Script.softness)
            );
            sidewaysInGravityMultiplier = baseMultiplier;
            sidewaysOutGravityMultiplier = baseMultiplier;
            SetGravityValues();
        }
    }
}
