using NAT.Common;
using NAT.Common.Contexts;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using System;
using System.Text;
using UnityEngine;

namespace NAT.TM.Handlers
{
    sealed class BreastJointPhysicsHandler : JointPhysicsHandler
    {
        readonly JSONStorableBool _xMotionEnabledJsb;

        float _lowMassInterp;
        float _midMassInterp;
        MorphWrapper _upAngleMorphS;
        MorphWrapper _upAngleMorphL;
        MorphWrapper _upAngleMorphXL;
        MorphWrapper _downAngleMorphS;
        MorphWrapper _downAngleMorphL;
        MorphWrapper _downAngleMorphXL;
        MorphWrapper _sidewaysInMorphS;
        MorphWrapper _sidewaysInMorphL;
        MorphWrapper _sidewaysInMorphXL;
        MorphWrapper _sidewaysOutMorphS;
        MorphWrapper _sidewaysOutMorphL;
        MorphWrapper _sidewaysOutMorphXL;

        internal BreastJointPhysicsHandler(ScriptBase script) : base(script)
        {
            control = Person.BreastControl;
            leftContext.bodyPart = Person.LeftBreast;
            rightContext.bodyPart = Person.RightBreast;

            leftContext.volumeVertexIndices = new[]
            {
                2403, 2411, 2592, 7197, 7208, 8902, 8904, 8905, 8908, 8909, 8911, 8915, 8917, 8922, 8924, 8925, 8926, 8928, 8930,
                8933, 8934, 8935, 8939, 8941, 8942, 8943, 8945, 8946, 8947, 8948, 8949, 8951, 8952, 8954, 8959, 8969, 8970,
            };
            rightContext.volumeVertexIndices = new[]
            {
                13233, 13241, 13410, 17904, 17915, 19568, 19570, 19571, 19574, 19575, 19577, 19581, 19583, 19588, 19590,
                19591, 19592, 19594, 19596, 19599, 19600, 19601, 19605, 19607, 19608, 19609, 19611, 19612, 19613, 19614,
                19615, 19617, 19618, 19620, 19625, 19635, 19636,
            };

            script.leftCalibrationTarget.centerVertexIndices = new[] { 8911, 8930, 8943, 8947 };
            script.rightCalibrationTarget.centerVertexIndices = new[] { 19577, 19596, 19609, 19625 };

            _xMotionEnabledJsb = new JSONStorableBool("Right/Left Motion Enabled", false, SetXMotionEnabled);
            LoadSettings();
        }

        void SetXMotionEnabled(bool value)
        {
            var motion = value ? ConfigurableJointMotion.Limited : ConfigurableJointMotion.Locked;
            leftContext.bodyPart.joint.xMotion = motion;
            rightContext.bodyPart.joint.xMotion = motion;
        }

        internal override void CalculateMass()
        {
            float leftVolume = CalculateVolume(leftContext.volumeVertexIndices);
            float rightVolume = CalculateVolume(rightContext.volumeVertexIndices);
            float volume = (leftVolume + rightVolume) / 2;

            // https://www.desmos.com/calculator/aw68taob0a
            float value = Mathf.Pow(0.614f * volume, 1.290f);
            float massBaseValue = Mathf.Clamp(value, BREAST_JOINT_MASS.MIN, BREAST_JOINT_MASS.MAX);
            UpdateMassValueAndAmounts(massBaseValue);
        }

        internal override void UpdateMassValueAndAmounts(float mass)
        {
            calibratedMass = mass;
            jointMass.UpdateMassValue(mass);
            jointMass.UpdateOffsetMinMax();
            const float min = BREAST_JOINT_MASS.MIN;
            const float max = BREAST_JOINT_MASS.MAX;
            normalizedMass = Mathf.InverseLerp(min, max, jointMass.Left.Value);
            normalizedRealMass = Mathf.InverseLerp(min, max, jointMass.Left.baseValueJsf.val);
            _lowMassInterp = Mathf.InverseLerp(0, 1 / 3f, normalizedRealMass);
            _midMassInterp = Mathf.InverseLerp(1 / 3f, 2 / 3f, normalizedRealMass);
        }

        float CalculateVolume(int[] vertexIndices)
        {
            var skinVerts = Person.Skin.rawSkinnedVerts;
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(Person.ChestRb, skinVerts[vertexIndices[i]]);
            }

            return 1000 * VolumeCalculator.CupVolume(positions, Person.ScaleJsf.val);
        }

        protected override void LoadSettings()
        {
            _upAngleMorphS = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Up Angle S"));
            _upAngleMorphL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Up Angle L"));
            _upAngleMorphXL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Up Angle XL"));
            _downAngleMorphS = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Down Angle S"));
            _downAngleMorphL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Down Angle L"));
            _downAngleMorphXL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Down Angle XL"));
            _sidewaysInMorphS = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Sideways In Angle S"));
            _sidewaysInMorphL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Sideways In Angle L"));
            _sidewaysInMorphXL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Sideways In Angle XL"));
            _sidewaysOutMorphS = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Sideways Out Angle S"));
            _sidewaysOutMorphL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Sideways Out Angle L"));
            _sidewaysOutMorphXL = new MorphWrapper(Person.GetBreastMorph("Angles/Breast Sideways Out Angle XL"));

            jointMass = new JointPhysicsParamPair(Script, BREAST_JOINT_MASS.NAME, Constant.BREAST_WEIGHT)
            {
                ValueFormat = BREAST_JOINT_MASS.FORMAT,
                Left = BreastMassParam(leftContext),
                Right = BreastMassParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(BREAST_JOINT_MASS.NAME)
                    .Append(" is the mass of the joint in kg.\n\n")
                    .Append("Since mass represents BodyPart size, other physics parameters are adjusted based on its value.")
                    .ToString(),
            };

            paramPairs.Add(centerOfGravity = new JointPhysicsParamPair(Script, CENTER_OF_GRAVITY.NAME)
            {
                ValueFormat = CENTER_OF_GRAVITY.FORMAT,
                Left = CenterOfGravityParam(leftContext),
                Right = CenterOfGravityParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(CENTER_OF_GRAVITY.NAME)
                    .Append(" is the position of the joint's center of mass.\n\n")
                    .Append("At 0, the center of mass is inside the chest at the pectoral joint. At 1, it is at the nipple.")
                    .ToString(),
            });

            paramPairs.Add(rotationSpring = new JointPhysicsParamPair(Script, ROTATION_SPRING.NAME)
            {
                ValueFormat = ROTATION_SPRING.FORMAT,
                Left = RotationSpringParam(leftContext),
                Right = RotationSpringParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(ROTATION_SPRING.NAME)
                    .Append(" pushes the joint towards its angle target.\n\n")
                    .Append("The angle target is defined by the Up/Down, Left/Right and Twist Angle Target parameters.")
                    .ToString(),
            });

            paramPairs.Add(rotationDamper = new JointPhysicsParamPair(Script, ROTATION_DAMPER.NAME)
            {
                ValueFormat = ROTATION_DAMPER.FORMAT,
                Left = RotationDamperParam(leftContext),
                Right = RotationDamperParam(rightContext),
                dependsOnPhysicsRate = true,
                infoText = new StringBuilder()
                    .AppendBold(ROTATION_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint's angle target.\n\n")
                    .Append("The higher the damper, the quicker breasts will stop swinging.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftSpring = new JointPhysicsParamPair(Script, RIGHT_LEFT_SPRING.NAME)
            {
                ValueFormat = RIGHT_LEFT_SPRING.FORMAT,
                Left = RightLeftSpringParam(leftContext),
                Right = RightLeftSpringParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_SPRING.NAME)
                    .Append(" pushes the joint towards its position target along the X axis.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftDamper = new JointPhysicsParamPair(Script, RIGHT_LEFT_DAMPER.NAME)
            {
                ValueFormat = RIGHT_LEFT_DAMPER.FORMAT,
                Left = RightLeftDamperParam(leftContext),
                Right = RightLeftDamperParam(rightContext),
                dependsOnPhysicsRate = true,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint's position target along the X axis.")
                    .ToString(),
            });

            paramPairs.Add(inOutSpring = new JointPhysicsParamPair(Script, IN_OUT_SPRING.NAME)
            {
                ValueFormat = IN_OUT_SPRING.FORMAT,
                Left = InOutSpringParam(leftContext),
                Right = InOutSpringParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(IN_OUT_SPRING.NAME)
                    .Append(" pushes the joint towards its position target along the Z axis.\n\n")
                    .Append("Directional force morphing along the forward-back axis depends on In/Out Spring being suitably low for the given BodyPart mass.")
                    .ToString(),
            });

            paramPairs.Add(inOutDamper = new JointPhysicsParamPair(Script, IN_OUT_DAMPER.NAME)
            {
                ValueFormat = IN_OUT_DAMPER.FORMAT,
                Left = InOutDamperParam(leftContext),
                Right = InOutDamperParam(rightContext),
                dependsOnPhysicsRate = true,
                infoText = new StringBuilder()
                    .AppendBold(IN_OUT_DAMPER.NAME)
                    .Append(" reduces oscillation around the joint's position target along the Z axis.")
                    .ToString(),
            });

            paramPairs.Add(upDownBaseAngle = new JointPhysicsParamPair(Script, BREAST_UP_DOWN_BASE_ANGLE.NAME, "Up/Down Angle")
            {
                ValueFormat = BREAST_UP_DOWN_BASE_ANGLE.FORMAT,
                Left = UpDownBaseAngleParam(leftContext),
                Right = UpDownBaseAngleParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(BREAST_UP_DOWN_BASE_ANGLE.NAME)
                    .Append(" adjusts the pectoral joint's vertical rotation using morphs.\n\n")
                    .Append("Use the calibration state to pause real-time morphing.")
                    .ToString(),
            });

            paramPairs.Add(upAngleLimit = new JointPhysicsParamPair(Script, UP_ANGLE_LIMIT.NAME)
            {
                ValueFormat = UP_ANGLE_LIMIT.FORMAT,
                Left = UpAngleLimitParam(leftContext),
                Right = UpAngleLimitParam(rightContext),
                needsCalibration = NeedsCalibration.IF_LDP_ENABLED,
                infoText = new StringBuilder()
                    .AppendBold(UP_ANGLE_LIMIT.NAME)
                    .Append(" is the maximum angle the joint can rotate up.")
                    .ToString(),
            });

            paramPairs.Add(downAngleLimit = new JointPhysicsParamPair(Script, DOWN_ANGLE_LIMIT.NAME)
            {
                ValueFormat = DOWN_ANGLE_LIMIT.FORMAT,
                Left = DownAngleLimitParam(leftContext),
                Right = DownAngleLimitParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(DOWN_ANGLE_LIMIT.NAME)
                    .Append(" is the maximum angle the joint can rotate down.")
                    .ToString(),
            });

            paramPairs.Add(upDownAngleTarget = new JointPhysicsParamPair(Script, BREAST_UP_DOWN_ANGLE_TARGET.NAME)
            {
                ValueFormat = BREAST_UP_DOWN_ANGLE_TARGET.FORMAT,
                Left = UpDownAngleTargetParam(leftContext),
                Right = UpDownAngleTargetParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(BREAST_UP_DOWN_ANGLE_TARGET.NAME)
                    .Append(" is the joint's vertical target angle. Negative values pull breasts down, positive values push them up.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftBaseAngle = new JointPhysicsParamPair(Script, RIGHT_LEFT_BASE_ANGLE.NAME, "Right/Left Angle")
            {
                ValueFormat = RIGHT_LEFT_BASE_ANGLE.FORMAT,
                Left = RightLeftBaseAngleParam(leftContext),
                Right = RightLeftBaseAngleParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_BASE_ANGLE.NAME)
                    .Append(" adjusts the pectoral joint's sideways rotation using morphs.\n\n")
                    .Append("Use the calibration state to pause real-time morphing.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftAngleLimit = new JointPhysicsParamPair(Script, RIGHT_LEFT_ANGLE_LIMIT.NAME)
            {
                ValueFormat = RIGHT_LEFT_ANGLE_LIMIT.FORMAT,
                Left = RightLeftAngleLimitParam(leftContext),
                Right = RightLeftAngleLimitParam(rightContext),
                needsCalibration = NeedsCalibration.IF_LDP_ENABLED,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_ANGLE_LIMIT.NAME)
                    .Append(" is the maximum angle the joint can rotate sideways.")
                    .ToString(),
            });

            paramPairs.Add(rightLeftAngleTarget = new JointPhysicsParamPair(Script, RIGHT_LEFT_ANGLE_TARGET.NAME)
            {
                ValueFormat = RIGHT_LEFT_ANGLE_TARGET.FORMAT,
                Left = RightLeftAngleTargetParam(leftContext),
                Right = RightLeftAngleTargetParam(rightContext),
                needsCalibration = NeedsCalibration.ALWAYS,
                rightInverted = true,
                infoText = new StringBuilder()
                    .AppendBold(RIGHT_LEFT_ANGLE_TARGET.NAME)
                    .Append(" is the joint's sideways target angle. A negative offset pulls breasts apart, while a positive offset pushes them together.")
                    .ToString(),
            });

            paramPairs.Add(twistAngleTarget = new JointPhysicsParamPair(Script, TWIST_ANGLE_TARGET.NAME)
            {
                ValueFormat = TWIST_ANGLE_TARGET.FORMAT,
                Left = TwistAngleTargetParam(leftContext, upDownAngleTarget.Left, rightLeftAngleTarget.Left),
                Right = TwistAngleTargetParam(rightContext, upDownAngleTarget.Right, rightLeftAngleTarget.Right),
                needsCalibration = NeedsCalibration.ALWAYS,
                rightInverted = true, // correct but has no effect because offset not adjusted directly
                infoText = new StringBuilder()
                    .AppendBold(TWIST_ANGLE_TARGET.NAME)
                    .Append(" is the joint's forward axis target angle.\n\n")
                    .Append("The final value depends on the person's pose. The offset determines the max angle when the person is upright.")
                    .Append(" The angle is inverted when upside down and zero when horizontal.")
                    .ToString(),
            });

            jointMass.SetCallbackFunctions();
            foreach(var pair in paramPairs)
            {
                pair.SetCallbackFunctions();
            }
        }

        JointPhysicsParam BreastMassParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            BREAST_JOINT_MASS.NAME,
            BREAST_JOINT_MASS.MIN,
            BREAST_JOINT_MASS.MAX,
            value => SyncMass(context.bodyPart.rigidbody, value)
        );

        JointPhysicsParam CenterOfGravityParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            CENTER_OF_GRAVITY.NAME,
            CENTER_OF_GRAVITY.MIN,
            CENTER_OF_GRAVITY.MAX,
            value => SyncCenterOfGravity(context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.51f,
                multiplierA = 0.30f,
                // TODO curveA
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = -0.06f,
                multiplierA = 0.24f,
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.06f,
                multiplierA = 0.24f,
            },
        };

        JointPhysicsParam RotationSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            ROTATION_SPRING.NAME,
            ROTATION_SPRING.MIN,
            ROTATION_SPRING.MAX,
            value => SyncRotationSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 97f,
                multiplierA = 0.17f,
                multiplierB = -0.65f,
                curveA = x => Calc.InverseSmoothStep(x, 0.46f, 0.10f),
                curveB = x => Calc.InverseSmoothStep(x, 0.10f, 0.13f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = 6f,
                multiplierA = 0.67f,
                curveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = -9f,
                multiplierA = 0.67f,
                curveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
            },
        };

        JointPhysicsParam RotationDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            ROTATION_DAMPER.NAME,
            ROTATION_DAMPER.MIN,
            ROTATION_DAMPER.MAX,
            value => SyncRotationDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 2.25f,
                multiplierA = 0.33f,
                multiplierB = -0.850f,
                curveA = x => Calc.InverseSmoothStep(x, 0.23f, 0.30f),
                curveB = x => Calc.InverseSmoothStep(x, 0.50f, 0.50f),
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = -0.08f,
                multiplierA = 8.375f,
                curveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.04f,
                multiplierA = 8.375f,
                curveA = x => Calc.InverseSmoothStep(x, 0.31f, 0.80f),
            },
        };

        // JointPhysicsParam AngularSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
        //     context.id,
        //     ANGULAR_SPRING.NAME,
        //     ANGULAR_SPRING.MIN,
        //     ANGULAR_SPRING.MAX,
        //     value => SyncAngularSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        // )
        // {
        //     PrimaryConfig = new MultiplierConfig(),
        // };

        // JointPhysicsParam AngularDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
        //     context.id,
        //     ANGULAR_DAMPER.NAME,
        //     ANGULAR_DAMPER.MIN,
        //     ANGULAR_DAMPER.MAX,
        //     value => SyncAngularDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        // )
        // {
        //     PrimaryConfig = new MultiplierConfig(),
        // };

        // JointPhysicsParam UpDownSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
        //     context.id,
        //     UP_DOWN_SPRING.NAME,
        //     UP_DOWN_SPRING.MIN,
        //     UP_DOWN_SPRING.MAX,
        //     value => SyncUpDownSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        // )
        // {
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 2000f,
        //     },
        // };

        // JointPhysicsParam UpDownDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
        //     context.id,
        //     UP_DOWN_DAMPER.NAME,
        //     UP_DOWN_DAMPER.MIN,
        //     UP_DOWN_DAMPER.MAX,
        //     value => SyncUpDownDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        // )
        // {
        //     PrimaryConfig = new MultiplierConfig
        //     {
        //         BaseValue = 10f,
        //     },
        // };

        JointPhysicsParam RightLeftSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_SPRING.NAME,
            RIGHT_LEFT_SPRING.MIN,
            RIGHT_LEFT_SPRING.MAX,
            value => SyncRightLeftSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 1500f,
                multiplierA = 0.50f,
                multiplierB = -0.77f,
                curveA = x => Calc.InverseSmoothStep(x, 0.53f, 0.70f),
                curveB = x => Calc.InverseSmoothStep(x, 0.60f, 0.62f),
            },
        };

        JointPhysicsParam RightLeftDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_DAMPER.NAME,
            RIGHT_LEFT_DAMPER.MIN,
            RIGHT_LEFT_DAMPER.MAX,
            value => SyncRightLeftDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 25f,
                multiplierB = -0.69f,
                curveB = x => Calc.InverseSmoothStep(x, 0.60f, 0.62f),
            },
        };

        JointPhysicsParam InOutSpringParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            IN_OUT_SPRING.NAME,
            IN_OUT_SPRING.MIN,
            IN_OUT_SPRING.MAX,
            value => SyncInOutSpring(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = -0.40f,
                curveA = x => 1 - x, // TODO smoothstep
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = 42f,
                multiplierA = -0.28f,
                curveA = x => 1 - x,
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = -63f,
                multiplierA = -0.28f,
                curveA = x => 1 - x,
            },
        };

        JointPhysicsParam InOutDamperParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            IN_OUT_DAMPER.NAME,
            IN_OUT_DAMPER.MIN,
            IN_OUT_DAMPER.MAX,
            value => SyncInOutDamper(context.bodyPart.joint, context.bodyPart.rigidbody, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 11.5f,
                multiplierB = -0.37f,
            },
            QuicknessConfig = new MultiplierConfig
            {
                baseValue = -1.33f,
                multiplierA = 1.33f,
            },
            SlownessConfig = new MultiplierConfig
            {
                baseValue = 1.33f,
                multiplierA = 1.33f,
            },
        };

        JointPhysicsParam UpDownBaseAngleParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            BREAST_UP_DOWN_BASE_ANGLE.NAME,
            BREAST_UP_DOWN_BASE_ANGLE.MIN,
            BREAST_UP_DOWN_BASE_ANGLE.MAX,
            context.id == Side.LEFT ? (Action<float>) SyncUpDownBaseAngle : null
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 9.0f,
            },
        };

        JointPhysicsParam UpAngleLimitParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            UP_ANGLE_LIMIT.NAME,
            UP_ANGLE_LIMIT.MIN,
            UP_ANGLE_LIMIT.MAX,
            value => SyncUpAngleLimit(context.bodyPart.joint, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 9.0f,
                multiplierA = 0.69f,
                multiplierB = -0.67f,
                curveA = x => Calc.InverseSmoothStep(x, -0.10f, 0.19f),
                curveB = x => Calc.InverseSmoothStep(1 - x, 0.37f, 0.72f),
            },
        };

        JointPhysicsParam DownAngleLimitParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            DOWN_ANGLE_LIMIT.NAME,
            DOWN_ANGLE_LIMIT.MIN,
            DOWN_ANGLE_LIMIT.MAX,
            value => SyncDownAngleLimit(context.bodyPart.joint, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = -12.7f,
                multiplierA = 0.69f,
                multiplierB = -0.66f,
                curveA = x => Calc.InverseSmoothStep(x, -0.17f, 0.15f),
                curveB = x => Calc.InverseSmoothStep(1 - x, 0.37f, 0.72f),
            },
        };

        static JointPhysicsParam UpDownAngleTargetParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            BREAST_UP_DOWN_ANGLE_TARGET.NAME,
            BREAST_UP_DOWN_ANGLE_TARGET.MIN,
            BREAST_UP_DOWN_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        JointPhysicsParam RightLeftBaseAngleParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_BASE_ANGLE.NAME,
            RIGHT_LEFT_BASE_ANGLE.MIN,
            RIGHT_LEFT_BASE_ANGLE.MAX,
            context.id == Side.LEFT ? (Action<float>) SyncRightLeftBaseAngle : null
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        JointPhysicsParam RightLeftAngleLimitParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_ANGLE_LIMIT.NAME,
            RIGHT_LEFT_ANGLE_LIMIT.MIN,
            RIGHT_LEFT_ANGLE_LIMIT.MAX,
            value => SyncRightLeftAngleLimit(context.bodyPart.joint, value)
        )
        {
            PrimaryConfig = new MultiplierConfig
            {
                baseValue = 12.0f,
                multiplierA = 0.67f,
                multiplierB = -0.74f,
                curveA = x => Calc.InverseSmoothStep(x, -0.10f, 0.12f),
                curveB = x => Calc.InverseSmoothStep(1 - x, 0.37f, 0.70f),
            },
        };

        static JointPhysicsParam RightLeftAngleTargetParam(JointPhysicsContext context) => new JointPhysicsParam(
            context.id,
            RIGHT_LEFT_ANGLE_TARGET.NAME,
            RIGHT_LEFT_ANGLE_TARGET.MIN,
            RIGHT_LEFT_ANGLE_TARGET.MAX
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        JointPhysicsParam TwistAngleTargetParam(
            JointPhysicsContext context,
            JointPhysicsParam upDownAngleTarget,
            JointPhysicsParam rightLeftAngleTarget
        ) => new JointPhysicsParam(
            context.id,
            TWIST_ANGLE_TARGET.NAME,
            TWIST_ANGLE_TARGET.MIN,
            TWIST_ANGLE_TARGET.MAX,
            value => SyncAngleTarget(
                context,
                upDownAngleTarget.Value,
                rightLeftAngleTarget.Value,
                value
            )
        )
        {
            PrimaryConfig = new MultiplierConfig(),
        };

        void SyncUpDownBaseAngle(float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            // TODO split morphs to left/right
            if(value > 0)
            {
                float morphValue = Mathf.InverseLerp(0, BREAST_UP_DOWN_BASE_ANGLE.MAX, value);
                _upAngleMorphS.SetValue(Calc.RoundToDecimals(morphValue * (1 - _lowMassInterp)));
                _upAngleMorphL.SetValue(Calc.RoundToDecimals(morphValue * _lowMassInterp * (1 - _midMassInterp)));
                _upAngleMorphXL.SetValue(Calc.RoundToDecimals(morphValue * _midMassInterp));
                _downAngleMorphS.Reset();
                _downAngleMorphL.Reset();
                _downAngleMorphXL.Reset();
            }
            else
            {
                float morphValue = Mathf.InverseLerp(0, -BREAST_UP_DOWN_BASE_ANGLE.MIN, -value);
                _downAngleMorphS.SetValue(Calc.RoundToDecimals(morphValue * (1 - _lowMassInterp)));
                _downAngleMorphL.SetValue(Calc.RoundToDecimals(morphValue * _lowMassInterp * (1 - _midMassInterp)));
                _downAngleMorphXL.SetValue(Calc.RoundToDecimals(morphValue * _midMassInterp));
                _upAngleMorphS.Reset();
                _upAngleMorphL.Reset();
                _upAngleMorphXL.Reset();
            }

            // TODO hacky
            upAngleLimit.SetExtraAdjustment(BREAST_UP_DOWN_BASE_ANGLE.NAME, -upDownBaseAngle.Left.offsetJsf.val);
        }

        void SyncRightLeftBaseAngle(float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            if(value > 0)
            {
                float morphValue = Calc.RoundToDecimals(Mathf.InverseLerp(0, RIGHT_LEFT_BASE_ANGLE.MAX, value));
                _sidewaysInMorphS.SetValue(Calc.RoundToDecimals(morphValue * (1 - _lowMassInterp)));
                _sidewaysInMorphL.SetValue(Calc.RoundToDecimals(morphValue * _lowMassInterp * (1 - _midMassInterp)));
                _sidewaysInMorphXL.SetValue(Calc.RoundToDecimals(morphValue * _midMassInterp));
                _sidewaysOutMorphS.Reset();
                _sidewaysOutMorphL.Reset();
                _sidewaysOutMorphXL.Reset();
            }
            else
            {
                float morphValue = Calc.RoundToDecimals(Mathf.InverseLerp(0, -RIGHT_LEFT_BASE_ANGLE.MIN, -value));
                _sidewaysOutMorphS.SetValue(Calc.RoundToDecimals(morphValue * (1 - _lowMassInterp)));
                _sidewaysOutMorphL.SetValue(Calc.RoundToDecimals(morphValue * _lowMassInterp * (1 - _midMassInterp)));
                _sidewaysOutMorphXL.SetValue(Calc.RoundToDecimals(morphValue * _midMassInterp));
                _sidewaysInMorphS.Reset();
                _sidewaysInMorphL.Reset();
                _sidewaysInMorphXL.Reset();
            }
        }

        void SyncAngleTarget(JointPhysicsContext context, float x, float y, float z)
        {
            if(!Script.enabled)
            {
                return;
            }

            bool noZRotation = twistAngleTargetValue == 0;
            if(context.rotationZZeroCount < 3)
            {
                if(noZRotation)
                {
                    context.rotationZZeroCount++;
                }
            }
            else if(!noZRotation)
            {
                context.rotationZZeroCount = 0;
            }

            SetBoneRotation(context.bodyPart.bone, x, y, z);
        }

        static void SetBoneRotation(DAZBone bone, float x, float y, float z)
        {
            var rotation = bone.baseJointRotation;
            rotation.x = -x;
            rotation.y = -y;
            rotation.z = z;
            if(rotation.IsEqual(bone.baseJointRotation))
            {
                return;
            }

            bone.baseJointRotation = rotation;
        }

        internal override void ResetAll()
        {
            _downAngleMorphS.Reset();
            _downAngleMorphL.Reset();
            _downAngleMorphXL.Reset();
            _upAngleMorphS.Reset();
            _upAngleMorphL.Reset();
            _upAngleMorphXL.Reset();
            _sidewaysInMorphS.Reset();
            _sidewaysInMorphL.Reset();
            _sidewaysInMorphXL.Reset();
            _sidewaysOutMorphS.Reset();
            _sidewaysOutMorphL.Reset();
            _sidewaysOutMorphXL.Reset();
        }

        internal override void Enable()
        {
            _xMotionEnabledJsb.val = true;
        }

        internal override void Disable()
        {
            if(Script.initialized != true)
            {
                return;
            }

            ResetAll();
            _xMotionEnabledJsb.val = false;
            RestoreOriginalPhysics();
            RestoreLimits(leftContext.bodyPart.joint);
            RestoreLimits(rightContext.bodyPart.joint);
        }

        static void RestoreLimits(ConfigurableJoint joint)
        {
            var highXLimit = joint.highAngularXLimit;
            var lowXLimit = joint.lowAngularXLimit;
            var yLimit = joint.angularYLimit;
            highXLimit.limit = 20;
            lowXLimit.limit = -15;
            yLimit.limit = 20;
            joint.highAngularXLimit = highXLimit;
            joint.lowAngularXLimit = lowXLimit;
            joint.angularYLimit = yLimit;
        }
    }
}
