using ColliderEditor;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.Common.UI.Windows;
using NAT.TM.Contexts;
using NAT.TM.Models;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using static NAT.SoftVerticesGroupName;

namespace NAT.TM.Handlers
{
    sealed class BreastSoftPhysicsHandler : SoftPhysicsHandler
    {
        readonly BreastSoftPhysicsContext _leftContext = new BreastSoftPhysicsContext(Side.LEFT);
        readonly BreastSoftPhysicsContext _rightContext = new BreastSoftPhysicsContext(Side.RIGHT);
        bool _pluginSoftPhysicsToggled;

        internal BreastSoftPhysicsHandler(ScriptBase script) : base(script)
        {
            physicsMesh = Person.BreastPhysicsMesh;
            leftContext = _leftContext;
            rightContext = _rightContext;
            SetupSoftVertices();
            softJointsFixedUpdate = new CustomFixedUpdate(Script.bootstrap.LdpUpdateRateJssc);
            softJointsFixedUpdate.actions += () => UpdateSoftJoints();
            SetupStorables();
            LoadSettings();
            SetupSoftJoints(_leftContext);
            SetupSoftJoints(_rightContext);
            colliderVisualizer = AddColliderVisualizer();
            softJointDrawService = AddDrawService();
            softPhysicsOn = softPhysicsOnJsb.val;
        }

        ColliderVisualizer AddColliderVisualizer() => Script.gameObject.AddComponent<ColliderVisualizer>(c =>
        {
            c.Init(
                _leftContext.allSoftJoints.SelectMany(softJoint => softJoint.GetColliders()),
                _rightContext.allSoftJoints.SelectMany(softJoint => softJoint.GetColliders()),
                new List<Group>
                {
                    new Group("Off", "$off"), //match nothing
                    new Group("Left breast", @"PhysicsMeshJointleft(areola|nipple|outer)?\d+$"),
                    new Group("Right breast", @"PhysicsMeshJointright(areola|nipple|outer)?\d+$"),
                    new Group("Both breasts", @"PhysicsMeshJoint(left|right)(areola|nipple|outer)?\d+$"),
                }
            );
            c.previewOpacityJSON.val = 0.40f;
            c.previewOpacityJSON.defaultVal = 0.40f;
            c.selectedPreviewOpacityJSON.val = 1;
            c.selectedPreviewOpacityJSON.defaultVal = 1;
            c.groupsJSON.label = "Show Colliders";
            c.groupsJSON.val = "Left breast";
            c.groupsJSON.defaultVal = "Left breast";
            c.highlightMirrorJSON.val = true;
            c.showPreviewsJSON.val = true;

            foreach(string option in new[] { Constant.SELECT, "Other", "All" })
            {
                c.groupsJSON.choices.Remove(option);
            }
        });

        void SetupSoftVertices()
        {
            SetupSoftVerticesGroups(_leftContext);
            SetupSoftVerticesGroups(_rightContext);
            SetupVertexIndices();
        }

        void SetupSoftVerticesGroups(BreastSoftPhysicsContext context)
        {
            var groups = physicsMesh.softVerticesGroups;
            string side = context.id;
            context.softVerticesGroups = new Dictionary<string, DAZPhysicsMeshSoftVerticesGroup>
            {
                { OUTER, groups.Find(group => group.name == $"{side}{OUTER}") },
                { MAIN, groups.Find(group => group.name == side) },
                { AREOLA, groups.Find(group => group.name == $"{side}{AREOLA}") },
                { NIPPLE, groups.Find(group => group.name == $"{side}{NIPPLE}") },
            };
        }

        void SetupSoftJoints(BreastSoftPhysicsContext context)
        {
            var allSJs = new List<SoftJoint>();
            var exceptNippleSJs = new List<SoftJoint>();
            var outerSJs = new List<SoftJoint>();
            var mainSJs = new List<SoftJoint>();
            var areolaSJs = new List<SoftJoint>();
            var nippleSJs = new List<SoftJoint>();

            foreach(var pair in context.softVerticesGroups)
            {
                string category = pair.Key;
                var group = pair.Value;
                foreach(var set in group.softVerticesSets)
                {
                    string id = $"{context.id}.{category}.{set.targetVertex.ToString()}";
                    var softJoint = new SoftJoint(id, set);
                    softJoint.SetPhysicsParams(
                        fatMass,
                        fatSpring,
                        fatDamper,
                        fatColliderRadius,
                        fatColliderDepth,
                        fatColliderFriction,
                        fatDistanceLimit
                    );
                    switch(category)
                    {
                        case OUTER:
                            outerSJs.Add(softJoint);
                            break;
                        case MAIN:
                            mainSJs.Add(softJoint);
                            break;
                        case AREOLA:
                            areolaSJs.Add(softJoint);
                            break;
                        case NIPPLE:
                            nippleSJs.Add(softJoint);
                            break;
                    }

                    allSJs.Add(softJoint);
                    if(category != NIPPLE)
                    {
                        exceptNippleSJs.Add(softJoint);
                    }
                }
            }

            context.allSoftJoints = allSJs.ToArray();
            context.outerSoftJoints = outerSJs.ToArray();
            context.mainSoftJoints = mainSJs.ToArray();
            context.centerSoftJoints = areolaSJs.ToArray();
            context.nippleSoftJoint = nippleSJs.First();

            var nearbyVerts = VertexIndices.NearbyVerts(context.id);
            var adjacentVerts = VertexIndices.AdjacentVerts(context.id);
            foreach(var softJoint in allSJs)
            {
                int target = softJoint.softVerticesSet.targetVertex;
                int[] vert;
                softJoint.nearbyJoints = nearbyVerts.TryGetValue(target, out vert)
                    ? allSJs.Where(x => vert.Contains(x.softVerticesSet.targetVertex)).ToArray()
                    : new SoftJoint[] { }; // nipple verts do not have nearby verts

                int[] adjacentVert;
                softJoint.adjacentJoints = adjacentVerts.TryGetValue(target, out adjacentVert)
                    ? allSJs.Where(x => adjacentVert.Contains(x.softVerticesSet.targetVertex)).ToArray()
                    : new SoftJoint[] { }; // areola and nipple verts do not have adjacent verts
            }

            var gradientGroupsList = new List<GradientGroup>();
            foreach(int[] indexGroup in VertexIndices.CenterToOuterVerts(context.id))
            {
                var gradientGroup = new BreastGradientGroup
                {
                    center = areolaSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[0])),
                    main1 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[1])),
                    main2 = mainSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[2])),
                    outer1 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[3])),
                };

                if(indexGroup.Length > 4)
                {
                    gradientGroup.outer2 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[4]));
                }

                if(indexGroup.Length > 5)
                {
                    gradientGroup.outer3 = outerSJs.Find(softJoint => MatchByVertex(softJoint, indexGroup[5]));
                }

                gradientGroupsList.Add(gradientGroup);
            }

            context.gradientGroups = gradientGroupsList;

            // TODO hard code
            context.nonGradientJoints = exceptNippleSJs
                .Where(softJoint => !context.gradientGroups.Any(x => x.HasJoint(softJoint)))
                .ToArray();
        }

        static bool MatchByVertex(SoftJoint softJoint, int vertex) => softJoint.softVerticesSet.targetVertex == vertex;

        void SetupStorables()
        {
            softPhysicsOnJsb = new StorableBool(Constant.BREAST_SOFT_PHYSICS_ENABLED, physicsMesh.on); // TODO refactor away
            bool ldpEnabled = physicsMesh.on && Script.bootstrap.BreastsLdpDefaultEnabledJsb.val;
            ldpEnabledJsb = new StorableBool("Local Distortion Physics Enabled", ldpEnabled);
            softPhysicsSelfCollideJsb = new StorableBool("Breast Soft Physics Self Collide", true);
            showColliderPreviewsJsb = new JSONStorableBool("Show Collider Previews", true);

            softPhysicsOnJsb.setCallbackFunction = OnToggleSoftPhysics;
            ldpEnabledJsb.setJSONCallbackFunction = OnToggleLdp;
            softPhysicsSelfCollideJsb.setCallbackFunction = SyncAllowSelfCollision;
            showColliderPreviewsJsb.setCallbackFunction = _ =>
            {
                if(Script.uiHandler.IsPhysicsTabActive() && Script.uiHandler.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    ((SoftParameterWindow) Script.uiHandler.GetActiveNestedWindow()).SetShowColliderPreviews();
                }
            };
        }

        void OnToggleSoftPhysics(bool value)
        {
            if(Script.calibrationStateHandler.CalibratingJsb.val)
            {
                Script.Loggr.LogMessage("Cannot enable/disable soft physics while calibrating.");
                softPhysicsOnJsb.valNoCallback = !value;
                return;
            }

            if(value)
            {
                ldpEnabledJsb.val = ldpEnabledJsb.val || Script.bootstrap.BreastsLdpDefaultEnabledJsb.val;
            }
            else
            {
                ldpEnabledJsb.val = false;
            }

            if(!_reverseSyncingSoftPhysics)
            {
                _pluginSoftPhysicsToggled = true;
                if(Script.IsEnabledAndActive())
                {
                    physicsMesh.on = value;
                }
                else
                {
                    Script.Loggr.LogMessage($"Enable the plugin to set {softPhysicsOnJsb.name}.");
                }

                if(Script.uiHandler.IsMainTabActive())
                {
                    Script.uiHandler.mainWindow.Refresh();
                }
            }

            _reverseSyncingSoftPhysics = false;
        }

        void OnToggleLdp(JSONStorableBool jsb)
        {
            if(!Script.IsEnabledAndActive())
            {
                Script.Loggr.LogMessage($"Enable the plugin to set {jsb.name}.");
                return;
            }

            var storable = (StorableBool) jsb;
            if(storable.val)
            {
                if(!softPhysicsOnJsb.val)
                {
                    Script.Loggr.LogMessage("Enable breast soft physics to use local distortion physics.");
                }
                else
                {
                    Script.SetCalibrationPending(storable);
                }
            }
            else
            {
                if(softPhysicsOn)
                {
                    Script.SetCalibrationPending(storable);
                }
            }

            softJointDrawService.visualizeLdpJsb.val = jsb.val;

            if(Script.uiHandler.IsMainTabActive())
            {
                Script.uiHandler.mainWindow.Refresh();
            }
        }

        protected override void LoadSettings()
        {
            softParamPairs.Add(fatSpring = new SoftPhysicsParamPair(Script, this, FAT_SPRING.NAME)
            {
                Left = FatSpringParam(_leftContext),
                Right = FatSpringParam(_rightContext),
                ValueFormat = FAT_SPRING.FORMAT,
                visualizeRelativeChange = true,
                visualizationLimit = 125f,
                clothingAdjustmentVisualizationLimit = 1000f,
                distortionLimit = 500f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_SPRING.NAME)
                    .AppendLine(" holds each soft joint in its target position.")
                    .AppendLine()
                    .Append("Low fat spring makes breast fat soft and slow. High fat spring makes it rigid and quick to return to its normal shape.")
                    .ToString(),
            });

            softParamPairs.Add(fatDamper = new SoftPhysicsParamPair(Script, this, FAT_DAMPER.NAME)
            {
                Left = FatDamperParam(_leftContext),
                Right = FatDamperParam(_rightContext),
                ValueFormat = FAT_DAMPER.FORMAT,
                visualizeRelativeChange = true,
                visualizationLimit = 2.50f,
                clothingAdjustmentVisualizationLimit = 5.00f,
                distortionLimit = 5.00f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_DAMPER.NAME)
                    .AppendLine(" reduces oscillation of each soft joint around its target position.")
                    .AppendLine()
                    .Append("Low fat damper makes breast fat jiggle more easily.")
                    .ToString(),
            });

            softParamPairs.Add(fatMass = new SoftPhysicsParamPair(Script, this, FAT_MASS.NAME)
            {
                Left = FatMassParam(_leftContext),
                Right = FatMassParam(_rightContext),
                ValueFormat = FAT_MASS.FORMAT,
                visualizeRelativeChange = true,
                visualizationLimit = 0.150f,
                clothingAdjustmentVisualizationLimit = 0.150f,
                distortionLimit = 0.300f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_MASS.NAME)
                    .Append(" determines the mass of each soft joint and the overall density of breast fat.")
                    .Append(" Increasing breast size without increasing fat mass reduces density.")
                    .ToString(),
            });

            softParamPairs.Add(fatColliderRadius = new SoftPhysicsParamPair(Script, this, FAT_COLLIDER_RADIUS.NAME)
            {
                Left = FatColliderRadiusParam(_leftContext),
                Right = FatColliderRadiusParam(_rightContext),
                ValueFormat = FAT_COLLIDER_RADIUS.FORMAT,
                usesRealMass = true,
                hasSoftColliderVisualization = true,
                visualizeRelativeChange = false,
                visualizationLimit = 0.0100f,
                clothingAdjustmentVisualizationLimit = 0.0100f,
                distortionLimit = 0.0200f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_RADIUS.NAME)
                    .AppendLine(" is a scaling factor for the size of the colliders connected to each soft joint.")
                    .AppendLine()
                    .Append("The actual size varies per collider and automatically scales with breast size.")
                    .ToString(),
            });

            softParamPairs.Add(fatColliderDepth = new SoftPhysicsParamPair(Script, this, FAT_COLLIDER_DEPTH.NAME)
            {
                ValueFormat = FAT_COLLIDER_DEPTH.FORMAT,
                Left = FatColliderDepthParam(_leftContext),
                Right = FatColliderDepthParam(_rightContext),
                hasSoftColliderVisualization = true,
                visualizeRelativeChange = false,
                visualizationLimit = 0.0020f,
                clothingAdjustmentVisualizationLimit = 0.0020f,
                distortionLimit = 0.0020f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_DEPTH.NAME)
                    .AppendLine(" offsets collider positions relative to skin surface")
                    .AppendLine()
                    .Append("Negative values pull colliders out from the breast, positive values push them into the breast.")
                    .ToString(),
            });

            softParamPairs.Add(fatColliderFriction = new SoftPhysicsParamPair(Script, this, FAT_COLLIDER_FRICTION.NAME)
            {
                ValueFormat = FAT_COLLIDER_FRICTION.FORMAT,
                Left = FatColliderFrictionParam(_leftContext),
                Right = FatColliderFrictionParam(_rightContext),
                hasSoftColliderVisualization = true,
                visualizeRelativeChange = true,
                visualizationLimit = 0.500f,
                clothingAdjustmentVisualizationLimit = 0.500f,
                distortionLimit = 1.000f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_COLLIDER_FRICTION.NAME)
                    .Append(" sets the stickiness of individual soft colliders. Click below to configure adaptive friction and to offset the amount of friction.")
                    .ToString(),
            });

            softParamPairs.Add(fatDistanceLimit = new SoftPhysicsParamPair(Script, this, FAT_DISTANCE_LIMIT.NAME)
            {
                ValueFormat = FAT_DISTANCE_LIMIT.FORMAT,
                Left = FatDistanceLimitParam(_leftContext),
                Right = FatDistanceLimitParam(_rightContext),
                usesRealMass = true,
                visualizeRelativeChange = true,
                visualizationLimit = 0.020f,
                clothingAdjustmentVisualizationLimit = 0.020f,
                distortionLimit = 0.020f,
                infoText = new StringBuilder()
                    .AppendBold(FAT_DISTANCE_LIMIT.NAME)
                    .Append(" is the maximum distance each soft joint can move away from its target position.")
                    .ToString(),
            });

            paramPairs.Add(fatBackForce = new JointPhysicsParamPair(Script, FAT_BACK_FORCE.NAME)
            {
                ValueFormat = FAT_BACK_FORCE.FORMAT,
                Left = FatBackForceParam(_leftContext),
                Right = FatBackForceParam(_rightContext),
                infoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE.NAME)
                    .Append(" applies a force to the pectoral joint based on the movement of soft joints. A high value can cause an out of control feedback loop.")
                    .ToString(),
            });

            paramPairs.Add(fatBackForceMaxForce = new JointPhysicsParamPair(Script, FAT_BACK_FORCE_MAX_FORCE.NAME)
            {
                ValueFormat = FAT_BACK_FORCE_MAX_FORCE.FORMAT,
                Left = FatBackForceMaxForceParam(_leftContext),
                Right = FatBackForceMaxForceParam(_rightContext),
                infoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE_MAX_FORCE.NAME)
                    .Append(" sets an upper limit on the magnitude of back force.")
                    .ToString(),
            });

            paramPairs.Add(fatBackForceThreshold = new JointPhysicsParamPair(Script, FAT_BACK_FORCE_THRESHOLD.NAME)
            {
                ValueFormat = FAT_BACK_FORCE_THRESHOLD.FORMAT,
                Left = FatBackForceThresholdParam(_leftContext),
                Right = FatBackForceThresholdParam(_rightContext),
                infoText = new StringBuilder()
                    .AppendBold(FAT_BACK_FORCE_THRESHOLD.NAME)
                    .AppendLine(" sets a minimum distance each soft joint needs to move for back force to be applied.")
                    .AppendLine()
                    .AppendFormat("Along with {0}, this can be used to prevent an out of control feedback loop.", FAT_BACK_FORCE_MAX_FORCE.NAME)
                    .ToString(),
            });

            foreach(var pair in softParamPairs)
            {
                pair.SetCallbackFunctions();
            }

            foreach(var pair in paramPairs)
            {
                pair.SetCallbackFunctions();
            }
        }

        internal override Rigidbody[] GetLeftPositionalRigidbodies(string group) => GetPositionalRigidbodies(_leftContext, group);
        internal override Rigidbody[] GetRightPositionalRigidbodies(string group) => GetPositionalRigidbodies(_rightContext, group);

        static Rigidbody[] GetPositionalRigidbodies(BreastSoftPhysicsContext context, string group)
        {
            var rigidbodies = new List<Rigidbody>();
            int[] indices = context.positionalIndices[group];
            foreach(var set in context.softVerticesGroups[group].softVerticesSets)
            {
                if(indices.Contains(set.targetVertex))
                {
                    rigidbodies.Add(set.jointRB);
                }
            }

            return rigidbodies.ToArray();
        }

        static float FatSpringSoftnessCurve(float x) => Calc.InverseSmoothStep(x, 0.35f, 0.11f);

        SoftPhysicsParam NewSoftPhysicsParameter(BreastSoftPhysicsContext context, string name, float min, float max) => new SoftPhysicsParam(
            Script.DomainName,
            context.id,
            name,
            min,
            max,
            (param, additionalOffset) => SyncGradientParam(context, param, additionalOffset)
        );

        SoftPhysicsParam FatSpringParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_SPRING.NAME, FAT_SPRING.MIN, FAT_SPRING.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.959f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 1000f,
                multiplierA = 0.00f,
                multiplierB = -0.954f,
                curveB = FatSpringSoftnessCurve,
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                baseValue = 50f,
                multiplierA = 0.00f,
                multiplierB = -0.75f,
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                baseValue = -125f,
                multiplierA = 0.00f,
                multiplierB = -0.96f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.40f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(2.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(2.25f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.30f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.85f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.30f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.85f);

            return parameter;
        }

        SoftPhysicsParam FatDamperParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_DAMPER.NAME, FAT_DAMPER.MIN, FAT_DAMPER.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 1.60f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 1.60f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                baseValue = -1.60f,
                multiplierA = 0.18f,
                multiplierB = -0.55f,
                curveA = x => Calc.Exponential2(x, 0.04f, 0.04f),
                curveB = x => Calc.Exponential1(x, 1.90f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(2.30f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.80f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(-0.40f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(-0.40f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        SoftPhysicsParam FatMassParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_MASS.NAME, FAT_MASS.MIN, FAT_MASS.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.045f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 0.045f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveA = x => Calc.Exponential1(x, 1.91f, 1.7f, 0.82f),
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.QuicknessConfig = new MultiplierConfig
            {
                baseValue = -0.038f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };
            parameter.SlownessConfig = new MultiplierConfig
            {
                baseValue = 0.038f,
                multiplierA = 1.31f,
                multiplierB = 0.60f,
                curveB = x => Calc.Exponential1(x, 2.30f, 1.74f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.50f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(-0.25f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        SoftPhysicsParam FatColliderRadiusParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_COLLIDER_RADIUS.NAME, FAT_COLLIDER_RADIUS.MIN, FAT_COLLIDER_RADIUS.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.0142f,
                multiplierA = 1.80f,
                multiplierB = 0.20f,
                curveA = x => Calc.Exponential1(x, 1.42f, 4.25f, 1.17f),
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 0.0153f,
                multiplierA = 1.80f,
                multiplierB = 0.20f,
                curveA = x => Calc.Exponential1(x, 1.42f, 4.25f, 1.17f),
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.10f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.25f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(0.33f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.40f;
            gradientConfig.middleToCenterMidPoint = 0.70f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.10f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.70f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.40f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.70f);

            return parameter;
        }

        SoftPhysicsParam FatColliderDepthParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_COLLIDER_DEPTH.NAME, FAT_COLLIDER_DEPTH.MIN, FAT_COLLIDER_DEPTH.MAX);
            parameter.PrimaryConfig = new MultiplierConfig();
            parameter.SecondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.30f;
            gradientConfig.middleToCenterMidPoint = 0.50f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(1.00f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.90f);

            return parameter;
        }

        SoftPhysicsParam FatColliderFrictionParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_COLLIDER_FRICTION.NAME, FAT_COLLIDER_FRICTION.MIN, FAT_COLLIDER_FRICTION.MAX);
            parameter.PrimaryConfig = new MultiplierConfig();
            parameter.SecondaryConfig = new MultiplierConfig();

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(0.00f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(1.00f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(1.25f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(1.50f);
            gradientConfig.outerToMiddleCurvature = 0.60f;
            gradientConfig.outerToMiddleMidPoint = 0.60f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.50f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(0.30f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.50f);

            return parameter;
        }

        SoftPhysicsParam FatDistanceLimitParam(BreastSoftPhysicsContext context)
        {
            var parameter = NewSoftPhysicsParameter(context, FAT_DISTANCE_LIMIT.NAME, FAT_DISTANCE_LIMIT.MIN, FAT_DISTANCE_LIMIT.MAX);
            parameter.PrimaryConfig = new MultiplierConfig
            {
                baseValue = 0.0155f,
                multiplierA = 3.68f,
                multiplierB = 0.52f,
            };
            parameter.SecondaryConfig = new MultiplierConfig
            {
                baseValue = 0.020f,
                multiplierA = 3.68f,
                multiplierB = 0.52f,
            };

            var gradientConfig = parameter.gradientConfig;
            gradientConfig.middleShiftJsf.SetValueAndDefault(-0.80f);
            gradientConfig.outerMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.centerMultiplierJsf.SetValueAndDefault(0.80f);
            gradientConfig.nippleMultiplierJsf.SetValueAndDefault(0.86f);
            gradientConfig.outerToMiddleCurvature = 0.30f;
            gradientConfig.outerToMiddleMidPoint = 0.50f;
            gradientConfig.middleToCenterCurvature = 0.30f;
            gradientConfig.middleToCenterMidPoint = 0.50f;

            var distortionConfig = parameter.distortionConfig;
            distortionConfig.compressionEffectJsf.SetValueAndDefault(0.00f);
            distortionConfig.compressionSensitivityJsf.SetValueAndDefault(0.80f);
            distortionConfig.stretchingEffectJsf.SetValueAndDefault(1.00f);
            distortionConfig.stretchingSensitivityJsf.SetValueAndDefault(0.80f);

            return parameter;
        }

        JointPhysicsParam FatBackForceParam(BreastSoftPhysicsContext context)
        {
            var parameter = new JointPhysicsParam(
                context.id,
                FAT_BACK_FORCE.NAME,
                FAT_BACK_FORCE.MIN,
                FAT_BACK_FORCE.MAX,
                value => SyncBackForce(context, value)
            )
            {
                PrimaryConfig = new MultiplierConfig
                {
                    baseValue = 0,
                },
                SlownessConfig = new MultiplierConfig
                {
                    baseValue = 0.72f,
                    multiplierA = 8.25f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        JointPhysicsParam FatBackForceMaxForceParam(BreastSoftPhysicsContext context)
        {
            var parameter = new JointPhysicsParam(
                context.id,
                FAT_BACK_FORCE_MAX_FORCE.NAME,
                FAT_BACK_FORCE_MAX_FORCE.MIN,
                FAT_BACK_FORCE_MAX_FORCE.MAX,
                value => SyncBackForceMaxForce(context, value)
            )
            {
                PrimaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                SecondaryConfig = new MultiplierConfig
                {
                    baseValue = 50.00f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        JointPhysicsParam FatBackForceThresholdParam(BreastSoftPhysicsContext context)
        {
            var parameter = new JointPhysicsParam(
                context.id,
                FAT_BACK_FORCE_THRESHOLD.NAME,
                FAT_BACK_FORCE_THRESHOLD.MIN,
                FAT_BACK_FORCE_THRESHOLD.MAX,
                value => SyncBackForceThresholdDistance(context, value)
            )
            {
                PrimaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
                SecondaryConfig = new MultiplierConfig
                {
                    baseValue = 0.001f,
                    multiplierA = 0.00f,
                    multiplierB = 0.00f,
                },
            };

            return parameter;
        }

        void SyncGradientParam(BreastSoftPhysicsContext context, SoftPhysicsParam param, float additionalOffset)
        {
            if(Script.initialized != true || !Script.enabled || Script.restoringFromJSON)
            {
                // Syncing parameter before first time calibration (e.g. due to offset storable callback) results in NaN errors
                return;
            }

            string name = param.paramName;
            float value = param.baseValueJsf.val + param.offsetJsf.val;
            foreach(var group in context.gradientGroups)
            {
                group.SetGradientBaseValues(name, param.gradientConfig, value);
                group.SetAdditionalOffsets(name, additionalOffset);
            }

            /* Calculate remaining physics values for joints not included in gradients.
             * Values are based on neighboring (immediately adjacent) joints on the same "ring".
             */
            foreach(var joint in context.nonGradientJoints)
            {
                float average = 0f;
                foreach(var adjacentJoint in joint.adjacentJoints)
                {
                    average += adjacentJoint.GetBaseValue(name);
                }

                average /= joint.adjacentJoints.Length;

                joint.SetBaseValue(name, average);
                joint.SetAdditionalOffset(name, additionalOffset);
            }

            context.nippleSoftJoint.SetBaseValue(name, value * param.gradientConfig.nippleMultiplierJsf.val);
            context.nippleSoftJoint.SetAdditionalOffset(name, additionalOffset);
        }

        void SyncBackForce(BreastSoftPhysicsContext context, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            foreach(var pair in context.softVerticesGroups)
            {
                pair.Value.jointBackForce = value;
            }
        }

        void SyncBackForceMaxForce(BreastSoftPhysicsContext context, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            foreach(var pair in context.softVerticesGroups)
            {
                pair.Value.jointBackForceMaxForce = value;
            }
        }

        void SyncBackForceThresholdDistance(BreastSoftPhysicsContext context, float value)
        {
            if(!Script.enabled)
            {
                return;
            }

            foreach(var pair in context.softVerticesGroups)
            {
                pair.Value.jointBackForceThresholdDistance = value;
            }
        }

        bool _reverseSyncingSoftPhysics;

        internal override void ReverseSyncSoftPhysicsOn()
        {
            if(!_pluginSoftPhysicsToggled)
            {
                _reverseSyncingSoftPhysics = true;
                softPhysicsOnJsb.valNoCallback = physicsMesh.on;
            }

            _pluginSoftPhysicsToggled = false;
        }

        internal override void ReverseSyncAllowSelfCollision() =>
            softPhysicsSelfCollideJsb.valNoCallback = physicsMesh.allowSelfCollision;

        internal override void SyncSoftPhysicsOn()
        {
            softPhysicsOn = Script.bootstrap.PersonMonitor.IsBreastSoftPhysicsOn();
            ldpOn = softPhysicsOn && ldpEnabledJsb.val;
            if(!softPhysicsSynced)
            {
                if(softPhysicsOn)
                {
                    Script.hardColliderHandler.OnSoftPhysicsEnabled(leftContext.allSoftJoints, _rightContext.allSoftJoints);
                }
                else
                {
                    Script.hardColliderHandler.OnSoftPhysicsDisabled();
                }

                Script.SetCalibrationTargetFunctions(softPhysicsOn);
            }

            SyncAllowSelfCollision(softPhysicsSelfCollideJsb.val);
            softPhysicsSynced = true;
        }

        void SyncAllowSelfCollision(bool value)
        {
            if(!Script.enabled)
            {
                Script.Loggr.LogMessage($"Enable the plugin to set {softPhysicsSelfCollideJsb.name}.");
                return;
            }

            physicsMesh.allowSelfCollision = value;
        }

        bool _originalAllowSelfCollision;

        internal override void Enable()
        {
            _originalAllowSelfCollision = physicsMesh.allowSelfCollision;
            EnableMultiplyFriction();
        }

        internal override void Disable()
        {
            if(Script.initialized != true)
            {
                return;
            }

            colliderVisualizer.enabled = false;
            softJointDrawService.enabled = false;
            physicsMesh.allowSelfCollision = _originalAllowSelfCollision;
            RestoreOriginalFriction();
            RestoreOriginalPhysics();
        }

        void SetupVertexIndices()
        {
            leftContext.positionalIndices[OUTER] = new[] { 17, 137, 2403, 2407, 2410, 2418, 2596, 7248, 8848, 8849 };
            leftContext.positionalIndices[MAIN] = new[]
            {
                7230, 7231, 7232, 7233, 7234, 7235, 7236, 7237, 7238, 7239, 7243, 7244, 7245,
                7246, 7247, 8925, 8926, 8927, 8928, 8929, 8930, 8931, 8932, 8933, 8934, 8935,
                8936, 8937, 8938, 8939, 8940, 8941, 8942, 8943, 8944, 8945, 8946, 8947, 8948,
            };
            leftContext.positionalIndices[AREOLA] = new[] { 7945, 7946, 7947, 7950, 7958, 7966 };

            rightContext.positionalIndices[OUTER] = new[] { 10945, 11065, 13233, 13237, 13240, 13248, 13412, 17954, 19519, 19520 };
            rightContext.positionalIndices[MAIN] = new[]
            {
                17936, 17937, 17938, 17939, 17940, 17941, 17942, 17943, 17944, 17945, 17949, 17950, 17951,
                17952, 17953, 19591, 19592, 19593, 19594, 19595, 19596, 19597, 19598, 19599, 19600, 19601,
                19602, 19603, 19604, 19605, 19606, 19607, 19608, 19609, 19610, 19611, 19612, 19613, 19614,
            };
            rightContext.positionalIndices[AREOLA] = new[] { 18623, 18624, 18625, 18628, 18636, 18644 };

            leftContext.mirrorVertexIndices = new Dictionary<int, int>
            {
                { 19601, 8935 },
                { 19603, 8937 },
                { 19602, 8936 },
                { 19606, 8940 },
                { 19597, 8931 },
                { 19592, 8926 },
                { 19594, 8928 },
                { 19596, 8930 },
                { 19595, 8929 },
                { 19593, 8927 },
                { 19591, 8925 },
                { 19614, 8948 },
                { 19613, 8947 },
                { 19612, 8946 },
                { 19611, 8945 },
                { 19610, 8944 },
                { 19607, 8941 },
                { 19609, 8943 },
                { 19608, 8942 },
                { 19605, 8939 },
                { 19604, 8938 },
                { 19598, 8932 },
                { 19599, 8933 },
                { 19600, 8934 },
                { 19523, 8852 },
                { 19522, 8851 },
                { 19520, 8849 },
                { 19521, 8850 },
                { 17954, 7248 },
                { 17952, 7246 },
                { 17951, 7245 },
                { 17950, 7244 },
                { 17949, 7243 },
                { 17953, 7247 },
                { 17936, 7230 },
                { 17937, 7231 },
                { 17938, 7232 },
                { 17939, 7233 },
                { 17940, 7234 },
                { 17941, 7235 },
                { 17942, 7236 },
                { 17943, 7237 },
                { 17944, 7238 },
                { 17945, 7239 },
                { 17946, 7240 },
                { 17948, 7242 },
                { 17947, 7241 },
                { 19519, 8848 },
                { 10944, 16 },
                { 19510, 8839 },
                { 19525, 8854 },
                { 19524, 8853 },
                { 19508, 8837 },
                { 19509, 8838 },
                { 10936, 8 },
                { 13234, 2404 },
                { 13233, 2403 },
                { 13231, 2401 },
                { 13232, 2402 },
                { 13235, 2405 },
                { 13239, 2409 },
                { 13243, 2413 },
                { 13241, 2411 },
                { 13244, 2414 },
                { 13246, 2416 },
                { 13249, 2419 },
                { 11066, 138 },
                { 11065, 137 },
                { 13409, 2591 },
                { 13412, 2596 },
                { 13413, 2597 },
                { 10945, 17 },
                { 11063, 135 },
                { 11064, 136 },
                { 13410, 2592 },
                { 13411, 2594 },
                { 13414, 2598 },
                { 13408, 2590 },
                { 13248, 2418 },
                { 13245, 2415 },
                { 13247, 2417 },
                { 13240, 2410 },
                { 13242, 2412 },
                { 13238, 2408 },
                { 13237, 2407 },
                { 13236, 2406 },
                { 11021, 93 },
                { 17923, 7216 },
                { 17924, 7218 },
                { 17925, 7219 },
                { 17926, 7220 },
                { 17927, 7221 },
                { 17928, 7222 },
                { 17929, 7223 },
                { 17932, 7226 },
                { 17930, 7224 },
                { 17931, 7225 },
                { 13672, 2871 },
                { 18622, 7944 },
                { 18624, 7946 },
                { 18643, 7965 },
                { 18628, 7950 },
                { 18630, 7952 },
                { 18644, 7966 },
                { 18642, 7964 },
                { 18636, 7958 },
                { 18632, 7954 },
                { 18625, 7947 },
                { 18637, 7959 },
                { 18623, 7945 },
                { 10928, 0 },
            };

            rightContext.mirrorVertexIndices = new Dictionary<int, int>
            {
                { 8935, 19601 },
                { 8937, 19603 },
                { 8936, 19602 },
                { 8940, 19606 },
                { 8931, 19597 },
                { 8926, 19592 },
                { 8928, 19594 },
                { 8930, 19596 },
                { 8929, 19595 },
                { 8927, 19593 },
                { 8925, 19591 },
                { 8948, 19614 },
                { 8947, 19613 },
                { 8946, 19612 },
                { 8945, 19611 },
                { 8944, 19610 },
                { 8941, 19607 },
                { 8943, 19609 },
                { 8942, 19608 },
                { 8939, 19605 },
                { 8938, 19604 },
                { 8932, 19598 },
                { 8933, 19599 },
                { 8934, 19600 },
                { 8852, 19523 },
                { 8851, 19522 },
                { 8849, 19520 },
                { 8850, 19521 },
                { 7248, 17954 },
                { 7246, 17952 },
                { 7245, 17951 },
                { 7244, 17950 },
                { 7243, 17949 },
                { 7247, 17953 },
                { 7230, 17936 },
                { 7231, 17937 },
                { 7232, 17938 },
                { 7233, 17939 },
                { 7234, 17940 },
                { 7235, 17941 },
                { 7236, 17942 },
                { 7237, 17943 },
                { 7238, 17944 },
                { 7239, 17945 },
                { 7240, 17946 },
                { 7242, 17948 },
                { 7241, 17947 },
                { 8848, 19519 },
                { 16, 10944 },
                { 8839, 19510 },
                { 8854, 19525 },
                { 8853, 19524 },
                { 8837, 19508 },
                { 8838, 19509 },
                { 8, 10936 },
                { 2404, 13234 },
                { 2403, 13233 },
                { 2401, 13231 },
                { 2402, 13232 },
                { 2405, 13235 },
                { 2409, 13239 },
                { 2413, 13243 },
                { 2411, 13241 },
                { 2414, 13244 },
                { 2416, 13246 },
                { 2419, 13249 },
                { 138, 11066 },
                { 137, 11065 },
                { 2591, 13409 },
                { 2596, 13412 },
                { 2597, 13413 },
                { 17, 10945 },
                { 135, 11063 },
                { 136, 11064 },
                { 2592, 13410 },
                { 2594, 13411 },
                { 2598, 13414 },
                { 2590, 13408 },
                { 2418, 13248 },
                { 2415, 13245 },
                { 2417, 13247 },
                { 2410, 13240 },
                { 2412, 13242 },
                { 2408, 13238 },
                { 2407, 13237 },
                { 2406, 13236 },
                { 93, 11021 },
                { 7216, 17923 },
                { 7218, 17924 },
                { 7219, 17925 },
                { 7220, 17926 },
                { 7221, 17927 },
                { 7222, 17928 },
                { 7223, 17929 },
                { 7226, 17932 },
                { 7224, 17930 },
                { 7225, 17931 },
                { 2871, 13672 },
                { 7944, 18622 },
                { 7946, 18624 },
                { 7965, 18643 },
                { 7950, 18628 },
                { 7952, 18630 },
                { 7966, 18644 },
                { 7964, 18642 },
                { 7958, 18636 },
                { 7954, 18632 },
                { 7947, 18625 },
                { 7959, 18637 },
                { 7945, 18623 },
                { 0, 10928 },
            };
        }
    }
}
