using NAT.Common;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using UnityEngine;

namespace NAT.TM.Models
{
    sealed class BreastGradientGroup : GradientGroup
    {
        internal SoftJoint center;
        internal SoftJoint main1;
        internal SoftJoint main2;
        internal SoftJoint outer1;
        internal SoftJoint outer2;
        internal SoftJoint outer3;

        // ratios of distances to total distance
        float _main1Ratio;
        float _main2Ratio;
        float _outer1Ratio;
        float _outer2Ratio;

        internal override void CalculateDistances()
        {
            float total;
            float areolaToMain1 = Vector3.Distance(center.jointRb.position, main1.jointRb.position);
            float main1ToMain2 = Vector3.Distance(main1.jointRb.position, main2.jointRb.position);
            float main2ToOuter1 = Vector3.Distance(main2.jointRb.position, outer1.jointRb.position);
            if(outer2 != null)
            {
                float outer1ToOuter2 = Vector3.Distance(outer1.jointRb.position, outer2.jointRb.position);
                if(outer3 != null)
                {
                    float outer2ToOuter3 = Vector3.Distance(outer2.jointRb.position, outer3.jointRb.position);
                    total = areolaToMain1 + main1ToMain2 + main2ToOuter1 + outer1ToOuter2 + outer2ToOuter3;
                }
                else
                {
                    total = areolaToMain1 + main1ToMain2 + main2ToOuter1 + outer1ToOuter2;
                }

                _outer2Ratio = (areolaToMain1 + main1ToMain2 + main2ToOuter1 + outer1ToOuter2) / total;
            }
            else
            {
                total = areolaToMain1 + main1ToMain2 + main2ToOuter1;
            }

            _main1Ratio = areolaToMain1 / total;
            _main2Ratio = (areolaToMain1 + main1ToMain2) / total;
            _outer1Ratio = (areolaToMain1 + main1ToMain2 + main2ToOuter1) / total;
        }

        /* Value is parameter's overall base value based on softness, quickness and breast mass.
         * Individual joints get adjusted values based on the gradient.
         */
        internal override void SetGradientBaseValues(string paramName, GradientConfig gradientConfig, float value)
        {
            center.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 0));
            main1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main1Ratio));
            main2.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _main2Ratio));

            if(outer2 == null)
            {
                outer1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
            }
            else
            {
                outer1.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _outer1Ratio));
                if(outer3 == null)
                {
                    outer2.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
                }
                else
                {
                    outer2.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, _outer2Ratio));
                    outer3.SetBaseValue(paramName, CalculateJointBaseValue(gradientConfig, value, 1));
                }
            }
        }

        // Gradient starts from outer (ratio 1) and ends in areola (ratio 0)
        static float CalculateJointBaseValue(GradientConfig config, float value, float ratio)
        {
            float middle = 1 -
                Mathf.Lerp(
                    0.01f,
                    0.99f,
                    Mathf.InverseLerp(config.middleShiftJsf.min, config.middleShiftJsf.max, config.middleShiftJsf.val)
                );

            if(ratio > middle)
            {
                float smoothRatio = Calc.InverseSmoothStepWithMax(
                    ratio - middle,
                    1 - middle,
                    config.outerToMiddleCurvature,
                    config.outerToMiddleMidPoint
                );
                return Mathf.Lerp(value, value * config.outerMultiplierJsf.val, smoothRatio);
            }
            else
            {
                float smoothRatio = 1 -
                    Calc.InverseSmoothStepWithMax(
                        ratio,
                        middle,
                        config.middleToCenterCurvature,
                        config.middleToCenterMidPoint
                    );
                return Mathf.Lerp(value, value * config.centerMultiplierJsf.val, smoothRatio);
            }
        }

        internal override void SetAdditionalOffsets(string paramName, float value)
        {
            center.SetAdditionalOffset(paramName, value);
            main1.SetAdditionalOffset(paramName, value);
            main2.SetAdditionalOffset(paramName, value);
            outer1.SetAdditionalOffset(paramName, value);
            outer2?.SetAdditionalOffset(paramName, value);
            outer3?.SetAdditionalOffset(paramName, value);
        }

        internal override bool HasJoint(SoftJoint joint) =>
            joint == center || joint == main1 || joint == main2 || joint == outer1 || joint == outer2 || joint == outer3;
    }
}
