using System;
using System.Collections;
using System.Collections.Generic;
using NAT;
using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using NAT.Common.UI;
using NAT.Common.UI.Windows;
using NAT.TM;
using NAT.TM.Handlers;
using SimpleJSON;
using UnityEngine;

namespace Naturalis
{
    sealed class TittyMagic : ScriptBase
    {
        public override bool ShouldIgnore() => false;

        protected override void OnUIEnabled()
        {
            softPhysicsHandler.ReverseSyncAllowSelfCollision();
            uiHandler.gravityControlsSection.Show(true);

            if(uiHandler.IsMainTabActive())
            {
                if(!uiHandler.IsNestedWindowActive())
                {
                    uiHandler.mainWindow.Refresh();
                }
                else if(uiHandler.GetActiveNestedWindow() is DevWindow)
                {
                    softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(enabled);
                }
            }
            else if(uiHandler.IsHardCollidersTabActive())
            {
                if(!enabled)
                {
                    uiHandler.GoToMainWindow();
                }

                hardColliderHandler.colliderVisualizer.SetEnabledNullSafe(enabled);
                hardColliderHandler.colliderDrawService.SetEnabledNullSafe(enabled);
            }
            else if(uiHandler.IsPhysicsTabActive())
            {
                if(!uiHandler.IsNestedWindowActive())
                {
                    if(softPhysicsHandler.softPhysicsSynced)
                    {
                        uiHandler.physicsWindow.Refresh();
                    }
                }
                else if(uiHandler.GetActiveNestedWindow() is JointAngleWindow)
                {
                    // jointPhysicsHandler.jointDrawService.SetEnabledNullSafe(enabled);
                }
                else if(uiHandler.GetActiveNestedWindow() is SoftParameterWindow)
                {
                    softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(enabled);
                    var softParameterWindow = (SoftParameterWindow) uiHandler.GetActiveNestedWindow();
                    softParameterWindow.Refresh();
                    softParameterWindow.SetShowColliderPreviews();
                }
            }
            else if(uiHandler.IsClothingTabActive())
            {
                softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(enabled);
            }
        }

        protected override void OnUIDisabled()
        {
            gravityPhysicsHandler.StopGravitySimulationAction.actionCallback();
            uiHandler.gravityControlsSection.Hide();

            if(bootstrap.CalibrateOnUIClosedJsb.val && calibrationStateHandler.Pending)
            {
                StartCalibration();
            }

            softPhysicsHandler.colliderVisualizer.SetEnabledNullSafe(false);
            softPhysicsHandler.softJointDrawService.SetEnabledNullSafe(false);
            hardColliderHandler.colliderVisualizer.SetEnabledNullSafe(false);
            hardColliderHandler.colliderDrawService.SetEnabledNullSafe(false);
            calibrationStateHandler.EnableCalibrationStateJsb.val = false;

            uiHandler.ClosePopups();
            if(uiHandler.IsMainTabActive())
            {
                if(uiHandler.GetActiveNestedWindow() is SettingsWindow)
                {
                    uiHandler.mainWindow.OnReturn();
                }
            }
        }

        internal override bool IsLdpUpdateOn => softPhysicsHandler.ldpOn && bootstrap.PersonMonitor.IsBreastSoftPhysicsOn();
        internal override bool IsSoftPhysicsOn => softPhysicsHandler.softPhysicsOn && bootstrap.PersonMonitor.IsBreastSoftPhysicsOn();

        public override void Init()
        {
            calibrationDataVersion = "1.2.0";
            SetupDomain(Constant.TITTY_MAGIC, Colors.funkyCyan);
            StartOrPostponeCoroutine(SetupRoutine());
        }

        IEnumerator SetupRoutine()
        {
            yield return InitBootstrapCo();
            if(bootstrap.Done == false)
            {
                initialized = false;
                yield break;
            }

            person = bootstrap.Person;
            leftCalibrationTarget = new CalibrationTarget(person.ChestRb, person.LeftBreast.rigidbody, person);
            rightCalibrationTarget = new CalibrationTarget(person.ChestRb, person.RightBreast.rigidbody, person);
            if(enabled)
            {
                person.EnableCustomChestPhysics();
                yield return person.CycleSoftPhysicsEnabled();
            }

            /* Used to communicate version to other plugin instances */
            var versionJss = new StorableString(Constant.NATURALIS_VERSION, "")
            {
                val = Bootstrap.NATURALIS_VERSION,
            };
            PreRegisterInternalParam(versionJss, false);

            angleCalculator = new AngleCalculator(person.LeftBreast.rigidbody, person.RightBreast.rigidbody, person.ChestRb);
            SetupHandlers();
            SetupBreastMorphListener();
            SetupStorables();
            SetupCallbacks();
            calibrationStateHandler = AddCalibrationStateHandler();

            bindings = bindings ?? new Bindings(this); // Might already be setup in OnBindingsListRequested.
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);

            uiModManager = new UIModManager(this);
            uiHandler = new UIHandler(this);
            RegisterParams();

            if(!restoringFromJSON)
            {
                presetHandler.LoadDefaultPreset(person.IsFemale, true);
                if(enabled)
                {
                    foreach(var handler in handlers)
                    {
                        handler.Enable();
                    }

                    StartCalibration(true);
                }
            }

            initialized = true;
        }

        IEnumerator InitBootstrapCo()
        {
            var bootyMagic = (BootyMagic) this.GetOwnScript("BootyMagic");
            bootstrap = bootyMagic.bootstrap;
            while(!bootstrap || bootstrap.Done == null)
            {
                yield return null;
            }
        }

        void SetupHandlers()
        {
            presetHandler = new PluginPresetHandler(this, $"{DomainName}Presets");

            handlers.Add(jointPhysicsHandler = new BreastJointPhysicsHandler(this));
            handlers.Add(softPhysicsHandler = new BreastSoftPhysicsHandler(this));
            handlers.Add(gravityPhysicsHandler = new BreastGravityPhysicsHandler(this));
            handlers.Add(forcePhysicsHandler = new BreastForcePhysicsHandler(this));
            handlers.Add(frictionHandler = new BreastFrictionHandler(this));
            handlers.Add(hardColliderHandler = new HardColliderHandler(this));
            handlers.Add(forceMorphHandler = new BreastForceMorphHandler(this));
            handlers.Add(nippleErectionHandler = new NippleErectionHandler(this));
            handlers.Add(clothingHandler = new ChestClothingHandler(this));

            bootstrap.ClothingStore.onItemsReloadedHandlers += clothingHandler.OnItemsReloaded;
            bootstrap.ClothingStore.onItemToggledHandlers += clothingHandler.OnItemToggled;
            bootstrap.ClothingStore.onItemAllowDetachToggledHandlers += clothingHandler.OnItemAllowDetachToggled;
        }

        void SetupBreastMorphListener()
        {
            morphListener = new MorphListener(
                new HashSet<int>
                {
                    8, 16, 17, 93, 135, 136, 137, 138, 2401, 2402, 2403, 2404, 2405, 2406, 2407, 2408, 2409, 2410, 2411, 2412,
                    2413, 2414, 2415, 2416, 2417, 2418, 2419, 2590, 2591, 2592, 2594, 2596, 2597, 2598, 2871, 2872, 2873, 7216,
                    7218, 7219, 7220, 7230, 7231, 7232, 7233, 7234, 7235, 7236, 7237, 7238, 7239, 7240, 7241, 7242, 7243, 7244,
                    7245, 7246, 7247, 7248, 8833, 8834, 8836, 8837, 8838, 8839, 8840, 8848, 8849, 8850, 8851, 8852, 8853, 8854,
                    10936, 10944, 10945, 11021, 11063, 11064, 11065, 11066, 13231, 13232, 13233, 13234, 13235, 13236, 13237,
                    13238, 13239, 13240, 13241, 13242, 13243, 13244, 13245, 13246, 13247, 13248, 13249, 13408, 13409, 13410,
                    13411, 13412, 13413, 13414, 13672, 13673, 13674, 17923, 17924, 17925, 17926, 17936, 17937, 17938, 17939,
                    17940, 17941, 17942, 17943, 17944, 17945, 17946, 17947, 17948, 17949, 17950, 17951, 17952, 17953, 17954,
                    19504, 19505, 19507, 19508, 19509, 19510, 19511, 19519, 19520, 19521, 19522, 19523, 19524, 19525, 19639,
                },
                new HashSet<string>
                {
                    "FBMFitnessDetails",
                    "FBMHeight",
                    "PBMShouldersSize",
                    "Alessa",
                    "Eisa",
                    "Izarra",
                    "PBMLegsLength",
                    "Shin Length",
                    "Thigh Length",
                    "Upper Body Length",
                    "G3F_Hip Abdomen Shape",
                    "bellybulge",
                    "deepthroat",
                    "Athletic",
                    "BodyShort",
                    "Hollow back",
                    "FBMBodyTone",
                    "FBMBodyBuilderDetails",
                    "Julian Body",
                    "Shoulders Width",
                    "Nipples Large",
                    "WeebU.My_morphs.3.SideBulge",
                    "Ribs Width",
                    "Rectus Width",
                    "Rectus Outer Detail",
                    "Distort 1",
                    "Distort 2",
                    "Amy AiMei",
                    "succubus body",
                    "Impact - Breast",
                    "Breath1",
                    "Ribcage undefined",
                    "Sternocleidomastoid",
                }
            );

            if(bootstrap.CheckForSizeChangesJsb.val)
            {
                ProcessMorphsForListener();
            }
        }

        void SetupStorables()
        {
            softnessJsf = new StorableFloat("Breast Softness", 70f, 0f, 100f)
            {
                setCallbackFunction = _ => SetCalibrationPending(softnessJsf),
            };
            quicknessJsf = new StorableFloat("Breast Quickness Offset", 0.00f, -1.00f, 1.00f)
            {
                setCallbackFunction = _ => SetCalibrationPending(quicknessJsf),
            };

            calibrationAction = new StorableAction(Constant.CALIBRATE, () => StartCalibration(true));
            var calibrateOnNextPendingAction = new StorableAction(
                Constant.CALIBRATE_ON_NEXT_PENDING,
                () => calibrationStateHandler.calibrateOnNextPending = true
            );

            PreRegisterGeneralParam(softnessJsf);
            PreRegisterGeneralParam(quicknessJsf);
            PreRegisterGeneralParam(calibrationAction);
            PreRegisterGeneralParam(calibrateOnNextPendingAction);
        }

        void SetupCallbacks()
        {
            person.onScaleChangedHandlers += OnAtomScaleChanged;
            SuperController.singleton.onBeforeSceneSaveHandlers += OnBeforeSceneSave;
            SuperController.singleton.onSceneSavedHandlers += OnSceneSaved;
        }

        internal override void SyncMorphsAngleOffsets(float? timer = null)
        {
            if(!enabled)
            {
                return;
            }

            var leftMorphsOffsets = clothingHandler.LeftMorphsOffsets + clothingHandler.LeftMorphsOffsetsCompensation;
            var rightMorphsOffsets = clothingHandler.RightMorphsOffsets + clothingHandler.RightMorphsOffsetsCompensation;

            if(timer != null && timer.Value >= Time.deltaTime)
            {
                var transition = CoroutineUtils.ExponentialVector3ArrayTransition(
                    timer.Value,
                    new Vector3TransitionParams(
                        leftMeasurementOffset,
                        leftMorphsOffsets,
                        value => leftMeasurementOffset = value
                    ),
                    new Vector3TransitionParams(
                        rightMeasurementOffset,
                        rightMorphsOffsets,
                        value => rightMeasurementOffset = value
                    )
                );
                StartCoroutine(transition);
            }
            else
            {
                leftMeasurementOffset = leftMorphsOffsets;
                rightMeasurementOffset = rightMorphsOffsets;
            }
        }

        internal override void SetCalibrationTargetFunctions(bool softPhysicsOn)
        {
            if(softPhysicsOn)
            {
                leftCalibrationTarget.SetBreastFunctionsWhenSoftPhysicsOn(
                    person,
                    softPhysicsHandler.GetLeftPositionalRigidbodies(SoftVerticesGroupName.OUTER),
                    softPhysicsHandler.GetLeftPositionalRigidbodies(SoftVerticesGroupName.MAIN),
                    softPhysicsHandler.GetLeftPositionalRigidbodies(SoftVerticesGroupName.AREOLA)
                );
                rightCalibrationTarget.SetBreastFunctionsWhenSoftPhysicsOn(
                    person,
                    softPhysicsHandler.GetRightPositionalRigidbodies(SoftVerticesGroupName.OUTER),
                    softPhysicsHandler.GetRightPositionalRigidbodies(SoftVerticesGroupName.MAIN),
                    softPhysicsHandler.GetRightPositionalRigidbodies(SoftVerticesGroupName.AREOLA)
                );
            }
            else
            {
                leftCalibrationTarget.SetBreastFunctionsWhenSoftPhysicsOff(person.LeftBreast.nippleRigidbody);
                rightCalibrationTarget.SetBreastFunctionsWhenSoftPhysicsOff(person.RightBreast.nippleRigidbody);
            }
        }

        void Update()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                if(!isUpdatePaused)
                {
                    UpdateBase();
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"Update: {e}");
                isUpdatePaused = true;
            }
        }

        void FixedUpdate()
        {
            if(initialized != true ||
                restoringFromJSON ||
                !person.Skin ||
                person.IsFreezeGrabbing ||
                bootstrap.IsGrabbingSubscene ||
                !SuperController.singleton.autoSimulation
            )
            {
                return;
            }

            if(isUpdatePaused)
            {
                if(inCalibrationCo && gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }

                return;
            }

            try
            {
                leftCalibrationTarget.Measure(leftMeasurementOffset);
                rightCalibrationTarget.Measure(rightMeasurementOffset);

                if(gravityPhysicsHandler.IsSimulatingGravity)
                {
                    SimulateGravity();
                }
                else
                {
                    angleCalculator.CalculateAngles();
                    gravityPhysicsHandler.FixedUpdate();

                    if(!bootstrap.ForceUpdatesPaused() && !calibrationStateHandler.EnableCalibrationStateJsb.val)
                    {
                        forcePhysicsHandler.OnFixedUpdate();
                        forceMorphHandler.OnFixedUpdate();
                    }
                }

                hardColliderHandler.OnFixedUpdate();
                clothingHandler.OnFixedUpdate();
                jointPhysicsHandler.OnFixedUpdate();
                softPhysicsHandler.OnFixedUpdate();
            }
            catch(Exception e)
            {
                Loggr.LogError($"FixedUpdate: {e}");
                isUpdatePaused = true;
            }
        }

        bool _updatingMass;

        void SimulateGravity()
        {
            string direction = gravityPhysicsHandler.GravitySimulationDirection;
            angleCalculator.OverrideAngles(direction);
            gravityPhysicsHandler.AddGravitySimulationForce(!inCalibrationCo);

            if(!_updatingMass)
            {
                if(inCalibrationCo && direction != Direction.DOWN)
                {
                    leftCalibrationTarget.Measure(Vector3.zero);
                    rightCalibrationTarget.Measure(Vector3.zero);
                }

                gravityPhysicsHandler.FixedUpdate(inCalibrationCo);
                forcePhysicsHandler.OnFixedUpdate();
                forceMorphHandler.OnFixedUpdate();
            }
        }

        protected override IEnumerator CalibrationCo(JSONClass calibrationData = null)
        {
            calibrationStateHandler.EnableCalibrationStateJsb.val = false;
            inCalibrationCo = true;
            isUpdatePaused = true;
            gravityPhysicsHandler.StopGravitySimulationAction.actionCallback();
            /* Ensure everything is ready for calibration to proceed in case of e.g. appearance preset load */
            {
                /* Ensures breast soft physics enabled value is correct on first calibration when soft physics enabled is synced to SoftPhysicsHandler */
                while(!bootstrap.PersonMonitor.Initialized)
                {
                    yield return null;
                }

                /* Can occur when loading a look and VaM pauses to load assets */
                while(!person.IsSelectedCharacterReady())
                {
                    yield return null;
                }

                /* Must be unfrozen before checking for hard collider rigidbodies */
                yield return calibrationStateHandler.WaitForAtomPhysicsUnfrozen();

                bootstrap.PersonMonitor.Check(true);
                bootstrap.PersonMonitor.enabled = false;

                /* Sanity check. Might occur (?) even after hard colliders are enabled by settings monitor */
                float timeout = Time.unscaledTime + 5;
                while(!hardColliderHandler.RigidbodiesFound() && Time.unscaledTime < timeout)
                {
                    yield return null;
                }

                if(!hardColliderHandler.RigidbodiesFound())
                {
                    Loggr.LogError("Calibration error: collider rigidbodies not found after 5 seconds of waiting");
                    yield break;
                }
            }

            bool doFullCalibration =
                calibrationData == null ||
                !calibrationData.HasKey("mass") ||
                !calibrationData.HasKey("left_breast") ||
                !calibrationData.HasKey("right_breast") ||
                !calibrationData.HasKey("soft_joints");

            if(!doFullCalibration)
            {
                try
                {
                    CalibrateFromJSON(calibrationData);
                }
                catch(Exception e)
                {
                    doFullCalibration = true;
                    Loggr.LogError($"Error loading calibration data from JSON, continuing with full calibration. Error: {e}");
                }
            }

            if(doFullCalibration)
            {
                yield return FullCalibrationCo();
            }

            hardColliderHandler.SyncAll();
            SaveCalibratedValues();
            if(!restoredFromJSON && !firstCalibrationDone)
            {
                clothingHandler.AutoLoadDefaultProfiles();
            }

            clothingHandler.SetAllAdjustmentValues();
            clothingHandler.SyncAdjustments();
            SyncMorphsAngleOffsets();

            calibrationStateHandler.updateMassWhenCalibrating = false;
            isUpdatePaused = false;
            inCalibrationCo = false;
            firstCalibrationDone = true;

            uiHandler.GetActiveWindow()?.Refresh();
        }

        void CalibrateFromJSON(JSONClass jc)
        {
            softPhysicsHandler.SyncSoftPhysicsOn();
            jointPhysicsHandler.UpdateMassValueAndAmounts(jc["mass"].AsFloat);

            softness = softnessJsf.val / 100f;
            quickness = quicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            leftCalibrationTarget.RestoreFromJSON(jc["left_breast"].AsObject);
            rightCalibrationTarget.RestoreFromJSON(jc["right_breast"].AsObject);
            softPhysicsHandler.CalibrateFromJSON(jc["soft_joints"].AsObject);

            calibrationStateHandler.EnableSettingsMonitor();
        }

        IEnumerator FullCalibrationCo()
        {
            softPhysicsHandler.softJointDrawService.ClearLines();
            hardColliderHandler.colliderDrawService.ClearLines();
            hardColliderHandler.colliderVisualizer.showPreviewsJSON.val = false;

            /* Must be called before calibration pose is applied because the current pose must be stored without duplicate morphs having >0 values. */
            ZeroDynamicAdjustments();
            person.ZeroAllDuplicateMorphs();

            yield return calibrationStateHandler.Begin();
            if(calibrationStateHandler.cancelling)
            {
                calibrationStateHandler.cancelling = false;
                yield break;
            }

            softPhysicsHandler.SyncSoftPhysicsOn();

            {
                gravityPhysicsHandler.EnableGravitySimulation(Direction.DOWN);
                yield return UpdateMassAndPhysicsCo();

                jointPhysicsHandler.SetGravitySimulationDamper();
                yield return new WaitForSeconds(0.25f);

                float minTime = Time.time + 0.40f;
                float maxTime = Time.time + 1.00f;
                const float tolerance = 0.003f;
                while(Time.time < minTime ||
                    ((leftCalibrationTarget.Moved(tolerance) || rightCalibrationTarget.Moved(tolerance)) && Time.time < maxTime)
                )
                {
                    yield return null;
                }

                leftCalibrationTarget.Calibrate();
                rightCalibrationTarget.Calibrate();
                if(IsLdpUpdateOn)
                {
                    softPhysicsHandler.CalibrateSoftJoints(Direction.DOWN);
                }
            }

            if(IsLdpUpdateOn)
            {
                foreach(string direction in new[] { Direction.FORWARD, Direction.UP, Direction.LEFT, Direction.RIGHT, Direction.BACK })
                {
                    gravityPhysicsHandler.EnableGravitySimulation(direction);
                    yield return new WaitForSeconds(0.25f);

                    float timeout = Time.time + Mathf.Lerp(1.00f, 2.00f, jointPhysicsHandler.normalizedRealMass);
                    while(softPhysicsHandler.AnySoftJointMoved() && Time.time < timeout)
                    {
                        yield return null;
                    }

                    softPhysicsHandler.CalibrateSoftJoints(direction);
                }
            }

            gravityPhysicsHandler.StopGravitySimulation();
            jointPhysicsHandler.UnsetGravitySimulationDamper();
            hardColliderHandler.colliderVisualizer.showPreviewsJSON.val = true;
            yield return calibrationStateHandler.Finish();
        }

        void ZeroDynamicAdjustments()
        {
            leftCalibrationTarget.Reset();
            rightCalibrationTarget.Reset();
            foreach(var handler in handlers)
            {
                handler.ResetAll();
            }
        }

        IEnumerator UpdateMassAndPhysicsCo()
        {
            _updatingMass = true;
            if(calibrationStateHandler.updateMassWhenCalibrating)
            {
                softness = 0;
                quickness = 0;
                jointPhysicsHandler.UpdateBaseValues(false);
                softPhysicsHandler.UpdatePhysics(false);

                yield return new WaitForSeconds(0.25f);
                /* Mass is calculated multiple times because each new mass value changes the exact breast
                 * shape and therefore the estimated volume.
                 */
                jointPhysicsHandler.CalculateMass();
                yield return Utils.WaitAndRepeat(
                    () => jointPhysicsHandler.CalculateMass(),
                    (int) Mathf.Lerp(4, 12, jointPhysicsHandler.normalizedRealMass),
                    0.13f
                );
            }

            softness = softnessJsf.val / 100f;
            quickness = quicknessJsf.val;
            SetMultipliersAndUpdatePhysics();

            _updatingMass = false;
        }

        protected override void SetMultipliersAndUpdatePhysics()
        {
            foreach(var handler in handlers)
            {
                handler.SetMultipliers();
            }

            jointPhysicsHandler.UpdateBaseValues();
            softPhysicsHandler.UpdatePhysics();
            nippleErectionHandler.UpdateNipple();
            frictionHandler.CalculateFriction();
        }

        void SaveCalibratedValues()
        {
            softnessJsf.SaveCalibrated();
            quicknessJsf.SaveCalibrated();
            morphListener.SaveCalibrated();
            calibratedScale = person.ScaleJsf.val;
            foreach(var handler in handlers)
            {
                handler.SaveCalibrated();
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if(_isSavingScene && firstCalibrationDone && !calibrationStateHandler.Pending && includePhysical && includeAppearance)
            {
                jc["stored_in_scene"].AsBool = true; // Unused, might be useful later
                var calibrationData = new JSONClass
                {
                    ["version"] = calibrationDataVersion,
                    ["mass"] = { AsFloat = jointPhysicsHandler.calibratedMass },
                    ["left_breast"] = leftCalibrationTarget.Serialize(),
                    ["right_breast"] = rightCalibrationTarget.Serialize(),
                    ["soft_joints"] = softPhysicsHandler.SerializeCalibrationData(),
                };
                jc["calibration_data"] = calibrationData;
            }

            var activeClothingProfiles = new JSONArray();
            foreach(var profile in clothingHandler.activeProfiles)
            {
                if(!string.IsNullOrEmpty(profile.loadedFullPath))
                {
                    activeClothingProfiles.Add(new JSONClass
                    {
                        ["presetName"] = profile.loadedFullPath,
                    });
                }
            }

            jc["Active Clothing Profiles"] = activeClothingProfiles;

            needsStore = true;
            return jc;
        }

        protected override void FinishRestoreFromJSON(JSONClass jc, string scenePackageId, JSONClass calibrationData)
        {
            if(!firstCalibrationDone)
            {
                // ReSharper disable once ExpressionIsAlwaysNull
                clothingHandler.AutoLoadProfiles(JSONUtils.GetJSONArrayEnumerable(jc, "Active Clothing Profiles"), scenePackageId);
            }

            if(enabled)
            {
                foreach(var handler in handlers)
                {
                    handler.Enable();
                }

                StartCalibration(true, calibrationData);
            }
        }

        void OnAtomScaleChanged(float value)
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            if(!calibrationStateHandler.Pending && !calibrationStateHandler.calibrateOnNextPending)
            {
                Loggr.LogMessage("Person atom scale changed - calibration set to pending.");
            }

            calibrationStateHandler.SetCalibrationPending(Constant.SCALE_CHANGED, !value.IsEqual(calibratedScale));
        }

        bool _isSavingScene;

        void OnBeforeSceneSave()
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            try
            {
                _isSavingScene = true;
                isUpdatePaused = true;
                forceMorphHandler.ResetAll();
                jointPhysicsHandler.ResetAll();
                nippleErectionHandler.Reset();
                clothingHandler.ResetMorphs();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnBeforeSceneSave: {e}");
            }
        }

        void OnSceneSaved()
        {
            if(initialized != true || !IsEnabledAndActive() || !_isSavingScene)
            {
                return;
            }

            try
            {
                _isSavingScene = false;
                if(!calibrationStateHandler.CalibratingJsb.val)
                {
                    isUpdatePaused = false;
                    nippleErectionHandler.UpdateNipple();
                    clothingHandler.SyncMorphsOffsetsCompensation();
                    SyncMorphsAngleOffsets();
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnSceneSaved: {e}");
            }
        }

        internal override void OnSoftPhysicsToggled(bool value)
        {
            if(initialized != true || !IsEnabledAndActive())
            {
                return;
            }

            try
            {
                softPhysicsHandler.ReverseSyncSoftPhysicsOn();

                if(!inCalibrationCo)
                {
                    isUpdatePaused = true;
                    calibrationStateHandler.SetCalibrationPending(Constant.SOFT_PHYSICS_CHANGED);
                    if(!calibrationStateHandler.calibrateOnNextPending)
                    {
                        Loggr.LogMessage($"Soft physics {(value ? "enabled" : "disabled")} - updates paused, calibration set to pending.");
                    }
                }

                softPhysicsHandler.softPhysicsSynced = false;

                if(uiHandler.IsPhysicsTabActive())
                {
                    if(uiHandler.GetActiveNestedWindow() is SoftParameterWindow)
                    {
                        if(!value)
                        {
                            uiHandler.physicsWindow.OnReturn();
                        }
                    }
                    else if(!value)
                    {
                        uiHandler.physicsWindow.Refresh();
                    }
                }

                if(!value)
                {
                    softPhysicsHandler.softJointDrawService.visualizeLdpJsb.val = false;
                }
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnSoftPhysicsToggled: {e}");
            }
        }

        void OnEnable()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                BaseOnEnable();
                person.EnableCustomChestPhysics();

                if(!restoringFromJSON)
                {
                    foreach(var handler in handlers)
                    {
                        handler.Enable();
                    }

                    StartCalibration(true);
                }

                uiModManager.Enable();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnEnable: {e}");
            }
        }

        void OnDisable()
        {
            if(initialized != true)
            {
                return;
            }

            try
            {
                if(calibrationCo != null)
                {
                    StopCoroutine(calibrationCo);
                }

                foreach(var handler in handlers)
                {
                    handler.Disable();
                }

                uiModManager.Disable();
                calibrationStateHandler.Finish(bootstrap.BootyMagic.IsEnabledNullSafe());
                person.DisableCustomChestPhysics();
            }
            catch(Exception e)
            {
                Loggr.LogError($"OnDisable: {e}");
            }
        }

        void OnDestroy()
        {
            try
            {
                BaseOnDestroy();
                foreach(var handler in handlers)
                {
                    handler.Destroy();
                }

                Destroy(bootstrap);
                bootstrap = null;

                SuperController.singleton.onSceneSavedHandlers -= OnSceneSaved;
                SuperController.singleton.onBeforeSceneSaveHandlers -= OnBeforeSceneSave;
                SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            }
            catch(Exception e)
            {
                if(initialized == true)
                {
                    SuperController.LogError($"OnDestroy: {e}");
                }
                else
                {
                    Debug.LogError($"OnDestroy: {e}");
                }
            }
        }
    }
}
