using NAT.Common;
using NAT.Common.Components;
using System;
using System.Collections;
using System.Text;
using UnityEngine;

namespace NAT.TM
{
    sealed class UIModManager : UIModManagerBase
    {
        readonly Bootstrap _bootstrap;

        internal UIModManager(ScriptBase script)
        {
            this.script = script;
            _bootstrap = script.bootstrap;
            frictionHandler = script.frictionHandler;

            uiMods["calibrationButton"] = new UIMod(script, null, CreateCalibrationButtonCo, true);
            uiMods[Constant.M_PECTORAL_PHYSICS] = new UIMod(script, "M Pectoral Physics", ReplaceWithPluginUIButtonCo);
            uiMods[Constant.F_BREAST_PHYSICS_1] = new UIMod(script, "F Breast Physics 1", ReplaceWithPluginUIButtonCo);
            uiMods[Constant.F_BREAST_PHYSICS_2] = new UIMod(script, "F Breast Physics 2", ReplaceWithPluginUIButtonCo);
            uiMods[Constant.F_BREAST_PRESETS] = new UIMod(script, "F Breast Presets", ReplaceWithPluginUIButtonCo);
            uiMods[Constant.SKIN_MATERIALS_2_ADD_BTN] = new UIMod(script, "Skin Materials 2", AddButtonToSkinMaterialsUICo);
            uiMods[Constant.SKIN_MATERIALS_2_FRICTION] = new UIMod(script, "Skin Materials 2", CreateFrictionUICo);

            foreach(var pair in uiMods)
            {
                pair.Value.DeferApply();
            }

            if(!script.enabled)
            {
                Disable();
            }

            initialized = true;
        }

        RectTransform _calibrationRect;
        UIDynamicButton _calibrationButton;

        IEnumerator CreateCalibrationButtonCo(UIMod uiMod)
        {
            while(script.initialized != true)
            {
                yield return null;
            }

            try
            {
                var customTransform = script.InstantiateButton(script.GetParentUI());
                _calibrationRect = customTransform.GetComponent<RectTransform>();
                _calibrationRect.pivot = Vector2.zero;
                _calibrationRect.anchoredPosition = script.person.IsFemale ? new Vector2(277, 155) : new Vector2(0, 155);
                _calibrationRect.sizeDelta = new Vector2(-837, 80);
                uiMod.AddCustomObject(customTransform.gameObject);

                _calibrationButton = customTransform.GetComponent<UIDynamicButton>();
                _calibrationButton.AddListener(() => script.StartCalibration());
                script.calibrationStateHandler.CalibrationButton = _calibrationButton;
            }
            catch(Exception e)
            {
                script.Loggr.LogError($"CreateCalibrationButtonCo: {e}");
            }
        }

        IEnumerator ReplaceWithPluginUIButtonCo(UIMod uiMod)
        {
            while(script.bindings == null && script.initialized != true)
            {
                yield return null;
            }

            try
            {
                uiMod.InactivateChildren();
                uiMod.AddCustomObject(OpenPluginUIButton(uiMod.Target).gameObject);
            }
            catch(Exception e)
            {
                script.Loggr.LogError($"ReplaceWithPluginUIButtonCo: {e}");
            }
        }

        IEnumerator CreateFrictionUICo(UIMod uiMod)
        {
            while(script.initialized != true)
            {
                yield return null;
            }

            if(script.person.HasDummySkin())
            {
                while(!frictionHandler.IsEnabled)
                {
                    yield return null;
                }
            }
            else if(!frictionHandler.IsEnabled)
            {
                yield break;
            }

            try
            {
                var leftSide = uiMod.Target.Find("LeftSide");
                var rightSide = uiMod.Target.Find("RightSide");

                /* Collider friction title */
                {
                    var fieldTransform = script.InstantiateTextField(leftSide);
                    var rectTransform = fieldTransform.GetComponent<RectTransform>();
                    rectTransform.pivot = Vector2.zero;
                    rectTransform.anchoredPosition = new Vector2(20, -930);
                    rectTransform.sizeDelta = new Vector2(rectTransform.sizeDelta.x + 10, 100);
                    uiMod.AddCustomObject(fieldTransform.gameObject);

                    var uiDynamic = fieldTransform.GetComponent<UIDynamicTextField>();
                    uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                    uiDynamic.text = $"{Constant.TITTY_MAGIC} Collider Friction".Size(32).Bold();
                    uiDynamic.textColor = Color.white;
                    uiDynamic.backgroundColor = Color.clear;
                }

                /* Collider friction info text area */
                {
                    var fieldTransform = script.InstantiateTextField(leftSide);
                    var rectTransform = fieldTransform.GetComponent<RectTransform>();
                    rectTransform.pivot = Vector2.zero;
                    rectTransform.anchoredPosition = new Vector2(20, -1290);
                    rectTransform.sizeDelta = new Vector2(rectTransform.sizeDelta.x + 10, 400);
                    uiMod.AddCustomObject(fieldTransform.gameObject);

                    var uiDynamic = fieldTransform.GetComponent<UIDynamicTextField>();
                    uiDynamic.text = new StringBuilder()
                        .AppendLine("Combined friction for hard colliders and soft colliders.")
                        .AppendLine()
                        .Append("Adaptive friction reduces friction when ")
                        .AppendItalic("Gloss")
                        .Append(" is increased. The higher the gloss, the more ")
                        .AppendItalic("Specular Bumpiness")
                        .AppendLine(" adds friction. Other skin material sliders are ignored.")
                        .AppendLine()
                        .Append("Dry skin friction represents the friction value when gloss and specular bumpiness are low.")
                        .ToString();
                    uiDynamic.textColor = Color.white;
                    uiDynamic.backgroundColor = Color.clear;
                }

                /* Adaptive friction toggle */
                {
                    var customTransform = CustomToggle(rightSide, new Vector2(0, -880));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var jsb = frictionHandler.AdaptiveFrictionJsb;
                    var uiDynamic = customTransform.GetComponent<UIDynamicToggle>();
                    uiDynamic.label = jsb.name;
                    jsb.toggle = uiDynamic.toggle;
                }

                /* Dry skin friction slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1020));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    drySkinFrictionSlider = customTransform.GetComponent<UIDynamicSlider>();
                    var jsf = frictionHandler.DrySkinFrictionJsf;
                    drySkinFrictionSlider.Configure(
                        jsf.name,
                        jsf.min,
                        jsf.max,
                        jsf.defaultVal,
                        jsf.constrained,
                        "F3"
                    );
                    jsf.slider = drySkinFrictionSlider.slider;
                    drySkinFrictionSlider.SetActiveStyle(frictionHandler.AdaptiveFrictionJsb.val, true);
                }

                /* Friction offset slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1160));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var uiDynamic = customTransform.GetComponent<UIDynamicSlider>();
                    var jsf = frictionHandler.FrictionOffsetJsf;
                    uiDynamic.Configure(
                        jsf.name,
                        jsf.min,
                        jsf.max,
                        jsf.defaultVal,
                        jsf.constrained,
                        "F3"
                    );
                    jsf.slider = uiDynamic.slider;
                }

                /* Soft collider friction value slider */
                {
                    var customTransform = CustomSlider(rightSide, new Vector2(0, -1300));
                    uiMod.AddCustomObject(customTransform.gameObject);

                    var uiDynamic = customTransform.GetComponent<UIDynamicSlider>();
                    var jsf = frictionHandler.FrictionValueJsf;
                    uiDynamic.Configure(
                        jsf.name,
                        jsf.min,
                        jsf.max,
                        jsf.defaultVal,
                        jsf.constrained,
                        "F3"
                    );
                    jsf.slider = uiDynamic.slider;
                    uiDynamic.SetActiveStyle(false, true);
                }

                if(frictionHandler.IsEnabled && script.enabled)
                {
                    HideOtherPluginFrictionUI();
                    skinMaterials2Modified = true;
                }
                else
                {
                    HideFrictionUI();
                }
            }
            catch(Exception e)
            {
                script.Loggr.LogError($"CreateFrictionUICo: {e}");
            }
        }

        void HideOtherPluginFrictionUI()
        {
            if(script.OtherPlugin != null)
            {
                script.OtherPlugin.uiModManager?.HideFrictionUI();
            }
        }

        IEnumerator AddButtonToSkinMaterialsUICo(UIMod uiMod)
        {
            while(script.initialized != true)
            {
                yield return null;
            }

            if(script.person.HasDummySkin())
            {
                while(!frictionHandler.IsEnabled)
                {
                    yield return null;
                }
            }
            else if(!frictionHandler.IsEnabled)
            {
                yield break;
            }

            try
            {
                var customTransform = CustomButton(uiMod.Target.Find("LeftSide"), new Vector2(20, -810));
                uiMod.AddCustomObject(customTransform.gameObject);
                var uiDynamicButton = customTransform.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = $"Switch To {script.DomainName}";
                uiDynamicButton.AddListener(SwitchToOwnFrictionUI);
                uiMod.Disable();
            }
            catch(Exception e)
            {
                script.Loggr.LogError($"AddButtonToSkinMaterialsUICo: {e}");
            }
        }

        Transform OpenPluginUIButton(Transform parent)
        {
            var button = script.InstantiateButton(parent);
            var uiDynamic = button.GetComponent<UIDynamicButton>();
            uiDynamic.label = new StringBuilder()
                .AppendFormat("<b>Open {0} UI</b>", Constant.TITTY_MAGIC)
                .ToString();
            uiDynamic.AddListener(() => script.bindings.actions["OpenUI"].actionCallback());
            return button;
        }

        Transform CustomToggle(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = script.InstantiateToggle(parent);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        Transform CustomSlider(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = script.InstantiateSlider(parent);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        Transform CustomButton(Transform parent, Vector2 anchoredPosition)
        {
            var customTransform = script.InstantiateButton(parent);
            var rectTransform = customTransform.GetComponent<RectTransform>();
            rectTransform.pivot = Vector2.zero;
            rectTransform.anchoredPosition = anchoredPosition;
            var sizeDelta = rectTransform.sizeDelta;
            rectTransform.sizeDelta = new Vector2(sizeDelta.x + 10, sizeDelta.y);
            return customTransform;
        }

        internal override void SetFrictionUIVisibility()
        {
            if(!script.enabled)
            {
                return;
            }

            if(frictionHandler.IsEnabled)
            {
                SwitchToOwnFrictionUI();
                _bootstrap.UIModManager.SetSkinMaterials2ModEnabled(true);
            }
            else
            {
                uiMods[Constant.SKIN_MATERIALS_2_FRICTION].Disable();
                uiMods[Constant.SKIN_MATERIALS_2_ADD_BTN].Disable();
                skinMaterials2Modified = false;
                if(!script.OtherPlugin.uiModManager.skinMaterials2Modified)
                {
                    _bootstrap.UIModManager.SetSkinMaterials2ModEnabled(false);
                }
            }
        }

        internal override void Enable()
        {
            if(!initialized)
            {
                return;
            }

            foreach(var pair in uiMods)
            {
                if(pair.Key == Constant.SKIN_MATERIALS_2_FRICTION)
                {
                    if(frictionHandler.IsEnabled)
                    {
                        pair.Value.Enable();
                        skinMaterials2Modified = true;
                    }
                }
                else if(pair.Key != Constant.SKIN_MATERIALS_2_ADD_BTN)
                {
                    pair.Value.Enable();
                }
            }

            _bootstrap.UIModManager.SetEnabled(frictionHandler.IsEnabled);
            HideOtherPluginFrictionUI();
        }

        internal override void Disable()
        {
            base.Disable();
            skinMaterials2Modified = false;
            if(script.OtherPlugin != null)
            {
                script.OtherPlugin.uiModManager?.ShowFrictionUI();
            }
            else
            {
                _bootstrap.UIModManager.Disable();
            }
        }
    }
}
